//-< OODBS.HPP >-----------------------------------------------------*--------*
// GOODS                     Version 1.0         (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Database management                                               *  /  \  *
//                           Created:      24-Feb-92  K.A. Knijnik   * / [] \ *
//                           Last update:   9-Oct-95  K.A. Knijnik   * GARRET *
//-------------------------------------------------------------------*--------*

#ifndef __OODBS_HPP__
 #define __OODBS_HPP__

 #include "ctprc.hpp"
 #include "store.hpp"

 #define  Dbs_Max_Stores         Dbs_Slot_Size // Maximum number of stores
 #define  Dbs_Space(ADDR)        (int(ADDR) & (Dbs_Slot_Size-1))
 #define  Dbs_Offset(ADDR)       (nat4(ADDR) & -Dbs_Slot_Size)
 #define  Dbs_Address(FILE,OFFS) ( (FILE) | (OFFS))

 struct Dbs_Exception : Ctk_Exception {
     Ctk_String*   File; // C++ file with primitive
     nat4          Line; // line in C file in which assertion was failed
     Dbs_Exception ( const char* File, int Line );
 };

 struct Dbs_Deadlock_Exception : Ctk_Exception { 
     Dbs_Deadlock_Exception() : Ctk_Exception( Dbs_Deadlock_Exception_Desc ) {}
 }; 

 struct Dbs_Mixing_Exception : Ctk_Exception { 
     Dbs_Mixing_Exception() : Ctk_Exception( Dbs_Mixing_Exception_Desc ) {}
 }; 

 struct Dbs_Store_Failed : Ctk_Exception { 
     Dbs_Store_Failed() : Ctk_Exception( Dbs_Store_Failed_Desc ) {}
 }; 
 
 struct Dbs_Zomby_Access : Ctk_Exception { 
     Dbs_Zomby_Access() : Ctk_Exception( Dbs_Zomby_Access_Desc ) {}
 }; 
 
 struct Dbs_Class_Not_Found : Ctk_Exception { 
     Ctk_String *Pkg, *Cls; 	
     Dbs_Class_Not_Found( const char* pkg_name, const char* cls_name )
     : Ctk_Exception( Dbs_Class_Not_Found_Desc, 0 ) 
     {	 
	Pkg = Ctk_String::Newc(pkg_name);
	Cls = Ctk_String::Newc(cls_name);
     }
 }; 
 
 struct Dbs_RPC_Error : Ctk_Exception {
     Dbs_RPC_Error() : Ctk_Exception( Dbs_RPC_Error_Desc ) {}
 };

 struct Dbs_DDS : Ctk_Exception { // data deterioration signal
   Dbs_DDS() : Ctk_Exception( Dbs_DDS_Desc ) {}
 };

 #define DBS_EXCEPTION  CTK_RAISE Dbs_Exception( __FILE__, __LINE__ )

 struct Ctk_Manager : Ctk_Resource {
     Ctk_Manager*  Next;
     Ctk_Manager*  Prev;
     static Ctk_Manager* Active;
     Ctk_Manager() : Ctk_Resource( Ctk_Manager_Desc, NULL ) 
	{ Next = Prev = this; }
 };

 class Ctk_Database : public Ctk_Manager {
   public:
     Ctk_Array_O*    	Store ;   // vector with base storages
     Ctk_String*       	Config;   // name of file with database configuration

     Ctk_Array_I4* 	Permutation; // used for database reformatting 	

     Dbs_Object*       	Chain ;   // chain of objects involved in transaction
     int                Within;   // counter of nested transactions
     int                Touch ;   // store taking part in transaction
     nat4               Loaded;   // total size of objects loaded between GC
     nat4               Timer;    // counter of loaded objects
     int                Deteriorated; // amount of deteriorated objects

     enum  { Signal = 1, Reload = 2 };
     int                DDS_Mask; // data deterioration signal mask

   private:
     Mutex          	Db_Mutex;     // only one database method can be 
				      // executed each time

     Mutex	        Trans_Mutex;  // mutex to prevent different 
				      // processes from starting parallel 
                                      // transaction: only one process can 
                                      // be involved in transaction and until
                                      // it preform Commit transaction no other
                                      // processes can modify any objects

     bool		Opened;       

     int		Involved;  // number of involved in transaction objects

     struct        {               // number and total size of objects from
         int       Num ;           // each store involved in transaction
         nat4      Size;
     } Total [ Dbs_Max_Stores ];

   public:
     Dbs_Object*         Hash[ 256 ]; // object hash table

     static   Dbs_DDS*   DDS;

     void		Check() 
      { if( !Opened ) { CTK_RAISE Ctk_Not_Available; Ctk_Process::Unwind(); } }

    //
    // Transaction control procedures
    //
   public:
     void                Begin  ( void );
     void                End    ( void );

     void                Commit ( void );
     void                Abort  ( void );
     void                Fix    ( Dbs_Object* Obj );
     void 		 Turn_Off_Transactions(); 

     // Perform mutual exclusion of processes running transactions
     void		 Transaction_Hold() { Trans_Mutex.enter(); }
     void		 Transaction_Release() { Trans_Mutex.release(); } 

     // Perform mutual exclusion of database methods
     void		 Database_Hold() { Db_Mutex.enter(); Check(); }
     void		 Database_Release() { Db_Mutex.leave(); } 	
	
      
    //
    // Object cache
    //
     void                Save   ( void );  // save all modified object in storages
   public:
     Dbs_Object*         Find   ( Dbs_Pointer Addr ); // find (or load) object
     Dbs_Object*         Get    ( Dbs_Pointer Addr ); // find object ( null if none )
     Dbs_Object*         Root   ( void );  // database root object

     Ctk_Boolean*        Empty  ( void );  // is database empty
     void		 UnLock ( int Mode ); // unlock all objects in database

     void                Flush  ( void );  // flush all storages
     void                Dump   ( void );  // dump database structures
     void                Throw  ( void );  // throw out unrefernced objects
     void                Inform ( void );  // inform server about swapped out o.
     void                Link   ( Dbs_Object* Obj ); // add new object in hash

     inline bool	 Valid_Object( Dbs_Object* Obj ) { 
	 // object is valid if it is not created within aborted transaction
	 // and is builtin (NIL,TRUE,FALSE) or belong to the same database
	 return !(Obj->State & Dbs_Object::Zomby) &&
                 (Obj->Addr < Dbs_Slot_Size || Obj->Store()->Base == this); 
     }
    //
    // Database control function
    //
     void                Open   ( void ); // open database
     void                Close  ( void ); // close database

     // wait until all remote servers are opened
     // !!! Internal function, don't call it !
     void 		 WaitAllOpened( int server ); 
		
     static   void       Init   ( void ); // database initialisation

     Ctk_Boolean*        Gc_Start();       // start garbage collection

     void	         Reformat( void ); // handle changes in database scheme
					   // (reallocate expanded objects)

     static int Permutation_Cmp( const void*, const void* ); 	
    //
    //  Updating deteriorated data
    //
     void                Update ( void );   // reread all deteriorated objects
     static void         Update_All ( void ); // update all active databases
     void                DDS_Enable ( void ); // enable data deterioration 
     					      // exception
     void                DDS_Disable( void ); // disable data deterioration 
     					      // exception
     void                Auto_Update( void ); // automatic update mode
     void                Lazy_Update( void ); // prohibit autoimatic update mode
     void		 Hold();
     void		 Release();

     static void Set_Deadlock_Detection_Delay( int delay ); 
     static int	Deadlock_Detection_Delay; 

     static Ctk_Database* New( Ctk_Array_O* Stores, Ctk_String* Cfg );
     Ctk_Database() { Desc = Ctk_Database_Desc; }
 };

 inline Dbs_Store* Dbs_Object::Get_Store() const { 
    Dbs_Store* s = Store(); 
    if( s->Base == NULL ) {
	CTK_RAISE Ctk_Not_Available; 
	Ctk_Process::Unwind(); 
    }
    return s;
 } 

 inline void Dbs_Object::Save  () {  // copy modified object on disk
     if( (State & (Write|Fixed)) == Write ) 
         State &= ~(Write|News), Get_Store()->Write(this); 
 }
 inline void Dbs_Object::Throw () {  // throw out unreferenced object
     Save(); 
     Get_Store()->Throw( this ); 
     if( State & Old ) Store()->Base->Deteriorated -= 1; 
 }

 inline Dbs_Object* Dbs_Object::Check( Dbs_Object* Obj ) const { 
     if( State & Obj->State & Zomby ) 
	 CTK_RAISE Dbs_Zomby_Access; 
     if( Obj->Addr < Dbs_Slot_Size ) // NIL, TRUE, FALSE	
	 return Obj; // builtin persistent objects are always correct
     if( Store()->Base != Obj->Store()->Base )
	 CTK_RAISE Dbs_Mixing_Exception;
     return Obj;
 }

 #define   FIND(a)    Get_Store()->Base->Find(a)

#endif
