#include "stdtp.hpp"
#include "numlb.hpp"
#include "dnmem.hpp"
#include "semph.hpp"
#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctlib.hpp"
#include "ctwin.hpp" 

Ctk_Box* Ctk_Box::New( int nx, int ny, int nw, int nh ) {
    x = nx, y = ny;
    width = nw, height = nh;
    return this;
}
 
void Ctk_Box::Coord( long Res[4] ) {
    Res[0] = x, Res[1] = y, Res[2] = width, Res[3] = height;  
}

Ctk_Boolean* Ctk_Box::Contain( int px, int py ) { 
    return unsigned(px - x) < unsigned(width)
        && unsigned(py - y) < unsigned(height)
	    ? CTK_TRUE : CTK_FALSE; 
}

Ctk_Arc* Ctk_Arc::New( int nx, int ny, int nw, int nh, int na1, int na2 ) {
    x = nx, y = ny;
    width = nw, height = nh;
    angle1 = na1, angle2 = na2;
    return this;
}

Ctk_Point* Ctk_Point::New( int nx, int ny ) {
    x = nx, y = ny; 
    return this;
}

Ctk_Segment* Ctk_Segment::New( int nx1, int ny1, int nx2, int ny2 ) {
    x1 = nx1, y1 = ny1;
    x2 = nx2, y2 = ny2;
    return this;
}

local Ctk_Primitive
    ps1( "Box::New", cnv(&Ctk_Box::New), "windows::box" ), 
    ps2( "Box::Coord", cnv(&Ctk_Box::Coord), "windows::box" ),  
    ps3( "Box::Contain", cnv(&Ctk_Box::Contain), "windows::box" ),
    ps4( "Point::New", cnv(&Ctk_Point::New), "windows::point" ), 
    ps5( "Segment::New", cnv(&Ctk_Segment::New), "windows::segment" ), 
    ps6( "Arc::New", cnv(&Ctk_Arc::New), "windows::arc" );

//-----------------------------------------------------------------------------

Ctk_Window_Interface_Descriptor* Ctk_Window_Interface_Descriptor::head = NULL;

Ctk_Window_Interface_Descriptor::Ctk_Window_Interface_Descriptor(
    const char* name, Win_Init_Func f, Win_Open_Func g, 
    Ctk_Window_Interface_Descriptor::Win_Pri pri)
{
    Ctk_Window_Interface_Descriptor** wip;

    // Initialize data fields
    Ctk_Window_Interface_Descriptor::name = name;
    open_func = g;
    init_func = f;
    Ctk_Window_Interface_Descriptor::pri = pri;

    // Link into the global window interface list
    for (wip = &head; *wip != NULL && (*wip)->pri > pri; wip = &(*wip)->next);
    next = *wip;
    *wip = this;
}

Ctk_Window_Interface_Descriptor::~Ctk_Window_Interface_Descriptor() {
    Ctk_Window_Interface_Descriptor** wip;

    // Unlink from the global list
    for (wip = &head; *wip != this; wip = &(*wip)->next);
    *wip = next;
}

//-----------------------------------------------------------------------------

Ctk_Display* Ctk_Window_System::head = NULL;

int         Ctk_Win_Input_Queue::n_elems;
Ctk_Event** Ctk_Win_Input_Queue::beg;
Ctk_Event** Ctk_Win_Input_Queue::end;
Ctk_Event*  Ctk_Win_Input_Queue::queue[Ctk_Win_Input_Queue::event_queue_size];
Semaphore*  Ctk_Win_Input_Queue::Input_Semph;

//-----------------------------------------------------------------------------

Ctk_Event* Ctk_Win_Input_Queue::get(){
    Ctk_Event* evt;

    Input_Semph->wait();	// Wait for input data arrival
    evt = *beg;
    beg = forw(beg);
    return evt;
}

//-----------------------------------------------------------------------------

void Ctk_Win_Input_Queue::init() {
    if (Input_Semph == NULL) {
	Input_Semph = new Semaphore;
	beg = end = queue;
    }
}

//-----------------------------------------------------------------------------

Ctk_Display* Ctk_Window_System::Display_List_Head() {

    // Check for empty list
    if (head == NULL)
	CTK_RAISE Ctk_UFO_Error;

    return head;
}

//-----------------------------------------------------------------------------

Ctk_Display* Ctk_Window_System::Open_Display(Ctk_String* name) {
    Ctk_Window_Interface_Descriptor* 	widp;
    Ctk_Display*		 	dpy;

    Ctk_Win_Input_Queue::init();

    // Try open display for a name given
    for (widp = Ctk_Window_Interface_Descriptor::head; 
	 widp != NULL; widp = widp->next)
	  {
	      if ((dpy = widp->open(name)) != NULL)
		  return dpy;
	  }

    // No valid display name, raise exception
    CTK_RAISE Ctk_UFO_Error;
    return NULL;
}

//-----------------------------------------------------------------------------

void Ctk_Window_System::Initialize() {
    Ctk_Window_Interface_Descriptor* wid;
    for( wid =  Ctk_Window_Interface_Descriptor::head; wid; wid = wid->next )
	 wid->init_func(); 
}

//-----------------------------------------------------------------------------

Ctk_Event*   Ctk_Window_System::Get_Event() {

    // Check for active input sources
    if (head == NULL) {
	CTK_RAISE Ctk_UFO_Error;
	return NULL;
    }

    Ctk_Event* Evt;
    while( (Evt = Ctk_Win_Input_Queue::get()) == NULL 
	&& !(Ctk_Event_Flag & CTK_INTERN) ); 

    return Evt; 
}

local Ctk_Primitive 
wns0("Window_System::Display_List_Head", cnv(&Ctk_Window_System::Display_List_Head),
     "windows::window_system"),
wns1("Window_System::Open_Display", cnv(&Ctk_Window_System::Open_Display),
     "windows::window_system"),
wns2("Window_System::Get_Event", cnv(&Ctk_Window_System::Get_Event),
     "windows::window_system"),
wns3("Window_System::Initialize", cnv(&Ctk_Window_System::Initialize),
     "windows::window_system");

//-----------------------------------------------------------------------------

import rtm_hook_list background_routines; 

static bool flush_display_hook( void* arg, rtm_hook_list* ) 
{ 
    ((Ctk_Display*)arg)->Flush();
    return FALSE;
}
 
Ctk_Display::Ctk_Display( nat2 Desc ) : Ctk_Resource( Desc ) {
    background_routines.add_hook( &flush_display_hook, this ); 
    next = Ctk_Window_System::head;
    Ctk_Window_System::head = this;
}

Ctk_Display::~Ctk_Display() {
    if (Desc == 0) { 
        return;
    }
    Ctk_Display** dpp;
    ms->dsp = NULL; 
    kbd->dsp = NULL; 
    for (Ctk_Screen *next_scr, *scr = scr_list; scr != NULL; scr = next_scr) {
        next_scr = scr->next;
	delete scr;
    }
    for (dpp = &Ctk_Window_System::head; *dpp != this; dpp = &(*dpp)->next);
    *dpp = next;
    next = NULL;		// to improve GC performance
    root_window = NULL;
    background_routines.rem_hook( &flush_display_hook, this ); 
}

Ctk_Screen* Ctk_Display::Scr( int no ) 
{ 
    Ctk_Screen * scr = scr_list;
    while( -- no >= 0 )
        if( (scr = scr->next) == NULL ) {
	    CTK_RAISE Ctk_Not_Available;
            return NULL;
    	}
    return scr;
}

Ctk_Boolean* Ctk_Display::Is_Active() 
  { return active() ? CTK_TRUE: CTK_FALSE; }


void Ctk_Display::Close() { if( Ok() ) Close(); }
void Ctk_Display::Flush() { if( Ok() ) Flush(); }
void Ctk_Display::Bell( int volume ) { if( Ok() ) Bell( volume ); }
void Ctk_Display::No_Op() { if( Ok() ) No_Op(); }
void Ctk_Display::Set_Close_Down_Mode(int mode) 
  { if( Ok() ) Set_Close_Down_Mode(mode); }
Ctk_Screen* Ctk_Display::Default_Screen() 
  { return Ok() ? Default_Screen() : NULL; } 
void Ctk_Display::Set_Selection_Owner( Ctk_String* selection,Ctk_Window* owner)
  { if( Ok() ) Set_Selection_Owner( selection, owner ); }
void Ctk_Display::UnSet_Selection_Owner( Ctk_String* selection)
  { if( Ok() ) UnSet_Selection_Owner( selection ); }
Ctk_Window* Ctk_Display::Get_Selection_Owner( Ctk_String* selection ) 
  { return Ok() ? Get_Selection_Owner(selection) : NULL; } 
void Ctk_Display::Convert_Selection( Ctk_String* selection, Ctk_String* type) 
  { if( Ok() ) Convert_Selection( selection, type ); }

local Ctk_Primitive 
wprim_a_01("Display::No_Op",
           mbr(Ctk_Display,No_Op),"windows::display"),
wprim_a_02("Display::Set_Close_Down_Mode",
	   mbr(Ctk_Display,Set_Close_Down_Mode),"windows::display"),
wprim_a_03("Display::Close",
	   mbr(Ctk_Display,Close),"windows::display"),
wprim_a_04("Display::Set_Selection_Owner",
	   mbr(Ctk_Display,Set_Selection_Owner),"windows::display"),
wprim_a_045("Display::UnSet_Selection_Owner",
	   mbr(Ctk_Display,UnSet_Selection_Owner),"windows::display"),
wprim_a_05("Display::Get_Selection_Owner",
	   mbr(Ctk_Display,Get_Selection_Owner),"windows::display"),
wprim_a_06("Display::Convert_Selection",
	   mbr(Ctk_Display,Convert_Selection),"windows::display"),
wprim_a_09("Display::Scr",
	   mbr(Ctk_Display,Scr),"windows::display"),
wprim_a_10("Display::Is_Active",
	   mbr(Ctk_Display,Is_Active),"windows::display"),
wprim_a_11("Display::Default_Screen",
	   mbr(Ctk_Display,Default_Screen),"windows::display"),
wprim_a_12("Display::Is_Active",
	   mbr(Ctk_Display,Is_Active),"windows::display"), 
wprim_a_13("Display::Flush",
	   mbr(Ctk_Display,Flush),"windows::display"), 
wprim_a_14("Display::Bell",
	   mbr(Ctk_Display,Bell),"windows::display"); 

//==========================================================================
// Ctk_Screen
//==========================================================================

Ctk_Boolean* Ctk_Winweb::Is_Active() 
{
    return active() ? CTK_TRUE : CTK_FALSE;
}

void Ctk_Winweb::Destroy() { if( Ok() ) Destroy(); } 

Ctk_Winweb::~Ctk_Winweb() {
    Ctk_Winweb** web;

    if (Scr != NULL) { 
        for( web = &Scr->web; *web != this; web = &(*web)->Next );
        *web = Next; 
	Scr = NULL;
	Next =  NULL; 
    }
}

local Ctk_Primitive wprim_a_16("Winweb::Destroy",
			       mbr(Ctk_Winweb,Destroy),"windows::winweb");


//==========================================================================

Ctk_Boolean* Ctk_Screen::Is_Active() 
{
    return active() ? CTK_TRUE : CTK_FALSE;
}

Ctk_Screen::Ctk_Screen(nat2 Desc, Ctk_Display* d) : Ctk_Virtual(Desc)
{
    dsp = d;			// Activate screen
    next = d->scr_list;		// insert self into the display screen list
    d->scr_list = this;
}

Ctk_Screen::~Ctk_Screen() {
    Ctk_Screen** spp;

    if (dsp == NULL) { 
        return;
    }
    // Link from the display screen list
    for (spp = &dsp->scr_list; *spp != this; spp = &(*spp)->next);
    *spp = next;

    Ctk_Winweb *wp, *next_wp;
    for (wp = web; wp != NULL; wp = next_wp ) 
	 next_wp = wp->Next; 
         delete wp; 

    dsp = NULL;			// disable this screen
    next = NULL;		// remove references, to improve the 
    root_window = NULL;		//  performance of conservative GC.
}

int Ctk_Screen::Black_Pixel() { return Ok() ? Black_Pixel() : 0; }
int Ctk_Screen::White_Pixel() { return Ok() ? White_Pixel() : 0; }
Ctk_Colormap* Ctk_Screen::Default_Colormap() 
  { return Ok() ? Default_Colormap() : (Ctk_Colormap*)NULL; }
int  Ctk_Screen::Default_Depth() { return Ok() ? Default_Depth() : 0; }
Ctk_Array_I4* Ctk_Screen::List_Depths() { return Ok() ? List_Depths() : NULL; }
Ctk_GC* Ctk_Screen::Default_GC()  { return Ok() ? Default_GC() : NULL; }

int Ctk_Screen::Display_Cells() { return Ok() ? Display_Cells() : 0; }
int Ctk_Screen::Height() { return Ok() ? Height() : 0; }
int Ctk_Screen::Width(){ return Ok() ? Width(): 0; }
int Ctk_Screen::Height_mm() { return Ok() ? Height_mm() : 0; }
int Ctk_Screen::Width_mm(){ return Ok() ? Width_mm(): 0; }

void Ctk_Screen::Query_Best_Tile(int width, int height, long Res[2] )
  { if( Ok() ) Query_Best_Tile(width, height, Res ); }

void Ctk_Screen::Query_Best_Stipple(int width, int height, long Res[2] )
  { if( Ok() ) Query_Best_Stipple(width, height, Res ); }

void Ctk_Screen::Query_Best_Cursor(int width, int height, long Res[2] ) 
  { if( Ok() ) Query_Best_Cursor(width, height, Res ); }

Ctk_Colormap* Ctk_Screen::Create_Colormap( Ctk_Boolean* alloc ) 
  { return Ok() ? Create_Colormap( alloc ) : NULL; }

Ctk_Cursor* Ctk_Screen::Create_Font_Cursor(int shape)
  { return Ok() ? Create_Font_Cursor(shape) : NULL; }

Ctk_Cursor* Ctk_Screen::Create_Glyph_Cursor(Ctk_Font* source_font, 
  Ctk_Font* mask_font, int src_char, int  mask_char, 
  int foreground, int background)
{ 
    return Ok() 
	? Create_Glyph_Cursor( source_font, mask_font, src_char, mask_char, 
			       foreground, background )
	: NULL;
} 

Ctk_Cursor* Ctk_Screen::Create_Pixmap_Cursor( Ctk_Pixmap* source, 
  Ctk_Pixmap* mask, int foreground, int background, int x, int y)
{
    return Ok() 
        ? Create_Pixmap_Cursor( source, mask, foreground, background, x, y )
	: NULL;
}

Ctk_Input_Window* Ctk_Screen::Create_Simple_Input_Window(
  Ctk_Input_Window* parent, int x, int y, int width, int height)
{
    if( Ok() ) {
	if( parent->Scr == this ) 
	    return Create_Simple_Input_Window( parent, x, y, width, height ); 
	CTK_RAISE Ctk_Bad_Value; 
    }
    return NULL;
}
		
Ctk_Input_Window* Ctk_Screen::Create_Input_Window( Ctk_Input_Window*
  parent, int x, int y, int width, int height, int win_gravity, int event_mask,
  int dont_propagate_mask, Ctk_Boolean* override_redirect)
{
    if( Ok() ) {
	if( parent->Scr == this ) 
	    return Create_Input_Window( parent, x, y, width, height, 
	      win_gravity, event_mask, dont_propagate_mask, override_redirect);
	CTK_RAISE Ctk_Bad_Value; 
    }
    return NULL;
}
		
Ctk_Window* Ctk_Screen::Create_Default_Window( Ctk_Window* p, int x, int y, 
int width, int height, int b_width)
{
    if( Ok() ) {
	if( p->Scr == this ) 
	    return Create_Default_Window( p, x, y, width, height, b_width );
	CTK_RAISE Ctk_Bad_Value; 
    }
    return NULL;
}



Ctk_Window* Ctk_Screen::Create_Simple_Window(
  Ctk_Window* parent, int x, int y, int width, int height, 
  int border_width, int border_pixel, int background_pixel )
{
    if( Ok() ) {
	if( parent->Scr == this ) 
	    return Create_Simple_Window( parent, x, y, width, height, 
                border_width, border_pixel, background_pixel );
	CTK_RAISE Ctk_Bad_Value; 
    }
    return NULL;
}

Ctk_Window* Ctk_Screen::Create_Window(Ctk_Window* parent, int x, int y, 
  int width, int height, int depth, int border_width, int border_pixel,
  int background_pixel, int win_gravity, int event_mask,
  int do_not_propagate_mask, Ctk_Boolean* override_redirect, 
  Ctk_Cursor* curs)
{
    if( Ok() ) {
	if( parent->Scr == this ) 
 	    return Create_Window(parent, x, y, width, height, depth, 
 	 	border_width, border_pixel, background_pixel, 
		win_gravity, event_mask, do_not_propagate_mask, 
		override_redirect, curs);
        CTK_RAISE Ctk_Bad_Value;
    }
    return NULL; 
}
	
Ctk_GC* Ctk_Screen::Create_Empty_Draw_Ctx(Ctk_Pixmap* drawable)
{
    if( Ok() ) {
	if( drawable->Scr == this )
	    return  Create_Empty_Draw_Ctx( drawable ); 
        CTK_RAISE Ctk_Bad_Value;
    }
    return NULL; 
}

Ctk_GC* Ctk_Screen::Create_Draw_Ctx(Ctk_Pixmap* drawable, 
  int foreground, int background, int funct, int plane_mask)
{
    if( Ok() ) {
	if( drawable->Scr == this )
	    return Create_Draw_Ctx( drawable, foreground, background, 
				    funct, plane_mask ); 
        CTK_RAISE Ctk_Bad_Value;
    }
    return NULL; 
}

Ctk_GC* Ctk_Screen::Create_Font_Draw_Ctx(Ctk_Pixmap* drawable, 
  int foreground, int background, int funct, int plane_mask, 
  Ctk_Font* font)
{
    if( Ok() ) {
	if( drawable->Scr == this && font->Scr == this ) 
	    return Create_Font_Draw_Ctx( drawable, foreground, background, 
					 funct, plane_mask, font ); 
        CTK_RAISE Ctk_Bad_Value;
    }
    return NULL; 
}

Ctk_Pixmap* Ctk_Screen::Create_Pixmap(int width, int height, int depth)
  { return Ok() ? Create_Pixmap( width, height, depth ) : NULL; }


// Fonts
Ctk_Font* Ctk_Screen::Load_Font(Ctk_String* name) 
  { return Ok() ? Load_Font(name) : NULL; }

Ctk_Array_O* Ctk_Screen::List_Fonts(Ctk_String* pattern, int limit)
  { return Ok() ? List_Fonts(pattern, limit) : NULL; }
  
local Ctk_Primitive
 wprim_a_50("Winweb::Is_Active", 
            mbr(Ctk_Winweb,Is_Active), "windows::winweb"), 
 wprim_a_51("Screen::Is_Active", 
            mbr(Ctk_Screen,Is_Active), "windows::screen"),
 wprim_b_00("Screen::Black_Pixel",
	    mbr(Ctk_Screen,Black_Pixel),"windows::screen"),
 wprim_b_01("Screen::White_Pixel",
	    mbr(Ctk_Screen,White_Pixel),"windows::screen"),
 wprim_b_02("Screen::Default_Colormap",
	    mbr(Ctk_Screen,Default_Colormap),"windows::screen"),
 wprim_b_03("Screen::Default_Depth",
	    mbr(Ctk_Screen,Default_Depth),"windows::screen"),
 wprim_b_04("Screen::List_Depths",
	    mbr(Ctk_Screen,List_Depths),"windows::screen"),
 wprim_b_05("Screen::Default_GC",
	    mbr(Ctk_Screen,Default_GC),"windows::screen"),
 wprim_b_07("Screen::Display_Cells",
	    mbr(Ctk_Screen,Display_Cells),"windows::screen"),
 wprim_b_08("Screen::Height",
	    mbr(Ctk_Screen,Height),"windows::screen"),
 wprim_b_09("Screen::Width",
	    mbr(Ctk_Screen,Width),"windows::screen"),
 wprim_b_10("Screen::Height_mm",
	    mbr(Ctk_Screen,Height_mm),"windows::screen"),
 wprim_b_11("Screen::Width_mm",
	    mbr(Ctk_Screen,Width_mm),"windows::screen"),
 wprim_b_12("Screen::Query_Best_Tile",
	    mbr(Ctk_Screen,Query_Best_Tile),"windows::screen"),
 wprim_b_13("Screen::Query_Best_Stipple",
	    mbr(Ctk_Screen,Query_Best_Stipple),"windows::screen"),
 wprim_b_14("Screen::Query_Best_Cursor",
	    mbr(Ctk_Screen,Query_Best_Cursor),"windows::screen"),
 wprim_b_15("Screen::Create_Font_Cursor",
	    mbr(Ctk_Screen,Create_Font_Cursor),"windows::screen"),
 wprim_b_16("Screen::Create_Glyph_Cursor",
	    mbr(Ctk_Screen,Create_Glyph_Cursor),"windows::screen"),
 wprim_b_17("Screen::Create_Pixmap_Cursor",
	    mbr(Ctk_Screen,Create_Pixmap_Cursor),"windows::screen"),
 wprim_b_18("Screen::Load_Font",
	    mbr(Ctk_Screen,Load_Font),"windows::screen"),
 wprim_b_19("Screen::List_Fonts",
	    mbr(Ctk_Screen,List_Fonts),"windows::screen"), 
 wprim_b_21("Screen::Create_Colormap", 
	    mbr(Ctk_Screen,Create_Colormap), "windows::screen" ),
 wprim_b_22("Screen::Create_Simple_Input_Window", 
	    mbr(Ctk_Screen,Create_Simple_Input_Window), "windows::screen" ),
 wprim_b_23("Screen::Create_Input_Window", 
	    mbr(Ctk_Screen,Create_Input_Window), "windows::screen" ),
 wprim_b_54("Screen::Create_Default_Window", 
	    mbr(Ctk_Screen,Create_Default_Window), "windows::screen" ),
 wprim_b_24("Screen::Create_Simple_Window", 
	    mbr(Ctk_Screen,Create_Simple_Window), "windows::screen" ),
 wprim_b_25("Screen::Create_Window", 
	    mbr(Ctk_Screen,Create_Window), "windows::screen" ),
 wprim_b_26("Screen::Create_Empty_Draw_Ctx", 
	    mbr(Ctk_Screen,Create_Empty_Draw_Ctx), "windows::screen" ),
 wprim_b_27("Screen::Create_Draw_Ctx", 
	    mbr(Ctk_Screen,Create_Draw_Ctx), "windows::screen" ),
 wprim_b_28("Screen::Create_Font_Draw_Ctx", 
	    mbr(Ctk_Screen,Create_Font_Draw_Ctx), "windows::screen" ),
 wprim_b_29("Screen::Create_Pixmap", 
	    mbr(Ctk_Screen,Create_Pixmap), "windows::screen" );
 
//-----------------------------------------------------------------------------

void   Ctk_Input_Window::Map(){ if( Ok() ) Map(); }
void   Ctk_Input_Window::Map_Raised() { if( Ok() ) Map_Raised(); }
void   Ctk_Input_Window::Map_Subwindows() { if( Ok() ) Map_Subwindows(); }
void   Ctk_Input_Window::Unmap() { if( Ok() ) Unmap(); }
void   Ctk_Input_Window::Unmap_Subwindows() { if( Ok() ) Unmap_Subwindows(); }
void   Ctk_Input_Window::Destroy_Subwindows()
  { if( Ok() ) Destroy_Subwindows(); }
void   Ctk_Input_Window::Move(int nx, int ny) { if( Ok() ) Move(nx, ny); }
void   Ctk_Input_Window::Resize_Window(int nwidth, int nheight)
  { if( Ok() ) Resize_Window(nwidth, nheight); }
void   Ctk_Input_Window::Move_Resize(int nx, int ny, int nwidth, int nheight)
  { if( Ok() ) Move_Resize(nx, ny, nwidth, nheight); }
void   Ctk_Input_Window::Origin( long Res[2] ) { if( Ok() ) Origin( Res ); }
void   Ctk_Input_Window::Size( long Res[2] ) { if( Ok() ) Size( Res ); }
void   Ctk_Input_Window::Raise() { if( Ok() ) Raise(); }
void   Ctk_Input_Window::Lower() { if( Ok() ) Lower(); }
void   Ctk_Input_Window::Circulate_Subwindows(int direction)
  { if( Ok() ) Circulate_Subwindows(direction); }
void   Ctk_Input_Window::Circulate_Subwindows_Up()
  { if( Ok() ) Circulate_Subwindows_Up(); }
void   Ctk_Input_Window::Circulate_Subwindows_Down()
  { if( Ok() ) Circulate_Subwindows_Down(); }
int Ctk_Input_Window::All_Event_Mask(){ return Ok() ? All_Event_Mask(): 0; }
int Ctk_Input_Window::Own_Event_Mask(){ return Ok() ? Own_Event_Mask(): 0; }
int Ctk_Input_Window::Not_Propagate_Mask()
  { return Ok() ? Not_Propagate_Mask(): 0; }
Ctk_Boolean* Ctk_Input_Window::Override_Redirect()
  { return Ok() ? Override_Redirect(): NULL; }
Ctk_Boolean* Ctk_Input_Window::Grab_Pointer( Ctk_Boolean* confine ) 
  { return Ok() ? Grab_Pointer( confine ) : NULL; }
void Ctk_Input_Window::Ungrab_Pointer() { if( Ok() ) Ungrab_Pointer(); }
void Ctk_Input_Window::Focus_In() { if( Ok() ) Focus_In() ; }
void Ctk_Input_Window::Focus_Out() { if( Ok() ) Focus_Out() ; }
void Ctk_Input_Window::Set_Bit_Gravity(int gravity)
  { if( Ok() ) Set_Bit_Gravity(gravity); }
void Ctk_Input_Window::Set_Win_Gravity(int gravity)
  { if( Ok() ) Set_Win_Gravity(gravity); }
int  Ctk_Input_Window::Bit_Gravity(){ return Ok() ? Bit_Gravity(): 0; }
int  Ctk_Input_Window::Win_Gravity(){ return Ok() ? Win_Gravity(): 0; }
void Ctk_Input_Window::Set_Not_Propagate_Mask(int mask) 
  { if( Ok() ) Set_Not_Propagate_Mask(mask) ; }
void Ctk_Input_Window::Set_Override_Redirect(Ctk_Boolean* redir)
  { if( Ok() ) Set_Override_Redirect(redir); }
void Ctk_Input_Window::Select_Input(int mask)
  { if( Ok() ) Select_Input(mask); }
void  Ctk_Input_Window::Put_Property(Ctk_String* name, Ctk_Object Obj)
  { if( Ok() ) Put_Property(name, Obj); }
Ctk_Object Ctk_Input_Window::Get_Property(Ctk_String* name) 
  { return Ok() ? Get_Property(name) : NULL; }
void Ctk_Input_Window::Remove_Property(Ctk_String* name) 
  { if( Ok() ) Remove_Property(name) ; }
Ctk_Array_O* Ctk_Input_Window::List_Properties()
  { return Ok() ? List_Properties(): NULL; }

local Ctk_Primitive
 wprim_c_01("Input_Window::Destroy_Subwindows",
	    mbr(Ctk_Input_Window,Destroy_Subwindows),"windows::input_window"),
 wprim_c_02("Input_Window::Map",
	    mbr(Ctk_Input_Window,Map),"windows::input_window"),
 wprim_c_03("Input_Window::Map_Raised",
	    mbr(Ctk_Input_Window,Map_Raised),"windows::input_window"),
 wprim_c_04("Input_Window::Map_Subwindows",
	    mbr(Ctk_Input_Window,Map_Subwindows),"windows::input_window"),
 wprim_c_05("Input_Window::Unmap",
	    mbr(Ctk_Input_Window,Unmap),"windows::input_window"),
 wprim_c_06("Input_Window::Unmap_Subwindows",
	    mbr(Ctk_Input_Window,Unmap_Subwindows),"windows::input_window"),
 wprim_c_07("Input_Window::Move",
	    mbr(Ctk_Input_Window,Move),"windows::input_window"),
 wprim_c_08("Input_Window::Resize",
	    mbr(Ctk_Input_Window,Resize_Window),"windows::input_window"),
 wprim_c_09("Input_Window::Move_Resize",
	    mbr(Ctk_Input_Window,Move_Resize),"windows::input_window"),
 wprim_c_12("Input_Window::Raise",
	    mbr(Ctk_Input_Window,Raise),"windows::input_window"),
 wprim_c_13("Input_Window::Lower",
	    mbr(Ctk_Input_Window,Lower),"windows::input_window"),
 wprim_c_14("Input_Window::Circulate_Subwindows",
	    mbr(Ctk_Input_Window,Circulate_Subwindows),"windows::input_window"),
 wprim_c_15("Input_Window::Circulate_Subwindows_Up",
	   mbr(Ctk_Input_Window,Circulate_Subwindows_Up),"windows::input_window"),
 wprim_c_16("Input_Window::Circulate_Subwindows_Down",
	 mbr(Ctk_Input_Window,Circulate_Subwindows_Down),"windows::input_window"),
 wprim_c_17("Input_Window::Put_Property",
	    mbr(Ctk_Input_Window,Put_Property),"windows::input_window"),
 wprim_c_18("Input_Window::Get_Property",
	    mbr(Ctk_Input_Window,Get_Property),"windows::input_window"),
 wprim_c_19("Input_Window::Remove_Property",
	    mbr(Ctk_Input_Window,Remove_Property),"windows::input_window"),
 wprim_c_20("Input_Window::List_Properties",
	    mbr(Ctk_Input_Window,List_Properties),"windows::input_window"),
 wprim_c_51("Input_Window::Size",
      mbr(Ctk_Input_Window,Size),"windows::input_window"),
 wprim_c_52("Input_Window::Origin",
      mbr(Ctk_Input_Window,Origin),"windows::input_window"),
 wprim_c_22("Input_Window::All_Event_Mask",
	    mbr(Ctk_Input_Window,All_Event_Mask),"windows::input_window"),
 wprim_c_23("Input_Window::Own_Event_Mask",
	    mbr(Ctk_Input_Window,Own_Event_Mask),"windows::input_window"),
 wprim_c_24("Input_Window::Not_Propagate_Mask",
	    mbr(Ctk_Input_Window,Not_Propagate_Mask),"windows::input_window"),
 wprim_c_25("Input_Window::Override_Redirect",
	    mbr(Ctk_Input_Window,Override_Redirect),"windows::input_window"),
 wprim_c_35("Input_Window::Grab_Pointer",
	    mbr(Ctk_Input_Window,Grab_Pointer),"windows::input_window"),
 wprim_c_36("Input_Window::Ungrab_Pointer",
	    mbr(Ctk_Input_Window,Ungrab_Pointer),"windows::input_window"),
 wprim_c_37("Input_Window::Focus_In",
	    mbr(Ctk_Input_Window,Focus_In),"windows::input_window"),
 wprim_c_38("Input_Window::Focus_Out",
	    mbr(Ctk_Input_Window,Focus_Out),"windows::input_window"),
 wprim_c_26("Input_Window::Set_Not_Propagate_Mask",
	    mbr(Ctk_Input_Window,Set_Not_Propagate_Mask),"windows::input_window"),
 wprim_c_27("Input_Window::Set_Override_Redirect",
	    mbr(Ctk_Input_Window,Set_Override_Redirect),"windows::input_window"),
 wprim_c_28("Input_Window::Not_Propagate_Mask",
	    mbr(Ctk_Input_Window,Not_Propagate_Mask),"windows::input_window"),
 wprim_c_29("Input_Window::Select_Input",
	    mbr(Ctk_Input_Window,Select_Input),"windows::input_window"),
 wprim_c_31("Input_Window::Bit_Gravity",
	    mbr(Ctk_Input_Window,Bit_Gravity),"windows::input_window"),
 wprim_c_32("Input_Window::Win_Gravity",
	    mbr(Ctk_Input_Window,Win_Gravity),"windows::input_window"),
 wprim_c_33("Input_Window::Set_Bit_Gravity",
	    mbr(Ctk_Input_Window,Set_Bit_Gravity),"windows::input_window"),
 wprim_c_34("Input_Window::Set_Win_Gravity",
	    mbr(Ctk_Input_Window,Set_Win_Gravity),"windows::input_window");

//--------------------------------------------------------------------------
// Input/Output X window
//--------------------------------------------------------------------------

void   Ctk_Window::Clear() { if( Ok() ) Clear(); }

void   Ctk_Window::Clear_Area(int x, int y, int width, int height, 
  Ctk_Boolean* exposures)
{
    if( Ok() ) Clear_Area( x, y, width, height, exposures );
}

void Ctk_Window::Configure(int nx, int ny, int nwidth, int nheight, 
			     int nborder_width)
{
    if( Ok() ) 
	Configure( nx, ny, nwidth, nheight, nborder_width ); 
}

Ctk_Colormap*	Ctk_Window::Get_Colormap()
  { return Ok() ? Get_Colormap(): NULL; }

int Ctk_Window::Depth() { return Depth(); }

void Ctk_Window::Set_Colormap(Ctk_Colormap* cmap)
{
    if( Ok() )
	if( cmap->Scr == Scr ) Set_Colormap(cmap);
	else CTK_RAISE Ctk_Bad_Value; 
}

void Ctk_Window::Set_WM_Properties( Ctk_String* Win_Name, 
  Ctk_String* Icon_Name, Ctk_Size_Hints* hints )
{
    if( Ok() ) Set_WM_Properties( Win_Name, Icon_Name, hints ); 
}

void  Ctk_Window::Set_Background_Pixel(int background)
  { if( Ok() ) Set_Background_Pixel(background); }
void  Ctk_Window::Set_Background_Pixmap(Ctk_Pixmap* pxm)
  { if( Ok() ) Set_Background_Pixmap(pxm); }
void  Ctk_Window::Set_Border_Pixel(int pixel)
  { if( Ok() ) Set_Border_Pixel(pixel); }
void  Ctk_Window::Set_Border_Pixmap(Ctk_Pixmap* pxm)
  { if( Ok() ) Set_Border_Pixmap(pxm); }
void   Ctk_Window::Set_Border_Width(int width)
  { if( Ok() ) Set_Border_Width(width); }
void Ctk_Window::Define_Cursor(Ctk_Cursor* cursor)
  { if( Ok() ) Define_Cursor(cursor); }
void Ctk_Window::Set_Backing_Store(int backing_store)
  { if( Ok() ) Set_Backing_Store(backing_store); } 
int  Ctk_Window::Get_Backing_Store()
  { return Ok() ? Get_Backing_Store() : 0; } 

void Ctk_Window::Undefine_Cursor()
  { if( Ok() ) Undefine_Cursor(); }

local Ctk_Primitive
 wprim_w_00("Window::Clear",
	    mbr(Ctk_Window,Clear),"windows::window"),
 wprim_w_01("Window::Clear_Area",
	    mbr(Ctk_Window,Clear_Area),"windows::window"),
 wprim_w_10("Window::Set_Border_Width",
	    mbr(Ctk_Window,Set_Border_Width),"windows::window"),
 wprim_w_11("Window::Configure",
	    mbr(Ctk_Window,Configure),"windows::window"),
 wprim_w_12("Window::Set_Background_Pixel",
	    mbr(Ctk_Window,Set_Background_Pixel),"windows::window"),
 wprim_w_13("Window::Set_Background_Pixmap",
	    mbr(Ctk_Window,Set_Background_Pixmap),"windows::window"),
 wprim_w_14("Window::Set_Border_Pixel",
	    mbr(Ctk_Window,Set_Border_Pixel),"windows::window"),
 wprim_w_15("Window::Set_Border_Pixmap",
	    mbr(Ctk_Window,Set_Border_Pixmap),"windows::window"),
 wprim_w_20("Window::Get_Colormap",
	    mbr(Ctk_Window,Get_Colormap),"windows::window"),
 wprim_w_21("Window::Set_Colormap",
	    mbr(Ctk_Window,Set_Colormap),"windows::window"),
 wprim_w_31("Window::Set_WM_Properties",
	    mbr(Ctk_Window,Set_WM_Properties),"windows::window"),
 wprim_w_22("Window::Define_Cursor",
	    mbr(Ctk_Window,Define_Cursor),"windows::window"),
 wprim_w_23("Window::Undefine_Cursor",
	    mbr(Ctk_Window,Undefine_Cursor),"windows::window"),
 wprim_w_24("Window::Set_Backing_Store",
	    mbr(Ctk_Window,Set_Backing_Store),"windows::window"),
 wprim_w_25("Window::Get_Backing_Store",
	    mbr(Ctk_Window,Get_Backing_Store),"windows::window"),
 wprim_w_26("Window::Depth",mbr(Ctk_Window,Depth),"windows::window");

//-----------------------------------------------------------------------------

Ctk_GC* Ctk_GC::Copy() { return Ok() ? Copy() : NULL; }
Ctk_Font*     Ctk_GC::Fnt()
  { return Ok() ? Fnt(): NULL; }
int Ctk_GC::Background(){ return Ok() ? Background(): 0; }
int Ctk_GC::Foreground(){ return Ok() ? Foreground(): 0; }
int Ctk_GC::Drawing_Function(){ return Ok() ? Drawing_Function(): 0; }
int Ctk_GC::Plane_Mask(){ return Ok() ? Plane_Mask(): 0; }
int Ctk_GC::Line_Width(){ return Ok() ? Line_Width(): 0; }
int Ctk_GC::Line_Style(){ return Ok() ? Line_Style(): 0; }
int Ctk_GC::Cap_Style(){ return Ok() ? Cap_Style(): 0; }
int Ctk_GC::Join_Style(){ return Ok() ? Join_Style(): 0; }
int Ctk_GC::Fill_Style(){ return Ok() ? Fill_Style(): 0; }
int Ctk_GC::Fill_Rule(){ return Ok() ? Fill_Rule(): 0; }
int Ctk_GC::Arc_Mode(){ return Ok() ? Arc_Mode(): 0; }
Ctk_Pixmap* Ctk_GC::Tile() { return Ok() ? Tile(): NULL; }
Ctk_Pixmap* Ctk_GC::Stipple() { return Ok() ? Stipple(): NULL; }
int Ctk_GC::TS_X_Origin(){ return Ok() ? TS_X_Origin(): 0; }
int Ctk_GC::TS_Y_Origin(){ return Ok() ? TS_Y_Origin(): 0; }
int Ctk_GC::Subwindow_Mode(){ return Ok() ? Subwindow_Mode(): 0; }
Ctk_Boolean*	Ctk_GC::Graphics_Exposures()
  { return Ok() ? Graphics_Exposures(): NULL; }
int Ctk_GC::Clip_X_Origin(){ return Ok() ? Clip_X_Origin(): 0; }
int Ctk_GC::Clip_Y_Origin(){ return Ok() ? Clip_Y_Origin(): 0; }


void Ctk_GC::Set_Fnt(Ctk_Font* Fnt)
{
    if( Ok() ) {
	if( Fnt->Scr == Scr ) 
	    Set_Fnt(  Fnt );
        else
	    CTK_RAISE Ctk_Bad_Value;
    }
}

void Ctk_GC::Set_Foreground(int foregr) { if( Ok() ) Set_Foreground(foregr); }
void Ctk_GC::Set_Background(int backgr) { if( Ok() ) Set_Background(backgr); }
void Ctk_GC::Set_Drawing_Function(int g_func)
  { if( Ok() ) Set_Drawing_Function(g_func); }
void Ctk_GC::Set_Plane_Mask(int plane_mask)
  { if( Ok() ) Set_Plane_Mask(plane_mask); }
void Ctk_GC::Set_Line_Width(int width) 
  { if( Ok() ) Set_Line_Width(width); }
void Ctk_GC::Set_Line_Style(int line_style) 
  { if( Ok() ) Set_Line_Style(line_style) ; }
void Ctk_GC::Set_Cap_Style(int cap_style)
  { if( Ok() ) Set_Cap_Style(cap_style); }
void Ctk_GC::Set_Join_Style(int join_style)
  { if( Ok() ) Set_Join_Style(join_style); }
void Ctk_GC::Set_Fill_Style(int fill_style)
  { if( Ok() ) Set_Fill_Style(fill_style); }
void Ctk_GC::Set_Fill_Rule(int fill_rule)
  { if( Ok() ) Set_Fill_Rule(fill_rule); }
void Ctk_GC::Set_Graphics_Exposures(Ctk_Boolean* graphics_exposures)
  { if( Ok() ) Set_Graphics_Exposures(graphics_exposures); }
void Ctk_GC::Set_Subwindow_Mode(int subwindow_mode)
  { if( Ok() ) Set_Subwindow_Mode(subwindow_mode); }
void Ctk_GC::Set_Arc_Mode(int arc_mode)
  { if( Ok() ) Set_Arc_Mode(arc_mode); }
void Ctk_GC::Set_Tile(Ctk_Pixmap* Tile)
{
    if( Ok() ) 
        if( Tile->Scr == Scr ) 
            Set_Tile( Tile );    
    	else 
	    CTK_RAISE Ctk_Bad_Value;
}

void Ctk_GC::Set_Stipple(Ctk_Pixmap* Stipple)
{
    if( Ok() ) 
        if( Stipple->Scr == Scr ) 
            Set_Stipple( Stipple );    
    	else 
	    CTK_RAISE Ctk_Bad_Value;
}

void Ctk_GC::Set_Clip_X_Origin(int x) { if( Ok() ) Set_Clip_X_Origin(x); }
void Ctk_GC::Set_Clip_Y_Origin(int y) { if( Ok() ) Set_Clip_Y_Origin(y); }
void Ctk_GC::Set_TS_X_Origin(int x)  { if( Ok() ) Set_TS_X_Origin(x); }
void Ctk_GC::Set_TS_Y_Origin(int y) { if( Ok() ) Set_TS_Y_Origin(y); }
void Ctk_GC::Set_Clip_Mask(Ctk_Pixmap* Msk) 
{ 
    if( Ok() ) 
	if( Msk->Scr == Scr ) Set_Clip_Mask(Msk);
        else CTK_RAISE Ctk_Bad_Value; 
}
void Ctk_GC::Set_Clip_Rectangle(Ctk_Box* Box)
  { if( Ok() ) Set_Clip_Rectangle(Box); }
void Ctk_GC::Set_Clip_Rectangles(Ctk_Array_O* Boxes)
  { if( Ok() ) Set_Clip_Rectangles(Boxes); }
void Ctk_GC::Set_Line_Attributes(int line_width, int line_style,
  int cap_style, int join_style)
{
    if( Ok() ) 
        Set_Line_Attributes( line_width, line_style, cap_style, join_style );
}

void Ctk_GC::Set_Dashes(Ctk_Array_I4* D) { if( Ok() ) Set_Dashes(D); }

local Ctk_Primitive 
 wprim_d_01("GC::Fnt",
	    mbr(Ctk_GC,Fnt),"windows::draw_ctx"),
 wprim_d_02("GC::Background",
	    mbr(Ctk_GC,Background),"windows::draw_ctx"),
 wprim_d_03("GC::Foreground",
	    mbr(Ctk_GC,Foreground),"windows::draw_ctx"),
 wprim_d_04("GC::Drawing_Function",
	    mbr(Ctk_GC,Drawing_Function),"windows::draw_ctx"),
 wprim_d_05("GC::Plane_Mask",
	    mbr(Ctk_GC,Plane_Mask),"windows::draw_ctx"),
 wprim_d_06("GC::Line_Width",
	    mbr(Ctk_GC,Line_Width),"windows::draw_ctx"),
 wprim_d_07("GC::Line_Style",
	    mbr(Ctk_GC,Line_Style),"windows::draw_ctx"),
 wprim_d_08("GC::Cap_Style",
	    mbr(Ctk_GC,Cap_Style),"windows::draw_ctx"),
 wprim_d_09("GC::Join_Style",
	    mbr(Ctk_GC,Join_Style),"windows::draw_ctx"),
 wprim_d_10("GC::Fill_Style",
	    mbr(Ctk_GC,Fill_Style),"windows::draw_ctx"),
 wprim_d_11("GC::Fill_Rule",
	    mbr(Ctk_GC,Fill_Rule),"windows::draw_ctx"),
 wprim_d_12("GC::Arc_Mode",
	    mbr(Ctk_GC,Arc_Mode),"windows::draw_ctx"),
 wprim_d_13("GC::Tile",
	    mbr(Ctk_GC,Tile),"windows::draw_ctx"),
 wprim_d_14("GC::Stipple",
	    mbr(Ctk_GC,Stipple),"windows::draw_ctx"),
 wprim_d_15("GC::TS_X_Origin",
	    mbr(Ctk_GC,TS_X_Origin),"windows::draw_ctx"),
 wprim_d_16("GC::TS_Y_Origin",
	    mbr(Ctk_GC,TS_Y_Origin),"windows::draw_ctx"),
 wprim_d_17("GC::Subwindow_Mode",
	    mbr(Ctk_GC,Subwindow_Mode),"windows::draw_ctx"),
 wprim_d_18("GC::Graphics_Exposures",
	    mbr(Ctk_GC,Graphics_Exposures),"windows::draw_ctx"),
 wprim_d_19("GC::Clip_X_Origin",
	    mbr(Ctk_GC,Clip_X_Origin),"windows::draw_ctx"),
 wprim_d_20("GC::Clip_Y_Origin",
	    mbr(Ctk_GC,Clip_Y_Origin),"windows::draw_ctx"),
 wprim_d_26("GC::Set_Fnt",
	    mbr(Ctk_GC,Set_Fnt),"windows::draw_ctx"),
 wprim_d_27("GC::Set_Background",
	    mbr(Ctk_GC,Set_Background),"windows::draw_ctx"),
 wprim_d_28("GC::Set_Foreground",
	    mbr(Ctk_GC,Set_Foreground),"windows::draw_ctx"),
 wprim_d_29("GC::Set_Drawing_Function",
	    mbr(Ctk_GC,Set_Drawing_Function),"windows::draw_ctx"),
 wprim_d_30("GC::Set_Plane_Mask",
	    mbr(Ctk_GC,Set_Plane_Mask),"windows::draw_ctx"),
 wprim_d_31("GC::Set_Line_Width",
	    mbr(Ctk_GC,Set_Line_Width),"windows::draw_ctx"),
 wprim_d_32("GC::Set_Line_Style",
	    mbr(Ctk_GC,Set_Line_Style),"windows::draw_ctx"),
 wprim_d_33("GC::Set_Cap_Style",
	    mbr(Ctk_GC,Set_Cap_Style),"windows::draw_ctx"),
 wprim_d_34("GC::Set_Join_Style",
	    mbr(Ctk_GC,Set_Join_Style),"windows::draw_ctx"),
 wprim_d_35("GC::Set_Fill_Style",
	    mbr(Ctk_GC,Set_Fill_Style),"windows::draw_ctx"),
 wprim_d_36("GC::Set_Fill_Rule",
	    mbr(Ctk_GC,Set_Fill_Rule),"windows::draw_ctx"),
 wprim_d_37("GC::Set_Arc_Mode",
	    mbr(Ctk_GC,Set_Arc_Mode),"windows::draw_ctx"),
 wprim_d_38("GC::Set_Tile",
	    mbr(Ctk_GC,Set_Tile),"windows::draw_ctx"),
 wprim_d_39("GC::Set_Stipple",
	    mbr(Ctk_GC,Set_Stipple),"windows::draw_ctx"),
 wprim_d_40("GC::Set_TS_X_Origin",
	    mbr(Ctk_GC,Set_TS_X_Origin),"windows::draw_ctx"),
 wprim_d_41("GC::Set_TS_Y_Origin",
	    mbr(Ctk_GC,Set_TS_Y_Origin),"windows::draw_ctx"),
 wprim_d_42("GC::Set_Subwindow_Mode",
	    mbr(Ctk_GC,Set_Subwindow_Mode),"windows::draw_ctx"),
 wprim_d_43("GC::Set_Graphics_Exposures",
	    mbr(Ctk_GC,Set_Graphics_Exposures),"windows::draw_ctx"),
 wprim_d_44("GC::Set_Clip_X_Origin",
	    mbr(Ctk_GC,Set_Clip_X_Origin),"windows::draw_ctx"),
 wprim_d_45("GC::Set_Clip_Y_Origin",
	    mbr(Ctk_GC,Set_Clip_Y_Origin),"windows::draw_ctx"),
 wprim_d_46("GC::Set_Clip_Mask",
	    mbr(Ctk_GC,Set_Clip_Mask),"windows::draw_ctx"),
 wprim_d_47("GC::Set_Clip_Rectagle",
	    mbr(Ctk_GC,Set_Clip_Rectangle),"windows::draw_ctx"),
 wprim_d_48("GC::Set_Clip_Rectangles",
	    mbr(Ctk_GC,Set_Clip_Rectangles),"windows::draw_ctx"),
 wprim_d_50("GC::Set_Dashes",
	    mbr(Ctk_GC,Set_Dashes),"windows::draw_ctx"),
 wprim_d_52("GC::Set_Line_Attributes",
	    mbr(Ctk_GC,Set_Line_Attributes),"windows::draw_ctx"),
 wprim_d_54("GC::Copy",
	    mbr(Ctk_GC,Copy),"windows::draw_ctx");

//-----------------------------------------------------------------------------

void Ctk_Cursor::Recolor(int foreground, int background)
  { if( Ok() ) Recolor(foreground, background); }

local Ctk_Primitive
 wprim_e_01("Cursor::Recolor",
	    mbr(Ctk_Cursor,Recolor),"windows::cursor");

//-----------------------------------------------------------------------------

Ctk_Colormap* Ctk_Colormap::Copy_And_Free() 
  { return Ok() ? Copy_And_Free() : NULL; }
int Ctk_Colormap::Alloc_Color(Ctk_Color* c)
  { return Ok() ? Alloc_Color(c) : 0; }
Ctk_Array_I4* Ctk_Colormap::Alloc_Colors( Ctk_Array_O* colors ) 
  { return Ok() ? Alloc_Colors( colors ) : NULL; }
Ctk_Array_I4* Ctk_Colormap::Alloc_Color_Cells( int n_colors ) 
  { return Ok() ? Alloc_Color_Cells( n_colors ) : NULL; }
void Ctk_Colormap::Alloc_Color_Planes( Ctk_Array_I4* pixels, 
  Ctk_Array_I4* planes, Ctk_Boolean* contig )
{
    if( Ok() ) Alloc_Color_Planes( pixels, planes, contig ); 
}
void Ctk_Colormap::Free_Colors(Ctk_Array_I4* colors, int planes)
{
    if( Ok() ) Free_Colors( colors, planes ); 
}
void Ctk_Colormap::Store_Color(int pixel, Ctk_Color* color)
  { if( Ok() ) Store_Color(pixel,color); }
void Ctk_Colormap::Store_Colors(Ctk_Array_I4* pixels, Ctk_Array_O* colors)
  { if( Ok() ) Store_Colors(pixels,colors); }

local Ctk_Primitive
 wprim_f_01("Colormap::Copy_And_Free",
	    mbr(Ctk_Colormap,Copy_And_Free),"windows::colormap"),
 wprim_f_09("Colormap::Alloc_Color",
	    mbr(Ctk_Colormap,Alloc_Color),"windows::colormap"),
 wprim_f_10("Colormap::Alloc_Colors",
	    mbr(Ctk_Colormap,Alloc_Colors),"windows::colormap"),
 wprim_f_11("Colormap::Alloc_Color_Cells",
	    mbr(Ctk_Colormap,Alloc_Color_Cells),"windows::colormap"),
 wprim_f_12("Colormap::Alloc_Color_Planes",
	    mbr(Ctk_Colormap,Alloc_Color_Planes),"windows::colormap"),
 wprim_f_17("Colormap::Free_Colors",
	    mbr(Ctk_Colormap,Free_Colors),"windows::colormap"),
 wprim_f_18("Colormap::Store_Color",
	    mbr(Ctk_Colormap,Store_Color),"windows::colormap"),
 wprim_f_19("Colormap::Store_Colors",
	    mbr(Ctk_Colormap,Store_Colors),"windows::colormap"); 


//-----------------------------------------------------------------------------

Ctk_Object Ctk_Font_Metrics::Get_Property( Ctk_String* Atom ) 
  { return Get_Property( Atom ); }

int Ctk_Font_Metrics::Width_String( Ctk_String* Str )
  { return Width_String( Str ); }

int Ctk_Font_Metrics::Width_String16( Ctk_Inter_String* Str )
  { return Width_String16( Str ); }

void Ctk_Font_Metrics::Extents_String( Ctk_String* Str, long Res[2] )
  { Extents_String( Str, Res ); }

void Ctk_Font_Metrics::Extents_String16( Ctk_Inter_String* Str, long Res[2])
  { Extents_String16( Str, Res ); }

void Ctk_Font_Metrics::Extents( long Res[2]) { Extents( Res ); }

local Ctk_Primitive
 wprim_fm_01("Font_Metrics::Width_String",
	     mbr(Ctk_Font_Metrics,Width_String),"windows::font_metrics"),
 wprim_fm_02("Font_Metrics::Width_String16",
	     mbr(Ctk_Font_Metrics,Width_String16),"windows::font_metrics"),
 wprim_fm_03("Font_Metrics::Extents_String",
	     mbr(Ctk_Font_Metrics,Extents_String),"windows::font_metrics"),
 wprim_fm_04("Font_Metrics::Extents_String16",
	     mbr(Ctk_Font_Metrics,Extents_String16),"windows::font_metrics"),
 wprim_fm_05("Font_Metrics::Extents",
	     mbr(Ctk_Font_Metrics,Extents),"windows::font_metrics"),
 wprim_fm_06("Font_Metrics::Get_Property",
	     mbr(Ctk_Font_Metrics,Get_Property),"windows::font_metrics");


//-----------------------------------------------------------------------------

Ctk_Font_Metrics* Ctk_Font::Metrics() { return Ok() ? Metrics() : NULL; }

local Ctk_Primitive
    wprim_g_02("Font::Metrics",mbr(Ctk_Font,Metrics),"windows::font");


//-----------------------------------------------------------------------------

int  Ctk_Pix_Image::Width() { return Width(); }

int  Ctk_Pix_Image::Height() { return Height(); }

int  Ctk_Pix_Image::Depth() { return Depth(); }


local Ctk_Primitive
pi2( "Pix_Image::Width", mbr(Ctk_Pix_Image,Width), "windows::pix_image" ),
pi3( "Pix_Image::Height", mbr(Ctk_Pix_Image,Height), "windows::pix_image" ),
pi4( "Pix_Image::Depth", mbr(Ctk_Pix_Image,Depth), "windows::pix_image" );

//-----------------------------------------------------------------------------

void Ctk_Pixmap::Copy_Area_From( Ctk_GC* ctx, int x, int y, int width, 
  int height, Ctk_Pixmap* drawable, int dst_x, int dst_y)
{
    if( Ok() && Ctx_Ok(ctx) )
	if( drawable->Scr == Scr )
            Copy_Area_From( ctx, x, y, width, height, drawable, dst_x, dst_y);
    	else 
	    CTK_RAISE Ctk_Bad_Value;
}
 
void   Ctk_Pixmap::Copy_Area(Ctk_GC* ctx, int x, int y, int width, 
				 int height, int dst_x, int dst_y)
{
    if( Ok() && Ctx_Ok(ctx) )
	Copy_Area( ctx, x, y, width, height, dst_x, dst_y);  
}


void   Ctk_Pixmap::Copy_Plane_From( Ctk_GC* ctx, int x, int y, int width, 
  int height, Ctk_Pixmap* drawable, int dst_x, int dst_y, int plane)
{
    if( Ok() && Ctx_Ok(ctx) )
	if( drawable->Scr == Scr )
            Copy_Plane_From( ctx, x, y, width, height, drawable,
			     dst_x, dst_y, plane);
    	else 
	    CTK_RAISE Ctk_Bad_Value;
}
 
void   Ctk_Pixmap::Copy_Plane(Ctk_GC* ctx, int x, int y, int width, 
				 int height, int dst_x, int dst_y, int plane)
{
    if( Ok() && Ctx_Ok(ctx) )
	Copy_Plane(ctx, x, y, width, height, dst_x, dst_y, plane);  
}

void   Ctk_Pixmap::Draw_Point(Ctk_GC* ctx, int x, int y) {
    if( Ok() && Ctx_Ok(ctx) ) Draw_Point( ctx, x, y);
}

void   Ctk_Pixmap::Draw_Points(Ctk_GC* ctx, Ctk_Array_O* a, 
				  Ctk_Boolean* rel)
{
    if( Ok() && Ctx_Ok(ctx) ) Draw_Points(ctx, a, rel ); 
}

void   Ctk_Pixmap::Draw_Line(Ctk_GC* ctx, int x_beg, int y_beg, 
				 int x_end, int y_end)
{
    if( Ok() && Ctx_Ok(ctx) )	Draw_Line(ctx, x_beg, y_beg, x_end, y_end);
}

void   Ctk_Pixmap::Draw_Lines(Ctk_GC* ctx, Ctk_Array_O* a, Ctk_Boolean* rel)
  { if( Ok() && Ctx_Ok(ctx) ) Draw_Lines( ctx, a, rel );  }

void   Ctk_Pixmap::Draw_Segments(Ctk_GC* ctx, Ctk_Array_O* a)
  { if( Ok() && Ctx_Ok(ctx) ) Draw_Segments( ctx, a ); }

void   Ctk_Pixmap::Draw_Rectangle(Ctk_GC* ctx, int x, int y, int width, 
				      int height)
{
    if( Ok() && Ctx_Ok(ctx) ) Draw_Rectangle( ctx, x, y, width, height );
}

void   Ctk_Pixmap::Draw_Rectangles(Ctk_GC* ctx, Ctk_Array_O* a)
  { if( Ok() && Ctx_Ok(ctx) ) Draw_Rectangles( ctx, a ); }

void   Ctk_Pixmap::Draw_Arc(Ctk_GC* ctx, int x, int y, int width, 
				int height, int angle1, int angle2)
{
    if( Ok() && Ctx_Ok(ctx) )
	Draw_Arc( ctx, x, y, width, height, angle1, angle2);
}

void   Ctk_Pixmap::Draw_Arcs(Ctk_GC* ctx, Ctk_Array_O* a)
  { if( Ok() && Ctx_Ok(ctx) ) Draw_Arcs( ctx, a ); }

void   Ctk_Pixmap::Fill_Rectangle(Ctk_GC* ctx, int x, int y, int width, 
 				  int height)
{
    if( Ok() && Ctx_Ok(ctx) ) Fill_Rectangle( ctx, x, y, width, height);
}

void   Ctk_Pixmap::Fill_Rectangles(Ctk_GC* ctx, Ctk_Array_O* a)
  { if( Ok() && Ctx_Ok(ctx) ) Fill_Rectangles( ctx, a ); }

void   Ctk_Pixmap::Fill_Polygon(Ctk_GC* ctx, Ctk_Array_O* a,
				    int shape, Ctk_Boolean* rel )
{
    if( Ok() && Ctx_Ok(ctx) ) Fill_Polygon( ctx, a, shape, rel ); 
}

void   Ctk_Pixmap::Fill_Arc(Ctk_GC* ctx, int x, int y, int width,
				int height, int angle1, int angle2)
{
    if( Ok() && Ctx_Ok(ctx) ) 
	Fill_Arc( ctx, x, y, width, height, angle1, angle2 );
}

void   Ctk_Pixmap::Fill_Arcs(Ctk_GC* ctx, Ctk_Array_O* a)
  { if( Ok() && Ctx_Ok(ctx) ) Fill_Arcs( ctx, a ); }


void   Ctk_Pixmap::Put_String(Ctk_GC* ctx,int x,int y,Ctk_String* Str)
  { if( Ok() && Ctx_Ok(ctx) ) Put_String( ctx, x, y, Str ); }

void Ctk_Pixmap::Put_String16(Ctk_GC* ctx, int x, int y, Ctk_Inter_String* Str)
  { if( Ok() && Ctx_Ok(ctx) ) Put_String16( ctx, x, y, Str ); }


void   Ctk_Pixmap::Put_Image_String(Ctk_GC* ctx,int x,int y, Ctk_String* Str)
  { if( Ok() && Ctx_Ok(ctx) ) Put_Image_String( ctx, x, y, Str ); }

void   Ctk_Pixmap::Put_Image_String16(Ctk_GC* ctx, int x, int y, 
					  Ctk_Inter_String* Str)
{ 
    if( Ok() && Ctx_Ok(ctx) ) Put_Image_String16( ctx, x, y, Str ); 
}

void   Ctk_Pixmap::Put_Image(Ctk_GC* ctx, Ctk_Pix_Image* xpix, 
    int src_x, int src_y, int dst_x, int dst_y,  int width, int height)
{
    if( Ok() && Ctx_Ok(ctx) ) 
	Put_Image( ctx, xpix, src_x, src_y, dst_x, dst_y, width, height );

}

Ctk_Pix_Image* Ctk_Pixmap::Get_Image(int x, int y, int width,int height,
					   int planes)
{
    return Ok() ? Get_Image( x, y, width, height, planes ) : NULL; 
} 

Ctk_Screen* Ctk_Pixmap::Get_Scr() { return Scr; } 

local Ctk_Primitive
 wprim_h_02("Drawable::Copy_Area_From",
	    mbr(Ctk_Pixmap,Copy_Area_From),"windows::drawable"),
 wprim_h_03("Drawable::Copy_Area",
	    mbr(Ctk_Pixmap,Copy_Area),"windows::drawable"),
 wprim_h_04("Drawable::Copy_Plane_From",
	    mbr(Ctk_Pixmap,Copy_Plane_From),"windows::drawable"),
 wprim_h_05("Drawable::Copy_Plane",
	    mbr(Ctk_Pixmap,Copy_Plane),"windows::drawable"),
 wprim_h_06("Drawable::Draw_Point",
	    mbr(Ctk_Pixmap,Draw_Point),"windows::drawable"),
 wprim_h_07("Drawable::Draw_Points",
	    mbr(Ctk_Pixmap,Draw_Points),"windows::drawable"),
 wprim_h_08("Drawable::Draw_Line",
	    mbr(Ctk_Pixmap,Draw_Line),"windows::drawable"),
 wprim_h_09("Drawable::Draw_Lines",
	    mbr(Ctk_Pixmap,Draw_Lines),"windows::drawable"),
 wprim_h_10("Drawable::Draw_Segments",
	    mbr(Ctk_Pixmap,Draw_Segments),"windows::drawable"),
 wprim_h_11("Drawable::Draw_Rectangle",
	    mbr(Ctk_Pixmap,Draw_Rectangle),"windows::drawable"),
 wprim_h_12("Drawable::Draw_Rectangles",
	    mbr(Ctk_Pixmap,Draw_Rectangles),"windows::drawable"),
 wprim_h_13("Drawable::Draw_Arc",
	    mbr(Ctk_Pixmap,Draw_Arc),"windows::drawable"),
 wprim_h_14("Drawable::Draw_Arcs",
	    mbr(Ctk_Pixmap,Draw_Arcs),"windows::drawable"),
 wprim_h_15("Drawable::Fill_Rectangle",
	    mbr(Ctk_Pixmap,Fill_Rectangle),"windows::drawable"),
 wprim_h_16("Drawable::Fill_Rectangles",
	    mbr(Ctk_Pixmap,Fill_Rectangles),"windows::drawable"),
 wprim_h_17("Drawable::Fill_Polygon",
	    mbr(Ctk_Pixmap,Fill_Polygon),"windows::drawable"),
 wprim_h_18("Drawable::Fill_Arc",
	    mbr(Ctk_Pixmap,Fill_Arc),"windows::drawable"),
 wprim_h_19("Drawable::Fill_Arcs",
	    mbr(Ctk_Pixmap,Fill_Arcs),"windows::drawable"),
 wprim_h_20("Drawable::Put_Image",
	    mbr(Ctk_Pixmap,Put_Image),"windows::drawable"),
 wprim_h_21("Drawable::Get_Image",
	    mbr(Ctk_Pixmap,Get_Image),"windows::drawable"),
 wprim_h_23("Drawable::Put_String",
	    mbr(Ctk_Pixmap,Put_String),"windows::drawable"),
 wprim_h_24("Drawable::Put_String16",
	    mbr(Ctk_Pixmap,Put_String16),"windows::drawable"),
 wprim_h_25("Drawable::Put_Image_String",
	    mbr(Ctk_Pixmap,Put_Image_String),"windows::drawable"),
 wprim_h_26("Drawable::Put_Image_String16",
	    mbr(Ctk_Pixmap,Put_Image_String16),"windows::drawable"),
 wprim_h_27("Drawable::Scr",
	    mbr(Ctk_Pixmap,Get_Scr),"windows::drawable");

//=============================================================================

void Ctk_Keyboard::Change_Control( int key_click_percent, int bell_percent,
  int bell_duration, Ctk_Boolean* autorepeat )
{
    if( Ok() ) 
	Change_Control( key_click_percent, bell_percent, bell_duration, 
			autorepeat ); 
}
 
local Ctk_Primitive 
  wprim_k_00("Keyboard::Change_Control",
	     mbr(Ctk_Keyboard,Change_Control),"windows::keyboard");

// the last statement in the file

