#ifndef __CTWIN_HPP__
#define __CTWIN_HPP__

// Forward declarations
class Ctk_Display;
class Ctk_Screen;
class Ctk_Event;
class Ctk_Drawable;
class Ctk_Window;
class Ctk_Keyboard;
class Ctk_Mouse;
class Ctk_Font;
class Ctk_Color;
class Ctk_Colormap;
class Ctk_Input_Window;
class Ctk_Window;
class Ctk_Pixmap;
class Ctk_Cursor;
class Ctk_GC;

//-----------------------------------------------------------------------------

struct Ctk_Box : Ctk_Header {
    int 	x, y;
    int 	width, height; 

    Ctk_Box* 	New( int x, int y, int width, int height ); 
    void     	Coord( long Res[4] ); 
    Ctk_Boolean* Contain( int x, int y ); 

  private: 
    Ctk_Box();
};

struct Ctk_Point : Ctk_Header { 
    int x, y;
    Ctk_Point* New( int x, int y ); 
  private:
    Ctk_Point() : Ctk_Header(0,No_Ptr){} 
};

struct Ctk_Arc : Ctk_Header {
    int 	x, y;
    int		width, height;
    int 	angle1, angle2;

    Ctk_Arc* 	New( int x, int y, int width, int height, int angle1, int angle2 ); 
  private:
    Ctk_Arc() : Ctk_Header(0,No_Ptr){} 
};    

struct Ctk_Segment : Ctk_Header { 
    int 	x1, y1;
    int 	x2, y2;

    Ctk_Segment* New( int x1, int y1, int x2, int y2 ); 
  private:
    Ctk_Segment() : Ctk_Header(0,No_Ptr){} 
};

struct Ctk_Color : Ctk_Header {
    int4 	r, g, b; 
    Ctk_String* name;
  private:
    Ctk_Color() : Ctk_Header(0,No_Ptr){} 
};

struct Ctk_Size_Hints : Ctk_Header { 
    int4 min_width, base_width, max_width, width_inc;
    int4 min_height, base_height, max_height, height_inc;
  private:
    Ctk_Size_Hints() : Ctk_Header(0,No_Ptr){} 
};

//-----------------------------------------------------------------------------
// Asyncroneous input queue

class Ctk_Win_Input_Queue {
    static const int event_queue_size = 256;
    static int n_elems;
    static Ctk_Event** beg;
    static Ctk_Event** end;
    static Ctk_Event*  queue[event_queue_size];
    static Semaphore*  Input_Semph;

    static bool   empty() {return n_elems == 0;}
    static bool   full()  {return n_elems == event_queue_size;}
    static Ctk_Event** forw(Ctk_Event** p) { 
	return p == queue + (event_queue_size - 1)
	    ? queue : p + 1;
    }

  public: 
    static bool put(Ctk_Event* evt) {
	if (full()) 
	    return FALSE;
	*end = evt;
	end = forw(end);
	Input_Semph->open();
	return TRUE;
    }
    
    static Ctk_Event* get();
    static void init();
};

//-----------------------------------------------------------------------------
// Ctk_Win_Open_Block --- handler, which is provided by concrete window
//			  system interface, to open the display 
//-----------------------------------------------------------------------------

class Ctk_Window_Interface_Descriptor {
  public:  
    typedef Ctk_Display* (*Win_Open_Func)(Ctk_String*);
    typedef void (*Win_Init_Func)();

    // enumeration type Win_Pri encodes the relative priorities of open
    // operations of different window systems. The window system with
    // higher priority will be tried first.
    enum Win_Pri {
	PM	 = 128,		// OS/2 Presentation manager
	MS_Win	 = 96,		// Microsloth windows
	X_Window = 64,		// X Window system
	GAW	 = 32		// Garret Window system
    };

  private:
    friend class Ctk_Window_System;
    static Ctk_Window_Interface_Descriptor*	head;
    
    Ctk_Window_Interface_Descriptor*	next;
    const char*				name;
    Win_Open_Func			open_func;
    Win_Init_Func			init_func;
    Win_Pri				pri;

  public:
    Ctk_Window_Interface_Descriptor
      (const char* name, Win_Init_Func f, Win_Open_Func g, Win_Pri pri);

    ~Ctk_Window_Interface_Descriptor();

    Ctk_Display* open(Ctk_String* str)
      {return open_func(str);}

    // Public accessor to a name component
    const char* get_name() const
      {return name;}
};

class Ctk_Window_System {
  protected :
    friend class Ctk_Display;
    static Ctk_Display* head;
  public:
    static Ctk_Display* Display_List_Head();
    static Ctk_Display* Open_Display(Ctk_String* name);
    static Ctk_Event*	Get_Event();
    static void 	Initialize();
};

// Display connection description

class Ctk_Display : public Ctk_Resource  
{
  protected:  
    bool			active() { return scr_list != NULL; }
    bool 			Ok() 
     { if( active() ) return TRUE; CTK_RAISE Ctk_Not_Available; return FALSE; }
  public:
    Ctk_Display(nat2 descr);
    Ctk_Display() : Ctk_Resource(0, NULL) {}
    ~Ctk_Display(); 

    Ctk_Window*         	root_window; // Root window for this display
    Ctk_Keyboard*       	kbd;         // Keyboard
    Ctk_Mouse*          	ms;          // Mouse input device, possibly nil
    Ctk_Display*        	next;        // Next item in list of displays
    Ctk_Screen*         	scr_list;    // All screens connected to this
                                             // display.
    int4			screen_count;// Total number of screens

    // General display information
    Ctk_String*         	win_protocol;
    Ctk_String*         	display_string;
    Ctk_String*         	server_vendor;
    int4                	protocol_version;
    int4                	protocol_revision;
    int4                	vendor_release;
    int4			consumers; 

    Ctk_Screen* 		Scr( int no ); 
    Ctk_Boolean*	        Is_Active(); 

    virtual void  		No_Op();
    virtual Ctk_Screen* 	Default_Screen(); 

    virtual void  		Set_Close_Down_Mode(int mode);

    virtual void 		Close();
    virtual void		Flush();
    virtual void		Bell( int volume );
   
    // Selections
    virtual void 		UnSet_Selection_Owner( Ctk_String* selection);
    virtual void 		Set_Selection_Owner( Ctk_String* selection, 
 				     Ctk_Window* owner);
    virtual Ctk_Window* 	Get_Selection_Owner( Ctk_String* selection );
    virtual void 		Convert_Selection( Ctk_String* selection, 
 				     Ctk_String* type);
};


class Ctk_Screen : public Ctk_Virtual {
    friend class Ctk_Winweb;
    friend class Ctk_Display; 
  public:
    Ctk_Window*			root_window;   	// Root window object
    Ctk_Screen*			next;          	// Next screen in list
    Ctk_Display*        	dsp;	       	// Display containing screen
    Ctk_Winweb* 		web;	     	// Screen objects

    virtual int  		Black_Pixel();
    virtual int 		White_Pixel();

    virtual Ctk_Colormap* 	Default_Colormap();
    virtual int  		Default_Depth();
    virtual Ctk_Array_I4* 	List_Depths();
    virtual Ctk_GC* 		Default_GC();
    virtual int 		Display_Cells();
 
    virtual int		 	Height();
    virtual int 		Width();
    virtual int 		Height_mm();
    virtual int 		Width_mm();

    virtual void                Query_Best_Tile(int width, int height, 
						long Res[2] );
    virtual void                Query_Best_Stipple(int width, int heigt, 
						   long Rs[2] );
    virtual void              	Query_Best_Cursor(int width, int height, 
						  long Res[2] );

    // Create colormap
    virtual Ctk_Colormap*       Create_Colormap( Ctk_Boolean* alloc ); 

    // Create cursors
    virtual Ctk_Cursor* 	Create_Font_Cursor(int shape);
    virtual Ctk_Cursor*	        Create_Glyph_Cursor(Ctk_Font* source_font, 
      Ctk_Font* mask_font, int  source_char, int  mask_char, 
      int foreground, int background);
    virtual Ctk_Cursor*         Create_Pixmap_Cursor( Ctk_Pixmap* source, 
      Ctk_Pixmap* mask, int foreground, int background, int x, int y);

    // Windows
    virtual Ctk_Input_Window*   Create_Simple_Input_Window
        ( Ctk_Input_Window* parent, int x, int y, int width, int height );
    virtual Ctk_Input_Window*   Create_Input_Window( Ctk_Input_Window* parent, 
	int x, int y, int width, int height, int win_gravity, int event_mask, 
	int do_not_propagate_mask, Ctk_Boolean* override_redirect);

    virtual Ctk_Window*   	Create_Default_Window
        ( Ctk_Window* parent, int x, int y, int width, int height, int bwidth);
    virtual Ctk_Window* 	Create_Simple_Window(Ctk_Window* parent, 
        int x, int y, int width, int height, int border_width, 
 	int border_pixel, int background_pixel);
    virtual Ctk_Window* 	Create_Window(Ctk_Window* parent, 
	int x, int y, int width, int height, int depth, int border_width, 
        int border_pixel, int background_pixel, int win_gravity, 
	int event_mask, int do_not_propagate_mask, 
	Ctk_Boolean* override_redirect, Ctk_Cursor* curs);

    // Graphic context
    virtual Ctk_GC* 		Create_Empty_Draw_Ctx(Ctk_Pixmap* drawable);
    virtual Ctk_GC* 		Create_Draw_Ctx(Ctk_Pixmap* drawable, 
	int foreground, int background, int funct, int plane_mask);
    virtual Ctk_GC* 		Create_Font_Draw_Ctx(Ctk_Pixmap* drawable, 
	int foreground, int background, int funct, int plane_mask, 
	Ctk_Font* fnt);

    // Pixmap
    virtual Ctk_Pixmap* 	Create_Pixmap(int width, int height,int depth);

    // Fonts
    virtual Ctk_Font* 		Load_Font(Ctk_String* name);
    virtual Ctk_Array_O*        List_Fonts(Ctk_String* pattern, int limit);

    // C_Talk primitive to check for disabled screen objects
    Ctk_Boolean* Is_Active();

    Ctk_Screen() : Ctk_Virtual(0) { dsp = NULL; }
    // Method used to `turn the screen of'
    ~Ctk_Screen();
  protected:
    // Predicate to check for disabled screens
    bool   	active() { return dsp != NULL; }
    bool	Ok() 
      { if(active()) return TRUE; CTK_RAISE Ctk_Not_Available; return FALSE; }
    Ctk_Screen(nat2 Desc, Ctk_Display* d);
};

class Ctk_Winweb : public Ctk_Virtual { 
  public:
    Ctk_Screen* Scr;
    Ctk_Winweb* Next;
 
    Ctk_Boolean* Is_Active();
    bool active() { return Scr != NULL; }
    bool Ok() 
      { if(active()) return TRUE; CTK_RAISE Ctk_Not_Available; return FALSE; }

    virtual void Destroy(); 

    Ctk_Winweb( nat2 Desc, Ctk_Screen* Scr ) : Ctk_Virtual( Desc ) 
        { this->Scr = Scr; Next = Scr->web; Scr->web = this; }
    Ctk_Winweb() : Ctk_Virtual(0) { 
        Scr = NULL; 
    }
    ~Ctk_Winweb();
};

class Ctk_Pix_Image : public Ctk_Destructed { 
  public:
    virtual int			Width();
    virtual int 		Height();
    virtual int			Depth();

    Ctk_Pix_Image( nat2 Desc = 0) : Ctk_Destructed( Desc ) {} 
}; 

class Ctk_Font_Metrics : public Ctk_Destructed 
{
  public:
    virtual Ctk_Object		Get_Property( Ctk_String* Str ); 
    virtual int 		Width_String( Ctk_String* Str ); 
    virtual int 		Width_String16( Ctk_Inter_String* Str ); 
    virtual void		Extents_String( Ctk_String* Str, long Res[2]); 
    virtual void		Extents_String16( Ctk_Inter_String* Str, 
						  long Res[2]); 
    virtual void		Extents( long Res[2] ); 
 
    Ctk_Font_Metrics( nat2 Desc = 0 ) : Ctk_Destructed( Desc ) {} 
};

class Ctk_Font : public Ctk_Winweb
{
  public:
    virtual Ctk_Font_Metrics* 	Metrics();
    Ctk_Font() {}
  protected: 
    Ctk_Font( nat2 Desc, Ctk_Screen* Scr ) : Ctk_Winweb( Desc, Scr ) {} 
};

class Ctk_Colormap : public Ctk_Winweb
{
  public: 
    virtual Ctk_Colormap* 	Copy_And_Free();
 
    virtual int 		Alloc_Color( Ctk_Color* color );
    virtual Ctk_Array_I4*	Alloc_Colors( Ctk_Array_O* colors );

    virtual Ctk_Array_I4*	Alloc_Color_Cells( int n_colors );
    virtual void		Alloc_Color_Planes(Ctk_Array_I4* pixels, 
			            Ctk_Array_I4* planes, Ctk_Boolean* contig);

    virtual void 		Free_Colors( Ctk_Array_I4* colors,int planes );
    virtual void 		Store_Color( int pixel, Ctk_Color* color );
    virtual void 		Store_Colors( Ctk_Array_I4* pixels,
					      Ctk_Array_O* colors);

    Ctk_Colormap() {}
  protected: 
    Ctk_Colormap( nat2 Desc, Ctk_Screen* Scr ) : Ctk_Winweb( Desc, Scr ) {} 
};

class Ctk_Cursor : public Ctk_Winweb 
{
  public:
    virtual void 		Recolor( int foreground, int background );
    Ctk_Cursor() {}
  protected: 
    Ctk_Cursor( nat2 Desc, Ctk_Screen* Scr ) : Ctk_Winweb( Desc, Scr ) {} 
};

class Ctk_GC : public Ctk_Winweb {
  public: 
    virtual Ctk_Font*		Fnt();
    virtual int			Background();
    virtual int			Foreground();
    virtual int 		Drawing_Function();
    virtual int 		Plane_Mask();
    virtual int 		Line_Width();
    virtual int 		Line_Style();
    virtual int 		Cap_Style();
    virtual int 		Join_Style();
    virtual int 		Fill_Style();
    virtual int 		Fill_Rule();
    virtual int 		Arc_Mode();
    virtual Ctk_Pixmap*		Tile();
    virtual Ctk_Pixmap*		Stipple();
    virtual int 		TS_X_Origin();
    virtual int 		TS_Y_Origin();
    virtual int 		Subwindow_Mode();
    virtual Ctk_Boolean*	Graphics_Exposures();
    virtual int 		Clip_X_Origin();
    virtual int 		Clip_Y_Origin();

    virtual void		Set_Fnt(Ctk_Font*  fnt);
    virtual void		Set_Background(int col);
    virtual void		Set_Foreground(int col);
    virtual void		Set_Drawing_Function(int f);
    virtual void		Set_Plane_Mask(int);
    virtual void		Set_Line_Width(int);
    virtual void		Set_Line_Style(int);
    virtual void		Set_Cap_Style(int);
    virtual void		Set_Join_Style(int);
    virtual void		Set_Fill_Style(int);
    virtual void		Set_Fill_Rule(int);
    virtual void		Set_Arc_Mode(int);
    virtual void		Set_Tile(Ctk_Pixmap* pixmap);
    virtual void		Set_Stipple(Ctk_Pixmap* pixmap);
    virtual void		Set_TS_X_Origin(int);
    virtual void		Set_TS_Y_Origin(int);
    virtual void		Set_Subwindow_Mode(int);
    virtual void		Set_Graphics_Exposures(Ctk_Boolean* report);
    virtual void		Set_Clip_X_Origin(int x);
    virtual void		Set_Clip_Y_Origin(int x);
    virtual void		Set_Clip_Mask(Ctk_Pixmap* pixmap);
    virtual void		Set_Clip_Rectangle(Ctk_Box* Box);
    virtual void		Set_Clip_Rectangles(Ctk_Array_O* Boxes);
    virtual void		Set_Dashes(Ctk_Array_I4* Arr);

    virtual void		Set_Line_Attributes(int line_width, 
				int line_style, int cap_style, int join_style);

    virtual Ctk_GC* 		Copy();

    Ctk_GC() {}
    Ctk_GC( nat2 Desc, Ctk_Screen* Scr ) : Ctk_Winweb( Desc, Scr ) {}
}; 

class Ctk_Pixmap : public Ctk_Winweb { 
  protected:
    bool			Ctx_Ok( Ctk_GC* gc ) 
      { if(gc->Scr == Scr) return TRUE; CTK_RAISE Ctk_Bad_Value; return FALSE;}

  public:
    // Drawing primitives
    virtual void 		Copy_Area_From(Ctk_GC* ctx, int x, int y, 
				  int width, int height, Ctk_Pixmap* dst, 
				  int dst_x, int dst_y);
    virtual void 		Copy_Area(Ctk_GC* ctx, int x, int y, 
				  int width, int height, int dst_x, int dst_y);
    virtual void 		Copy_Plane_From(Ctk_GC* ctx, int x, int y, 
				  int width, int height, Ctk_Pixmap* dst, 
				  int dst_x, int dst_y, int plane);
    virtual void 		Copy_Plane(Ctk_GC* ctx, int x, int y, 
				  int width, int height, 
				  int dst_x, int dst_y, int plane);
    virtual void 		Draw_Point(Ctk_GC* ctx, int x, int y);
    virtual void 		Draw_Points(Ctk_GC* ctx, Ctk_Array_O* points,
				    Ctk_Boolean* rel);
    virtual void 		Draw_Line(Ctk_GC* ctx, int x1, int y1, int x2,
				  int y2);
    virtual void 		Draw_Lines(Ctk_GC* ctx, Ctk_Array_O* lines, 
				   Ctk_Boolean* rel);
    virtual void 		Draw_Segments(Ctk_GC* ctx, Ctk_Array_O* segs);
    virtual void 		Draw_Rectangle(Ctk_GC* ctx, int x, int y, 
				       int width, int height);
    virtual void 		Draw_Rectangles(Ctk_GC* ctx,Ctk_Array_O* rects);
    virtual void 		Draw_Arc(Ctk_GC* ctx, int x, int y, 
				 int width, int height,int angle1, int angle2);
    virtual void 		Draw_Arcs(Ctk_GC* ctx, Ctk_Array_O* arcs);
    virtual void 		Fill_Rectangle(Ctk_GC* ctx, int x, int y, 
				       int width, int height);
    virtual void 		Fill_Rectangles(Ctk_GC* ctx,Ctk_Array_O* rects);
    virtual void 		Fill_Polygon(Ctk_GC* ctx, Ctk_Array_O* points, 
				     int mode, Ctk_Boolean* rel);
    virtual void 		Fill_Arc(Ctk_GC* ctx, int x, int y, 
				 int width, int height,int angle1, int angle2);
    virtual void 		Fill_Arcs(Ctk_GC* ctx, Ctk_Array_O* arcs);

    // Image input/output primitives
    virtual void 		Put_Image(Ctk_GC* ctx, Ctk_Pix_Image* img,
				  int src_x, int src_y, int dst_x, int dst_y,
				  int width, int height);
    virtual Ctk_Pix_Image*	Get_Image(int x, int y, int width, int height, 
				  int plane_mask);

    // Text output primitives
    virtual void 		Put_String(Ctk_GC* ctx,int x,int y, 
				  Ctk_String* s);
    virtual void 		Put_String16(Ctk_GC* ctx, int x, int y, 
				  Ctk_Inter_String* s);
    virtual void 		Put_Image_String(Ctk_GC* ctx,int x,int y, 
				  Ctk_String* s);
    virtual void 		Put_Image_String16(Ctk_GC* ctx, int x, int y, 
				  Ctk_Inter_String* s);
    Ctk_Screen* 		Get_Scr(); 
    Ctk_Pixmap() {}
  protected:
    Ctk_Pixmap( nat2 Desc, Ctk_Screen* Scr ) : Ctk_Winweb( Desc, Scr ) {} 
}; 

class Ctk_Input_Window : public Ctk_Pixmap { 
  public:
    virtual void   		Destroy_Subwindows();

    // map/unmap
    virtual void   		Map();
    virtual void   		Map_Raised();
    virtual void   		Map_Subwindows();
    virtual void   		Unmap();
    virtual void   		Unmap_Subwindows();

    // configuring window
    virtual void   		Move(int x, int y);
    virtual void   		Resize_Window(int width, int height);
    virtual void   		Move_Resize(int x, int y,int width,int height);
    virtual void 		Origin( long Res[2] );  // x, y 
    virtual void		Size( long Res[2] );   	// width. height

    // Changing Window Stacking Order
    virtual void   		Raise();
    virtual void   		Lower();
    virtual void   		Circulate_Subwindows(int direction);
    virtual void   		Circulate_Subwindows_Up();
    virtual void   		Circulate_Subwindows_Down();    

    // events handling
    virtual int 		All_Event_Mask();
    virtual int 		Own_Event_Mask();
    virtual int  		Not_Propagate_Mask();
    virtual Ctk_Boolean*  	Override_Redirect();

    virtual Ctk_Boolean* 	Grab_Pointer( Ctk_Boolean* confine ); 
    virtual void		Ungrab_Pointer();

    virtual void		Focus_In(); 
    virtual void		Focus_Out(); 

    //gravity
    virtual void		Set_Bit_Gravity(int gravity);
    virtual void		Set_Win_Gravity(int gravity);
    virtual int 		Bit_Gravity();
    virtual int  		Win_Gravity();

    virtual void 		Set_Not_Propagate_Mask(int mask); 
    virtual void 		Set_Override_Redirect(Ctk_Boolean* redir); 
    virtual void   		Select_Input(int mask);
    
    // Properties
    virtual void   		Put_Property(Ctk_String* name, Ctk_Object val);
    virtual Ctk_Object		Get_Property(Ctk_String* name);
    virtual void   		Remove_Property(Ctk_String* name);
    virtual Ctk_Array_O*	List_Properties();

    Ctk_Input_Window() {}
  protected:
    Ctk_Input_Window( nat2 Desc, Ctk_Screen* Scr ) : Ctk_Pixmap( Desc, Scr) {} 
}; 

class Ctk_Window : public Ctk_Input_Window { 
  public:
    virtual void 		Clear();
    virtual void 		Clear_Area(int x, int y, int width, int heght, 
                                   Ctk_Boolean* exposures);

    virtual void		Configure(int x, int y, int width, int height, 
				  int border_width);
    virtual int                 Depth();

    virtual void		Set_Background_Pixel(int background);
    virtual void		Set_Background_Pixmap(Ctk_Pixmap* pxm);
    virtual void		Set_Border_Pixel(int pixel);
    virtual void		Set_Border_Pixmap(Ctk_Pixmap* pxm);
    virtual void		Set_Border_Width(int width); 

    virtual Ctk_Colormap*	Get_Colormap();
    virtual void		Set_Colormap(Ctk_Colormap* cmp);

    virtual void		Set_WM_Properties( Ctk_String* win_name, 
					   Ctk_String* icon_name, 
					   Ctk_Size_Hints* Hints ); 

    virtual void                Set_Backing_Store(int backing_store);
    virtual int                 Get_Backing_Store(); 


    // Pointer device cursor
    virtual void		Define_Cursor(Ctk_Cursor* cur);
    virtual void	       	Undefine_Cursor();
    
    Ctk_Window() {}
  protected:
    Ctk_Window( nat2 Desc, Ctk_Screen* Scr ) : Ctk_Input_Window( Desc, Scr) {} 
};

class Ctk_Keyboard : public Ctk_Virtual {
  protected:
    friend class Ctk_Display;
    Ctk_Display*	dsp;	// Display to which keyboard belongs
    bool Ok() { 
	if( dsp ) return TRUE; 
	CTK_RAISE Ctk_Not_Available; 
	return FALSE;
    } 
  public:
    void Change_Control( int key_click_percent, int bell_percent,
			 int bell_duration, Ctk_Boolean* autorepeat ); 
    Ctk_Keyboard( nat2 Desc=0, Ctk_Display* Disp=NULL) : Ctk_Virtual(Desc)
      { dsp = Disp; }
};

class Ctk_Mouse : public Ctk_Virtual {
  protected:
    friend class Ctk_Display;
    Ctk_Display* 	dsp;	// Display to which mouse belongs 
    bool Ok() { 
	if( dsp ) return TRUE; 
	CTK_RAISE Ctk_Not_Available; 
	return FALSE;
    } 
  public:
    Ctk_Mouse( nat2 Desc, Ctk_Display* Disp) : Ctk_Virtual(Desc)
      { dsp = Disp; }
};

class Ctk_Event : public Ctk_Header {
  public:
    int4		tag;	// Generic event tag
    int4		type;	// Window system specific event type
    int4		serial;	// Serial number of event
    int4 		repeat; // Repeat counter 
    Ctk_Boolean*	send_e; // True if this event was user generated
    Ctk_Display*	dsp;	// Display from which the event was obtained
    Ctk_Input_Window*	wnd;	// Window for which this event was generated
       				// (may be left unused for some kinds of 
       				//  requests)

    enum {
         key_press,		// Key pressed
	 key_release,		// Key depressed
	 button_press,		// Mouse button pressed
	 button_release,	// Mouse button depressed
	 motion_notify,		// Mouse moved
	 enter_notify,		// Mouse entered the window *
	 leave_notify,		// Mouse leaved the window *
	 focus_in,		// Input focus is switched to window *
	 focus_out,		// Input focus is switched from window *
	 expose,		// Exposure request from window system
	 resize_notify,		// Window was resized
	 map_notify,		// Notification in window mapping *
	 unmap_notify,		// Notification in window unmapping *
	 destroy_notify,	// Notification in window deletion *
	 selection_clear,	// Selection owner has been changed *
	 selection_request,	// Selection value request from clent *
	 selection_notify,	// Selection reply to the request of selection*
	 win_sys_specific	// Window system specific event, distunguished
				// by window system specific code *
    // Items marked by * are represented as an event object (in core protocol)
     };

    // Masks to specify in window accept masks
    enum {
         no_event_mask 		= 0x00000000,
         key_press_mask 	= 0x00000001,
	 key_release_mask 	= 0x00000002,
	 button_press_mask 	= 0x00000004,
	 button_release_mask 	= 0x00000008,
	 motion_notify_mask 	= 0x00000010,
	 enter_notify_mask 	= 0x00000020,
	 leave_notify_mask 	= 0x00000040,
	 button1_motion_mask 	= 0x00000100,
	 button2_motion_mask 	= 0x00000200,
	 button3_motion_mask 	= 0x00000400,
	 button_motion_mask 	= 0x00002000,
         expose_mask 		= 0x00008000,
	 resize_notify_mask 	= 0x00020000,
         focus_change_mask	= 0x00100000
     };
  protected:
    Ctk_Event(nat2 Descr) : Ctk_Header(Descr, 0) { repeat = 1; }
};

class Ctk_Input_Event : public Ctk_Event {
  public:
    Ctk_Window* root;		// Parent window
    int x_root, y_root;		// Coordinates relative to root window
    int x, y;			// Coordinates relative to window
    int keycode;		// Keycode, or button
    int state;			// Keyboard/mouse buttons state
    int time;			// Time stamp in milliseconds

    enum {
	ShiftMask     = 0x0001,    
	LockMask      = 0x0002,    
	ControlMask   = 0x0004,    
	Mod1Mask      = 0x0008,
	Mod2Mask      = 0x0010,
	Mod3Mask      = 0x0020,
	Mod4Mask      = 0x0040,
	Mod5Mask      = 0x0080,
	Button1Mask   = 0x0100,     
	Button2Mask   = 0x0200,     
	Button3Mask   = 0x0400,     
	Button4Mask   = 0x0800,     
	Button5Mask   = 0x1000,     
	AnyButtonMask = 0x1F00 
    };
  protected:
    Ctk_Input_Event(nat2 Descr) : Ctk_Event(Descr) {;}
};

class Ctk_Expose_Event : public Ctk_Event {
  public:
    int     x, y;		// Base coordinates of redraw rectangle
    int     width, height;	// Dimensions of redraw rectangle
    int     count;		// Counter in sequence (if not 0, then more
				// requests to follow
  protected:
    Ctk_Expose_Event(nat2 Descr) : Ctk_Event(Descr) {;}
};

class Ctk_Configure_Event : public Ctk_Event {
  public:
    int      x, y;		// New base coordinates of window
    int      width, height;	// New dimensions of window
    int      border_width;	// New width of border
  protected:
    Ctk_Configure_Event(nat2 Descr) : Ctk_Event(Descr) {;}
};

#endif

