//-----------------------------------------------------------------------------
//   Main module for Garret C-Talk
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>

#include <unistd.h>

#include "dynld.hpp"
#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctcol.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "fifth.hpp"
#include "cttyp.hpp"
#include "ctnum.hpp"
#include "ctlib.hpp"
#include "ctprc.hpp"
#include "ctlst.hpp"
#include "ctlsp.hpp"
#include "ctios.hpp"
#include "store.hpp"
#include "fstor.hpp"
#include "ctdbs.hpp"
#include "oodbs.hpp"
#include "ctmat.hpp"
#include "ctlib.hpp"
#include "editr.hpp"
#include "multp.hpp"
#include "ctdtm.hpp"

//-----------------------------------------------------------------------------

extern	const	char   ctk_version[];

local  Text_With_Undo  Program;
local  Text            Output;

local  bool	       Ctk_Batch_Mode = FALSE;
inline void 	       Append( char* Buffer ){ 
    if( !Ctk_Batch_Mode ) 
	Output.Append( Buffer ); 
}

//-----------------------------------------------------------------------------

global  void  Ctk_Print (const char* Format, ...)  {
    char Buffer [ 4096 ];
    va_list Args;
    va_start ( Args, Format );
    vsprintf ( Buffer, Format, Args );
    printf ( Buffer );
    Append ( Buffer );
    va_end( Args );
}

//-----------------------------------------------------------------------------

global  void Ctk_Echo ( char* Str ) {
    Append ( Str );
}

//-----------------------------------------------------------------------------

global  int  Ctk_Printf ( Ctk_Stream* Out, const char* Format, ... )  {
    char Buffer [ 4096 ];
    int  Code;
    va_list Args;

    va_start ( Args, Format );
#ifdef  __IOSTREAM__ // if existed good stream library
    if ( Out->Stream == &cout || Out->Stream == &cerr ) {
#else
    if ( Out->Stream == stdout || Out->Stream == stderr ) {
#endif
	Code = vsprintf ( Buffer, Format, Args );
	fprintf ( Out->Stream, Buffer );
	Append ( Buffer );
    } else 
	Code = vfprintf ( Out->Stream, Format, Args );
    va_end( Args );
    return Code;
}

//-----------------------------------------------------------------------------

local Ctk_Class* New_Class (nat2 Desc, nat2 Size, nat2 Stat, const char* Frm){
    Ctk_Class* Class = new Ctk_Class;
    Class->Info = Ctk_Header::No_Ptr;
    Class->Obj_Info = 0;
    Class->Obj_Size = Size;
    Class->Hdr_Templ.Desc = Desc;
    Class->Status |= Stat;
    if (Frm == NULL && Desc != Ctk_Vector_O_Desc || (Stat & Ctk_Class::Database))
	Class->Hdr_Templ.Info = Ctk_Header::No_Ptr;
    if ( Frm != NULL ) {
	nat2 Segment [ 64 ], *Seg = Segment;
	int base = sizeof(Ctk_Header);
	while ( TRUE ) {
	    while (*Frm++ == '-') 
		base += 4;
	    if (*--Frm == '+')  {
		int len; 
                #ifdef STRICT_ALIGN
                    base = ALIGN (base, sizeof(Ctk_Object));
	        #endif
	        *Seg++ = base;
		for( len = 0; *Frm++ == '+'; base += sizeof(Ctk_Object))
		    len++;
		*Seg++ = len;
		Frm -= 1;
	    } else if (*Frm == '*') { 
                #ifdef STRICT_ALIGN
                    base = ALIGN (base, sizeof(Ctk_Object));
	        #endif
	        base += sizeof(void*);
	        Frm += 1;
	    } else {
		break;
	    }
	}
	Class->Ptr_Segm = Ctk_Vector_U2::New ( Seg - Segment );
	scpy ( Class->Ptr_Segm->Items, Segment, Seg - Segment );
	rtm_gc_build_descriptor(Segment, (Seg - Segment)/2, Desc);
    } else {
	if (Desc == Ctk_Vector_O_Desc)
	    rtm_gc_build_vector_descriptor(Desc);
    }
    Class->Info &= ~Ctk_Header::No_Ptr;
    return Class;
}

//-----------------------------------------------------------------------------

void Load_Descriptors () {
    Ctk_Class* Hdr;

    #define CTK_CLASS( ID, NAME, FORMAT, STATUS ) \
        Ctk_Descriptor::Built_In[ NAME##_Desc ] = New_Class( NAME##_Desc, \
        sizeof( NAME ), STATUS, FORMAT );
        #include "ctcls.dpp"

    Hdr = CTK_DESC(Ctk_Header);
    Hdr->Pro_Tbl = Hdr->Members = Hdr->Bases = Ctk_Vector_O::New( 0 );
    rtm_mem_gc_ban = FALSE; // Now we are old enough to collect garbage

    // This is ugly smoke test for garbage collector
    #if 0
         int counter = 0;
	 while(1) {
	     Ctk_Vector_O* p;

	     p = Ctk_Vector_O::New(10000);
	     for (int i = 0; i < 10000; i++) {
		 p->Items[i] = new Ctk_String(i%100);
		 for (int j = 0; j  <= i; j++) {
		     if (p->Items[j]->Info != 0 
			 || (((Ctk_String*)p->Items[j])->Vector->Info 
			     != Ctk_Header::No_Ptr)) {
			 *(int*)0 = 0;
		     }
		 }
	     }
	     Trace("ok %d\n", counter++);
	 }
    #endif
}

global jmp_buf Ctk_Jmp_Buf;

import char* yyptr;
import char  yybuf[];
import bool  yyparse_error; 

import bool  Ctk_Fail_Flag; // program is terminated either by compilation 
                            // error either by exception

global char* Ctk_Stack_Top; // remember lower bound of stack to determinate
                            // scope of local variables

global bool Ctk_All_Warnings = FALSE;
global bool Ctk_Sys_Init = TRUE;
global char Ctk_Profile_Flag = 0;

global Task*      Ctk_Debugger;
import void       Ctk_Heap_Profile( void );

//-----------------------------------------------------------------------------

local	void	Program_Error (int Signal) {
    printf (MSG("program error"), Signal);

    #if 0
    int pid;

    if( (pid = fork()) != 0  ) { 
	char buf[ 256 ];
	sprintf (buf, "%u", pid); 
	execlp ("gdb", "gdb", Ctk_Prog_Path, buf, NULL);
    } else {
	pause(); 
    }
    #endif

// dump core 
//    exit (1); 
   abort(); 
}

//-----------------------------------------------------------------------------

local void Output_Profile () {
    Ctk_Class** Ptr = (Ctk_Class**) Ctk_Class::Packages->Items;
    for ( int i = Ctk_Class::Packages->N_Elems; -- i >= 0; ) {
        Ctk_Class* Pk = *Ptr++;
        if ( Pk->Stat->Visit ) {
           nat4* Prof = Pk->Stat->Visit->Items, *Cur = Prof + 1;
           nat4* Last = Prof + Pk->Stat->Visit->N_Elems;
           char Map_Name [ 256 ]; // FIXME:: constant sized arrays forbidden !
           strcat (strcpy (Map_Name, Ctk_Get_Symbol_Name (Pk->Ident)), ".map");
           FILE* Out = fopen ( Map_Name, "w" );
           switch ( Ctk_Profile_Flag ) {
              case 1:
                 fprintf( Out, "Not attended parts of package:\n" );
                 while( Cur < Last ) {
                    while ( *Cur++ != 0 );
                    if ( Cur >= Last ) break;
		    *Last = 1;
		    int Beg = Cur - Prof - 1; 
                    while ( *Cur++ == 0 );
                    *Last = 0;     
                    int End = Cur - Prof - 1; 
                    fprintf( Out, "%u - %u\n", Beg, End );
                 }
                 break;
              case 2:
                 fprintf(Out,"Execution profile ( line number : passes )\n");
                 while( Cur < Last ) { 
                    int Times = *Cur++;
		    if( Times )
                        fprintf( Out, "%u\t: %u\n", (int)(Cur-Prof-1),Times );
		 }
           }
           fclose ( Out );
       }
   }
}

local  Ctk_Proc*  Ctk_Main_Proc;
local  Ctk_Break_Exception*  Ctk_Break;

//-----------------------------------------------------------------------------

local void FPE_Handler( int ) {
    CTK_RAISE Ctk_Math_Exception( "", "", 0.0, 0.0 );
    signal( SIGFPE, FPE_Handler );
}

//-----------------------------------------------------------------------------

local void Break_Handler( int Sig ) {
    Ctk_Event_Flag |= CTK_INTERN;
    Ctk_Except_Obj = Ctk_Break;
    signal( Sig, Break_Handler );
}

//-----------------------------------------------------------------------------

void Ctk_Run_Interpreter( Ctk_Proc* Main ) {
    typedef void (*handler)(int);
    handler Save_Handler[3];

    Save_Handler[0] = signal(SIGFPE,	FPE_Handler);
    Save_Handler[1] = signal(SIGINT,	Break_Handler);
    Save_Handler[2] = signal(SIGQUIT,	Break_Handler);

    (void)new Ctk_Process(new Ctk_Closure( Main, new Ctk_Frame ), TRUE);

    signal(SIGFPE, Save_Handler[0]);
    signal(SIGINT, Save_Handler[1]);
    signal(SIGQUIT, Save_Handler[2]);

    Ctk_Main_Proc = Main;
}

//-----------------------------------------------------------------------------
extern "C" int	unexec(char*	new_name, 
		       char*	a_name,
		       void*	data_start,
		       void*	bss_start,
		       void*	entry_address);

static	char*	unexec_name = NULL;
static	bool	ctk_initialized = FALSE;

//-----------------------------------------------------------------------------

local int C_Talk_Loop(int argc, char* argv[]) {
    int	n;
    int	Page_Pool_Size = 4;
    int	i; 

    n = argc - 1;
    for (i = 1; i <= n && argv[i][0] == '-'; i ++) {
	switch (argv[i][1]) {
	  case '\0': 
            Ctk_Batch_Mode = TRUE;
	    goto End_Of_Options;
#ifdef USE_UNEXEC
	  case 'u' :
	    if (i < n) {
		i += 1;
		unexec_name = argv[i];
	    }
	    continue;
#endif
	  case 'B':
            Ctk_Batch_Mode = TRUE;
            continue;
	  case 'I':
            void(new Ctk_Path (&argv[i][2]));
            continue;
	  case 'M':
            Ctk_Profile_Flag = 1;
            continue;
	  case 'P':
            Ctk_Profile_Flag = 2;
            continue;
	  case 'W':
	    Ctk_All_Warnings = TRUE;
            continue;
	  case 'S':
            Page_Pool_Size = atoi(&argv[i][2]);
            continue;
          case 'L':
	    Ctk_Database::Set_Deadlock_Detection_Delay( atoi( &argv[i][2] ));
	    continue;
	  default:
            puts ("Command line options:\n"
		  "\t-B\t\tbatch mode\n"
		  "\t-M\t\toutput information about unvisited parts of program\n"
		  "\t-P\t\toutput profile information\n"
		  "\t-W\t\tenable all warning message\n"
		  "\t-Ipath\t\tadd path for package search\n"
		  "\t-Snumb\t\tspecify size of page pool (in pages)\n" 
		  "\t-Ltime\t\tspecify database deadlock detection interval\n"
#ifdef USE_UNEXEC
		  "\t-u file-name\t\tsave program snapshot in file\n"
#endif
		);
            return 1;
	}
    }
  End_Of_Options:
    
    char* path = getenv("CTK_LIB_PATH"); 
    if( path != NULL ) { 
	while( *path != '\0' ) { 
	    char* end = strchr(path,':'); 
            if( end != NULL ) { 
		*end = '\0';
		void(new Ctk_Path(path));
		path = end + 1;
	    } else { 
		void(new Ctk_Path(path));
		break; 
	    }
	}
    }

#ifdef USE_UNEXEC
    rtm_dld::init(argv[0]);
#endif

    signal (SIGINT, SIG_IGN);
    signal (SIGQUIT, SIG_IGN);
    signal (SIGILL, &Program_Error);
    signal (SIGBUS, &Program_Error);
    signal (SIGSEGV, &Program_Error);

    // Mark resource list in order to free all requested resources after
    // exit from C-Talk loop
    Ctk_Resource* All_Resources = Ctk_Resource::Mark();

    // Do system data initialization
    if (!ctk_initialized) {
	Ctk_Stack_Top = (char*)&argc - 40;

	printf(
	    "C-Talk interpreter. Version %s (c) Garret,"
	    " 1993,1994,1995,1996.\n",
	    ctk_version);
	setbuf (stdout, 0);
	printf ("Loading system library...\n");
	Dbs_F_Store::Page_Cache = new Dbs_Pool( Page_Pool_Size );
	Ctk_Descriptor::Init();
	Load_Descriptors();
	Ctk_Break = new Ctk_Break_Exception;
	Ctk_Database::Init();

	if( setjmp( Ctk_Jmp_Buf ) ) { // if error has occured
	    All_Resources->Free(); 
	    return 1;
	}

	Initialize();
	printf ( "\nDone\n" );
    }

    Ctk_Sys_Init = FALSE;
    if( i > n ) { 
	argv[i] = (char*)"data.ctk"; 
	n += 1;
    }
    if (!ctk_initialized) {
	ctk_initialized = TRUE;
	while ( i < n ) 
	    Ctk_Compile( argv[i++] );
    }
#ifdef USE_UNEXEC
    // Handle unexec option 
    if (unexec_name != NULL) {
	char*	p;

	p = unexec_name;
	unexec_name = NULL;
	if (unexec(p, argv[0], 0, 0, 0) != 0) {
	    fprintf (stderr, "Unexec failed\n");
	    exit (1);
	}
	exit (0);
    }
#endif
    
    Ctk_Resource* Client_Resources = Ctk_Resource::Mark(); 

    if( Ctk_Batch_Mode ) { 
	Ctk_Compile( argv[n] ); 
	if( *argv[n] == '-' ) { 
	    if( setjmp( Ctk_Jmp_Buf ) ) { // if terminated by error
		Client_Resources->Free();
	    }
	    while( !feof(stdin) ) {
		(Package = World)->Load_Ctx();
		fflush ( stdin );
		Ctk_Compile("-");
	    }
	}
	All_Resources->Free();
	if ( Ctk_Profile_Flag ) 
	    Output_Profile();
	return 0;
    }
    char* File = argv[n];
    char* Suffix = strrchr( File, '.' ); 
    if( !Suffix || Suffix[1] == '/' ) {
	char* Dst = (char*)alloca( strlen(File) + 5 ); 
	File = strcat(strcpy( Dst,File),".ctk"); 
    }
    Program.Load ( File );
    bool Recompile = TRUE;
    while ( TRUE ) {
	if( setjmp(Ctk_Jmp_Buf) ) {  // abnormal program termination 
	    Client_Resources->Free();
            Ctk_Fail_Flag = TRUE;
	}

        int Code = (Ctk_Fail_Flag || Package == World)
	        ? Editor2::Edit(&Program, &Output, 
			    yyparse_error ? yyptr - yybuf - 1 : 0,
			    Ctk_Cur_Line - 1, 15 )
	        : Editor2::Edit( &Program, &Output, -1, -1, 15 );
	if ( !(Code & Editor::Run) ) 
	    break;
	if ( Code & Editor::Chg ) 
	    Recompile = TRUE;
	Output.Clear();
	Output.Max_Line_No = 500;
	setbuf( stdout, 0 );
	Ctk_Fail_Flag = FALSE;
	if ( Recompile ) {
	    (Package = World)->Load_Ctx();
	    World->Members->Ch_Size(0);
	    World->N_Inst_Vars = 0;
	    Ctk_Sweep_Descriptors();
	    Ctk_Compile ( File );
	    if ( Ctk_Class::Packages->N_Elems == 1 ) 
		Recompile = FALSE;
	} else {
	    Ctk_Run_Interpreter( Ctk_Main_Proc );
	}
	Client_Resources->Free();
	printf ( "\n" );
	fflush ( stdin );
    }
    All_Resources->Free();
    if ( Ctk_Profile_Flag ) 
	Output_Profile();
    return 0;
}

//-----------------------------------------------------------------------------

global rtm_hook_list background_routines; 

local int Null_Proc( int, char** ) {
    while ( TRUE ) {
	if( !background_routines.run_hooks() ) {   
	    Multiplex_Channels();
	    if( Ctk_Event_Flag & CTK_EXTERN ) 
		Ctk_Event_Manager.Exec();
	} 
	Task::reschedule();
    }
    return 0;
}

//-----------------------------------------------------------------------------

int main (int argc, char* argv[]) {
    if (!ctk_initialized) {
	Task::init();
	Ctk_Process::init();
	Task::run (&Null_Proc, 0, NULL, FALSE, 0);
    }
    return C_Talk_Loop(argc, argv);
}




