//-----------------------------------------------------------------------------
//   List representation to bytecode compiler -- byte code generator
//   Copyright (C)  1993,1994,1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include "stdtp.hpp"
#include "numlb.hpp"
#include "memmv.hpp"
#include "l2lst.hpp"
#include "ermsg.hpp"
#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "cttyp.hpp"
#include "ctnum.hpp"
#include "ctlst.hpp"
#include "ctlsp.hpp"
#include "ctlib.hpp"
#include "ctprc.hpp"
#include "cttab.hpp"
#include "fifth.hpp"

#include "ctdbs.hpp"

//-----------------------------------------------------------------------------
//
// Dynamic buffer for byte code
//

class Ctk_Code_Buffer {
  private:
    nat1* Beg;
    nat1* Cur;
    nat1* End;

  public:
    nat1* Get () { return Beg; }

    nat1* Put ( int Size )  {
	nat1 *Pos = Cur, *Inc = Pos + Size;
	if (Inc > End) {
	    int   Dim = (End - Beg) << 1;
	    nat1* New = (nat1*)realloc( Beg, Dim );
	    Pos = New + (Cur - Beg);
	    Inc = Pos + Size;
	    Beg = New;
	    End = New + Dim;
	}
	Cur = Inc;
	return Pos;
    }

    void Reset () { Cur = Beg; }

    Ctk_Code_Buffer ( int size = 1024 ) {
	Beg = Cur = new nat1 [ size ];
	End = Beg + size;
    }

    ~Ctk_Code_Buffer () { delete Beg; }
};

local Ctk_Code_Buffer Code_Buf;
local int             Pc; // FIXME:: make this a component of code buffer

#define   SELF fifth::LDL1 // self pointer is passed as first parameter

//-----------------------------------------------------------------------------

import	Ctk_Object	Mark(Ctk_Object Expr, Ctk_Object Profile = NULL);

#define	WITHIN(x, a, b)		(unsigned(x) - (a) <= unsigned(b) - (a))

import Ctk_Proc*        Procedure;       // pointer to compiled procedure
import Ctk_Sym_Table    Sym_Tab;
local  TABLE( int4 )    Try  ( 128 );    // bounds of try blocks

#define  MTH_LEVEL  0   // level at which methods are compiled

#if BYTES_IN_WORD == 8 
    #define DOUBLE(expr)
#else
    #define DOUBLE(expr) expr
#endif
    
//-----------------------------------------------------------------------------
// Gen
//-----------------------------------------------------------------------------

local void Gen (nat1 Cop){
   *Code_Buf.Put(1) = Cop;
   Pc += 1;
}

//-----------------------------------------------------------------------------
// Gen_Byte
//-----------------------------------------------------------------------------

local void Gen_Byte (nat1 Cop, int1 Val){
    nat1* Dst;

    Dst = Code_Buf.Put(2);
    *Dst++ = Cop;
    *Dst++ = Val;
    Pc += 2;
}

//-----------------------------------------------------------------------------
// Gen_Offs
//-----------------------------------------------------------------------------

local void Gen_Offs ( nat1 Cop, int2 Offs ) {
    #ifdef STRICT_ALIGN
        if ( ( Pc & 1 ) == 0 ) {
	    *Code_Buf.Put(1) = fifth::NOP1;
	    Pc += 1;
	}
    #endif
    nat1* Dst = Code_Buf.Put ( 3 );
    *Dst++ = Cop;
    *((int2*)Dst)++ = Offs;
    Pc += 3;
}

//-----------------------------------------------------------------------------
// Gen_Offs_Offs
//-----------------------------------------------------------------------------

local void Gen_Offs_Offs ( nat1 Cop, int2 Offs1, int2 Offs2 ) {
   #ifdef STRICT_ALIGN
       if ( ( Pc & 1 ) == 0 ) {
	   *Code_Buf.Put(1) = fifth::NOP1;
	   Pc += 1;
       }
   #endif
   nat1* Dst = Code_Buf.Put ( 5 );
   *Dst++ = Cop;
   *((int2*)Dst)++ = Offs1;
   *((int2*)Dst)++ = Offs2;
   Pc += 5;
}

//-----------------------------------------------------------------------------
// Gen_Imm
//-----------------------------------------------------------------------------

local void Gen_Imm ( nat1 Cop, int4 Val ) {
    #ifdef STRICT_ALIGN
        int i = 4 - ((Pc + 1) & 3);

        if ( i < 4 ) {
	    *Code_Buf.Put(i) = fifth::NOP1 + i - 1;
	    Pc += i;
	}
    #endif

    nat1* Dst = Code_Buf.Put ( 5 );
    *Dst++ = Cop;
    *((int4*)Dst)++ = Val;
    Pc += 5;
}

//-----------------------------------------------------------------------------
// Gen_Double
//-----------------------------------------------------------------------------

local void Gen_Double ( nat1 Cop, real Val ) {
   #ifdef STRICT_ALIGN
      int i = 4 - (( Pc + 1 ) & 3 );
      if ( i < 4 ) *Code_Buf.Put(i) = fifth::NOP1 + i - 1, Pc += i;
   #endif
   nat1* Dst = Code_Buf.Put ( 9 );
   *Dst++ = Cop;
   *((real*)Dst)++ = Val;
   Pc += 9;
}

//=============================================================================

struct Branch;
enum   Ctk_Target { RVALUE, LVALUE, IGNORE, COND };
local  bool Compile ( Ctk_Object List, Ctk_Target Target, Branch* Goto=NULL );

struct Branch {
    nat1  Cop ;  // compare operation
    nat2  Then;  // number of THEN label
    nat2  Else;  // number of ELSE label

    static void If_True(Ctk_Object Expr, 
			nat2 Then, nat2 Else=Sym_Tab.Tmp_Lab())
    {
	Branch Br = { fifth::IF_TRUE, Then, Else };
	Compile ( Expr, COND, &Br );
	Gen_Offs( Br.Cop, Then );
	Sym_Tab.Set_Addr( Else, Pc );
    }

    static void If_False(Ctk_Object Expr,
			 nat2 Then, nat2 Else=Sym_Tab.Tmp_Lab())
    {
	Branch Br = { fifth::IF_FALSE, Else, Then };
	Compile ( Expr, COND, &Br );
	Gen_Offs( Br.Cop, Then );
	Sym_Tab.Set_Addr( Else, Pc );
    }
};

//=============================================================================
// Binary operations: + - * / % & | ^ << >> == != > >= < <=

local nat1 const Bool_Op [] =
  { 0, 0, 0, 0, 0, fifth::LAND, fifth::LOR, fifth::LXOR, 0, 0, fifth::PEQ,
    fifth::PNE, 0, 0, 0, 0 };
local nat1 const Sint_Op [] =
  { fifth::ADD, fifth::SUB, fifth::MUL, fifth::DIV, fifth::MOD, fifth::AND, fifth::OR,
    fifth::XOR, fifth::SHL, fifth::SHR, fifth::EQ, fifth::NE, fifth::GT, fifth::GE,
    fifth::LT, fifth::LE };
// local nat1 const Uint_Op [] =
// { ADD, SUB, MUL, DIV, MOD, AND, OR, XOR, SHL, SHR, EQ, NE, HI, HS, LO, LS };
local nat1 const Real_Op [] =
  { fifth::FADD, fifth::FSUB,fifth:: FMUL, fifth::FDIV, 0, 0, 0, 0, 0, 0, 
    fifth::FEQ, fifth::FNE,fifth::FGT, fifth::FGE, fifth::FLT, fifth::FLE };

//-----------------------------------------------------------------------------
// Reverse -- emit code for argument list in reversal order
//-----------------------------------------------------------------------------

local void Reverse ( Ctk_Object List ) { // load arguments in reverse order
   if ( List != NIL ) {
      Reverse ( CDR( List ) );
      Compile ( List, RVALUE );
   }
}

//-----------------------------------------------------------------------------

#define Put_Mark(LINE) Gen_Offs_Offs(fifth::MARK,Package->Hdr_Templ.Desc,LINE)

//-----------------------------------------------------------------------------
// Relocate
//-----------------------------------------------------------------------------

local void Relocate ( void ) {
    nat1* Beg = Code_Buf.Get();
    nat1* End = Beg + Pc;
    nat1* Ip  = Beg;

    while (Ip < End) {
	unsigned Cop = *Ip++;
	// control flow instruction
	if ( Cop - fifth::BR <= fifth::UNTIL - fifth::BR ) {
	    int2 Addr = Sym_Tab.Get_Addr ( *(nat2*)Ip );
#ifdef STRICT_ALIGN
	    unsigned Cop1 = Beg[Addr] - fifth::NOP1; 
	    if ( Cop1 <= fifth::NOP3 - fifth::NOP1 )
		Addr += Cop1 + 1; // optimize branch to NOP instruction
#endif
	    // replace IF_TRUE with WHILE or IF_FALSE with UNTIL
	    if ( ( Addr -= int2( Ip - Beg )) < 0 && Cop != fifth::BR ) 
		*(Ip-1) += 2;
	    *(int2*)Ip = Addr;
	}
	Ip += fifth::insn_ext_len [ Cop ];
    }
}

//-----------------------------------------------------------------------------
// Send -- emit send command 
//-----------------------------------------------------------------------------

local	void	Send (Ctk_Object Method) {
    int4	Index;

    // Search for a method in local constant pool, and add it if not found
    Index = Procedure->Const->Index (Method);
    if (Index < 0) {
	Index = Procedure->Const->N_Elems;
	Procedure->Const = Procedure->Const->Push (Method);
    }

    // Emit SEND with byte displacement if possible
    if (Index < 256) 
	Gen_Byte (fifth::SEND, Index);
    else 
	Gen_Offs (fifth::SEND2, Index);
}

//-----------------------------------------------------------------------------
// Generate move instruction
//-----------------------------------------------------------------------------

#define Moveb( Cop, Offs )                                		\
   if ( Offs < sizeof(Ctk_Header) + 8*4 )                 		\
      Gen( Cop##1 + ((Offs - sizeof(Ctk_Header)) >> 2) ); 		\
   else if ( Offs < 256 ) Gen_Byte( Cop##n, Offs );       		\
   else Gen_Offs( Cop##N, Offs )

#if defined(STRICT_ALIGN) || BYTES_IN_WORD == 4
  #define Move( Cop, Offs )                                 		\
    if ( Offs < sizeof(Ctk_Header) + 8*sizeof(Ctk_Object) )		\
       Gen( Cop##1 + (Offs - sizeof(Ctk_Object)) / sizeof(Ctk_Object) );\
    else if ( Offs < 256 ) Gen_Byte( Cop##n, Offs );       		\
    else Gen_Offs( Cop##N, Offs )
#else
  #define Move( Cop, Offs )                                 		\
    if ( (Offs & (sizeof(Ctk_Object)-1)) == 0 				\
      && Offs < sizeof(Ctk_Header) + 8*sizeof(Ctk_Object) )		\
       Gen( Cop##1 + (Offs - sizeof(Ctk_Object)) / sizeof(Ctk_Object) );\
    else if ( Offs < 256 ) Gen_Byte( Cop##n, Offs );       		\
    else Gen_Offs( Cop##N, Offs )
#endif

#define Short_Cmd( Cop, Offs ) 						\
    if( Offs < 256 ) Gen_Byte( Cop##n, Offs ); 				\
    else Gen_Offs( Cop##N, Offs )


//-----------------------------------------------------------------------------
// Compile -- compile attributed list to byte code
// Compile returns TRUE if operator (block) return VALUE.
//-----------------------------------------------------------------------------

local	bool	Compile (Ctk_Object Expr, Ctk_Target Target, Branch* Goto) {
    static nat2 Continue = 0; // loop continue label
    static nat2 Break = 0;    // loop break label
    bool   Return = FALSE;    // if last operator is return operator

    Ctk_Object List = CAR(Expr);
    switch ( List->Desc )  {
      case Ctk_Member_Desc:
        {  
	    Ctk_Member* Member = (Ctk_Member*) List;
	    Ctk_Class* Class;
	    Ctk_Object Type = Member->Type;
	    if ( Member->Atr & Ctk_Member::Instance ) {
		Class = Procedure->Object;
		ASSERT( Class );
		int BaseOffs = Class->Base_Offset( Member->Cls );
		if ( Member->Atr & Ctk_Member::Method ) {
		    Gen( Sym_Tab.Level == MTH_LEVEL ? SELF :fifth::THIS );
		    if ( BaseOffs != 0 
			 && !(Member->Atr & Ctk_Member::Protocol) )
		    {
			Gen_Offs ( fifth::INC, BaseOffs );
		    }
		    break;
		}
		unsigned Offs = Member->Off;
		bool Virtual = TRUE;
		if ( !(Member->Atr & Ctk_Member::Protocol) )
		    Virtual = FALSE, Offs += BaseOffs;
		if ( Target == LVALUE )  {
		    if ( Member->Atr & Ctk_Member::Constant )
			Ctk_Error( MSG("assignment to constant component") );
		    if ( Class->Status & Ctk_Class::Database )
			if ( !IS_DATABASE(Type) && !IS_BUILTIN(Type) )
			    Ctk_Error( MSG("non-persistent component") );
		    if ( !Virtual && Sym_Tab.Level == MTH_LEVEL ) {
			if( Class->Status & Ctk_Class::Database ) Gen( fifth::SMOD );
			if( (Class->Status & Ctk_Class::Database) 
			    && IS_DATABASE( Type ) ) { 
			    Short_Cmd( fifth::STSD, Offs ); 
			} else if( IS_BUILTIN( Type ) ) {
			    if( IS_REAL( Type ) ) Short_Cmd( fifth::STSF, Offs ); 
			    else Moveb( fifth::STSB, Offs );
			} else Move( fifth::STS, Offs );
		    } else {
			Gen( Sym_Tab.Level == MTH_LEVEL ? SELF : fifth::THIS );
			if ( Virtual ) 
			    Gen_Offs( fifth::CNV, Member->Cls );
			if ( Class->Status & Ctk_Class::Database ) 
			    Gen(fifth::MODF );
			if ( (Class->Status & Ctk_Class::Database) 
			     && IS_DATABASE( Type ) ) { 
			    Short_Cmd( fifth::STOD, Offs ); 
			} else if( IS_BUILTIN( Type ) ) {
			    if( IS_REAL( Type ) ) Short_Cmd( fifth::STOF, Offs );
			    else Moveb( fifth::STOB, Offs );
			} else Move( fifth::STO, Offs );
		    }
		} else if ( Target != IGNORE ) { // it is not always right
		    if ( IS_BUILTIN( Type ) ) {
			if ( !Virtual && Sym_Tab.Level == MTH_LEVEL ) {
			    if ( Class->Status & Ctk_Class::Database ) 
				Gen( fifth::SSYNC ); 
			    if( IS_REAL( Type ) ) Short_Cmd( fifth::LDSF, Offs ); 
			    else Moveb( fifth::LDSB, Offs );
			} else {
			    Gen( Sym_Tab.Level == MTH_LEVEL ? SELF : fifth::THIS );
			    if ( Class->Status & Ctk_Class::Database ) 
				Gen(fifth::SYNC); 
			    if ( Virtual ) 
				Gen_Offs( fifth::CNV, Member->Cls );
			    if( IS_REAL( Type ) ) 
				Short_Cmd(fifth::LDOF, Offs ); 
			    else Moveb( fifth::LDOB, Offs );
			}
		    } else {
			if ( !Virtual && Sym_Tab.Level == MTH_LEVEL )  {
			    if ( Class->Status & Ctk_Class::Database ) {
				if ( IS_DATABASE(Type) ) {
				    Gen ( fifth::SDB );
				    Short_Cmd( fifth::LDSD, Offs );
				    Gen ( fifth::FIND );
				} else {
				    Ctk_Error( MSG("non-persistent component") );
				}
			    } else Move( fifth::LDS, Offs );
			} else {
			    Gen( Sym_Tab.Level == MTH_LEVEL ? SELF : fifth::THIS );
			    if ( Virtual ) 
				Gen_Offs ( fifth::CNV, Member->Cls );
			    if ( Class->Status & Ctk_Class::Database ) {
				if ( IS_DATABASE(Type) ) {
				    Gen ( fifth::DB );
				    Short_Cmd( fifth::LDOD, Offs ); 
				    Gen ( fifth::FIND );
				} else {
				    Ctk_Error( MSG("non-persistent component") );
				}
			    } else Move( fifth::LDO, Offs );
			}
		    }
		}
	    } else if ( Target != IGNORE ) { // class variable
		if ( Member->Atr & Ctk_Member::Method ) 
		    break;
		Class = Ctk_Class::Get( Member->Cls );
		if ( Member->Atr & Ctk_Member::Constant ) {
		    if ( Target == LVALUE )
			Ctk_Error( MSG("assignment to constant component") );
		    if ( Type == META ) {
			Ctk_Class* Class = (Ctk_Class*)Member->Val;
			Gen_Offs( fifth::LCLS, Class->Hdr_Templ.Desc );
			break;
		    }
		    if ( IS_BUILTIN( Type ) ) {
			if ( IS_REAL( Type ) ) { 
			    real Val;

			    Val =*(real*)(Class->Cls_B_Vars->Items
					  + Member->Off);
			    if ( Val == 0.0 ) {
				Gen( fifth::LDI_0 ); 
				DOUBLE( Gen( fifth::LDI_0 ); );
			    } else Gen_Double( fifth::LDD, Val ); 
			} else { 
			    int4 Val = Class->Cls_B_Vars->Items[ Member->Off ];
			    if ( Val >= -1 && Val <= 1 )
				Gen ( fifth::LDI_0 + int1(Val) );
			    else if ( Val >= -128 && Val < 127 )
				Gen_Byte( fifth::LDB, Val );
			    else Gen_Imm ( fifth::LDI, Val );
			}
			break;
		    }
		}
		unsigned Offs = Member->Off;
		if ( IS_BUILTIN( Type ) )  {
		    Gen_Offs ( fifth::LBCV, Member->Cls );
		    if ( Target == LVALUE ) {
			if( IS_REAL( Type ) ) 
			    Short_Cmd( fifth::STOF, 
				       offsetof(Ctk_Vector_F,Items[Offs]) );
			else
			    Moveb( fifth::STOB, offsetof(Ctk_Vector_I4,Items[Offs]) );
		    } else { 
			if( IS_REAL( Type ) ) 
			    Short_Cmd( fifth::LDOF, 
				       offsetof(Ctk_Vector_F,Items[Offs]) );
			else
			    Moveb( fifth::LDOB, offsetof(Ctk_Vector_I4,Items[Offs]) );
		    }  
		} else {
		    Gen_Offs ( fifth::LOCV, Member->Cls );
		    if ( Target == LVALUE )
			Move( fifth::STO, offsetof(Ctk_Vector_O,Items[Offs]) );
		    else 
			Move( fifth::LDO, offsetof(Ctk_Vector_O,Items[Offs]) );
		}
	    }
	    break;
	}
      case Ctk_Atom_Desc:
        {  
	    int i = ATOM(List)->Index;
	    Ctk_Variable Var = Sym_Tab.Var(i);
	    unsigned     Offs = Var.Disp;
	    if ( Target == LVALUE ) {
		if ( Var.Attr & Ctk_Variable::Const )
		    Ctk_Error( MSG("assignment to read-only variable"), 
			       Ctk_Get_Symbol_Name ( ATOM(List)->Ident ) );
		if ( Var.Frame == Sym_Tab.Level ) {
		    Move( fifth::STL, Offs );
		    DOUBLE( if( IS_REAL( Var.Type ) ) Move( fifth::STL, Offs + 4 ); );
		} else {
		    if ( Var.Frame < 2 && Offs < 256 ) 
			Gen_Byte( fifth::STF1 + Var.Frame, Offs );
		    else                  
			Gen_Offs_Offs( fifth::STFn, 
				       Var.Frame*sizeof(Ctk_Frame*), Offs );
		    DOUBLE( 
			if ( IS_REAL( Var.Type ) ) { 
			    if ( Var.Frame < 2 && Offs < 256 - 4 ) 
				Gen_Byte( fifth::STF1 + Var.Frame, Offs + 4 );
			    else                  
				Gen_Offs_Offs( fifth::STFn, 
					       Var.Frame*sizeof(Ctk_Frame*),
					       Offs + 4 );
			}
		    );
		}
	    } else if ( Target != IGNORE )  {
		if ( Var.Attr & Ctk_Variable::Literal ) {
		    if ( IS_REAL( Var.Type ) ) { 
			real Val = *(real*)&Sym_Tab.Imm( Offs );
			if ( Val == 0.0 ) { 
			    Gen( fifth::LDI_0 );
			    DOUBLE( Gen( fifth::LDI_0 ); );
			} else Gen_Double( fifth::LDD, Val );    
		    } else { 
			int4 Val = Sym_Tab.Imm( Offs );
			if ( Val >= -1 && Val <= 1 ) 
			    Gen ( fifth::LDI_0 + int1(Val) );
			else if ( Val >= -128 && Val < 127 ) 
			    Gen_Byte( fifth::LDB, Val );
			else Gen_Imm ( fifth::LDI, Val );
		    }
		} else {
		    if ( Var.Frame == Sym_Tab.Level ) {
			if ( Var.Attr & Ctk_Variable::Init ) {
			    DOUBLE( 
				if( IS_REAL( Var.Type ) ) 
				Move( fifth::LDL, Offs+4 ); 
			    );
			    Move(fifth::LDL,Offs ); 
			} else Move(fifth::LDLC, Offs );
		    } else {
			if ( Var.Attr & Ctk_Variable::Init ) {
			    DOUBLE( 
				if ( IS_REAL( Var.Type ) ) { 
				    if ( Var.Frame < 2 && Offs < 256 - 4 ) 
					Gen_Byte( fifth::LDF1 + Var.Frame, Offs + 4);
				    else 
					Gen_Offs_Offs( fifth::LDFn, 
						       Var.Frame
						       *sizeof(Ctk_Frame*), 
						       Offs + 4 );
				}
			    );
			    if ( Var.Frame < 2 && Offs < 256 ) 
				Gen_Byte( fifth::LDF1 + Var.Frame, Offs );
			    else 
				Gen_Offs_Offs( fifth::LDFn, 
					       Var.Frame*sizeof(Ctk_Frame*),
					       Offs ); 
			} else {
			    if ( Var.Frame < 2 && Offs < 256 ) 
				Gen_Byte( fifth::LDFC1 + Var.Frame, Offs );
			    else 
				Gen_Offs_Offs( fifth::LDFCn, 
					       Var.Frame*sizeof(Ctk_Frame*),
					       Offs ); 
			}
		    }
		}
	    }
        }
        break;
      case Ctk_Integer_Desc:
      case Ctk_Char_Desc:
	if ( Target != IGNORE ) {
	    int4 Val = ((Ctk_Integer*)List)->Val;
	    if ( Val >= -1 && Val <= 1 ) 
		Gen ( fifth::LDI_0 + int1(Val) );
	    else if ( Val >= -128 && Val < 127 ) 
		Gen_Byte ( fifth::LDB, Val );
	    else Gen_Imm ( fifth::LDI, Val );
	}
	break;
      case Ctk_Float_Desc:
	if ( Target != IGNORE ) {
	    real Val = ((Ctk_Float*)List)->Val;
	    if ( Val == 0.0 ) { 
		Gen( fifth::LDI_0 ); 
		DOUBLE( Gen( fifth::LDI_0 ); );
            } else Gen_Double(fifth::LDD, Val );
	}
        break;
      case Ctk_String_Desc:
        if ( Target != IGNORE ) {
	    nat4 Index = Procedure->Const->N_Elems;
	    if ( Index < 256 ) 
		Gen_Byte ( fifth::LDC, Index );
	    else Gen_Offs (fifth::LDC2, Index << 2 );
	    Procedure->Const = Procedure->Const->Push ( List );
        }
        break;
      case Ctk_Node_Desc: {
	  nat2 Name = ATOM(CAR(List))->Ident;
	  Ctk_Object Node = CDR ( List );
	  switch ( Name ) {
	    case S_POINT: 
	    case S_CTX:
	    { 
		Ctk_Class* Class = (Ctk_Class*) NODE( Node )->Type;
                Ctk_Member* Member = (Ctk_Member*) NODE( List )->Type;
                Ctk_Object Type = Member->Type;
                if ( Member->Atr & Ctk_Member::Instance ) {
		    int BaseOffs = Class->Base_Offset( Member->Cls );
		    if ( Member->Atr & Ctk_Member::Method )  {
			if (BaseOffs && !(Member->Atr & Ctk_Member::Protocol))
			    NODE(Node)->Modf += BaseOffs;
			Compile ( Node, RVALUE );
			break;
		    }
		    if ( Class->Status & Ctk_Class::Database ) {
			if ( !IS_DATABASE(Type) && !IS_BUILTIN(Type) )
			    Ctk_Error( MSG("non-persistent component") );
		    }
		    Compile ( Node, RVALUE );
		    unsigned Offs = Member->Off;
		    if ( Member->Atr & Ctk_Member::Protocol )
			Gen_Offs ( fifth::CNV, Member->Cls );
		    else Offs += BaseOffs;
		    if ( Target == LVALUE )  {
			if ( Member->Atr & Ctk_Member::Constant )
			    Ctk_Error( MSG("assignment to constant component") );
			if ( Class->Status & Ctk_Class::Database ) 
			    Gen( fifth::MODF );
			if ( IS_BUILTIN( Type ) ) {
			    if( IS_REAL( Type ) ) 
				Short_Cmd(fifth:: STOF, Offs );  
			    else 
				Moveb( fifth::STOB, Offs ); 
			} else if( (Class->Status & Ctk_Class::Database) 
				   && IS_DATABASE( Type ) ) {
			    Short_Cmd( fifth::STOD, Offs ); 
			} else Move( fifth::STO, Offs );
		    } else if ( Target != IGNORE ) { // it is not always right
			if ( !IS_BUILTIN( Type ) ) {
			    if ( Class->Status & Ctk_Class::Database ) {
				Gen( fifth::DB );
				Short_Cmd(fifth:: LDOD, Offs );
				Gen( fifth::FIND );
			    } else Move(fifth:: LDO, Offs );
			} else {
			    if( Class->Status & Ctk_Class::Database )
				Gen( fifth::SYNC );  
			    if ( IS_REAL( Type ) ) 
				Short_Cmd( fifth::LDOF, Offs ); 
			    else Moveb(fifth::LDOB, Offs ); 
			}
		    } else Gen ( fifth::DROP ); // remove address from stack
                } else if ( Target != IGNORE ) { // class variable
		    if ( Member->Atr & Ctk_Member::Method ) 
			break;
		    if ( Member->Atr & Ctk_Member::Constant ) {
			if ( Target == LVALUE )
			    Ctk_Error( MSG("assignment to constant component") );
			if ( Type == META ) {
			    Ctk_Class* Class = (Ctk_Class*)Member->Val;
			    Gen_Offs( fifth::LCLS, Class->Hdr_Templ.Desc );
			    break;
			}
			if ( IS_BUILTIN( Type ) ) {
			    if ( IS_REAL( Type ) ) { 
				real Val = *(real*)(Class->Cls_B_Vars->Items 
						    + Member->Off);
				if ( Val == 0.0 ) {
				    Gen ( fifth::LDI_0 );
				    DOUBLE( Gen ( fifth::LDI_0 ); );
				} else Gen_Double( fifth::LDD, Val );  
			    } else {
				int4 Val;
				Val = Class->Cls_B_Vars->Items[Member->Off];
				if ( Val >= -1 && Val <= 1 ) 
				    Gen ( fifth::LDI_0 + char(Val) ); 
				else if ( Val >= -128 && Val < 127 )
				    Gen_Byte( fifth::LDB, Val );
				else Gen_Imm ( fifth::LDI, Val );
			    }
			    break;
			}
		    }
		    unsigned Offs = Member->Off;
		    if ( IS_BUILTIN( Type ) )  {
			Gen_Offs ( fifth::LBCV, Member->Cls );
			if ( Target == LVALUE ) {
			    if ( IS_REAL( Type ) ) 
				Short_Cmd( fifth::STOF, 
					   offsetof(Ctk_Vector_F,Items[Offs]));
			    else
				Moveb( fifth::STOB, 
				       offsetof(Ctk_Vector_I4,Items[Offs]) );
			} else { 
			    if ( IS_REAL( Type ) ) { 
				Short_Cmd( fifth::LDOF, 
					   offsetof(Ctk_Vector_F,
						    Items[Offs])); 
			    } else {
				Moveb( fifth::LDOB, 
				       offsetof(Ctk_Vector_I4,Items[Offs]) );
			    }
			}
		    } else {
			Gen_Offs( fifth::LOCV, Member->Cls );
			if ( Target == LVALUE )
			    Move( fifth::STO, offsetof(Ctk_Vector_O,Items[Offs]) );
			else
			    Move( fifth::LDO, offsetof(Ctk_Vector_O,Items[Offs]) );
		    }
                }
                break;
	    }
	    case S_NIL:   
	      if ( Target != IGNORE ) 
		  Gen( fifth::LDnil); 
	      break;
	    case S_TRUE:  
	      if ( Target != IGNORE ) 
		  Gen( fifth::LDt  ); 
	      break;
	    case S_FALSE: 
	      if ( Target != IGNORE ) 
		  Gen( fifth::LDf  ); 
	      break;
	    case S_SELF:
              if ( Target != IGNORE )
                 Gen( Sym_Tab.Level == MTH_LEVEL ? SELF : fifth::THIS );
              break;
	    case S_RAISE:
              Compile ( Node, RVALUE );
              Gen ( fifth::RAS );
              Return = TRUE;
              goto Exit;
	    case S_TRY:
	        { 
		    static int Catch_Depth = 0;
		    static nat2 Catch_Label [ 128 ];
		    Catch_Label [ Catch_Depth++ ] = Sym_Tab.Tmp_Lab();
		    nat2 Skip = Sym_Tab.Tmp_Lab();
		    nat2 Begin = Pc;
		    Return = Compile ( Node, IGNORE );
		    Node = CDR( Node ); 
		    if( CAR( Node ) != NIL ) { // finally block
			nat2 Br = Sym_Tab.Tmp_Lab();
			Try.Alloc(), *Try = Begin; // begin of the block
			Try.Alloc(), *Try = Pc;    // end of the block
			Gen_Offs( fifth::PATCH, Br ); 
			Sym_Tab.Set_Addr( Catch_Label[ --Catch_Depth ], Pc-3 );
			Compile( CDAR( Node ), IGNORE ); 
			Sym_Tab.Set_Addr( Br, Pc ); 
			Gen_Offs( fifth::BR, Skip );
		    } else { 
			Gen_Offs ( fifth::BR, Skip );
			Sym_Tab.Set_Addr ( Catch_Label[ --Catch_Depth ], Pc );
			Try.Alloc(), *Try = Begin; // begin of the block
			Try.Alloc(), *Try = Pc;    // end of the block
		    } 
		    while ( ( Node = CDR( Node )) != NIL ) {
			Ctk_Object Handler = CDAR( Node );
			Ctk_Class* Type = (Ctk_Class*) NODE(Handler)->Type;
			nat2 Next = Sym_Tab.Tmp_Lab();
			Gen ( fifth::SIG );
			Gen_Offs ( fifth::DER, Type->Hdr_Templ.Desc );
			Gen_Offs ( fifth::IF_FALSE, Next );
			Ctk_Atom* Param = ATOM( CADR( Handler ));
			if ( Param->Ident != S_NONAME ) {
			    unsigned Offs = Sym_Tab.Var( Param->Index ).Disp;
			    Gen ( fifth::SIG );
			    Move( fifth::STL, Offs );
			}
			Return &= Compile ( CDDR ( Handler ), IGNORE );
			Gen_Offs ( fifth::BR, Skip );
			Sym_Tab.Set_Addr ( Next, Pc );
		    }
		    if ( Catch_Depth == 0 ) {
#if 0
			if ( CADDDR( Procedure->Profile ) != NIL ) {
			    // procedure has explicit raise specification
			    Gen ( fifth::SIG );
			    Gen_Offs_Offs ( fifth::INVOKE, 1, Ctk_Exception_Desc );
			    // invokation of `unexpected' method
			}
#endif
			Gen ( fifth::RER );
		    } else Gen_Offs ( fifth::BR, Catch_Label[ Catch_Depth - 1 ] );
		    Sym_Tab.Set_Addr ( Skip, Pc );
		}
		goto Exit;
	    case S_OF:
              if ( Target != IGNORE ) {
		  Ctk_Class* Class = (Ctk_Class*) NODE(Node)->Type;
		  Gen_Offs ( fifth::LCLS, Class->Hdr_Templ.Desc );
              }
              break;
	    case S_IF:
              { 
		  nat2 Else = Sym_Tab.Tmp_Lab();
		  Branch::If_False ( Node, Else );
		  Node = CDR ( Node );
		  if( Ctk_Profile_Flag ) 
		      Put_Mark( Ctk_Cur_Line );
		  Return = Compile ( Node, IGNORE );
		  if ( ( Node = CDR(Node) ) != NIL ) {
		      nat2 Endif = Sym_Tab.Tmp_Lab();
		      Gen_Offs ( fifth::BR, Endif );
		      Sym_Tab.Set_Addr ( Else, Pc );
		      if ( Ctk_Profile_Flag ) 
			  Put_Mark( Ctk_Cur_Line );
		      Return &= Compile ( Node, IGNORE );
		      Sym_Tab.Set_Addr ( Endif, Pc );
		  } else {
		      Return = FALSE;
		      Sym_Tab.Set_Addr ( Else, Pc );
		  }
		  if ( Ctk_Profile_Flag ) Put_Mark( Ctk_Cur_Line );
              }
              goto Exit;
	    case S_WHILE:
              { 
		  nat2 ExBreak = Break, ExContinue = Continue;
		  nat2 Loop = Sym_Tab.Tmp_Lab();
		  Break    = Sym_Tab.Tmp_Lab();
		  Continue = Sym_Tab.Tmp_Lab();
		  Gen_Offs ( fifth::BR, Continue );
		  Sym_Tab.Set_Addr ( Loop, Pc );
		  int Line = Ctk_Cur_Line + 1;
		  Compile ( CDR ( Node ), IGNORE );
		  Sym_Tab.Set_Addr ( Continue, Pc );
		  if ( Ctk_Profile_Flag ) 
		      Put_Mark( Ctk_Cur_Line > Line ? Line : Ctk_Cur_Line );
		  Branch::If_True( Node, Loop, Break );
		  Break = ExBreak; 
		  Continue = ExContinue;
		  Ctk_Cur_Line = NODE(List)->Line;
		  if ( Ctk_Profile_Flag ) 
		      Put_Mark( Ctk_Cur_Line );
              }
              goto Exit;
	    case S_DOWHILE:
	      { 
		  nat2 ExBreak = Break, ExContinue = Continue;
		  nat2 Loop = Sym_Tab.Tmp_Lab();
		  Break    = Sym_Tab.Tmp_Lab();
		  Continue = Sym_Tab.Tmp_Lab();
		  Sym_Tab.Set_Addr ( Loop, Pc );
		  int Line = Ctk_Cur_Line + 1; 
		  Compile ( Node, IGNORE );
		  Sym_Tab.Set_Addr ( Continue, Pc );
		  if ( Ctk_Profile_Flag ) 
		      Put_Mark( Ctk_Cur_Line > Line ? Line : Ctk_Cur_Line );
		  Branch::If_True( CDR( Node ), Loop, Break );
		  Break = ExBreak; 
		  Continue = ExContinue;
		  Ctk_Cur_Line = NODE(List)->Line;
		  if (Ctk_Profile_Flag) 
		      Put_Mark(Ctk_Cur_Line);
              }
              goto Exit;
	    case S_CAND:
              if ( Target == COND ) {
		  Branch::If_False( Node, Goto->Else );
		  Compile( CDR ( Node ), COND, Goto );
              } else {
		  nat2 False = Sym_Tab.Tmp_Lab();
		  Compile ( Node, RVALUE );
		  Gen ( fifth::DUP );
		  Gen_Offs ( fifth::IF_FALSE, False );
		  Gen ( fifth::DROP );
		  Compile ( CDR ( Node ), RVALUE );
		  Sym_Tab.Set_Addr ( False, Pc );
		  if ( Target == IGNORE ) 
		      Gen ( fifth::DROP );
              }
              break;
	    case S_COR:
              if ( Target == COND )   {
		  Branch::If_True( Node, Goto->Then );
		  Compile ( CDR(Node), COND, Goto );
              } else {
		  nat2 True = Sym_Tab.Tmp_Lab();
		  Compile ( Node, RVALUE );
		  Gen ( fifth::DUP );
		  Gen_Offs ( fifth::IF_TRUE, True );
		  Gen ( fifth::DROP );
		  Compile ( CDR ( Node ), RVALUE );
		  Sym_Tab.Set_Addr ( True, Pc );
		  if ( Target == IGNORE ) 
		      Gen ( fifth::DROP );
              }
              break;
	    case S_NOT:
              if ( Target == COND )  {
		  nat2 Label = Goto->Then; // inverse branch
		  Goto->Then = Goto->Else; 
		  Goto->Else = Label;
                  //inverse condition
		  Goto->Cop = fifth::IF_TRUE + fifth::IF_FALSE - Goto->Cop; 
		  Compile ( Node, COND, Goto );
              } else {
		  Compile ( Node, RVALUE );
		  if ( Target == IGNORE ) 
		      Gen ( fifth::DROP );
		  else 
		      Gen ( fifth::NOT );
              }
              break;
	    case S_PEQ:
              { 
		  Ctk_Object Opd1=NODE(Node)->Type, Opd2=NODE(CDR(Node))->Type;
		  Compile ( Node, RVALUE ), Compile ( CDR( Node ), RVALUE );
		  if ( IS_INT( Opd1 ) && IS_INT( Opd2 ) )  
		      Gen( fifth::EQ );
		  else if (IS_REAL(Opd1) && IS_REAL(Opd2))  
		      Gen(fifth::FEQ);
		  else if ( IS_REAL( Opd1 ) || IS_REAL( Opd2 ) ) {
		      Ctk_Warning( MSG("always false") );
		      Gen ( fifth::DROP ); 
		      Gen ( fifth::DROP );
		      DOUBLE( Gen ( fifth::DROP ); );
		      Gen ( fifth::LDf );
		  } else if ( IS_INT( Opd1 ) || IS_INT( Opd2 ) )  {
		      Ctk_Warning( MSG("always false") );
		      Gen ( fifth::DROP ); 
		      Gen ( fifth::LDf );
		  } else { 
		      if( Ctk_Implicit_Convert(Opd1, Opd2) < 0  &&
		          Ctk_Implicit_Convert(Opd2, Opd1) < 0 )
                      { 
			  Ctk_Warning( MSG("always false") );
		      } 
		      Gen ( fifth::PEQ );
		  }     
		  if ( Target == IGNORE ) {
		      Ctk_Warning ( MSG("suspicious =") );
		      Gen ( fifth::DROP );
		  }
              }
              break;
	    case S_PNE:
              { 
		  Ctk_Object Opd1=NODE(Node)->Type, Opd2=NODE(CDR(Node))->Type;
		  Compile ( Node, RVALUE ), Compile ( CDR( Node ), RVALUE );
		  if ( IS_INT( Opd1 ) && IS_INT( Opd2 ) )  
		      Gen( fifth::NE );
		  else if (IS_REAL(Opd1) && IS_REAL(Opd2))  
		      Gen( fifth::FNE );
		  else if ( IS_REAL( Opd1 ) || IS_REAL( Opd2 ) ) {
		      Ctk_Warning( MSG("always true") );
		      Gen (fifth:: DROP ); 
		      Gen( fifth::DROP );
		      DOUBLE( Gen ( fifth::DROP ); );
		      Gen ( fifth::LDt );
		  } else if ( IS_INT( Opd1 ) || IS_INT( Opd2 ) )  {
		      Ctk_Warning( MSG("always true") );
		      Gen ( fifth::DROP ); 
		      Gen ( fifth::LDt );
		  } else { 
		      if( Ctk_Implicit_Convert(Opd1, Opd2) < 0  &&
		          Ctk_Implicit_Convert(Opd2, Opd1) < 0 )
                      { 
			  Ctk_Warning( MSG("always true") );
		      } 
		      Gen ( fifth::PNE );
		  }
		  if ( Target == IGNORE ) {
		      Ctk_Warning ( MSG("suspicious =") ); //FIXME:: suspicious 
		      Gen ( fifth::DROP );
		  }
              }
              break;
	    case S_GOTO:
              Gen_Offs ( fifth::BR, ATOM( CAR( Node ))->Index );
              goto Exit;
	    case S_LABEL:
              Sym_Tab.Set_Addr ( ATOM( CAR( Node ))->Index, Pc );
              if ( Ctk_Profile_Flag ) 
		  Put_Mark (Ctk_Cur_Line);
              Return = Compile ( CDR( Node ), IGNORE );
              goto Exit;
	    case S_BREAK:
              if ( !Break ) 
		  Ctk_Error( MSG("break outside loop") );
              Gen_Offs ( fifth::BR, Break );
              goto Exit;
	    case S_CONTINUE:
              if ( !Continue ) 
		  Ctk_Error( MSG("continue outside loop") );
              Gen_Offs ( fifth::BR, Continue );
              goto Exit;
	    case S_ASSERT:
              Compile ( Node, RVALUE );
              Gen ( fifth::ASSERT );
              return IS_NODE(CAR(Node)) && ATOM(CAAR(Node))->Ident == S_FALSE;
	    case S_PUT_AT:
              { 
		  Ctk_Member* Method = (Ctk_Member*)NODE(List)->Type;
		  Reverse ( Node );
		  unsigned Class = Method->Cls;
		  if ( WITHIN( Class, Ctk_Array_I1_Desc, Ctk_Array_O_Desc ))
		      Gen ( fifth::STV_I1 + Class - Ctk_Array_I1_Desc );
		  else {
		      if (Ctk_Is_Array(Method->Class())) 
			  Gen( fifth::STV_O );
		      else if ( Method->Atr & Ctk_Member::Protocol )
			  Gen_Offs_Offs( fifth::INVOKE, Method->Off, 
					 Method->Cls );
		      else Send ( Method->Val );
		  }
		  goto Exit;
              }
	    case S_VAR:
              { 
		  DOUBLE( Ctk_Object Type = CAR(Node); );
		  while ( (Node = CDR ( Node )) != NIL ) {
		      Ctk_Object Decl = CAR( Node );
		      unsigned Offs;
		      if ( !IS_ATOM( Decl ) ) {
			  Compile ( CDR( Decl ), RVALUE );
			  Decl = CAR( Decl );
		      } else {
			  Gen( fifth::LDI_0 );
			  DOUBLE( if( IS_REAL( Type ) ) Gen( fifth::LDI_0 ); );
		      }
		      Offs = Sym_Tab.Var(ATOM(Decl)->Index).Disp;
		      Move( fifth::STL, Offs );
		      DOUBLE( 
			     Ctk_Object Var_Type = (Type == NIL) 
			         ? NODE(CDAR(Node))->Type : Type; 
			     if( IS_REAL( Var_Type ) ) 
			         Move( fifth::STL, Offs + 4 ); 
                      )
		  } 
              }
              goto Exit;
	    case S_CONST:
	      while ( Node != NIL ) {
		  Ctk_Object Decl = CAR( Node );
		  if ( IS_ATOM(CAR(Decl)) && !IS_BUILTIN(NODE(Decl)->Type) ) {
		      nat4 Index = Procedure->Const->N_Elems;
		      if ( Index < 256 ) 
			  Gen_Byte( fifth::LDC, Index );
		      else 
			  Gen_Offs( fifth::LDC2, Index << 2 );
		      Procedure->Const = Procedure->Const->Push ( CDR(Decl) );
		      unsigned Offs = Sym_Tab.Var(ATOM(CAR(Decl))->Index).Disp;
		      Move( fifth::STL, Offs );
		  }
		  Node = CDR( Node );
              }
              goto Exit;
	    case S_POW: // generate fast code for case  x ** 2
              { 
		  Ctk_Object Type = NODE(Node)->Type;
		  Compile ( Node, RVALUE );
		  if ( IS_INT( Type ) ) {
		      if ( CADR( Node )->Desc == Ctk_Integer_Desc &&
			   ((Ctk_Integer*)CADR( Node ))->Val == 2 )
			  Gen( fifth::DUP ), Gen( fifth::MUL );
		      else Compile( CDR(Node), RVALUE ), Gen(fifth::IPOW);
		      if ( Target == IGNORE ) 
			  Gen( fifth::DROP );
		  } else {
		      Compile( Node = CDR( Node ), RVALUE );
		      Gen ( IS_INT( NODE(Node)->Type ) ? fifth::FPOWI : fifth::FPOWF );
		      if ( Target == IGNORE ) {
			  Gen( fifth::DROP ); 
			  DOUBLE( Gen( fifth::DROP ); );
		      }
		  }
              }
              break;
	    case S_ADD: case S_SUB: case S_MUL: case S_DIV: case S_MOD:
	    case S_SHL: case S_SHR: case S_AND: case S_OR: case S_XOR:
              { 
		  Ctk_Object Type = NODE(List)->Type;
		  Compile ( Node, RVALUE ); 
		  Compile ( CDR( Node ), RVALUE );
		  nat1 Op = 0;
		  if ( IS_REAL(Type) )     
		      Op = Real_Op[ Name - S_ADD ];
		  else if( IS_SINT(Type) ) 
		      Op = Sint_Op[ Name - S_ADD ];
		  // else if( IS_UINT(Type) ) Op = Uint_Op[ Name - S_ADD ];
		  else if( IS_BOOL(Type) ) 
		      Op = Bool_Op[ Name - S_ADD ];
		  if ( !Op ) 
		      Ctk_Error( MSG("bad operands for binary operator"),
				 Ctk_Get_Symbol_Name(Name));
		  Gen ( Op );
		  if ( Target == IGNORE ) {
		      Gen ( fifth::DROP );
		      DOUBLE( if( IS_REAL( Type ) ) Gen ( fifth::DROP ); );
		  }
              }
              break;
	    case S_EQ: case S_NE: case S_GT: case S_GE: case S_LT: case S_LE:
              { 
		  Ctk_Object Type = NODE(List)->Type;
		  Compile ( Node, RVALUE ); 
		  Compile ( CDR( Node ), RVALUE );
		  nat1 Op = 0;
		  if ( IS_REAL(Type) )     
		      Op = Real_Op[ Name - S_ADD ];
		  else if( IS_SINT(Type) ) 
		      Op = Sint_Op[ Name - S_ADD ];
		  // else if( IS_UINT(Type) ) Op = Uint_Op[ Name - S_ADD ];
		  else if( IS_BOOL(Type) ) 
		      Op = Bool_Op[ Name - S_ADD ];
		  if (!Op) {
		      Ctk_Error( MSG("bad operands in comparison"),
				 Ctk_Get_Symbol_Name(Name));
		  }
		  Gen ( Op );
		  if ( Target == IGNORE ) 
		      Gen ( fifth::DROP );
              }
              break;
	    case S_NEG:
              Compile ( Node, RVALUE );
              if ( Target == IGNORE ) { 
		  Gen ( fifth::DROP );
		  DOUBLE( if( IS_REAL( NODE(List)->Type ) ) Gen ( fifth::DROP ); );
              } else Gen( IS_REAL( NODE(List)->Type ) ? fifth::FNEG : fifth::NEG );
              break;
	    case S_COM:
              Compile ( Node, RVALUE );
              Gen ( Target == IGNORE ? fifth::DROP : fifth::COM );
              break;
	    case S_RETURN:
              if ( Procedure->Ident == S_NEW ) 
		  Gen (SELF );
              else Reverse ( Node );
              Gen( IS_SELF_RETURN(CADDR(Procedure->Profile)) ? fifth::SRET 
                   : Procedure->Level ? fifth::RET : fifth::FRET );
              Return = TRUE;
              goto Exit;
	    case S_LET:
              Reverse ( CADR( Node ));
              for ( Node = CAR( Node ); Node != NIL; Node = CDR( Node ))
                  Compile( Node, LVALUE );
              goto Exit;
	    case S_FUNCTION:
              { 
		  unsigned Offs = Sym_Tab.Var( ATOM(CAR(Node))->Index ).Disp;
		  nat4 Index = Procedure->Const->N_Elems;
		  Procedure->Const = Procedure->Const->Push(NODE(List)->Type);
		  if ( NODE(Node = CDDR(Node))->Type == STRING ) {
		      Compile ( Node, RVALUE );
		      Gen_Offs ( fifth::COMP, Index );
		  } else {
		      if ( Index < 256 ) Gen_Byte ( fifth::FCLU, Index );
		      else Gen_Offs ( fifth::FCLU2, Index );
		  }
		  Move( fifth::STL, Offs );
              }
              return FALSE;
	    case S_LAMBDA:
              { 
		  nat4 Index = Procedure->Const->N_Elems;
		  Procedure->Const = Procedure->Const->Push(NODE(List)->Type);
		  if ( NODE(Node = CDR(Node))->Type == STRING ) {
		      Compile ( Node, RVALUE );
		      Gen_Offs ( fifth::COMP, Index );
		  } else {
		      if ( Index < 256 ) 
			  Gen_Byte ( fifth::FCLU, Index );
		      else 
			  Gen_Offs ( fifth::FCLU2, Index );
		  }
		  if ( Target == IGNORE ) 
		      Gen ( fifth::DROP );
              }
              break;
	    case S_PACKAGE:
	      for (Node = CADR (CDDR (Node)); Node != NIL; Node = CDR (Node))
                  Compile (Node, RVALUE);
	      // nobreak ?
	    case S_MDEF:
              { 
		  Ctk_Member* Init = (Ctk_Member*) NODE(List)->Type;
		  if ( Init ) 
		      Send ( Init->Val );
		  return FALSE;
              }
	    case S_REDEFINE:
	    case S_CLASS:
	    case S_PROTOCOL:
              return FALSE;
	    case S_CNV:
              { 
		  Ctk_Object From = NODE(Node)->Type, To = CDR(Node);
		  Compile ( Node, RVALUE );
		  if ( IS_BUILTIN(To) && IS_NUMBER(From) ) {
		      if ( IS_INT(To) ) {
			  Gen_Offs ( fifth::CNV, Ctk_Integer_Desc ); 
			  Gen ( fifth::LDOB1 );
		      } else if ( IS_REAL(To) ) {
			  Gen_Offs ( fifth::CNV, Ctk_Float_Desc ); 
			  Gen_Byte( fifth::LDOFn, 4 );
		      }
		  }
		  else if( IS_REAL(From) && IS_INT(To) )  
		      Gen( fifth::FTOI );
		  else if( IS_INT(From) && IS_REAL(To) )  
		      Gen( fifth::ITOF );
		  else if( Ctk_Implicit_Convert( From, To ) >= 0 ) {
		      if ( IS_PROC( To ) ) {
			  int Index = Procedure->Const->N_Elems;
			  Procedure->Const = Procedure->Const->Push( To );
			  Gen_Offs( fifth::CNVP, Index );
		      } else Gen_Offs( fifth::CNV, ((Ctk_Class*)To)->Hdr_Templ.Desc );
		  } else  Ctk_Error( MSG("illegal conversion") );
		  if ( Target == IGNORE ) {
		      Gen (fifth::DROP );
		      DOUBLE( if( IS_REAL( To ) ) Gen( fifth::DROP );  );
		  }
              }
              break;
	    case S_BLOCK:
              while ( Node != NIL ) {
		  Return = Compile (Node, IGNORE);
		  Node = CDR( Node );
              }
              return Return;
            case S_INIT:
              { 
		  Ctk_Class* Class = Procedure->Object;
		  while ( Node != NIL ) {
		      Ctk_Object Init = CAR( Node );
		      nat2 Base;

		      Base = ((Ctk_Class*)NODE(Init)->Type)->Hdr_Templ.Desc;
		      Reverse ( CDR( Init ) );
		      Gen ( SELF );
		      int Offs = Class->Base_Offset ( Base );
		      if ( Offs != 0 ) 
			  Gen_Offs ( fifth::INC, Offs );
		      Send ( ((Ctk_Member*)CAR(Init))->Val );
		      Gen ( fifth::DROP );
		      Node = CDR( Node );
		  }
              }
              goto Exit;
	    case S_NEW:
              { 
		  Ctk_Class* Class = (Ctk_Class*) NODE(Node)->Type;
		  Ctk_Member* Cons = (Ctk_Member*) CAR(Node);
		  Reverse ( CDR( Node ) );
		  if ( !( Class->Status & Ctk_Class::Builtin ) )  {
		      if ( Class->Status & Ctk_Class::No_Derive ) {
			  if ( Class->Name == S_DBSLST 
			       || Class->Name == S_DBSVEC 
			       && unsigned(Class->Hdr_Templ.Desc) 
			       - Dbs_Vector_I1_Desc 
			       >= Dbs_Vector_O_Desc - Dbs_Vector_I1_Desc )
			      Gen_Offs ( fifth::SPEC, Class->Hdr_Templ.Desc );
			  Send ( Cons->Val );
		      } else {
			  Gen_Offs ( ( Class->Status & Ctk_Class::Database )
				     ? fifth::DBSNEW : fifth::NEW, Class->Hdr_Templ.Desc );
			  if ( Cons ) Send ( Cons->Val );
		      }
		  }
		  if ( Target == IGNORE ) 
		      Gen ( fifth::DROP );
		  break;
              }
	    case S_CALL:
              { 
		  Ctk_Object Prof = NODE(List)->Type;
		  Reverse ( Node );
		  if ( !IS_PROC( Prof ) ) { // method invokation
		      Ctk_Member* Method = (Ctk_Member*) Prof;
		      Prof = Method->Type;
		      unsigned Class = Method->Cls;

		      if( Method->Class()->With == System ) {
			  nat2 Class_Name = Method->Class()->Name;
			  switch( Method->Slc ) {
			    case S_SIZE:
			      if( Class_Name == S_ARRAY ) {
				  Move( fifth::LDO,
					offsetof(Ctk_Array_O,Vector));
				  Gen( fifth::LDOB1 );
				  goto ReturnValues;
			      } else if ( Class_Name == S_DBSVEC ) {
				  Moveb( fifth::LDOB,
					 offsetof(Dbs_Vector_O,N_Elems) );
				  goto ReturnValues;
			      }
			      break;
			    case S_INDEX:
			      if (WITHIN(Class,
					 Ctk_Array_I1_Desc, Ctk_Array_O_Desc)){
				  Gen(fifth::LDV_I1+Class-Ctk_Array_I1_Desc);
				  goto ReturnValues;
			      } else if (Class_Name == S_ARRAY) {
				  Gen(fifth::LDV_O);
				  goto ReturnValues;
			      }
			      break;
			    case S_DEFINED:
			      if( Class == Ctk_Header_Desc ) {
				  Gen( fifth::IFDEF );
				  goto ReturnValues;
			      }
			  }
		      } // endif system class
		      
		      if (Method->Atr & Ctk_Member::Protocol)
			  Gen_Offs_Offs( fifth::INVOKE, Method->Off,
					 Method->Cls);
		      else
			  Send( Method->Val );
		  } else {
		      Gen( fifth::CALL ); // function call
		  }
		ReturnValues:
		  Ctk_Object Result = CADDR( Prof );
		  if ( Target == IGNORE ) {
		      if ( IS_STRUCT( Result ) ) {
			  while ( ( Result = CDR( Result ) ) != NIL ) {
			      Gen( fifth::DROP );
			      DOUBLE( if( IS_REAL( CAR( Result ) ) )
				      Gen( fifth::DROP ); 
				  );
			  }
		      } else if ( Result != VOID ) { 
			  Gen( fifth::DROP );
			  DOUBLE( if( IS_REAL(Result) ) Gen( fifth::DROP ); );
		      }
		  }
		  break;
              }
	    case S_ARRAY:
	      { 
		  Ctk_Class* Array = (Ctk_Class*) NODE(List)->Type;
		  unsigned n; 
		  for ( n = 0; Node != NIL; n ++ ) { 
		      Compile ( Node, RVALUE );
		      Node = CDR( Node );
		  } 
		  if ( n < 128 ) 
		      Gen_Byte ( fifth::LDB, n );
		  else Gen_Imm( fifth::LDI, n );
		  if ( Array->Hdr_Templ.Desc < Ctk_Array_O_Desc )
		      Gen (fifth::NEW_V1 + Array->Hdr_Templ.Desc - Ctk_Array_I1_Desc);
		  else Gen_Offs ( fifth::NEW_VO, Array->Hdr_Templ.Desc );
		  if ( Target == IGNORE ) 
		      Gen ( fifth::DROP );
		  break;
              }
           default: 
	     return FALSE;
	  } // end expressions switch
	  break;
      }
      default: 
	return FALSE;
    }
    
    { 
	nat2 Cvt = NODE( Expr )->Modf;
	switch( Cvt ) { 
	  case Ctk_Direct: break;
	  case Ctk_Integer_Obj: Gen( fifth::ITO ); break;
	  case Ctk_Float_Obj: Gen( fifth::FTO ); break;
	  case Ctk_Multiple:
            { 
		nat2 *Cnv = ((Ctk_Vector_U2*)NODE( Expr )->Type)->Items;
		while( (Cvt = *Cnv++) != 0 ) { 
		    nat2 Off = *Cnv++; 
		    if( Cvt == Ctk_Integer_Obj ) 
			Gen_Byte( fifth::ITOn, Off );
		    else if( Cvt == Ctk_Float_Obj ) 
			Gen_Byte( fifth::FTOn, Off );
		    else
			Gen_Offs_Offs( fifth::INCn, Cvt, Off );
		}
	    }
	    break;
	  default: 
	    Gen_Offs( fifth::INC, Cvt );
	}
    }
  Exit:
    { 
	Ctk_Cur_Line = IS_NODE(List) ? NODE(List)->Line : NODE(Expr)->Line;
	nat4 Num = Ctk_Cur_Line - Procedure->Start;
	if (Num < Procedure->Offs->N_Elems) 
	    Procedure->Offs->Items[ Num ] = Pc;
    }
    return Return;
}

//-----------------------------------------------------------------------------
// Ctk_Proc::Compile -- compile lambda expression. The lambda expression
//			is represented in a list form
//-----------------------------------------------------------------------------

global	void	Ctk_Proc::Compile (Ctk_Object Lambda) {
    int Pos[4];
    Sym_Tab.Enter (Pos);
    Ctk_Proc* Ex_Proc = Procedure;

    // Scan profile of procedure, recognize amount, types and size of arguments
    int First = Sym_Tab.Disp;
    int Insp = Sym_Tab.Deb.Used;
    Ctk_Object Parm = CADR ( CAR ( Lambda ) );
    if ( (Attr & Instance) && Sym_Tab.Level == MTH_LEVEL )
	Sym_Tab.Push_Var ( S_SELF, ANY, Ctk_Variable::Init );
    while ( Parm != NIL ) {
	Ctk_Object Decl = CAR( Parm );  // formal parameters
	Sym_Tab.Push_Var( CDR(Decl) != NIL ? ATOM(CAR(CDR(Decl)))->Ident
			  : int(S_NONAME), CAR(Decl), Ctk_Variable::Init );
	Parm = CDR( Parm );
    }
    // Count total size (in long words) of all parameters
    int N_Param_Words = ( Sym_Tab.Disp - First ) / sizeof(Ctk_Object); 
    if( N_Param_Words >= 256 ) 
	Ctk_Error( "function with too many parameters" );
    Params = N_Param_Words;

    // Mark tree (at this stage Start is used as temporary labels counter,
    //            and Level is counter of nested procedures)
    Start = 0; // number of temporary labels
    Level = 0; 
    Procedure = this;
    ::Mark( CDR(Lambda) );
    Sym_Tab.Cross_Table( Start ); // number of temporary labels
    Const = Ctk_Vector_O::New( 0 );

    // Prepare debugging information
    Start = NODE(CDAR(Lambda))->Line;
    Lambda = CDR( Lambda );
    Offs = Ctk_Vector_U2::New( NODE(Lambda)->Line - Start + 1 );
    Loc_Var = Ctk_Vector_U2::New ( Sym_Tab.Deb.Used - Insp );
    scpy ( Loc_Var->Items, &Sym_Tab.Deb(Insp), Sym_Tab.Deb.Used - Insp );
    Sym_Tab.Deb.Used = Insp;

    // Generate prologue for database methods and compile procedure
    Code_Buf.Reset(), Pc = 0, Try.Used = 0;
    Ctk_Cur_Line = Start;
    if (Ctk_Profile_Flag) 
	Put_Mark( Start );

    if((Attr & Mutator) && Object && (Object->Status & Ctk_Class::Database) 
       && Sym_Tab.Level == MTH_LEVEL && Ident != S_NEW)
    {
        // force locking of modified object in order to provide
	// data consistency: x := x + 1
        Gen(fifth::SMOD); 
    }
    if (::Compile (Lambda, IGNORE) == FALSE && CADDR (Profile) != VOID)
	Ctk_Error (MSG("no return"));
    if (Ident == S_NEW) 
	Gen (SELF);
    Gen (Level ? fifth::RET : fifth::FRET);
    Relocate ();
    Code = Ctk_Vector_U1::New (Pc);
    memcpy (Code->Items, Code_Buf.Get(), Pc);
    if (Try.Used > 0) {
	Catch = Ctk_Vector_I4::New (Try.Used);
	icpy (Catch->Items, Try.Item, Try.Used);
    } else {
	Catch = NULL;
    }
    Frm_Size = Sym_Tab.Disp + offsetof (Ctk_Frame, Local);
    Level = Sym_Tab.Level;
    Procedure = Ex_Proc;
    Frm_Desc = Sym_Tab.Leave (Pos);

    #if 0
        if (!Ctk_Sys_Init) 
	    Dump ();
    #endif
}

