//-----------------------------------------------------------------------------
//   Compiler tables
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <stdlib.h>
#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctcls.hpp"
#include "cttyp.hpp"
#include "ctsym.hpp"
#include "cttab.hpp"
#include "fifth.hpp"

//-----------------------------------------------------------------------------
// Ctk_Sym_Table::Find_Var -- search for a variable
//-----------------------------------------------------------------------------

nat2	Ctk_Sym_Table::Find_Var (nat2 Ident) {
    Ctk_Variable_Name* Var;	

    Var = &Sym (Sym.Used);
    Sym(Bot).Ident = Ident;
    while (Ident != (--Var)->Ident);
    return Var->Index;
}

//-----------------------------------------------------------------------------
// Ctk_Sym_Table::Check_Var -- check for multiple definition of a variable 
//-----------------------------------------------------------------------------

void	Ctk_Sym_Table::Check_Var (nat2 Ident) {
    Ctk_Variable_Name *Beg, *Var;

    Beg = &Sym (Var_1); 
    Var = &Sym (Sym.Used);
    while (Var > Beg) {
	if (Ident == (--Var)->Ident) {
	    Ctk_Error (MSG("multiple definition"), 
		       Ctk_Get_Symbol_Name (Ident));
	}
    }
}

//-----------------------------------------------------------------------------
// Ctk_Sym_Table::Enter -- open the new block on symbol table
//-----------------------------------------------------------------------------

void	Ctk_Sym_Table::Enter (int* Save) {
    Save[0] = Var.Used; 
    Save[1] = Disp; 
    Save[2] = Lab_1; 
    Save[3] = Begin();
    Level += 1;
    Disp = (Level + 1)*sizeof(Ctk_Object);
    Lab_1 = Lab.Used;
}

//-----------------------------------------------------------------------------
// Ctk_Sym_Table::Leave -- close the block 
//-----------------------------------------------------------------------------

nat2	Ctk_Sym_Table::Leave (int* Restore) {

    // Check for references to undefined and definitions of unused labels
    {
	Ctk_Label* Label;

	Label = &Lab(Lab_1);
	for ( int n = Lab.Used - Lab_1; -- n >= 0; Label += 1 ) {
	    if (!Label->Defined) {
		Ctk_Error (MSG("undefined label"),
			   Ctk_Get_Symbol_Name (Label->Ident));
	    } else {
		if (!Label->Referenced) {
		    Ctk_Warning (MSG("unused label"), 
				 Ctk_Get_Symbol_Name (Label->Ident));
		}
	    }
	}
    }

    // Build pointer segments
    nat2 Segment [256]; 
    nat2 *Seg = Segment;
    Ctk_Variable *Loc = &Var( Restore[0] - 1 ), *Last = &Var( Var.Used );
    int Len = Level + 3;
   
    *Seg++ = offsetof( Ctk_Frame, Proc );
    while (TRUE) {
	if (++Loc >= Last) { 
	    *Seg++ = Len; 
	    goto Stop; 
	}
	if (IS_BUILTIN (Loc->Type)) {
	    *Seg++ = Len;
	    do { 
		if (++Loc >= Last) 
		    goto Stop;
	    } while (IS_BUILTIN (Loc->Type));
	    *Seg++ = Loc->Disp + offsetof(Ctk_Frame, Local);
	    Len = 1;
	} else Len += 1;
    }
  Stop:
    int Frame_Size = offsetof(Ctk_Frame,Local) + Disp;
    int Ptr_Num = Seg - Segment;

    // Pop context from symbol table
    Var.Used = Restore[0]; 
    Disp = Restore[1];
    Lab.Used = Lab_1; 
    Lab_1 = Restore[2];
    End (Restore[3]);
    Level -= 1;

    // Try to reuse frame descriptor
    Ctk_Descriptor** dtp = Ctk_Descriptor::Table;
    Ctk_Descriptor** end = dtp + Ctk_Descriptor::T_Size;
    Ctk_Descriptor*  d;
    while (dtp < end) {
	d = *dtp++;
	if (d != NULL 
	    && d->Desc == Ctk_Descriptor_Desc
	    && d->Obj_Size == Frame_Size 
	    && d->Ptr_Segm->N_Elems == Ptr_Num
	    && memcmp((int2*)d->Ptr_Segm->Items, (int2*)Segment, 2*Ptr_Num)==0)
	{
             return d->Hdr_Templ.Desc;
	}
    }

    // Build new frame descriptor
    d = new Ctk_Descriptor;
    d->Obj_Info = 0; 
    d->Obj_Size = Frame_Size;
    d->Ptr_Segm = Ctk_Vector_U2::New(Ptr_Num);
    scpy ((int2*)d->Ptr_Segm->Items, (int2*)Segment, Ptr_Num);
    d->Hdr_Templ.Info = 0;
    rtm_gc_build_descriptor(Segment, Ptr_Num/2, d->Hdr_Templ.Desc);

    return d->Hdr_Templ.Desc;
}

//-----------------------------------------------------------------------------
// Ctk_Sym_Table::Push_Var -- process variable declaration
//-----------------------------------------------------------------------------

nat2 Ctk_Sym_Table::Push_Var ( nat2 Ident, Ctk_Object Type, nat1 Attr ) {

    if (Ident != 0) {
	Check_Var(Ident);
    }

    Sym.Alloc(); 
    Sym->Ident = Ident; 
    Sym->Index = Var.Used;

    Var.Alloc(); 
    Var->Attr = Attr; 
    Var->Disp = Disp;
    Var->Frame = Level; 
    Var->Type = Type;

    Deb.Alloc(); 
    *Deb = Ident;
    Deb.Alloc(); 
    *Deb = (Type->Desc == Ctk_Class_Desc)
	? ((Ctk_Class*)Type)->Hdr_Templ.Desc : nat2(Ctk_Proc_Desc);

    Disp += IS_REAL( Type ) ? sizeof(real) : sizeof(Ctk_Object);
    return Var.Used - 1;
}

//-----------------------------------------------------------------------------
// Ctk_Sym_Table::Push_Iimm -- process reference to integer constant
//-----------------------------------------------------------------------------

void Ctk_Sym_Table::Push_Iimm ( nat2 Ident, int4 Value ) {
    Check_Var (Ident);

    Sym.Alloc(); 
    Sym->Ident = Ident; 
    Sym->Index = Var.Used;

    Var.Alloc(); 
    Var->Disp = Imm.Used; 
    Var->Type = INT;
    Var->Attr = Ctk_Variable::Literal|Ctk_Variable::Const|Ctk_Variable::Init;

    Imm.Alloc(); 
    *Imm = Value;
}

//-----------------------------------------------------------------------------
// Ctk_Sym_Table::Push_Fimm -- process reference to floating point constant
//-----------------------------------------------------------------------------

void Ctk_Sym_Table::Push_Fimm ( nat2 Ident, real Value ) {
    Check_Var (Ident);

    Sym.Alloc(); 
    Sym->Ident = Ident; 
    Sym->Index = Var.Used;

    Var.Alloc(); 
    Var->Disp = Imm.Used; 
    Var->Type = REAL;
    Var->Attr = Ctk_Variable::Literal|Ctk_Variable::Const|Ctk_Variable::Init;

    Imm.Alloc(); 
    *Imm = *(int4*)&Value; 
    Imm.Alloc(); 
    *Imm = *((int4*)&Value+1);
}

//-----------------------------------------------------------------------------
// Ctk_Sym_Table::Ref_Lab -- process reference to label (currently goto only)
//-----------------------------------------------------------------------------

nat2 Ctk_Sym_Table::Ref_Lab ( nat2 Ident ) {
    Ctk_Label* Label = &Lab( Lab_1 );
    int n = Lab.Used - Lab_1; 

    for (int i = 0; i < n; i++, Label++) {
	if ( Label->Ident == Ident ) {
	    int j;    
	    for( j = Sym.Used; Sym(--j).Index > Label->Scope; );
	    if ( Label->Defined && Sym(j).Index != Label->Scope )
		Ctk_Error ( MSG("jump over definition") );
	    Label->Scope = Sym(j).Index;
	    Label->Referenced = TRUE;
	    return i;
	}
    }

    Lab.Alloc ();
    Lab->Ident = Ident; 
    Lab->Scope = Sym->Index;
    Lab->Defined = FALSE; 
    Lab->Referenced = TRUE;
    return n;
}

//-----------------------------------------------------------------------------
// Ctk_Sym_Table::Set_Lab -- process label definition (label:)
//-----------------------------------------------------------------------------

nat2 Ctk_Sym_Table::Set_Lab (nat2 Ident) {
    Ctk_Label* Label;
    int n;

    Label = &Lab( Lab_1 );
    n = Lab.Used - Lab_1; 

    // Search for a label already in table
    for ( int i = 0; i < n; i++, Label++ ) {
	if ( Label->Ident == Ident ) {
	    if ( !Label->Defined ) { 	// label undefined
		if (Sym->Index > Label->Scope) {
		    Ctk_Error ( MSG("jump over definition") );
		}
		Label->Defined = TRUE;
		Label->Scope = Sym->Index;
	    } else {
		// Redefinition
		Ctk_Error (MSG("label redefinition"), 
			   Ctk_Get_Symbol_Name(Ident));
	    }
	    return i;
	}
    }

    Lab.Alloc ();
    Lab->Ident = Ident; 
    Lab->Scope = Sym->Index;
    Lab->Defined = TRUE; 
    Lab->Referenced = FALSE;
    return n;
}

//-----------------------------------------------------------------------------
// Ctk_Sym_Table::Cross_Table
//-----------------------------------------------------------------------------

void 	Ctk_Sym_Table::Cross_Table (int n) {
    n += Lab.Used; 	// temporary + user labels
    if (Ref.Size < n) {
	delete Ref.Item;
	Ref.Item = new int2 [n];
	Ref.Size = n;
    }
    Ref.Used = Lab.Used;
}



