//-----------------------------------------------------------------------------
//   Garbage collector for core memory objects 
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <string.h>
#include <stdio.h>
#include "ctalk.hpp"
#include "mmlib.hpp"
#include "mmlow.hpp"
#include "gcobj.hpp"

#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "cttyp.hpp"
#include "ctnum.hpp"
#include "ctlst.hpp"
#include "ctlsp.hpp"
 
//---- Options ---------------------------------------------------------------

// Uncomment the following line if you want comprehencive debugging facility
//#define GC_DEBUG

// Uncomment the following line if you want to collect statistics
//#define GC_STATISTICS

//#define RTM_NO_OVERSTACK_OPERATIONS 1
//-----------------------------------------------------------------------------

global size_t rtm_free_object::st = 0;

//-----------------------------------------------------------------------------
// rtm_gc_info_block -- statistics on gc memory allocator
//-----------------------------------------------------------------------------

struct rtm_gc_info_block {
    size_t	allocated;	// Number of bytes allocated for gc objects
    size_t	used;		// Number of bytes survived last gc
    size_t	used_before;	// Number of bytes survived last previous gc
    size_t	requested;	// Number of bytes requested since last gc
    size_t	reclaimed;	// Number of bytes reclaimed by last gc
    size_t	to_alloc;	// Limit for number of bytes before next GC

    // Account for garbage collection
    void	gc(size_t allocated, size_t used, size_t reclaimed) {
	used_before = rtm_gc_info_block::used;
	rtm_gc_info_block::allocated = allocated;
	rtm_gc_info_block::used = used;
	rtm_gc_info_block::reclaimed = reclaimed;
    }

    // Interpolate linearly the amount of next allocation, in order to get
    //	d1/d2 ratio of used/allocated. (We assume the linear behavior of
    //	used(requested) function)
    size_t	lerp_allocation(size_t d1, size_t d2);

    // Make debugging info dump
    void	Dump();
};

//-----------------------------------------------------------------------------

// The only instance of rtm_gc_info_block
global	rtm_gc_info_block	rtm_gc_info;

global	void	rtm_gc_info_block::Dump() {
    Trace("Rtm Mem Gc Info: alloc=%#x used=%#x used_before=%#x req=#%x\n",
	  allocated, used, used_before);
    Trace("reclaimed=%#x to_alloc=%#x\n", reclaimed, to_alloc);
}

//-----------------------------------------------------------------------------

global	size_t	rtm_gc_info_block::lerp_allocation(size_t d1, size_t d2) {
    #if 0
        // Level 1 approximation
        // FIXME: implement real linear extrapolation here
    #else
        static int j = 3;
        // Level 0 approximation
        size_t a;
	const size_t limit = 128*1024;

	if (j > 0) {
	    j -= 1;
	    return 1024*1024;
	}
	a = (used * (d2-d1))/d1;
	return (a < limit) ? limit : a;
    #endif
}

local	bool	rtm_gc_low_hook(void*, rtm_hook_list*) {

    rtm_gc_info.requested += rtm_free_object::st;
    rtm_free_object::st = 0;

    if (rtm_gc_info.requested > rtm_gc_info.to_alloc && !rtm_mem_gc_ban) {
	rtm_gc_info.requested = 0;
	rtm_mem_gc();
	// FIXME: change 3/5 to more parameterized expression
	rtm_gc_info.to_alloc = rtm_gc_info.lerp_allocation (2, 3);
	return TRUE;
    }
    return FALSE;
}

//-----------------------------------------------------------------------------
// Type for garbage collector instruction code. We try to use THR code,
// but this is relied upon nonportable GCC-specific technique, and we support
// byte code interpreter also

#ifndef RTM_MEM_GC_BYTECODE
    typedef	void*	gc_cmd;
#else
    typedef	char	gc_cmd;
#endif

//-----------------------------------------------------------------------------
// gc_prog -- garbage collector's program for scanning object
//
// We keep all programs for garbage collector in special objects, we
// avoid duplication and use reference counters for gc program reclamation,
// introducing thereof extra level of garbage collection: garbage collection
// of garbage collector's programs.  
// 
// To speedup the matching for duplicates we keep hash-table. Access to 
// programs by identifier is performed by table of pointers to programs 
// indexed by object identifiers. To slightly speedup the access to gc 
// commands, pointers to first command in program are stored in table instead 
// of pointers to gc program objects. The gc program object could be obtained
// from pointer to first command by prog() method. 
//-----------------------------------------------------------------------------

class	gc_prog	{
    gc_prog*	next;		// Pointer to next program in hash chain
    nat		n_refs;		// Number of classes which use this program
    nat		len;		// Number of gc commands in this program
    unsigned	v_hash;		// Value of hash function
    gc_cmd	code[1];	// First command of gc program. Other commands
				//   (if any) follow this one.
    
    // Hash table. Statically sized.

    static const int h_table_size = 127;
    static gc_prog*	h_table[h_table_size];

    unsigned	hash(); 	// Calculate hash function value for object
    static unsigned	hash(nat n, gc_cmd* c);	// and a separate program

    void	insert();       // Insert this program into appropriate 
				//   hash bucket
    static void	clear_hash();	// Reinit hash table afte garbage collection

    // Gc programs table. Dynamically sized.

    static int		gc_table_size;
    static gc_cmd**	gc_table;
    static void	ensure(int i);  // Ensure that gc_table has size enough to 
                                // contain item with index i

    // Obtain gc program for a given first command's address
    static gc_prog* prog(gc_cmd* first_cmd);

    // Allocation related methods
    static size_t size(nat n);  // Calculate allocation size for gc_prog 
                                //   with n commands
    size_t	size();		// Calculate allocation size for this object
    gc_prog*	forward();	// Get address of gc program after this

    friend class gc_prog_cluster;
    friend void	rtm_gc_recycle_descriptor(nat desc);
    friend void	rtm_gc_build_vector_descriptor(nat desc);
    friend void	rtm_mem_dump_program(gc_cmd* c);
  public:
    // Get first command of gc program for object with class id = desc
    static gc_cmd*	get(nat desc)	{ return gc_table[desc]; }

    static void build(nat len, gc_cmd* cmds, int descr);
};

//-----------------------------------------------------------------------------
// gc_prog_cluster -- memory block for allocation of gc programs
//
// gc program clusters are included in single linked list. 
//-----------------------------------------------------------------------------

class	gc_prog_cluster {
    gc_prog_cluster*	next;	// Next	cluster in global chain
    char*		free;	// Address of first free byte in cluster
    char*		last;	// Address of first byte after the cluster
    
    static gc_prog_cluster*	head;	// Head of all blocks list
    static gc_prog_cluster*	tail;	// Tail of all blocks list
    static size_t		b_room;	// Total of all deallocated bytes
					//   since last gc 

    gc_prog*	first_prog()	// Get address of first program in prog cluster
      { return (gc_prog*)(this + 1); }

    static gc_prog*	alloc(int n_ptrs);
    static void		release(gc_prog* gp);
    static void		mark(void);
    static void		relocate(void);
    static void		sweep(void);

    friend void	rtm_gc_recycle_descriptor(nat desc);
    friend class gc_prog;
};

//=============================================================================

gc_prog*	gc_prog::h_table[gc_prog::h_table_size];
gc_cmd**	gc_prog::gc_table;
int		gc_prog::gc_table_size;

//-----------------------------------------------------------------------------
// gc_prog::hash -- calculate the value of hash function for gc program
//-----------------------------------------------------------------------------

inline	unsigned gc_prog::hash(nat n, gc_cmd* c) {
    unsigned	h;
    int		i;

    for (h = i = n; --i >= 0; ) {
	#ifdef RTM_MEM_GC_BYTECODE
	    h <<= 1;
	#endif
	h += (unsigned)*c++;
    }
    return h % h_table_size;
}

inline	unsigned gc_prog::hash() {
    return v_hash;
}

//-----------------------------------------------------------------------------
// gc_prog::insert -- add gc program to hash table
//-----------------------------------------------------------------------------

inline	void	gc_prog::insert() {
    int i;

    i = hash();
    next = h_table[i];
    h_table[i] = this;
}

//-----------------------------------------------------------------------------
// clear_hash -- reinitialize hash table
//-----------------------------------------------------------------------------

inline	void	gc_prog::clear_hash() {
    memset (h_table, 0, sizeof(h_table));
}

//-----------------------------------------------------------------------------
// gc_prog::ensure -- ensure that gc programs table contains entry for item
//		      with index i
//-----------------------------------------------------------------------------

void	gc_prog::ensure(int i) {
    if (gc_table_size <= i) {
	// Expand table
	const size_t gcs = sizeof(gc_cmd*);
	size_t	s;

	if (i == 0)
	    i = 1;
	s = ALIGN(i, RTM_MEM_CLUSTER_SIZE/gcs);
	if (gc_table_size == 0) {
	    gc_table = (gc_cmd**)rtm_blk_alloc(s*gcs);
	} else {
	    gc_table = (gc_cmd**)rtm_blk_realloc((char*)gc_table, s*gcs);
	}
	memset(gc_table + gc_table_size, 0, (s - gc_table_size)*gcs);
	gc_table_size = s;
    }
}

//-----------------------------------------------------------------------------
// gc_prog::prog -- translate address of first command in the gc program to
//		    address of gc program block
//-----------------------------------------------------------------------------

inline	gc_prog*     gc_prog::prog(gc_cmd* first_cmd) {
    return (gc_prog*)((char*)first_cmd - offsetof(gc_prog, code));
}

//-----------------------------------------------------------------------------
// gc_prog::size -- Calculate allocation size in bytes
//-----------------------------------------------------------------------------

inline size_t gc_prog::size(nat n_cmd) {
    size_t	b_size;

    b_size = sizeof(gc_prog) + (n_cmd - 1)*sizeof(gc_cmd);
    b_size = ALIGN(b_size, sizeof(void*));
    return b_size;
}

inline size_t gc_prog::size() {
    return size(len);
}

//-----------------------------------------------------------------------------
// gc_prog::forward -- get address of next program
//-----------------------------------------------------------------------------

inline gc_prog* gc_prog::forward() {
    return (gc_prog*)((char*)this + size());
}

//-----------------------------------------------------------------------------
// gc_prog::build -- create gc program block
//-----------------------------------------------------------------------------

void	gc_prog::build(nat len, gc_cmd* cmds, int descr) {
    unsigned	h;
    gc_prog*	gcp;

    ensure(descr);

    // If we are redefining program, decrement reference counter for old
    // program
    if (gc_table[descr] != NULL)
	prog(gc_table[descr])->n_refs -= 1;

    // Search for program with exactly the same length and commands
    h = hash(len, cmds);
    for (gcp = h_table[h]; gcp != NULL; gcp = gcp->next) {
	if (len == gcp->len && !memcmp(cmds, gcp->code, len*sizeof(gc_cmd))) {
	    goto reuse;
	}
    }

    // Build new gc program block
    gcp = gc_prog_cluster::alloc(len);
    gcp->len = len;
    gcp->n_refs = 0;
    gcp->v_hash = h;
    memcpy(gcp->code, cmds, len*sizeof(gc_cmd));
    gcp->next = h_table[h];
    h_table[h] = gcp;

  reuse:
    // Make reference to program block
    gcp->n_refs += 1;
    gc_prog::gc_table[descr] = gcp->code;
}

//=============================================================================

gc_prog_cluster*	gc_prog_cluster::head;
gc_prog_cluster*	gc_prog_cluster::tail;
size_t			gc_prog_cluster::b_room;

//-----------------------------------------------------------------------------
// gc_prog_cluster::alloc --- allocate space for gc program block with
//			      n_cmd commands
//-----------------------------------------------------------------------------

gc_prog*	gc_prog_cluster::alloc(int n_cmd) {
    gc_prog_cluster*	clu;
    gc_prog*		gcp;
    size_t		b_size;

    static size_t	bytes_since_gc = 0;

    // Calculate allocation size in bytes
    b_size = gc_prog::size(n_cmd);
    
    ASSERT(b_size <= (RTM_MEM_CLUSTER_SIZE - sizeof(gc_prog_cluster)));

    bytes_since_gc += b_size;
    
    // Search in allocation list
    for (clu = tail; clu != NULL; clu = clu->next) {
	if ((clu->last - clu->free) >= b_size) {
	    gcp = (gc_prog*)clu->free;
	    clu->free += b_size;
	    tail = clu;		// Remember cluster where we found free memory
	    return gcp;
	}
    }

    // Try to reuse space
    if (bytes_since_gc >= RTM_MEM_CLUSTER_SIZE && b_room >= b_size) {
        bytes_since_gc = 0;
	b_room = 0;
	mark();
	gc_prog::clear_hash();
	relocate();
	sweep();
	return alloc(n_cmd);
    }

    // Allocate new cluster
    clu = (gc_prog_cluster*)::xmalloc(RTM_MEM_CLUSTER_SIZE);

    // Link the new cluster into cluster list
    clu->next = NULL;
    if (head == NULL)
	head = clu;
    if (tail != NULL)
	tail->next = clu;
    tail = clu;

    // Initialize free space planning components and allocate gc program block
    clu->last = (char*)clu + RTM_MEM_CLUSTER_SIZE;
    gcp = clu->first_prog();
    clu->free = (char*)gcp + b_size;

    return gcp;
}

//-----------------------------------------------------------------------------
// gc_prog_cluster::release -- account the deallocation of gc program pointed
//			       by gcp
//-----------------------------------------------------------------------------

inline	void	gc_prog_cluster::release(gc_prog* gcp) {
    b_room += gcp->size();
}

//-----------------------------------------------------------------------------
// gc_prog_cluster::mark -- first pass of metacollector
//
// scans the allocation region, assigning new addresses for gc programs with
// non null reference counters. Note that this pass destroys chains of hash
// table by storing new addresses in "next" field of gc_prog. Hash table will
// be reconstructed at the last pass of metacollector. In addition, the
// the information stored in field "last" of gc programs clusters is also
// destroyed. New allocation limits are stored in it. The canonical value will
// be also restored at last pass. 
//-----------------------------------------------------------------------------

void	gc_prog_cluster::mark(void) {
    gc_prog_cluster*	from_cluster;
    gc_prog*		from_program;
    gc_prog_cluster*	to_cluster;
    gc_prog*		to_program;
    size_t		b_size;

    from_cluster = to_cluster = head;
    from_program = to_program = to_cluster->first_prog();

    while (from_cluster != NULL) {
	b_size = from_program->size();
	if (from_program->n_refs != 0) {
	    if (((char*)to_program + b_size) >= to_cluster->last) {
		to_cluster->last = (char*)to_program + b_size;
		to_cluster = to_cluster->next;
		to_program = to_cluster->first_prog();
	    }
	    from_program->next = to_program;
	    to_program = (gc_prog*)((char*)to_program + b_size);
	}
	from_program = (gc_prog*)((char*)from_program + b_size);
	if ((char*)from_program >= from_cluster->free) {
	    from_cluster = from_cluster->next;
	    from_program = from_cluster->first_prog();
	}
    }

    if (to_cluster != NULL) {
	to_cluster->last = (char*)to_program;
	while ((to_cluster = to_cluster->next) != NULL)

	    to_cluster->last = (char*)to_cluster->first_prog();
    }
}

//-----------------------------------------------------------------------------
// gc_prog_cluster::relocate -- middle pass of collector
//-----------------------------------------------------------------------------

void	gc_prog_cluster::relocate(void) {
    gc_cmd	**curr,	**stop;
    gc_cmd	*cmd;

    stop = gc_prog::gc_table;
    curr = stop + gc_prog::gc_table_size;

    while (--curr >= stop) {
	if ((cmd = *curr) != NULL) {
	    *curr = gc_prog::prog(cmd)->next->code;
	}
    }
}

//-----------------------------------------------------------------------------
// gc_prog_cluster::sweep -- last pass of metacollector
//-----------------------------------------------------------------------------

void 	gc_prog_cluster::sweep(void) {
    gc_prog_cluster*	clu;
    gc_prog*		pro;
    size_t		b_size;

    clu = head;
    pro = head->first_prog();
    while (clu != NULL) {
	b_size = pro->size();
	if (pro->n_refs != 0) {
	    memcpy (pro->next, pro, b_size);
	    pro->next->insert();
	}
	pro = (gc_prog*)((char*)pro + b_size);
	if ((char*)pro == clu->free) {
	    clu->free = clu->last;
	    clu->last = (char*)clu + RTM_MEM_CLUSTER_SIZE;
	    clu = clu->next;
	    pro = clu->first_prog();
	}
    }
    // FIXME: here remove excess free clusters
    tail = head;
}

//=============================================================================

#ifdef RTM_MEM_TYPE_GENERATIONAL_GC
// rtm_gc_ptr_class represents "can point to" relation

enum	rtm_gc_set {
    treatened,
    bystander,
    immune
};

class	rtm_gc_ptr_pair;

class	rtm_gc_ptr_class {
    friend class rtm_gc_ptr_pair;

    rtm_gc_ptr_class*	t_next;	// Next item in global list
    rtm_gc_ptr_class*	s_next;	// Next item in stack order
    rtm_gc_ptr_class* 	f_next; // Next item in factor leaders list
    rtm_gc_ptr_class* 	l_next; // 
    rtm_gc_ptr_class*	leader; // Leader node for this node
    rtm_gc_ptr_pair*	pset;	// List of classes we can point to
    rtm_gc_ptr_pair*	bases;	// List of base classes
    Ctk_Descriptor*	meta;	// Pointer to class metaobject
    int			numb;	// Number used for numeration (0 for
				// not visited nodes)
    int			d_link;	// DownLink(this) value
    bool		in_stack;// True iff class is pushed in the stack

    static rtm_gc_ptr_class* t_list; // List of all classes
    static rtm_gc_ptr_class* s_list; // Stack of classes
    static rtm_gc_ptr_class* f_list; // Header of factor node leaders list
    static int		     counter;// Counter of visited nodes

    // Stack primitives
    static void push(rtm_gc_ptr_class* c);
    static rtm_gc_ptr_class* pop();

    void merge(rtm_gc_ptr_class* cls);
    void calc_inheritance(rtm_gc_ptr_class* cls);
    void search();
    static void	factor();

    void Dump_Cluster();
    static void Dump_Clusters();
  public:
    void append(rtm_gc_ptr_class* cls);
    void append_base(rtm_gc_ptr_class* cls);
    rtm_gc_ptr_class(Ctk_Descriptor* meta);
};

class	rtm_gc_ptr_pair {
    friend class rtm_gc_ptr_class;

    rtm_gc_ptr_pair*	next;
    rtm_gc_ptr_class*	cls;
  public:
    rtm_gc_ptr_pair(rtm_gc_ptr_class* cls, rtm_gc_ptr_pair* own);
    rtm_gc_ptr_pair(rtm_gc_ptr_class* cls, rtm_gc_ptr_pair* own, bool base);
};

//-----------------------------------------------------------------------------

rtm_gc_ptr_class* rtm_gc_ptr_class::t_list;
rtm_gc_ptr_class* rtm_gc_ptr_class::s_list;
rtm_gc_ptr_class* rtm_gc_ptr_class::f_list;
int		  rtm_gc_ptr_class::counter;

//-----------------------------------------------------------------------------

void rtm_gc_ptr_class::Dump_Cluster() {
    rtm_gc_ptr_class*	p;

    if (p->meta->Desc == Ctk_Descriptor_Desc) {
	Trace ("Frame descriptor %#x\n", p->meta->Desc);
    } else {
	Ctk_Class* cls;

	cls = (Ctk_Class*)p->meta;
	
    }
}

//-----------------------------------------------------------------------------

void rtm_gc_ptr_class::Dump_Clusters() {
}

//-----------------------------------------------------------------------------
// rtm_gc_ptr_pair::rtm_gc_ptr_pair -- constructor
//-----------------------------------------------------------------------------

rtm_gc_ptr_pair::rtm_gc_ptr_pair(rtm_gc_ptr_class* cls, rtm_gc_ptr_pair* own) {
    rtm_gc_ptr_pair::cls = cls;
    next = own->pset;
    own->pset = this;
}

rtm_gc_ptr_pair::rtm_gc_ptr_pair(
    rtm_gc_ptr_class* cls, rtm_gc_ptr_pair* own, bool base) 
{
    rtm_gc_ptr_pair::cls = cls;
    next = own->bases;
    own->bases = this;
}

//-----------------------------------------------------------------------------
// rtm_gc_ptr_class::rtm_gc_ptr_class -- constructor
//-----------------------------------------------------------------------------

rtm_gc_ptr_class::rtm_gc_ptr_class(Ctk_Descriptor* meta) {
    t_next = t_list;
    t_list = this;
    rtm_gc_ptr_class::meta = meta;
    pset = NULL;
    bases = NULL;
    in_stack = FALSE;
}

//-----------------------------------------------------------------------------
// rtm_gc_ptr_class::push -- push class node on the top of stack
//-----------------------------------------------------------------------------

inline void rtm_gc_ptr_class::push(rtm_gc_ptr_class* c) {
    c->s_next = s_list;
    s_list = c;
    in_stack = TRUE;
}

//-----------------------------------------------------------------------------
// rtm_gc_ptr_class::pop -- pop top node from the stack
//-----------------------------------------------------------------------------

inline rtm_gc_ptr_class* rtm_gc_ptr_class::pop() {
    rtm_gc_ptr_class* c;

    c = s_list;
    s_list = c->s_next;
    in_stack = FALSE;
}

//-----------------------------------------------------------------------------
// rtm_gc_ptr_class::merge
//-----------------------------------------------------------------------------

void	rtm_gc_ptr_class::merge(rtm_gc_ptr_class* cls) {
    while(cls != NULL) {
	append(cls);
	cls = cls->next;
    }
|

//-----------------------------------------------------------------------------
// rtm_gc_ptr_class::append -- account for possible can point relation
//-----------------------------------------------------------------------------

void	rtm_gc_ptr_class::append(rtm_gc_ptr_class* cls) {
    rtm_gc_ptr_pair*	pp;

    for (pp = pset; pp != NULL; pp = pp->next) {
	if (pp->cls == cls)
	    return;
    }
    pp = new rtm_gc_ptr_pair(cls, this);
}

//-----------------------------------------------------------------------------
// rtm_gc_ptr_class::append_base
//-----------------------------------------------------------------------------

void	rtm_gc_ptr_class::append_base(rtm_gc_ptr_class* cls) {
    rtm_gc_ptr_pair*	pp;

    for (pp = bases; pp != NULL; pp = pp->next) {
	if (pp->cls == cls)
	    return;
    }
    pp = new rtm_gc_ptr_pair(cls, this, TRUE);
}

//-----------------------------------------------------------------------------
// rtm_gc_ptr_class::calc_inheritance
//-----------------------------------------------------------------------------

void rtm_gc_ptr_class::calc_inheritance(rtm_gc_ptr_class* cls) {
    rtm_gc_ptr_class*	c;

    if (cls != pset)
	append(cls);

    for (c = bases; c != NULL; c = c->next)
	c->calc_inheritance(cls);
}

//-----------------------------------------------------------------------------
// The following two functions implement the algorithm for finding strongly
// connected components in directed graph. 
// See A.Aho, J.Hopcroft, J.Ullman "The design and
// analysis of computer algorithms" Addison-Wesley 1976.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// rtm_gc_ptr_class::search -- calculates the value of DownLink function for
//		               this node
//-----------------------------------------------------------------------------

void	rtm_gc_ptr_class::search() {
    rtm_gc_ptr_pair*	p;
    rtm_gc_ptr_class*	c;

    d_link = numb = counter++;
    push(this);
    for (p = pset; p != NULL; p = p->next) {
	c = p->cls;
	if (c->numb == 0) {
	    c->search();
	    if (c->d_link < d_link)
		d_link = c->d_link;
	} else {
	    if (c->numb < numb && c->in_stack && c->numb < d_link)
		d_link = c->numb;
	}
    }

    if (d_link == numb) {
	// Build component
	l_next = NULL;
	do {
	    c = pop();
	    c->leader = this;
	    c->l_next = l_next;
	    l_next = c;
	} while (c != this);
	f_next = f_list;
	f_list = this;
    } 
}

//-----------------------------------------------------------------------------
// rtm_gc_ptr_class::factor -- calculate strongly connected components for 
//			       (class, can point to) graph.
//-----------------------------------------------------------------------------

void	rtm_gc_ptr_class::factor() {
    rtm_gc_ptr_class*	c;

    // Make inheritance related pointer closure
    for (c = t_list; c != NULL; c = c->t_next)
	c->calc_inheritance(c->pset);

    counter = 1;
    for (c = t_list; c != NULL; c = c->t_next)
	c->numb = 0;
    f_list = NULL;
    for (c = t_list; c != NULL; c = c->t_next) {
	if (c->numb == 0)
	    c->search();
    }
}

#endif

//=============================================================================

#ifdef RTM_MEM_TIME_GENERATIONAL_GC
#endif

//=============================================================================

#ifdef RTM_MEM_CONCURRENT_GC
#endif

//=============================================================================

class	rtm_mem_region : public Ctk_Header {
  public:
    char*		s_addr;	// Start address of a region
    size_t		r_size;	// Size of region in bytes
    rtm_mem_region* 	next;	// Next item in static chain
    int			flags;
    
    enum {
	dynamic = 0x01
    };

    // Static chain of regions
    static rtm_mem_region* s_chain;

    void   unlink();
    void   scan();
    
    rtm_mem_region(bool conservative, char* addr, size_t size, bool dynamic) 
      : Ctk_Header(conservative ? Ctk_Short_Desc : Ctk_Integer_Desc)
    {
	s_addr = addr;
	r_size = size;
	if (dynamic) {
	    next = NULL;
	    flags = rtm_mem_region::dynamic;
	} else {
	    next = s_chain;
	    s_chain = this;
	    flags = 0;
	}
    }

    ~rtm_mem_region() {
	if ((flags & dynamic) == 0)
	    unlink();
    }

  protected:
    rtm_mem_region(char* addr, size_t size, bool dynamic, nat2 Descr) 
      : Ctk_Header(Descr)
    {
	s_addr = addr;
	r_size = size;
	if (dynamic) {
	    next = NULL;
	    flags = rtm_mem_region::dynamic;
	} else {
	    next = s_chain;
	    s_chain = this;
	    flags = 0;
	}
    }
};

#ifdef RTM_MEM_DYNAMIC_INFERENCE

class	rtm_mem_scan_region : public rtm_mem_region {
  public:
    char*	r_map;		// Region map

    rtm_mem_scan_region (har* addr, size_t size, bool dynamic)
      : rtm_mem_region (addr, size, dynamic, Ctk_Float_Desc)
    {;}  
};

#endif

//=============================================================================

rtm_mem_region* rtm_mem_region::s_chain;

void	rtm_mem_region::unlink() {
    rtm_mem_region** p;
    rtm_mem_region*  r;

    for (p = &s_chain; (r = *p) != this; p = &r->next);
    *p = next;
}

//=============================================================================
// Gc machine
typedef Ctk_Header* object;

enum gc_insn_codes {
    #define DEF_GC_INSN(mnemo, code, class, size) mnemo,
    #include "gcobj.dpp"
    gc_last_insn
};

#ifndef RTM_MEM_GC_BYTECODE
    local void** insn_labs_p;
    #define GC_INSN(code) insn_labs_p[code]
#else
    #define GC_INSN(code) code
#endif

//-----------------------------------------------------------------------------
local	const	char*	const	rtm_gc_cmd_mnem[gc_last_insn] = {
    #define DEF_GC_INSN(mnemo, code, class, size) #mnemo,
    #include "gcobj.dpp"
};

//-----------------------------------------------------------------------------
// rtm_mem_gc_cmd_mnem
//-----------------------------------------------------------------------------

local	char const*	rtm_mem_gc_cmd_mnem(gc_cmd cmd) {
    static char buf [64];
    #ifndef RTM_MEM_GC_BYTECODE
        int i;
	
	for (i = gc_last_insn; --i >= 0; )
	    if (cmd == insn_labs_p[i])
		return rtm_gc_cmd_mnem[i];
    #else
	if (cmd < gc_last_insn)
	    return rtm_gc_cmd_mnem[cmd];
    #endif
    sprintf(buf, "illegal insn code %#lx", (long)cmd);
    return buf;
}

//-----------------------------------------------------------------------------
// rtm_mem_dump_program
//-----------------------------------------------------------------------------

global	void	rtm_mem_dump_program(gc_cmd* c) {
    gc_prog*	prg = gc_prog::prog(c);

    Trace("Dump for program %p\n", c);
    for (int i = 0; i < prg->len; i++)
	Trace(rtm_mem_gc_cmd_mnem(prg->code[i]));
}

//-----------------------------------------------------------------------------
// rtm_gc_conservative_check_1
//-----------------------------------------------------------------------------

#ifndef RTM_MEM_OBJ_ADDR_MASK
#define RTM_MEM_OBJ_ADDR_MASK 0x3
#endif

global	object	rtm_gc_conservative_check_1(object ob) {
    rtm_tag		tg;
    long 		offset;
    rtm_mem_obj_page*   pg;
    int			vl;

    tg = RTM_TAG(ob);
    vl = RTM_TAG_VAL(tg);
    if ((vl & RTM_TAG_SCAN) == 0) {
	return NULL;
    }
    if ((vl & RTM_TAG_CLUSTER) != 0) {
	ob = ob->To_Base();
	if (ob->Info & Ctk_Header::In_Fre)
	    return NULL;
	pg = (rtm_mem_obj_page*)RTM_TAG_ADDR(tg);
	if (ob >= pg->sent)
	    return NULL;
	offset = (char*)ob - (char*)(pg + 1);
	if (offset == 0 || ((offset > 0) && (offset % pg->size) == 0))
	    return ob;
    } else {
	if (((char*)ob == ((char*)((~(RTM_MEM_BLOCK_SIZE - 1))&(long)ob) +
			   sizeof(rtm_mem_obj_page)))
	    && ((vl & RTM_TAG_EDGE) != 0))
	{
	    // FIXME: support for multiple inheritance
	    return ob;
	}
    }
    return NULL;
}

//-----------------------------------------------------------------------------
// rtm_gc_conservative_check
//-----------------------------------------------------------------------------

inline object	rtm_gc_conservative_check(object ob) {
    if (!rtm_mem.in_pool(ob) || ((large)(ob) & RTM_MEM_OBJ_ADDR_MASK) != 0)
	return NULL;
    else
	return	rtm_gc_conservative_check_1(ob);
}

//-----------------------------------------------------------------------------
// rtm_gc_dbg_conservative_check
//-----------------------------------------------------------------------------

global object rtm_gc_dbg_conservative_check(object ob) { 
    if (ob == &Ctk_Nil || ob == &Ctk_True || ob == &Ctk_False
	|| (ob >= Ctk_Atom_Table && ob < Ctk_Atom_Table+S_LAST))
	return ob;
    return rtm_gc_conservative_check(ob);
}

//-----------------------------------------------------------------------------
// start scanning of new object 
//-----------------------------------------------------------------------------

#define gc_process_header() \
pc = gc_prog::get(ob->Desc); pp = (object*)(ob + 1)

//-----------------------------------------------------------------------------
// Object stack operations
//-----------------------------------------------------------------------------

#ifdef __i386__
    #define gc_dcl()
    #define gc_init()
    #define gc_push(val)	asm("pushl %0": :"r"(val))
    #define gc_pop()		asm("popl  %0":"=r"(ob)); gc_process_header()
#else
#ifndef RTM_NO_OVERSTACK_OPERATIONS
    #define gc_dcl()		object qq, *sp
    #ifndef RTM_STACK_GOES_UPWARD
        #define gc_init()	sp = &qq - 4096
        #define gc_push(val)	*--sp = (val)

        #define gc_pop()	ob = *sp++; gc_process_header()
    #else
        #define gc_init()	sp = &qq + 4096
        #define gc_push(val)	*sp++ = (val)
        #define gc_pop()	ob = *--sp; gc_process_header()
    #endif
#else
    #define gc_dcl()		object sp, gc_stack[RTM_GC_STACK_SIZE]
    #ifndef RTM_GC_STACK_SIZE
        #define RTM_GC_STACK_SIZE (8*1024)
    #endif
    #define gc_init()		sp = gc_stack
    #define gc_push(val) \
        if (sp == gc_stack + RTM_GC_STACK_SIZE) gc_mark_loop(val); \
	else *sp++ = (val);
    #define gc_pop() ob = *--sp; gc_process_header()
#endif
#endif

//-----------------------------------------------------------------------------
// Pointer scanning
//-----------------------------------------------------------------------------

#ifdef GC_DEBUG
#define gc_scan_debug() if (!rtm_gc_dbg_conservative_check(ob))*(int*)0=0;
#else
#define gc_scan_debug()
#endif

#define gc_scan() \
  if ((ob = *pp++) != NULL) { \
      gc_scan_debug(); \
      h = *ob; \
      if (h.Info > (Ctk_Header::Mark | Ctk_Header::No_Ptr)) \
	  h = *(ob = ob->To_Base()); \
      if ((h.Info & Ctk_Header::Mark) == 0) { \
	  ob->Info |= Ctk_Header::Mark; \
	  if ((h.Info & Ctk_Header::No_Ptr) == 0) \
	      gc_push(ob);  \
      } \
  }

#define gc_conservative_scan() \
  if ((ob = rtm_gc_conservative_check(*pp++)) != NULL \
      && (ob->Info & Ctk_Header::Mark) == 0) \
  {\
      ob->Info |= Ctk_Header::Mark;\
      if ((ob->Info & Ctk_Header::No_Ptr) == 0){\
	  gc_push(ob);\
      }\
  }


//-----------------------------------------------------------------------------
// gc_mark_loop -- main loop of garbage collector machine
//
// Arguments:
//   object ob	-- "root" object to be scanned 
//-----------------------------------------------------------------------------

local	Ctk_Header	gc_ret(Ctk_Char_Desc);

local	void	gc_mark_loop(object ob) {
    Ctk_Header	h(0);
    object*	pp;
    gc_cmd*	pc;
    gc_dcl();			// Declarations for object stack

    #ifndef RTM_MEM_GC_BYTECODE
        static void* insn_labs[gc_last_insn] = {
	    #define DEF_GC_INSN(mnemo, code, class, size) &&L_##mnemo,
	    #include "gcobj.dpp"
	};
	#define next_gc_insn()	goto **pc++

	// Code used for initialisation
	if (ob == NULL) {
	    insn_labs_p = insn_labs;
	    return;
	}
    #else
	#define next_gc_insn()  continue
    #endif

    gc_init();
    gc_push(&gc_ret);
    gc_process_header();

    while (TRUE) {
	#ifndef RTM_MEM_GC_BYTECODE
	    next_gc_insn();
	    #define DEF_GC_INSN(mnemo, code, class, size) L_##mnemo: code;
	    #include "gcobj.dpp"
	#else
	    switch (*pc++) {
		#define DEF_GC_INSN(mnemo, code, class, size) case mnemo: code;
		#include "gcobj.dpp"
	    }
	#endif
    }
}

//-----------------------------------------------------------------------------
// rtm_gc_conservative_scan -- scan conservatively memory area
//-----------------------------------------------------------------------------

global	void	rtm_gc_conservative_scan(void* base, long len) {
    rtm_mem_region r(TRUE, (char*)base, len*sizeof(object), TRUE);
    r.scan();
}

//-----------------------------------------------------------------------------
// init_insn_codes -- initializer for some static hacks, internal to garbage 
//		      collector. These hacks are pretty ugly. But alternatives
//		      I see now, are even more ugly. 
//
//		      We use some internal use only formats for "objects".
//		      This formats are associated with no pointer (and 
//		      even "no object") classes, objects of which never
//		      will be scanned by collector.
//
//	The following descriptors of builtin types are used in this way:
//      Ctk_Char_Desc, Ctk_Integer_Desc, Ctk_Short_Desc, Ctk_Float_Desc
//-----------------------------------------------------------------------------

local	bool	gc_machine_inited = FALSE;

local	void	init_insn_codes() {
    gc_cmd cmd;

    // Initialize array of label pointers. (We need to do this before any
    // execution of GC_INSN(x)
    #ifndef RTM_MEM_GC_BYTECODE    
	gc_mark_loop (NULL);
    #endif

    // "Return from scanning loop" pseudo op is hided under Ctk_Char
    cmd = GC_INSN(RETURN);
    gc_prog::build (1, &cmd, Ctk_Char_Desc);

    // "Scan pointer area" pseudo op
    cmd = GC_INSN(SCANP);
    gc_prog::build (1, &cmd, Ctk_Integer_Desc);

    // "Scan area conservatively" pseudo op
    cmd = GC_INSN(SCANC);
    gc_prog::build (1, &cmd, Ctk_Short_Desc);

    // "Scan area conservatively with dynamic inference" pseudo op
    cmd = GC_INSN(SCANCI);
    gc_prog::build (1, &cmd, Ctk_Float_Desc);

    // Set "already initialized" flag
    gc_machine_inited = TRUE;

    // Setup garbage collector as low memory hook
    rtm_morecore_hooks.add_hook(rtm_gc_low_hook);
}

//-----------------------------------------------------------------------------
// rtm_mem_region::scan()
//-----------------------------------------------------------------------------

inline void	rtm_mem_region::scan() {
    gc_mark_loop(this);
}

//-----------------------------------------------------------------------------
// rtm_gc_compile_descriptor -- "optimizing" compiler for pointer segment map
// assume that we bump scan pointer automatically at the start of object
//-----------------------------------------------------------------------------

local	void	rtm_gc_compile_descriptor(nat* p_segms, int n_p_segms, nat id)
{
    nat		offset;		// Byte offset from the start of object
    gc_cmd	a_prog[RTM_MEM_CLUSTER_SIZE/sizeof(void*)];
    int		n_insn;
    nat		n;

    #define push_insn(a) a_prog[n_insn++] = a

    if (gc_machine_inited == FALSE)
	init_insn_codes();
    n_insn = 0;
    offset = sizeof(Ctk_Header);		
    while (--n_p_segms >= 0) {

	// Compile skipping of the non-pointer part
	n = (*p_segms - offset) / 4;
	offset = *p_segms++;
	while (n > 16) {
	    push_insn (GC_INSN(SKIP_16));
	    n -= 16;
	}
	if (n > 0)
	    push_insn (GC_INSN(SKIP_1 + 1 - n));

	// Compile pointer part
	n = *p_segms++;
	offset += n*sizeof(object);
	while (n > 16) {
	    push_insn (GC_INSN(SCAN_16));
	    n -= 16;
	}

	// If it is the last pointer segment in object, combine the scan of
	// final pointer with popping object from the stack
	push_insn (GC_INSN ( (n_p_segms ? SCAN_1 + 1 : SCANP_1 + 1) - n));
    }

    gc_prog::build (n_insn, a_prog, id);
    #undef push_insn
}

//-----------------------------------------------------------------------------
// rtm_gc_build_descriptor -- build and register gc descriptor for class with
//			      identifier desc, and pointer map p_map.
//			      pointer map is zero terminated string, with byte
//			      corresponding to each slot. Pointer slots are 
//			      marked with '+'.
//-----------------------------------------------------------------------------

global	void	rtm_gc_build_descriptor (const char* p_map, nat desc) {
    nat        *p_segms, *p_curr;
    size_t	offset, s_offset;
    char	c;

    s_offset = 0;
    offset = sizeof(Ctk_Header);
    p_curr = p_segms = (nat*)alloca (strlen(p_map));
    do {
	c = *p_map++;
	if (c == '+') {
	    if (s_offset == 0)
		s_offset = offset;
	} else {
	    if (s_offset != 0) {
		*p_curr++ = s_offset;
		*p_curr++ = offset - s_offset;
		s_offset = 0;
	    }
	}
	offset += sizeof(void*);
    } while (c != 0);
    rtm_gc_compile_descriptor (p_segms, (p_curr - p_segms)/2, desc); 
}

//-----------------------------------------------------------------------------

global	void	rtm_gc_build_descriptor(nat* segms, int n_segms, nat desc) 
{
    rtm_gc_compile_descriptor (segms, n_segms, desc);
}

//-----------------------------------------------------------------------------
// rtm_gc_build_vector_descriptor -- special routine to compile vector of
//				     objects descriptor. For all types of
//				     objects we keep only one program, which is
//				     at first call of this routine.
//-----------------------------------------------------------------------------

export  void	rtm_gc_build_vector_descriptor(nat desc) {
    // Keep state in this static variables
    static bool		vec_cmd_already_created = FALSE;
    static gc_prog* 	vec_prog = NULL;

    if (vec_cmd_already_created) {
	gc_prog::ensure(desc);
	gc_prog::gc_table[desc] = vec_prog->code;
	vec_prog->n_refs += 1;
    } else {
	gc_cmd cmd = GC_INSN(SCANV);
	gc_prog::build (1, &cmd, desc);
	vec_prog = gc_prog::prog (gc_prog::gc_table[desc]);
	vec_prog->n_refs += 1;	// Additional reference fixes program in memory
				// even if all vector classes will be destroyed
	vec_cmd_already_created = TRUE;
    }
}

//-----------------------------------------------------------------------------
// rtm_gc_recycle_descriptor -- note that descriptor desc is no more in use.
//				It can be reused later for other class, 
//				probably with different gc program
//-----------------------------------------------------------------------------

global	void	rtm_gc_recycle_descriptor(nat desc) {
    gc_prog*	gcp;

    ASSERT(desc < gc_prog::gc_table_size && gc_prog::gc_table[desc] != NULL);
    gcp = gc_prog::prog(gc_prog::gc_table[desc]);
    if (--gcp->n_refs == 0) {
	gc_prog_cluster::release(gcp);
    }
    gc_prog::gc_table[desc] = NULL;
}

//=============================================================================

local rtm_mem_obj_page*	rtm_mem_large_objects;
global	rtm_mem_obj_page_queue obj_page_queue[RTM_MEM_BLOCK_SIZE + 1];
local	bool	obj_page_queues_initialized;

global	rtm_mem_obj_page_queue* const	rtm_free_object::entry[RTM_MEM_BLOCK_SIZE + 1] 
= {
#include "gcobjt.dpp"
};

//-----------------------------------------------------------------------------
// init_obj_page_queues -- initialize page queue structures
//-----------------------------------------------------------------------------

local	void	init_obj_page_queues() {
    size_t	pqn;
    size_t	eff;

    obj_page_queues_initialized = TRUE;
    eff = RTM_MEM_CLUSTER_SIZE - sizeof(Ctk_Header) - sizeof(rtm_mem_obj_page);
    for (pqn = 1; pqn <= RTM_MEM_BLOCK_SIZE; pqn += 1) {
	obj_page_queue[pqn].o_size = pqn;
	obj_page_queue[pqn].o_sent = eff - eff%pqn + sizeof(rtm_mem_obj_page);
    }
}

//-----------------------------------------------------------------------------
// rtm_free_object::small_obj_alloc -- allocation from the pool of small
//				       objects
//-----------------------------------------------------------------------------

global rtm_free_object*	rtm_free_object::small_obj_alloc(size_t size) {
    rtm_mem_obj_page*		opg;
    rtm_mem_obj_page_queue*	opgq;
    rtm_free_object		*p, *q, *r;

    if (obj_page_queues_initialized == FALSE)
	init_obj_page_queues();

    #if 0
        rtm_mem_gc(); // Sometimes we need this for debugging
    #endif

    opg = (rtm_mem_obj_page*)rtm_blk_alloc(RTM_MEM_CLUSTER_SIZE);
    rtm_gc_info.requested += RTM_MEM_CLUSTER_SIZE;
    opgq = entry[size];

    // Build tags
    rtm_tag_set_segm(opg, RTM_MEM_CLUSTER_SIZE/RTM_MEM_BLOCK_SIZE,
		     RTM_CONS_TAG(opg, (RTM_TAG_USED | RTM_TAG_CLUSTER
					| RTM_TAG_SCAN)));

    // Link into page list
    opg->next = opgq->p_free;
    size = opg->size = opgq->o_size;
    opgq->p_free = opg;

    // Build free list
    opg->sent = r = (rtm_free_object*)((char*)opg + opgq->o_sent);
    opgq->o_free = p = (rtm_free_object*)(opg + 1);
    while ((q = (rtm_free_object*)((char*)p + size)) != r) {
	p->Info = Ctk_Header::In_Fre;
	p = p->next = q;
    }
    p->next = NULL;
    p->Info = Ctk_Header::In_Fre;

    return (rtm_free_object*)(opg + 1);
}

//-----------------------------------------------------------------------------
// rtm_free_object::obj_alloc -- allocate large object
//-----------------------------------------------------------------------------

global	Ctk_Header*	rtm_free_object::obj_alloc(size_t size) {
    char*	h;
    rtm_tag	tg;		
    size_t	a_size;		// Allocation size

    #if 0
        rtm_mem_gc();
    #endif

    // Calculate allocation size
    a_size = size + sizeof(rtm_mem_obj_page);
    a_size = ALIGN(a_size, RTM_MEM_BLOCK_SIZE);

    // Request memory
    h = rtm_blk_alloc(a_size);
    rtm_gc_info.requested += a_size;

    // Initialize object area
    memset (h + 1, 0, a_size - sizeof(rtm_mem_obj_page));

    // Set tags for allocated region
    tg = RTM_CONS_TAG(a_size, (RTM_TAG_USED | RTM_TAG_SCAN));
    if (a_size > RTM_MEM_CLUSTER_SIZE) {
	rtm_tag_set_segm(h, RTM_MEM_CLUSTER_SIZE/RTM_MEM_BLOCK_SIZE, tg);
	RTM_TAG(h + a_size - 1) = tg;
	RTM_TAG(h) |= RTM_TAG_EDGE;
    } else {
	rtm_tag_set_segm(h, a_size/RTM_MEM_BLOCK_SIZE, tg);
	RTM_TAG(h) |= RTM_TAG_EDGE;
    }

    // Initialize page header
    ((rtm_mem_obj_page*)h)->next = rtm_mem_large_objects;
    ((rtm_mem_obj_page*)h)->size = size;
    rtm_mem_large_objects = (rtm_mem_obj_page*)h;

    return (Ctk_Header*)(h + sizeof(rtm_mem_obj_page));
}

//-----------------------------------------------------------------------------
// rtm_free_object::realloc
//-----------------------------------------------------------------------------

Ctk_Header*  rtm_free_object::realloc(Ctk_Header* obj, size_t new_size) {
    rtm_tag	tg;
    rtm_mem_obj_page*	opg;
    size_t	a_size;
    Ctk_Header *h;

    tg = RTM_TAG(obj);
    if ((RTM_TAG_VAL(tg) & RTM_TAG_CLUSTER) != 0) { // Small objects
	opg = (rtm_mem_obj_page*)RTM_TAG_ADDR(tg);
	if (opg->size >= new_size) {
	    return obj;
	}
    } else {			// Large objects
	rtm_mem_obj_page* npg;
	size_t 		  n_size;

	opg = (rtm_mem_obj_page*)((char*)obj - sizeof(rtm_mem_obj_page));
	a_size = (size_t)RTM_TAG_ADDR(tg) - sizeof(rtm_mem_obj_page);
	if (a_size >= new_size) {
	    opg->size = new_size;
	    return obj;
	}
	n_size = ALIGN(new_size, RTM_MEM_BLOCK_SIZE);
	npg = (rtm_mem_obj_page*)rtm_blk_realloc ((char*)opg, n_size);
	npg->size = new_size;

	// If reallocation moved memory area
	if (npg != opg) {
	    tg = RTM_CONS_TAG(n_size, (RTM_TAG_USED | RTM_TAG_SCAN));
	    if (n_size > RTM_MEM_CLUSTER_SIZE) {
		rtm_tag_set_segm(npg, RTM_MEM_CLUSTER_SIZE/RTM_MEM_BLOCK_SIZE,
				 tg);
		RTM_TAG((char*)npg + n_size - 1) = tg;
		RTM_TAG(npg) |= RTM_TAG_EDGE;
	    } else {
		rtm_tag_set_segm(npg, n_size/RTM_MEM_BLOCK_SIZE, tg);
		RTM_TAG(npg) |= RTM_TAG_EDGE;
	    }
	    free (obj);
	    npg->next = rtm_mem_large_objects;
	    rtm_mem_large_objects = npg;
	    return (Ctk_Header*)(npg + 1);
	}
	return obj;
    }

    // Brute force reallocation
    h = rtm_free_object::alloc (new_size, *obj);
    memcpy (h, obj, opg->size);
    free (obj);
    return h;
}

//-----------------------------------------------------------------------------
// rtm_free_object::free
//-----------------------------------------------------------------------------

global void rtm_free_object::free(void* p) {
    rtm_tag	tg;

    tg = RTM_TAG(p);
    if (RTM_TAG_VAL(tg) & RTM_TAG_CLUSTER) {
	rtm_mem_obj_page_queue* q;
	rtm_free_object*	f;

	f = (rtm_free_object*)p;
	q = obj_page_queue + ((rtm_mem_obj_page*)RTM_TAG_ADDR(tg))->size;
	f->Info = Ctk_Header::In_Fre;
	f->next = q->o_free;
	q->o_free = f;
    } else {
	rtm_mem_obj_page*		opg;
	rtm_mem_obj_page*		opgd;
	rtm_mem_obj_page**		opgp;

	opgd = (rtm_mem_obj_page*)((long)p & (~RTM_MEM_BLOCK_SIZE));

	// hopefully it will be a short search, ...
	// FIXME: we can do it more efficiently
	opgp = &rtm_mem_large_objects; 
	while((opg = *opgp) != opgd) 
	    opgp = &opg->next;
	*opgp = opg->next;
	rtm_blk_free ((char*)p, (size_t)RTM_TAG_ADDR(tg));
    }
}

//-----------------------------------------------------------------------------
// rtm_free_object::alloc_size
//-----------------------------------------------------------------------------

global size_t rtm_free_object::alloc_size(void* p) { 
    rtm_tag tg;

    tg = RTM_TAG(p);
    if (RTM_TAG_VAL(tg)&RTM_TAG_CLUSTER)
	return ((rtm_mem_obj_page*)RTM_TAG_ADDR(tg))->size;
    return (size_t)RTM_TAG_ADDR(tg); 
}

//-----------------------------------------------------------------------------
// rtm_mem_gc_sweep -- the last phase of garbage collector. Unmarks all
//		       marked objects, and deallocates all unmarked.
//		       For large unmarked objects, and clusters with all
//		       small objects umarked, memory is returned to block
//		       pool. All unmarked small objects in nonreturned
//		       clusters are linked into lists, separate list for
//		       each object size.
//-----------------------------------------------------------------------------

void	rtm_mem_gc_sweep() {
    rtm_mem_obj_page*		opg;
    rtm_mem_obj_page**		opgp;
    rtm_mem_obj_page_queue*	opgq;
    rtm_free_object		*p, *q, *r;
    size_t			size;

    rtm_gc_info.allocated = 0;
    rtm_gc_info.used = 0;

    // Scan lists of small block clusters
    for (opgq = obj_page_queue; 
	 opgq <= obj_page_queue + RTM_MEM_BLOCK_SIZE;
	 opgq += 1)
 	{
	    if (opgq->p_free == NULL)
		continue;
	    q = NULL;
	    size = opgq->o_size;
	    for (opg = *(opgp = &opgq->p_free); opg != NULL; opg = *opgp) {
		r = (rtm_free_object*)opg->sent;
		p = (rtm_free_object*)(opg + 1);

		// Special processing for segment of unmarked objects at the
		// start of cluster. If it spans till sentinel the whole
		// cluster should be deallocated

		r->Info = Ctk_Header::Mark; // Setup sentinel
		while ((p->Info&Ctk_Header::Mark) == 0) {
		    p = (rtm_free_object*)((char*)p + size);
		}

		if (p == r) {		// Deallocate cluster
		    *opgp = opg->next;  // Unlink from the cluster chain
		    rtm_blk_free ((char*)opg, RTM_MEM_CLUSTER_SIZE);
		    continue;
		}

		p = (rtm_free_object*)(opg + 1);
		// Something left in the cluster
		do {
		    // Scan segment of consecutive marked objects
		    r->Info = 0;
		    while ((p->Info & Ctk_Header::Mark) != 0) {
			p->Info &= ~Ctk_Header::Mark;
			p = (rtm_free_object*)((char*)p + size);
		    }
		    if (p == r)
			break;

		    // Scan segment of consecutive not marked objects
		    r->Info = Ctk_Header::Mark;
		    do {
			p->next = q;
			q = p;
			p->Info = Ctk_Header::In_Fre;
			p = (rtm_free_object*)((char*)p + size);
		    } while ((p->Info & Ctk_Header::Mark) == 0);
		} while (p != r);
		rtm_gc_info.used += RTM_MEM_CLUSTER_SIZE;
		opgp = &opg->next;
	    }
	    opgq->o_free = q;
	}

    // Scan list of large blocks
    for (opgp = &rtm_mem_large_objects; (opg = *opgp) != NULL; ) {
	p = (rtm_free_object*)(opg + 1);
	size = (size_t)RTM_TAG_ADDR(RTM_TAG(opg));
	if (p->Info & Ctk_Header::Mark) {
	    // Marked object
	    p->Info &= ~Ctk_Header::Mark;
	    rtm_gc_info.used += size;
	    opgp = &opg->next;
	} else {
	    // Not marked object
	    *opgp = opg->next;
	    rtm_blk_free((char*)opg, size);
	}
    }
}

//-----------------------------------------------------------------------------
// interface to low level gc related subroutines

import "C"	void	rtm_low_cpu_regs_scan();
import "C"	void	rtm_low_data_section_scan();
import "C"	void	rtm_low_stack_scan();
import "C"	void	rtm_low_thread_stacks_scan();

//-----------------------------------------------------------------------------
// rtm_mem_gc_mark -- mark all accessible objects
//-----------------------------------------------------------------------------

local	void	rtm_mem_gc_mark() {
    // It is significant to scan process area first because  
    // we should perform Save_Context for current process before GC 
    rtm_low_thread_stacks_scan();

    // Scan descriptor table
    rtm_mem_region r(FALSE, (char*)Ctk_Descriptor::Table,
		     Ctk_Descriptor::T_Size*sizeof(object), TRUE);
    r.scan();

    // Conservatively scan ambigious root areas
    rtm_low_cpu_regs_scan();
    rtm_low_data_section_scan();
    rtm_low_stack_scan();
}
global	bool	rtm_mem_gc_ban = TRUE;

global	rtm_hook_list	rtm_gc_pre_hook;
global	rtm_hook_list	rtm_gc_in_hook;
global	rtm_hook_list	rtm_gc_post_hook;

//-----------------------------------------------------------------------------
// rtm_mem_gc -- garbage collector for memory objects
//-----------------------------------------------------------------------------

global	void	rtm_mem_gc() {
    if (rtm_mem_gc_ban)
	return;
    
    #if 0
        Trace("Garbage collecting ....\n");
    #endif
    rtm_gc_pre_hook.run_hooks();
    rtm_mem_gc_mark();
//->>>> FIXME:
    Ctk_Dbs_Gc_Beg();

    rtm_gc_in_hook.run_hooks();
    rtm_mem_gc_sweep();
    rtm_gc_post_hook.run_hooks();

//->>>> FIXME:
    Ctk_Dbs_Gc_End();
    #if 0
        Trace("...Done\n");
    #endif
}

//-----------------------------------------------------------------------------
// rtm_mem_mark_used_descriptors -- build map of descriptor usage
//		       
//   builds map, each entry of which represents descriptor identifier,
//   and contains value != 0, iff there are allocated objects with such
//   descriptor

global void  rtm_mem_mark_used_descriptors(char* map, size_t map_size) {
    rtm_mem_obj_page*		opg;
    rtm_mem_obj_page_queue*	opgq;
    Ctk_Header			*p, *q;
    size_t			size;

    // Initialize the map
    bclr (map, map_size); 

    // Scan lists of small block clusters
    for (opgq = obj_page_queue; 
	 opgq <= obj_page_queue + RTM_MEM_BLOCK_SIZE;
	 opgq += 1)
    {
	size = opgq->o_size;

	// Visit each cluster
	for (opg = opgq->p_free; opg != NULL; opg = opg->next) {
	    q = (Ctk_Header*)opg->sent;
	    p = (Ctk_Header*)(opg + 1);

	    // And each object
	    do {
		if ((p->Info & Ctk_Header::In_Fre) == 0) {
		    #if 1
		    if (p->Desc >= map_size) {
			abort();
		    }
		    #endif

		    // Mark allocated object
		    map[p->Desc] = 1;
		}
		p = (Ctk_Header*)((char*)p + size);
	    } while (p != q);
	}
    }

    // Scan list of large blocks, all objects in large block list are
    // allocated, so we do not check for free objects
    for (opg = rtm_mem_large_objects; opg != NULL; opg = opg->next) {
	p = (Ctk_Header*)(opg + 1);

	#if 1
	if (p->Desc >= map_size) {
	    abort();
	}
	#endif

	map[p->Desc] = 1;
    }
}

//-----------------------------------------------------------------------------
// for debugging

extern "C" size_t dbg_allocated(void* addr) {
    return ((Ctk_Header*)addr)->Allocated();
}

// Last statement in file 
