//-----------------------------------------------------------------------------
//   Lisp reader 
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "stdtp.hpp"
#include "numlb.hpp"
#include "l2lst.hpp"
#include "memmv.hpp"
#include "ermsg.hpp"
#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "ctlst.hpp"
#include "ctlsp.hpp"
#include "ctnum.hpp"

// TODO move this somewhere
global Ctk_Symbol* lsp_symbol_quote;
global Ctk_Symbol* lsp_symbol_quasiquote;
global Ctk_Symbol* lsp_symbol_unquote_splicing;
global Ctk_Symbol* lsp_symbol_unquote;
global Ctk_Symbol* lsp_symbol_dot;

//-----------------------------------------------------------------------------
// Local variables

#define LSP_BUF_SIZE	256
local	FILE*	lsp_in;
local	const	char* 	(*lsp_read_line)(char*,int,FILE*);
local	const	char*	lsp_read_ptr;
local	const	char*	lsp_base_ptr;
local	char	lsp_buf[LSP_BUF_SIZE];
global	Ctk_Object	lsp_reader ();

//-----------------------------------------------------------------------------
// lsp_string_reader -- 
//-----------------------------------------------------------------------------

local	const 	char*	lsp_string_reader(char*,int,FILE*) {
    const char*	ptr;
    char	chr;

    ptr = lsp_base_ptr;
    while ((chr = *ptr++) != '\n') {
	if (chr == '\0')
	    return NULL;
    }
    lsp_base_ptr = ptr;
    return ptr;
}

//-----------------------------------------------------------------------------
// lsp_skip_whitespace -- skip whitespace characters in the input stream.
//			  returns TRUE if end of file was detected
//-----------------------------------------------------------------------------

local	int	lsp_skip_whitespace() {
    unsigned char c;

    while (TRUE) {
	c = *lsp_read_ptr++;

	// Comment or end of line 
	if (c == ';' || c == 0) {
	    if ((lsp_read_ptr = lsp_read_line(NULL,0,NULL)) == NULL) {
		return -1;
	    } else {
		continue;
	    }
	}

	// Whitespace characters
	if (c != ' ' && c != '\t' && c != '\n' && c != '\r') {
	    return c;
	}
    }
}

//-----------------------------------------------------------------------------
// lsp_read_token
//-----------------------------------------------------------------------------

local	Ctk_Object	lsp_read_token(char first_char) {
    char const* p; char* q; int len;
    // FIXME: make a real token scanner
    lsp_read_ptr -= 1;
    switch (first_char) {
      case '0' : case '1' : case '2' : case '3' : case '4' :
      case '5' : case '6' : case '7' : case '8' : case '9' :
	for (p = lsp_read_ptr; strchr("0123456789.", *p); p ++);
	len = p - lsp_read_ptr;
	q = new char [len + 1];
	memcpy (q, lsp_read_ptr, len);
	q[len] = 0;
	lsp_read_ptr = p;
	if (strchr(q, '.')) {
	    Ctk_Float* flt;

	    flt = new Ctk_Float(0.0);
	    sscanf (q, "%f", &flt->Val);
	    delete q;
	    return flt;
	} else {
	    Ctk_Integer* integ;
	    
	    integ = new Ctk_Integer(0);
	    sscanf (q, "%d", &integ->Val);
	    delete q;
	    return integ;
	}
      default :
	for (p = lsp_read_ptr; 
	     *p != 0 && *p != ' ' && *p != '\t' && *p != '\n'
		 && *p != ')' && *p != ']' && *p != '(' && *p != '['; 
	     p ++);
	len = p - lsp_read_ptr;
	q = new char [len + 1];
	memcpy (q, lsp_read_ptr, len);
	q[len] = 0;
	lsp_read_ptr = p;
	{
	    Ctk_Symbol* sym;

	    sym = Ctk_Symbol::Make_Symbol(q);
	    delete q;
	    return sym;
	}
    }
}

//-----------------------------------------------------------------------------
// lsp_read_string_constant
//-----------------------------------------------------------------------------

local	Ctk_String*	lsp_read_string_constant() {
    char*	p;
    Ctk_String*	str;
    // FIXME: true string constant reading
    
    if ((p = strchr(lsp_read_ptr, '"')) == NULL)
	Ctk_Error ("unterminated string constant");
    str = new Ctk_String(p - lsp_read_ptr);
    memcpy (str->Items(), lsp_read_ptr, p - lsp_read_ptr);
    lsp_read_ptr = p + 1;
    return str;
}

//-----------------------------------------------------------------------------
// lsp_read_list -- reader for lists 
//-----------------------------------------------------------------------------

local	Ctk_Object	lsp_read_list(char delim) {
    Ctk_Pair*	head;		// Head of list being built
    Ctk_Pair*	tail;		// Tail of the list
    Ctk_Object	temp;		// Current item of the list
    int		c;		// Current symbol
    
    head = tail = NULL;
    while (TRUE) {
	c = lsp_skip_whitespace();
	if (c == -1) {
	    Ctk_Error ("Unclosed list\n");
	}
	if (c == delim) { 
	    return head;
	}
	if (c == ')' || c == ']') {
	    Ctk_Error ("list terminator '%c' does not match start\n");
	}
	--lsp_read_ptr;
	temp = lsp_reader();
	if (temp == lsp_symbol_dot) {
	    temp = lsp_reader();
	    c = lsp_skip_whitespace();
	    if (c != delim) {
		Ctk_Error ("Missing '%c'\n", delim);
	    }
	    return tail;
	}
	if (head == NULL) {
	    tail = head = new Ctk_Pair (temp, NULL);
	} else {
	    tail->Cdr = new Ctk_Pair (temp, NULL);
	    tail = (Ctk_Pair*)tail->Cdr;
	}
    }
}

//-----------------------------------------------------------------------------
// lsp_read -- reader for lisp expressions
//-----------------------------------------------------------------------------

global	Ctk_Object	lsp_reader () {
    int		c;
    int		i;
    Ctk_Pair*	pair;
    Ctk_Object	temp;
    Ctk_Array_O*a;

    switch (c = lsp_skip_whitespace()) {
      case -1 :
	    return CTK_NIL;
      case '(' : 
	return lsp_read_list (')');
      case '[' :
	return lsp_read_list (']');
      case '#' :
	switch (c = *lsp_read_ptr++) {
	  case 't' :
	  case 'T' :
	    return CTK_TRUE;
	  case 'f' :
	  case 'F' :
	    return CTK_FALSE;
	  case '(' :
	    // Get elements as a list
	    temp = lsp_read_list (')');

	    // Calculate length of the list
	    for (i = 0, pair = (Ctk_Pair*)temp; 
		 pair != NULL;
		 pair = (Ctk_Pair*)pair->Cdr)
	      {
		 i += 1;
		 if (pair->Desc != Ctk_Pair_Desc)
		     break;
	      }

	    // Convert list to array
	    a = Ctk_Array_O::New(i);
	    for (pair = (Ctk_Pair*)temp, i = 0; pair != NULL;
		 pair = (Ctk_Pair*)pair->Cdr)
	      {
		 if (pair->Desc == Ctk_Pair_Desc) {
		     a->Elem(i) = pair->Car;
		     i += 1;
		 } else {
		     a->Elem(i) = pair;
		     break;
		 }
	      }
	    return a;
	  case '\\' :
	    
	  default :
	    ; // FIXME:: ...
	}
      case '\'':
	return new Ctk_Pair(lsp_symbol_quote, 
			    new Ctk_Pair (lsp_reader(), NULL));
      case '`' :
	return new Ctk_Pair(lsp_symbol_quasiquote, 
			    new Ctk_Pair (lsp_reader(), NULL));
      case ',' :
	if (*lsp_read_ptr == '@') {
	    return new Ctk_Pair(lsp_symbol_unquote_splicing,
				new Ctk_Pair (lsp_reader(), NULL));
	} else {
	    return new Ctk_Pair(lsp_symbol_unquote, 
				new Ctk_Pair (lsp_reader(), NULL));
	}
      case '"' :
	return lsp_read_string_constant();
      default:
	return lsp_read_token(c);
    }
}

//-----------------------------------------------------------------------------

global  void	lsp_open_file_for_reading (const char* F_Name) {
    if ((lsp_in = fopen(F_Name, "r")) == NULL)
	return;
    lsp_read_ptr = lsp_base_ptr = lsp_buf;
    lsp_read_line = (const char * (*)(char *, int, FILE *))fgets;
    lsp_buf[0] = '\0';
}

//-----------------------------------------------------------------------------

global 	void	lsp_open_string_for_reading (Ctk_String* Str) {
    lsp_in = NULL;
    lsp_read_ptr = lsp_base_ptr = Str->Items();
    lsp_read_line = lsp_string_reader;
}

//-----------------------------------------------------------------------------

global	void	lisp_init() {
   lsp_symbol_quote = Ctk_Symbol::Make_Symbol("quote");
   lsp_symbol_quasiquote = Ctk_Symbol::Make_Symbol("quasiquote");
   lsp_symbol_unquote_splicing = Ctk_Symbol::Make_Symbol("unquote-splicing");
   lsp_symbol_unquote = Ctk_Symbol::Make_Symbol("unquote");
   lsp_symbol_dot = Ctk_Symbol::Make_Symbol(".");
}


local Ctk_Primitive
   l0("lisp::Initialize", (void*)&lisp_init, "scheme::lisp"),
   l1("lisp::open_file",  (void*)&lsp_open_file_for_reading, "scheme::lisp"),
   l2("lisp::open_string", (void*)&lsp_open_string_for_reading, 
      "scheme::lisp"),
   l3("lisp::read", 	(void*)&lsp_reader, "scheme::lisp");
