//-----------------------------------------------------------------------------
//   C-Talk main include file
//   Copyright (C) 1993, 1994, 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#ifndef __MTALK_HPP__
#define __MTALK_HPP__

//=============================================================================

// Type for metaobject descriptors. Current setting limits us to 64K
// metaobjects. I do not think this is a problem. But I do not expect it
// will be very difficult to switch to 4 byte (i.e 4G metaobjects)
// representation. 
typedef	nat2	mob_id_t;

// Type for low level object header information. Should be integral type,
// at least 2 bytes. Encoding is explained in mtk_header class definition.
typedef nat2	obj_info_t;

// Common base for all C-Talk objects. If only single inheritance used,
// each C-Talk object contains only one header in the beginning of
// memory area used for the object. In the case of multiple inheritance,
// each minimal element in inheritance lattice has its own base. Conversion
// from any such base to first base could be done using offset in obj_info
// field. Other bits in obj_info field are used only in the first base.
class	mtk_header {
    mob_id_t	mob_id;		// Identifier of this object`s metaobject
    obj_info_t	obj_info;	// Additional low level object information

    // Classes allowed to touch internals mtk_header
    friend class rtm_obj_mem;	// Object memory manager
  public:
    // Bit definitions for obj_info header field. The major part of this
    // field is used for byte offset to first header of object in the case of
    // multilpe inheritance. Since object headers in C-Talk are aligned
    // to at least 4 byte boundary, we can use two low bits. Since it is not
    // enough, we take also the high bit, thus limiting the offset to
    // 32k bytes.
    enum {
	marked	= 0x0001,	// Object is marked by garbage collector.
				// No code outside memory management should
				// use this bit !
	no_ptr	= 0x0002,	// Object contains no pointers to other
				// objects.
	in_free	= 0x8000,	// Object is included in free list. No
				// application accessible objects have this
				// bit set.
	// All special bits in object info
	state	= (marked | no_ptr | in_free),
	offset  = ~state	// Bits used for offset to object origin
    };

    // Get first header of object
    mtk_header*	to_base(void) {
	return (mtk_header*)((char*)this - (info & offset));
    }

    // Get number of bytes allocated for this object
    size_t	allocated(void);

    // Operator new for ordinary objects
    void*	operator new(size_t Size);

    // Disable destructor, since we rely on garbage collector
    void	operator delete(void*) {;}

  protected:
    // Operator new for vector objects
    void*	operator new(size_t Size, size_t Size_V);

    // Constructors
    mtk_header(mtk_header& hdr) {
	*this = hdr;
    }

    mtk_header(mob_id_t	mob_id_val, obj_info_t obj_info_val = 0) {
	mob_id = mob_id_val;
	obj_info = obj_info_val;
    } 

  private:
    // Interface with memory management

    // Test if object is marked
    bool	is_marked(void) {
	return	obj_info & marked;
    }

    // Test if object is in free list or marked
    bool	is_marked_or_free(void) {
	return	(obj_info & (marked | in_free));
    }
};

#include "gcobj.hpp"

inline	size_t	mtk_header::allocated(void) {
    return (size_t)rtm_obj_mem::alloc_size(this);
}

inline	void*	mtk_header::operator new(size_t obj_size) {
    return rtm_obj_mem::alloc(ALIGN(obj_size, sizeof(mtk_header*)));
}

inline	void*	mtk_Header::operator new(size_t Size, size_t Size_V) {
    return rtm_obj_mem::alloc(ALIGN(Size+Size_V, sizeof(mtk_header*)));
}

//=============================================================================

#endif
