//-< TXMOD.HPP >------------------------------------------------------*--------*
// Garret Simple Text Editor                      (c) 1993  GARRET    *     ?  *
//                                                                    *   /\|  *
//                                                                    *  /  \  *
//      Informational Text Model                   Moscow, RUSSIA     * / [] \ *
//                                                                    * GARRET *
//--------------------------------------------------------------------*--------*
// Informational Text Model
// Copyright (C) 1993 Garret.
//
// This file is part of IV - Garret Simple Text Editor
//
// IV is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// IV is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with IV; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: txmod.hpp,v 1.4 1995/12/21 22:11:17 abramow Exp $
//

#ifndef __TXMOD_HPP__
    #define __TXMOD_HPP__

    // Maximal sting size
    #define STR_SIZE        79

    // Forward declarations
    class   View;
    class   Editor;

    // Text string
    struct  Txt_String {
        char    Str[STR_SIZE + 1];      // Contents of the string

        Txt_String();
        Txt_String(Txt_String* s);
        Txt_String(char const* p);
    };

    // Text
    class   Text {
      protected:
        friend class    View;

        Txt_String**    Txt;            // Array of the strings
        int             All;            // Number of allocated strings
        int             Used;           // Number of used strings
        char*           Fnam;           // File where text is placed
        View*           Wnd;            // Text viewer
        int             FFB;            // Fist free byte (for stream model)

        // Invariants:
        //   0 <= All; 0 <= Used <= All; Txt[0], ... Txt[Used-1] - text strings

        virtual void    Flush_Undo();   // Dummy undo method
      public:
        int             Changed;        // 0 if no changes happened,1 otherwise
	int		Max_Line_No;	// Limit for number of lines 

        Text();
        virtual ~Text();

        // Make text empty
        void    Clear();

        // Load text from specified file
        int     Load(char const* Fname);

        // Save text into specified file (initial by default)
        int     Save(char const* Fname = NULL);

        // Insert line
        virtual void    Ins_Line(int Before, Txt_String* s);

        // Delete line
        virtual void    Del_Line(int Line);

        // Insert character
        virtual void    Ins_Char(int Line, int Column, char c);

        // Replace character
        virtual void    Rep_Char(int Line, int Column, char c);

        // Delete character
        virtual void    Del_Char(int Line, int Column);

        // Undo last action
        virtual void    Undo(Editor* e);

        // Redo last undo
        virtual void    Redo(Editor* e);

        // Get number of lines
        int     Lines(){ return Used; }

        // Get symbol
        char    Sym(int Line, int Column){ return Txt[Line]->Str[Column]; }

        // Get line
        Txt_String* Str(int Line){ return Txt[Line]; }

        // Get number of characters in line
        int     Eol(int Line);

        // Append something (stream text model)
        void    Append(char const* p);
    };

    // Text with Undo ability
    class   Text_With_Undo : public Text {
        virtual void    Flush_Undo();   // Undo method
        enum    {ins, rep, del};        // Operation types

        // Size of "undo" buffer (circular implementation)
        #define UNDO_BUF_SIZE   64
        // |<-------------- UNDO_BUF_SIZE ------------>|
        // |<---Before---->|<------After------>|-------|
        // |---------------|-------------------|-------|
        // ^      ^        ^         ^
        // Beg    |        Cur       |
        //        |                  |
        //        |                  |
        //       Undo               Redo
        //      Buffer             Buffer
        //
        // Every element of this circular buffer consists from
        //      1) string contents before operation,
        //      2) Y-coordinate of cursor in the text before operation,
        //      3) X-coordinate of cursor in the text before operation,
        //      4) type of operation ({ins, rep, del}).
        //
        Txt_String*     U_Buf[UNDO_BUF_SIZE];
        int             L_No [UNDO_BUF_SIZE];
        char            X_Pos[UNDO_BUF_SIZE];
        char            Op   [UNDO_BUF_SIZE];
        int             Beg, Cur, Before, After;

        // Put next editor action in the "Undo" buffer
        void            Put_Rec(char Code, int Y, int X);

        // Apply i-th element of "Undo" buffer
        void            Apply(int i, Editor* e);

      public:
        Text_With_Undo();
        ~Text_With_Undo();

        // Editing methods (see class Editor)
        virtual void    Ins_Line(int Before, Txt_String* s);
        virtual void    Del_Line(int Line);
        virtual void    Ins_Char(int Line, int Column, char c);
        virtual void    Rep_Char(int Line, int Column, char c);
        virtual void    Del_Char(int Line, int Column);

        virtual void    Undo(Editor* e);
        virtual void    Redo(Editor* e);
    };

#endif
