//-----------------------------------------------------------------------------
//   Dynamic loading for major a.out formats
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <stdio.h>
#include "stdtp.hpp"
#include "mmlib.hpp"
#include "dldin.hpp"
#include "aout.hpp"

class	rtm_dld_aout_section;

//-----------------------------------------------------------------------------

class	rtm_dld_aout_format : public rtm_dld_format {
    char*	fmt_name;
  public:
    virtual bool detect (char* 	buf_addr, size_t  buf_size, char* name);
    virtual rtm_dld_module* load(FILE* file, char* name, int what);
    virtual const char*	name();

    rtm_dld_aout_format(bool is_native);
};

//-----------------------------------------------------------------------------

class	rtm_dld_aout_module : public rtm_dld_module {
  protected:
    aout_file_header	f_hdr;		// File header
    size_t		text_offset;
    size_t		text_size;
    size_t		data_offset;
    size_t		trel_offset;
    size_t		drel_offset;
    size_t		sym_offset;	// Offset in file to symbol table
    size_t		str_offset;	// Offset in the file to string table
    
    rtm_dld_aout_section* s_data;
    rtm_dld_aout_section* s_text;
    rtm_dld_aout_section* s_bss;

    int		load_symbol_info (int what, FILE* fs);
    int		load_sections (int what, FILE* fs);

    // Run constructors
    virtual	int		run_constructors();

    // Run destructors
    virtual	int		run_destructors();

    friend class rtm_dld_aout_section;
  public:
    // Load additional information
    virtual	int	load(FILE* file, int what);

    rtm_dld_aout_module();
};

//-----------------------------------------------------------------------------

class	rtm_dld_aout_section : public rtm_dld_section {
  protected:
    size_t		rel_offset; // Offset of relocation table within file
    size_t		rel_items;  // Number of items in relocation table
    int			aout_flags; // a.out specific flags

    int		load_reloc_table (FILE* fs);
    int		load_ext_reloc_table(FILE* fs);
  public:
    // Values for a.out specific flags
    enum {
	ext_reloc = 0x0001,	    // Extended relocation format is used
	is_bss	  = 0x0002	    // Section is bss
    };
    rtm_dld_aout_section(rtm_dld_module* v_module, char* v_name, 
			 int v_s_num, int v_f_offs, int v_f_size,
			 size_t v_rel_offset, size_t v_rel_items,
			 int v_aout_flags);

    virtual int		load (FILE* fs, int what);
};

//=============================================================================

//-----------------------------------------------------------------------------

rtm_dld_aout_section::rtm_dld_aout_section(
    rtm_dld_module* v_module, char* v_name, 
    int v_s_num, int v_f_offs, int v_f_size,
    size_t v_rel_offset, size_t v_rel_items,
    int v_aout_flags)
: rtm_dld_section (v_module, v_name, v_s_num, v_f_offs, v_f_size)
{
    rel_offset = v_rel_offset;
    rel_items  = v_rel_items;
    aout_flags = v_aout_flags;
}

//-----------------------------------------------------------------------------

int rtm_dld_aout_section::load_reloc_table (FILE* fs) {
    aout_relocation_info	r_info;
    rtm_dld_symbol**		p_symb;
    size_t			r_add;
    int				i;
    int				r_type;

    static const short	rc_tab [8] = {
	rtm_dld_relocation::rel8_u,	
	rtm_dld_relocation::rel16_u,
	rtm_dld_relocation::rel32_u,
	rtm_dld_relocation::bad_reloc,
	rtm_dld_relocation::rel8_ru,
	rtm_dld_relocation::rel16_ru,
	rtm_dld_relocation::rel32_ru,
	rtm_dld_relocation::bad_reloc
    };

    if (r_alloc < rel_items)
	extend_reloc_table (rel_items - r_alloc);
    r_used = 0;

    if (fseek (fs, rel_offset, SEEK_SET) != 0)
	return -1;

    for (i = rel_items; --i >= 0; ) {
	if (fread (&r_info, 1, sizeof (r_info), fs) != sizeof (r_info))
	    return -1;
	if (r_info.r_extern) {
	    p_symb = NULL;
	    switch (r_info.r_symbolnum & (~N_EXT)) {
	      case N_TEXT:
		r_add = (size_t)((rtm_dld_aout_module*)module)->s_text;
		break;
	      case N_DATA:
		r_add = (size_t)((rtm_dld_aout_module*)module)->s_data;
		break;
	      case N_BSS:
		r_add = (size_t)((rtm_dld_aout_module*)module)->s_bss;
		break;
	      case N_ABS:
		r_add = (size_t)rtm_dld_section::s_abs;
		break;
	      default:
		fprintf (stderr, "Bad fmt\n");
		return -1;
	    }
	} else {
	    r_add = 0;
	    p_symb = module->sym_table + r_info.r_symbolnum;
	}
	r_type = rc_tab [r_info.r_length + (r_info.r_pcrel ? 4 : 0)];
	(void) new (this) rtm_dld_relocation (p_symb, 
					      r_info.r_address, 0, r_type, 0);
    }
}

//-----------------------------------------------------------------------------

int rtm_dld_aout_section::load_ext_reloc_table (FILE* fs) {
    aout_ext_relocation_info	r_info;
    rtm_dld_symbol**		p_symb;
    int				i;
    int				r_type;

    static const short	rc_tab [] = {
     	// simple relocations
	//   RELOC_8,		// data[0:7] = addend + sv
	rtm_dld_relocation::rel8,
	//   RELOC_16,		// data[0:15] = addend + sv
	rtm_dld_relocation::rel16,
	//   RELOC_32,		// data[0:31] = addend + sv
	rtm_dld_relocation::rel32,
	// pc-rel displacement
	//   RELOC_DISP8,	// data[0:7] = addend - pc + sv
	rtm_dld_relocation::rel8_r,
	//   RELOC_DISP16,	// data[0:15] = addend - pc + sv
	rtm_dld_relocation::rel16_r,
	//   RELOC_DISP32,	// data[0:31] = addend - pc + sv
	rtm_dld_relocation::rel32_r,
	// Special
	//   RELOC_WDISP30	// data[0:29] = (addend + sv - pc)>>2
	rtm_dld_relocation::rel30_wr,
	//   RELOC_WDISP22	// data[0:21] = (addend + sv - pc)>>2
	rtm_dld_relocation::rel22_wr,
	//   RELOC_HI22,	// data[0:21] = (addend + sv)>>10
	rtm_dld_relocation::rel22_hi,
	//   RELOC_22,		// data[0:21] = (addend + sv)
	rtm_dld_relocation::rel22,
	//   RELOC_13,		// data[0:12] = (addend + sv)
	rtm_dld_relocation::rel13,
	//   RELOC_LO10,	// data[0:9] = (addend + sv)
	rtm_dld_relocation::rel10,
	//   RELOC_SFA_BASE,		
	rtm_dld_relocation::bad_reloc,
	//   RELOC_SFA_OFF13,
	rtm_dld_relocation::bad_reloc,
	// P.I.C. (base-relative)
	//   RELOC_BASE10,
	rtm_dld_relocation::bad_reloc,
	//   RELOC_BASE13,
	rtm_dld_relocation::bad_reloc,
	//   RELOC_BASE22,
	rtm_dld_relocation::bad_reloc,
	// for some sort of pc-rel P.I.C. (?)
	//   RELOC_PC10,
	rtm_dld_relocation::bad_reloc,
	//   RELOC_PC22,
	rtm_dld_relocation::bad_reloc,
	// P.I.C. jump table
	//   RELOC_JMP_TBL,
	rtm_dld_relocation::bad_reloc,
	// reputedly for shared libraries somehow
	//   RELOC_SEGOFF16,
	rtm_dld_relocation::bad_reloc,
	//   RELOC_GLOB_DAT,
	rtm_dld_relocation::bad_reloc,
	//   RELOC_JMP_SLOT,
	rtm_dld_relocation::bad_reloc,
	//   RELOC_RELATIVE,
	rtm_dld_relocation::bad_reloc,

	//   RELOC_11,	
	rtm_dld_relocation::bad_reloc,
	//   RELOC_WDISP2_14,
	rtm_dld_relocation::bad_reloc,
	//   RELOC_WDISP19,
	rtm_dld_relocation::bad_reloc,
	//   RELOC_HHI22,	// data[0:21] = (addend + sv) >> 42
	rtm_dld_relocation::bad_reloc,
	//   RELOC_HLO10,	// data[0:9] = (addend + sv) >> 32
	rtm_dld_relocation::bad_reloc,
	// 29K relocation types
	//   RELOC_JUMPTARG,
	rtm_dld_relocation::bad_reloc,
	//   RELOC_CONST,
	rtm_dld_relocation::bad_reloc,
	//   RELOC_CONSTH,
	rtm_dld_relocation::bad_reloc,

	// for sparc v9
	//   RELOC_64,		// data[0:63] = addend + sv
	rtm_dld_relocation::bad_reloc,
	//   RELOC_DISP64,	// data[0:63] = addend - pc + sv
	rtm_dld_relocation::bad_reloc,
	//   RELOC_WDISP21,	// data[0:20] = (addend + sv - pc)>>2
	rtm_dld_relocation::bad_reloc,
	//   RELOC_DISP21,	// data[0:20] = addend - pc + sv
	rtm_dld_relocation::bad_reloc,
	//   RELOC_DISP14,	// data[0:13] = addend - pc + sv
	rtm_dld_relocation::bad_reloc,

	//   NO_RELOC
	rtm_dld_relocation::rel_nop,
    };

    if (r_alloc < rel_items)
	extend_reloc_table (rel_items - r_alloc);
    r_used = 0;

    if (fseek (fs, rel_offset, SEEK_SET) != 0)
	return -1;

    for (i = rel_items; --i >= 0; ) {
	if (fread (&r_info, 1, sizeof (r_info), fs) != sizeof (r_info))
	    return -1;
	p_symb = module->sym_table + r_info.r_symbolnum;
	r_type = rc_tab [r_info.r];
	(void) new (this) rtm_dld_relocation (p_symb, 
					      r_info.r_address, 0, r_type, 0);
    }
}

//-----------------------------------------------------------------------------

int rtm_dld_aout_section::load (FILE* fs, int what) {
    int	r_code;

    // Load section contence 
    if (what & text) {
	size_t 	sect_size;

	// Determine in memory section size
	if (aout_flags & is_bss) {
	    sect_size = ((rtm_dld_aout_module*)module)->f_hdr.a_bss;
	} else {
	    sect_size = f_size;
	}

	// Allocate / reallocate enough memory
	if (data == NULL) {
	    m_size = sect_size;
	    data = (char*)xmalloc (sect_size);
	} else {
	    if (m_size != sect_size) {
		m_size = sect_size;
		data = (char*)xrealloc (data, sect_size);
	    }
	}

	// Load contence
	if (aout_flags & is_bss) {
	    memset (data, 0, sect_size);
	} else {
	    if (fseek (fs, f_offs, SEEK_SET) != 0)
		return -1;
	    if (fread (data, 1, sect_size, fs) != sect_size)
		return -1;
	}
    }

    // Load relocation information
    if (what & relocated) {
	if (aout_flags & ext_reloc)
	    r_code = load_ext_reloc_table();
	else
	    r_code = load_reloc_table();
    }

    return r_code;
}

//=============================================================================

bool rtm_dld_aout_format::detect (char* buf_addr, size_t  buf_size, char* name)
{
    aout_file_header*	f_hdr;

    f_hdr = (aout_file_header*)buf_addr;
    return f_hdr->good_magic();
}

//-----------------------------------------------------------------------------

rtm_dld_module* rtm_dld_aout_format::load (FILE* file, char* name, int what) {
    rtm_dld_aout_module* module;

    module = new rtm_dld_aout_module;
    module->load (file, name, what);
    return module;
}

//-----------------------------------------------------------------------------

const char*	rtm_dld_aout_format::name () {
    return fmt_name;
}

//-----------------------------------------------------------------------------

rtm_dld_aout_format::rtm_dld_aout_format (bool is_native)
: rtm_dld_format (is_native)
{
    name = "a.out";
}

//-----------------------------------------------------------------------------

int	rtm_dld_aout_module::load (FILE* file, int what) {
    size_t	rel_size;
    int		rel_flg;

    if ((mask & headers) == 0)
	what |= headers;

    // Load header
    if (what & headers) {
	rewind (file);
	if (fread (&f_hdr, 1, sizeof(f_hdr), file) != sizeof(f_hdr)) {
	    return -1;
	}
	mask |= headers;
    }

    // Calculate addresses of file components
    text_offset = N_TXTOFF(f_hdr);
    text_size = N_TXTSIZE(f_hdr);
    data_offset = text_offset + text_size;
    trel_offset = data_offste + f_hdr.a_data;
    drel_offset = trel_offset + f_hdr.a_trsize;
    sym_offset = drel_offset + f_hdr.a_drsize;
    str_offset = sym_offset + f_hdr.a_syms;

    // FIXME:
    //rel_flg = rtm_dld_out_section::ext_reloc;
    //rel_size = sizeof (aout_ext_relocation_info);
    rel_size= sizeof(aout_relocation_info);
    rel_flg = 0;

    // Create section objects
    if (s_text == NULL) {
	s_text = new rtm_dld_aout_section (this, ".text", 0,
					   text_offset, text_size,
					   trel_offset, 
					   f_hdr.a_trsize/rel_size,
					   rel_flg);
	s_data = new rtm_dld_aout_section (this, ".data", 1,
					   data_offset, f_hdr.a_data,
					   drel_offset, 
					   f_hdr.a_drsize/rel_size,
					   rel_flg);
	s_bss  = new rtm_dld_aout_section (this, ".bss", 2,
					   0, f_hdr.a_bss,
					   0, 0,
					   rtm_dld_aout_section::is_bss);
    }

    // Load symbols
    if (what & symbols) {
	if (load_symbol_info (what, fs) != 0)
	    return -1;
    }

    // Load section contens
    if (what & text) {
	if (s_text->load (file, what) != 0
	    || s_data->load (file, what) != 0
	    || s_bss->load  (file, what))
	{
	    return -1;
	}
    }

    // Relocate
    if ((what & relocated) != 0
	|| resolve () != 0
	|| relocate() != 0)
    {
	return -1;
    }

    // Run constructors
    if (what & constructed) {
	if (run_constructors() != 0)
	    return -1;
    }
    
    return 0;
}

//-----------------------------------------------------------------------------

typedef void (*func_ptr)();

int rtm_dld_aout_module::run_constructors() {
    rtm_dld_symbol_name*	s_name;
    rtm_dld_symbol*		p_symb;
    func_ptr			fptr;

    s_name = loc_names.search (CTOR_LIST_NAME);
    if (s_name != NULL) {
	for (p_symb = s_name->syms; p_symb != NULL; p_symb = p_symb->next) {
	    if (p_symb->get_section() != rtm_dld_section::s_abs) {
		fptr = (func_ptr)p_symb->value();
		fptr();
	    }
	}
    }
    return 0;
}

int rtm_dld_aout_module::run_destructors() {
    rtm_dld_symbol_name*	s_name;
    rtm_dld_symbol*		p_symb;
    func_ptr			fptr;

    s_name = loc_names.search (DOR_LIST_NAME);
    if (s_name != NULL) {
	for (p_symb = s_name->syms; p_symb != NULL; p_symb = p_symb->next) {
	    if (p_symb->get_section() != rtm_dld_section::s_abs) {
		fptr = (func_ptr)p_symb->value();
		fptr();
	    }
	}
    }
    return 0;
}

//-----------------------------------------------------------------------------

int rtm_dld_aout_module::load_symbol_info (int what, FILE* fs) {
    rtm_dld_section*	sect;
    aout_nlist*		p_nlist;
    aout_nlist*		q_nlist;
    int			i;
    short		v_attr;
    short		v_type;
    int			str_table_size;
    char*		str_table;

    if (fseek (fs, sym_offset, SEEK_SET) != 0)
	return -1;
    p_nlist = xmalloc (f_hdr.a_syms);
    if (fread (p_nlist, 1, f_hdr.a_syms, fs) != f_hdr.a_syms)
	goto cleanup;
    if (fseek (fs, str_offset, SEEK_SET) != 0)
	goto cleanup;
    if (fread (&str_table_size, 1, sizeof(int), fs) != sizeof (int))
	goto cleanup;
    str_table_size -= sizeof (int);
    str_table = xmalloc (str_table_size);
    if (fread (str_table, 1, str_table_size, fs) != str_table_size) {
	free (str_table);
	goto cleanup;
    }

    for (i = f_hdr.a_syms/sizeof(aout_nlist), q_nlist = p_nlist;
	 --i >= 0; q_nlist++)
    {
	v_type = 0;
	sect = 0;
	switch (q_nlist->n_type) {
	  case N_UNDF:
	    v_attr = is_undefined;
	    break;
	  case N_UNDF | N_EXT :
	    v_attr = is_undefined | is_global;
	    break;
	  case N_ABS :
	    v_attr = is_local;
	    sect = rtm_dld_section::s_abs;
	    break;
	  case N_ABS  | N_EXT :
	    v_attr = is_global;
	    sect = rtm_dld_section::s_abs;
	    break;
	  case N_TEXT :
	    v_attr = is_local;
	    sect = s_text;
	    break;
	  case N_TEXT | N_EXT :
	    v_attr = is_global;
	    sect = s_text;
	    break;
	  case N_DATA :
	    v_attr = is_local;
	    sect = s_data;
	    break;
	  case N_DATA | N_EXT :
	    v_attr = is_global;
	    sect = s_data;
	    break;
	  case N_BSS :
	    v_attr = is_local;
	    sect = s_bss;
	    break;
	  case N_BSS  | N_EXT :
	    v_attr = is_global;
	    sect = s_bss;
	    break;
	  case N_SETA :
	    v_attr = is_local | is_constructor;
	    sect = rtm_dld_section::s_abs;
	    break;
	  case N_SETT :
	    v_attr = is_local | is_constructor;
	    sect = s_text;
	    break;
	  case N_SETD :
	    v_attr = is_local | is_constructor;
	    sect = s_data;
	    break;
	  case N_SETB :
	    v_attr = is_local | is_constructor;
	    sect = s_bss;
	    break;
	  case N_SETV :
	    continue;
	  case N_COMM :
	  case N_COMM | N_EXT : //???
	    
	  case N_TYPE :
	  case N_INDR :

	  case N_WARNING :
	  case N_WEAKU :
	  case N_WEAKA :
	  case N_WEAKT :
	  case N_WEAKD :
	  case N_WEAKB :

	  case N_FN :
	  case N_FN_SEQ :
	  default :
	    if ((q_nlist->n_type & N_STAB) 
		|| q_nlist->n_type == N_FN
		|| q_nlist->n_type == N_FN_SEQ) 
	    {
		if (what & rtm_dld_module::debug) {

		}
	    }
	    continue;
	}
	// NOT SURE ...
	name = str_table + q_nlist->n_strx;
	(void)new (this, v_attr) rtm_dld_symbol (name, v_attr, v_type,
						 q_nlist->n_value, this, sect);
    }
    
    free (str_table);
    free (p_nlist);
    return 0;

  cleanup:
    free (p_nlist);
    return -1;
}

