//-----------------------------------------------------------------------------
//   Description of a.out family formats
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

// This file is remake of GNU BFD a.out support

#ifndef __AOUT_HPP__
#define __AOUT_HPP__

// This is the layout on disk of the 32-bit or 64-bit exec header.
struct aout_file_header {
    unsigned	int	a_info;	  // Magic info packed
    unsigned	long	a_text;	  // Length of text, in bytes
    unsigned	long	a_data;	  // Length of data, in bytes
    unsigned	long	a_bss;	  // Length of uninitialized data, in bytes
    unsigned	long	a_syms;	  // Length of symbol table data, in bytes
    unsigned	long	a_entry;  // Start address
    unsigned	long	a_trsize; // Length of reloc. info for text, in bytes
    unsigned	long	a_drsize; // Length of reloc. info for data, in bytes
};

#define EXEC_BYTES_SIZE	sizeof(aout_file_header)

enum aout_machine_type {
    M_OLDSUN2	= 0,
    M_68010	= 1,
    M_68020	= 2,
    M_SPARC	= 3,
    M_386	= 100,
    M_29K	= 101,		// AMD 29000
    M_386_DYNIX = 102,		// Sequent running dynix
    M_386_NETBSD= 134,		// NetBSD/386 binary
    M_MIPS1	= 151,		// MIPS R2000/R3000 binary
    M_MIPS2	= 152,		// MIPS R4000/R6000 binary
    M_HP200	= 200,		// HP 200 (68010) BSD binary
    M_HP300	= (300%256),	// HP 300 (68020+68881) BSD binary
    M_HPUX	= (0x20c%256)	// HP 200/300 HPUX binary
};

// Magic numbers for a.out files
#if RTM_ARCH_SIZE == 64
enum aout_magic {
    OMAGIC	= 0x1001,	// Code indicating object file
    ZMAGIC	= 0x1002,	// Code indicating demand-paged executable
    NMAGIC	= 0x1003	// Code indicating pure executable
};

#define N_BADMAG(x)	  (N_MAGIC(x) != OMAGIC		\
			&& N_MAGIC(x) != NMAGIC		\
  			&& N_MAGIC(x) != ZMAGIC)
#else
enum aout_magic {
    OMAGIG	= 0407,		// object file or impure executable
    NMAGIC	= 0410,		// pure executable
    ZMAGIC	= 0413,		// demand-paged executable
    QMAGIC	= 0314,		// demand-paged executable with the header 
				// in the text. The first page is unmapped
                                // to help trap NULL pointer references
    CMAGIC	= 0421		// core file
};

#define N_BADMAG(x)	  (N_MAGIC(x) != OMAGIC		\
			&& N_MAGIC(x) != NMAGIC		\
  			&& N_MAGIC(x) != ZMAGIC		\
		        && N_MAGIC(x) != QMAGIC)
#endif

#ifdef QMAGIC
#define N_IS_QMAGIC(x) (N_MAGIC (x) == QMAGIC)
#else
#define N_IS_QMAGIC(x) (0)
#endif

// The difference between PAGE_SIZE and N_SEGSIZE is that PAGE_SIZE is
// the finest granularity at which you can page something, thus it
// controls the padding (if any) before the text segment of a ZMAGIC
// file.  N_SEGSIZE is the resolution at which things can be marked as
// read-only versus read/write, so it controls the padding between the
// text segment and the data segment (in memory; on disk the padding
// between them is PAGE_SIZE).  PAGE_SIZE and N_SEGSIZE are the same
// for most machines, but different for sun3.

// By default, segment size is constant.  But some machines override this
// to be a function of the a.out header (e.g. machine type).

#ifndef	N_SEGSIZE
#define	N_SEGSIZE(x)	SEGMENT_SIZE
#endif

//=============================================================================
// Virtual memory address of the text section.
// This is getting very complicated.  A good reason to discard a.out format
// for something that specifies these fields explicitly.  But til then...
//
// * OMAGIC and NMAGIC files:
//     (object files: text for "relocatable addr 0" right after the header)
//     start at 0, offset is EXEC_BYTES_SIZE, size as stated.
// * The text address, offset, and size of ZMAGIC files depend
//   on the entry point of the file:
//   * entry point below TEXT_START_ADDR:
//     (hack for SunOS shared libraries)
//     start at 0, offset is 0, size as stated.
//   * If N_HEADER_IN_TEXT(x) is true (which defaults to being the
//     case when the entry point is EXEC_BYTES_SIZE or further into a page):
//     no padding is needed; text can start after exec header.  Sun
//     considers the text segment of such files to include the exec header;
//     for BFD's purposes, we don't, which makes more work for us.
//     start at TEXT_START_ADDR + EXEC_BYTES_SIZE, offset is EXEC_BYTES_SIZE,
//     size as stated minus EXEC_BYTES_SIZE.
//   * If N_HEADER_IN_TEXT(x) is false (which defaults to being the case when
//     the entry point is less than EXEC_BYTES_SIZE into a page (e.g. page
//     aligned)): (padding is needed so that text can start at a page boundary)
//     start at TEXT_START_ADDR, offset PAGE_SIZE, size as stated.
//
//  Specific configurations may want to hardwire N_HEADER_IN_TEXT,
//  for efficiency or to allow people to play games with the entry point.
//  In that case, you would #define N_HEADER_IN_TEXT(x) as 1 for sunos,
//  and as 0 for most other hosts (Sony News, Vax Ultrix, etc).
//  (Do this in the appropriate bfd target file.)
//  (The default is a heuristic that will break if people try changing
//  the entry point, perhaps with the ld -e flag.)
//
//  * QMAGIC is always like a ZMAGIC for which N_HEADER_IN_TEXT is true,
//  and for which the starting address is PAGE_SIZE (or should this be
//  SEGMENT_SIZE?) (TEXT_START_ADDR only applies to ZMAGIC, not to QMAGIC).
    
// This macro is only relevant for ZMAGIC files; QMAGIC always has the header
// in the text.
#ifndef N_HEADER_IN_TEXT
#define N_HEADER_IN_TEXT(x) (((x).a_entry & (PAGE_SIZE-1)) >= EXEC_BYTES_SIZE)
#endif

// Sun shared libraries, not linux.  This macro is only relevant for ZMAGIC
// files.
#ifndef N_SHARED_LIB
#define N_SHARED_LIB(x) ((x).a_entry < TEXT_START_ADDR)
#endif

// Returning 0 not TEXT_START_ADDR for OMAGIC and NMAGIC is based on
// the assumption that we are dealing with a .o file, not an
// executable.  This is necessary for OMAGIC (but means we don't work
// right on the output from ld -N); more questionable for NMAGIC.

#ifndef N_TXTADDR
#define N_TXTADDR(x) \
    (/* The address of a QMAGIC file is always one page in, */ \
     /* with the header in the text.  */ \
     N_IS_QMAGIC (x) ? PAGE_SIZE + EXEC_BYTES_SIZE : \
     N_MAGIC(x) != ZMAGIC ? 0 :	/* object file or NMAGIC */\
     N_SHARED_LIB(x) ? 0 :	\
     N_HEADER_IN_TEXT(x)  ?	\
	    TEXT_START_ADDR + EXEC_BYTES_SIZE :	/* no padding */\
	    TEXT_START_ADDR			/* a page of padding */\
    )
#endif

// If N_HEADER_IN_TEXT is not true for ZMAGIC, there is some padding
// to make the text segment start at a certain boundary.  For most
// systems, this boundary is PAGE_SIZE.  But for Linux, in the
// time-honored tradition of crazy ZMAGIC hacks, it is 1024 which is
// not what PAGE_SIZE needs to be for QMAGIC.

#ifndef ZMAGIC_DISK_BLOCK_SIZE
#define ZMAGIC_DISK_BLOCK_SIZE PAGE_SIZE
#endif

#define N_DISK_BLOCK_SIZE(x) \
  (N_MAGIC(x) == ZMAGIC ? ZMAGIC_DISK_BLOCK_SIZE : PAGE_SIZE)

// Offset in an a.out of the start of the text section.
#ifndef N_TXTOFF
#define N_TXTOFF(x)	\
    (/* For {O,N,Q}MAGIC, no padding.  */ \
     N_MAGIC(x) != ZMAGIC ? EXEC_BYTES_SIZE : \
     N_SHARED_LIB(x) ? 0 : \
     N_HEADER_IN_TEXT(x) ?	\
	    EXEC_BYTES_SIZE :			/* no padding */\
	    ZMAGIC_DISK_BLOCK_SIZE		/* a page of padding */\
    )
#endif
// Size of the text section.  It's always as stated, except that we
// offset it to `undo' the adjustment to N_TXTADDR and N_TXTOFF
// for ZMAGIC files that nominally include the exec header
// as part of the first page of text.  (BFD doesn't consider the
// exec header to be part of the text segment.)
#ifndef N_TXTSIZE
#define	N_TXTSIZE(x) \
    (/* For QMAGIC, we don't consider the header part of the text section.  */\
     N_IS_QMAGIC (x) ? (x).a_text - EXEC_BYTES_SIZE : \
     (N_MAGIC(x) != ZMAGIC || N_SHARED_LIB(x)) ? (x).a_text : \
     N_HEADER_IN_TEXT(x)  ?	\
	    (x).a_text - EXEC_BYTES_SIZE:	/* no padding */\
	    (x).a_text				/* a page of padding */\
    )
#endif

//=============================================================================
// Symbols

struct aout_nlist {
    unsigned long	n_strx;	// index into string table of name 
    unsigned char	n_type;	// type of symbol 
    unsigned char	n_other;// misc info (usually empty)
    unsigned short	n_desc;	// description field 
    unsigned long	n_value;// value of symbol
};

// The n_type field is the symbol type, containing:

enum aout_nlist_type {
    N_UNDF	= 0,	// Undefined symbol
    N_ABS 	= 2,	// Absolute symbol -- defined at particular addr
    N_TEXT 	= 4,	// Text sym -- defined at offset in text seg
    N_DATA 	= 6,	// Data sym -- defined at offset in data seg
    N_BSS 	= 8,	// BSS  sym -- defined at offset in zero'd seg
    N_COMM	= 0x12,	// Common symbol (visible after shared lib dynlink)
    N_FN	= 0x1f,	// File name of .o file
    N_FN_SEQ	= 0x0C,	// N_FN from Sequent compilers (sigh)

    // Note: N_EXT can only be usefully OR-ed with N_UNDF, N_ABS, N_TEXT,
    // N_DATA, or N_BSS.  When the low-order bit of other types is set,
    // (e.g. N_WARNING versus N_FN), they are two different types.
    N_EXT 	= 1,	// External symbol (as opposed to local-to-this-file)

    N_TYPE	= 0x1e,
    N_STAB 	= 0xe0,	// If any of these bits are on, it's a debug symbol

    N_INDR	= 0x0a,

    // The following symbols refer to set elements.
    // All the N_SET[ATDB] symbols with the same name form one set.
    // Space is allocated for the set in the text section, and each set
    // elements value is stored into one word of the space.
    // The first word of the space is the length of the set (num of elements).
    //
    // The address of the set is made into an N_SETV symbol
    // whose name is the same as the name of the set.
    // This symbol acts like a N_DATA global symbol
    // in that it can satisfy undefined external references.

    // These appear as input to LD, in a .o file.
    N_SETA	= 0x14,		// Absolute set element symbol
    N_SETT	= 0x16,		// Text set element symbol
    N_SETD	= 0x18,		// Data set element symbol
    N_SETB	= 0x1A,		// Bss set element symbol

    // This is output from LD.
    N_SETV	= 0x1C,		// Pointer to set vector in data area.

    // Warning symbol. The text gives a warning message, the next symbol
    // in the table will be undefined. When the symbol is referenced, the
    // message is printed.
    N_WARNING	= 0x1e,

    // Weak symbols.  These are a GNU extension to the a.out format.  The
    // semantics are those of ELF weak symbols.  Weak symbols are always
    // externally visible.  The N_WEAK? values are squeezed into the
    // available slots.  The value of a N_WEAKU symbol is 0.  The values
    // of the other types are the definitions.
    N_WEAKU	= 0x0d,		// Weak undefined symbol.
    N_WEAKA	= 0x0e,		// Weak absolute symbol.
    N_WEAKT	= 0x0f,		// Weak text symbol.
    N_WEAKD	= 0x10,		// Weak data symbol.
    N_WEAKB	= 0x11,		// Weak bss symbol.
};

//=============================================================================
// Relocations 
//
//  There are two types of relocation flavours for a.out systems,
//  standard and extended. The standard form is used on systems where the
//  instruction has room for all the bits of an offset to the operand, whilst
//  the extended form is used when an address operand has to be split over n
//  instructions. Eg, on the 68k, each move instruction can reference
//  the target with a displacement of 16 or 32 bits. On the sparc, move
//  instructions use an offset of 14 bits, so the offset is stored in
//  the reloc field, and the data in the section is ignored.
//=============================================================================

//-----------------------------------------------------------------------------
// This structure describes a single relocation to be performed.
// The text-relocation section of the file is a vector of these structures,
// all of which apply to the text section.
// Likewise, the data-relocation section applies to the data section.
//-----------------------------------------------------------------------------

struct aout_relocation_info {
    // Address (within segment) to be relocated.
    long	r_address;

    // The meaning of r_symbolnum depends on r_extern.
    unsigned	r_symbolnum : 24;

    // Nonzero means value is a pc-relative offset
    // and it should be relocated for changes in its own address
    // as well as for changes in the symbol or section specified.
    unsigned	r_pcrel	    : 1;

    // Length (as exponent of 2) of the field to be relocated.
    // Thus, a value of 2 indicates 1<<2 bytes.
    unsigned	r_length    : 2;

    // 1 => relocate with value of symbol.
    // r_symbolnum is the index of the symbol in files the symbol table.
    // 0 => relocate with the address of a segment.
    // r_symbolnum is N_TEXT, N_DATA, N_BSS or N_ABS
    // (the N_EXT bit may be set also, but signifies nothing).
    unsigned	r_extern    : 1;

    // The next three bits are for SunOS shared libraries, and seem to
    // be undocumented.
    unsigned	r_baserel   : 1; // Linkage table relative
    unsigned    r_jmptable  : 1; // pc-relative to jump table
    unsigned    r_relative  : 1; // "relative relocation"

    // unused
    unsigned	r_pad	    : 1; // Padding -- set to zero
};

//-----------------------------------------------------------------------------
// Extended relocations

struct aout_ext_relocation_info {
    long	r_address;	 // Offset of of data to relocate
    unsigned	r_index	    :24; // Symbol table index of symbol
    unsigned	r_extern    : 1; // 
    unsigned	r_pad	    : 2; // 
    unsigned	r_type	    : 5; // Relocation type
    long	r_addend;	 // datum addend
};

// Values of r_type
enum aout_ext_reloc_type {
    // simple relocations
    RELOC_8,			// data[0:7] = addend + sv
    RELOC_16,			// data[0:15] = addend + sv
    RELOC_32,			// data[0:31] = addend + sv
    // pc-rel displacement
    RELOC_DISP8,		// data[0:7] = addend - pc + sv
    RELOC_DISP16,		// data[0:15] = addend - pc + sv
    RELOC_DISP32,		// data[0:31] = addend - pc + sv
    // Special
    RELOC_WDISP30,		// data[0:29] = (addend + sv - pc)>>2
    RELOC_WDISP22,		// data[0:21] = (addend + sv - pc)>>2
    RELOC_HI22,			// data[0:21] = (addend + sv)>>10
    RELOC_22,			// data[0:21] = (addend + sv)
    RELOC_13,			// data[0:12] = (addend + sv)
    RELOC_LO10,			// data[0:9] = (addend + sv)
    RELOC_SFA_BASE,		
    RELOC_SFA_OFF13,
    // P.I.C. (base-relative)
    RELOC_BASE10,  		// Not sure - maybe we can do this the
    RELOC_BASE13,		// right way now
    RELOC_BASE22,
    // for some sort of pc-rel P.I.C. (?)
    RELOC_PC10,
    RELOC_PC22,
    // P.I.C. jump table
    RELOC_JMP_TBL,
    // reputedly for shared libraries somehow
    RELOC_SEGOFF16,
    RELOC_GLOB_DAT,
    RELOC_JMP_SLOT,
    RELOC_RELATIVE,

    RELOC_11,	
    RELOC_WDISP2_14,
    RELOC_WDISP19,
    RELOC_HHI22,		// data[0:21] = (addend + sv) >> 42
    RELOC_HLO10,		// data[0:9] = (addend + sv) >> 32
    // 29K relocation types
    RELOC_JUMPTARG,
    RELOC_CONST,
    RELOC_CONSTH,

    // for sparc v9
    RELOC_64,			// data[0:63] = addend + sv
    RELOC_DISP64,		// data[0:63] = addend - pc + sv
    RELOC_WDISP21,		// data[0:20] = (addend + sv - pc)>>2
    RELOC_DISP21,		// data[0:20] = addend - pc + sv
    RELOC_DISP14,		// data[0:13] = addend - pc + sv

    NO_RELOC
};

#define CTOR_LIST_NAME "__CTOR_LIST__"
#define DTOR_LIST_NAME "__DTOR_LIST__"
#endif


