//-----------------------------------------------------------------------------
//   Dynamic loader
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>

#include "stdtp.hpp"
#include "ermsg.hpp"
#include "numlb.hpp"
#include "memmv.hpp"
#include "mmlib.hpp"
#include "dldld.hpp"

//=============================================================================

void	rtm_dld_symbol_name::push (rtm_dld_symbol* symb) {
    symb->next = n_syms;
    n_syms = symb;
}

//-----------------------------------------------------------------------------

void	rtm_dld_symbol_name::drop (rtm_dld_symbol* symb) {
    rtm_dld_symbol**	s_ref;
    rtm_dld_symbol*	s_ptr;

    for (s_ref = &n_syms; (s_ptr = *s_ref) != symb; s_ref = &s_ptr->next) {
	if (s_ptr == NULL)
	    abort();
    }
    *s_ref = symb->next;
}

//=============================================================================

#ifndef RTM_DLD_GLOBAL_NAME_HASH_TABLE_SIZE
#define RTM_DLD_GLOBAL_NAME_HASH_TABLE_SIZE	127
#endif

rtm_dld_symbol_name_table rtm_dld_global(RTM_DLD_GLOBAL_NAME_HASH_TABLE_SIZE);

//-----------------------------------------------------------------------------

inline	unsigned rtm_dld_hash(const char* p) {
    unsigned	hv;
    char	c;

    for (hv = 0; (c = *p++) != '\0'; hv <<= 1) {
	hv ^= c;
    }
    return hv;
}

//-----------------------------------------------------------------------------

rtm_dld_symbol_name_table::rtm_dld_symbol_name_table (int v_h_size, 
						      Dnm_Stack* stk) 
: stor (BFR_SMALL)
{
    ASSERT (v_h_size > 0);
    h_size = v_h_size;
    h_table = (rtm_dld_symbol_name**)xmalloc(v_h_size * sizeof(void*));
    pzero (h_table, v_h_size);
    stor = stk;
}

//-----------------------------------------------------------------------------

rtm_dld_symbol_name_table::~rtm_dld_symbol_name_table() {
    free (h_table);
}

//-----------------------------------------------------------------------------
// Add a new name to name table, or reuse the existing one

rtm_dld_symbol_name*	rtm_dld_symbol_name_table::add(char* name) {
    unsigned			idx;
    rtm_dld_symbol_name*	sym;
    size_t			siz;

    // Get number of bucket
    idx = rtm_dld_hash (name) % h_size;

    // Look for a symbol
    for (sym = h_table [idx]; sym != NULL; sym = sym->next) {
	if (strcmp (name, sym->name) == 0)
	    return sym;
    }

    // Build symbol
    siz = strlen (name);
    sym = new (siz, stor) rtm_dld_symbol_name (h_table[idx], siz, name);
    h_table [siz] = sym;
    
    return sym;
}

//-----------------------------------------------------------------------------
// Search for a name

rtm_dld_symbol_name*	rtm_dld_symbol_name_table::search(char* name) {
    unsigned			idx;
    rtm_dld_symbol_name*	sym;

    // Get number of bucket
    idx = rtm_dld_hash (name) % h_size;

    // Look for a symbol
    for (sym = h_table [idx]; sym != NULL; sym = sym->next) {
	if (strcmp (name, sym->name) == 0)
	    break;
    }

    return sym;
}

//=============================================================================

void rtm_dld_section_table::pop (rtm_dld_section* s) {
    rtm_dld_section**	s_ref;
    rtm_dld_section*	s_ptr;

    for (s_ref = &s_sect; (s_ptr = *s_ptr) != s; s_ref = &s_ptr->next)
	if (s_prt == NULL)
	    abort();
    *s_ref = s;
}

//-----------------------------------------------------------------------------

int	rtm_dld_section_table::relocate() {
    rtm_dld_section*	s;
    int			r_code;

    r_code = 0;
    for (s = s_sect; s != NULL; s = s->next) {
	if ((r_code = s->relocate()) != 0)
	    break;
    }

    return r_code;
}

//-----------------------------------------------------------------------------

rtm_dld_section* rtm_dld_section_table::operator [] (char const* name) {
    rtm_dld_section*	sect;

    for (sect = s_sect; sect != NULL; sect = sect->next) {
	if (strcmp (sect->name, name) == 0)
	    break;
    }

    return sect;
}

//=============================================================================

void rtm_dld_symbol_table::size_hint (size_t size) {
    size_t	bytes_to_alloc;

    if (size <= a_size)
	return;
    bytes_to_alloc = sizeof (void*) * size;
    if (a_size == 0)
	s_table = (rtm_dld_symbol**)xmalloc (bytes_to_alloc);
    else
	s_table = (rtm_dld_symbol**)xrealloc (s_table, bytes_to_alloc);
    a_size = size;
}

//-----------------------------------------------------------------------------

#ifndef RTM_DLD_SYMBOL_TABLE_QUANT
#define RTM_DLD_SYMBOL_TABLE_QUANT	64
#endif

size_t rtm_dld_symbol_table::push (rtm_dld_symbol* symb) {
    if (n_used == a_size)
	size_hint (a_size + RTM_DLD_SYMBOL_TABLE_QUANT);
    
    s_table[n_used] = symb;
    n_used += 1;

    return n_used;
}

//-----------------------------------------------------------------------------

int	rtm_dld_symbol_table::find (rtm_dld_symbol* symb) {
    size_t	idx;

    for (idx = 0; idx < n_used; idx += 1)
	if (s_table [idx] == symb)
	    return (int)idx;

    return -1;
}

//-----------------------------------------------------------------------------

int	rtm_dld_symbol_table::resolve () {
    rtm_dld_symbol*	symb;
    rtm_dld_symbol*	sdef;
    int	i;

    for (i = 0; i < n_used; i++) {
	symb = s_table[i];
	if (symb->is_defined()) {
	    symb->prepare_value();
	} else {
	    for (sdef = symb->name->syms(); sdef != NULL; sdef = sdef->next) {
		if (sdef->is_defined())
		    break;
	    }
	    if (sdef == NULL) {
		fprintf (stderr, "Uresolved reference to \"%s\"\n",
			 symb->name->name);
		return -1;
	    }
	    s_table[i] = sdef;
	}
    }
    return 0;
}

//=============================================================================

rtm_dld_module::~rtm_dld_module() {
    run_destructors();
}

//-----------------------------------------------------------------------------

rtm_dld_module::rtm_dld_module (rtm_dld_module* v_master, 
				rtm_dld_format* v_form,
				const char* v_name, 
				const char* v_path,
				size_t v_f_base, 
				size_t v_f_size)
: rtm_dld (v_master, v_form, v_name, v_path),
  loc_data (BFR_SMALL), glb_data(BFR_SMALL), loc_names (&glb_data)
{
    // Initialize file data
    f_base = v_f_base;
    f_size = v_f_size;
} 

//-----------------------------------------------------------------------------

void rtm_dld_module::dump () {
    rtm_dld::dump();
    Trace (" f_base = %#lx, f_size = %#lx\n", f_base, f_size);
}

//=============================================================================

void rtm_dld_relocation_table::push_relocation (rtm_dld_relocation* rel) {
    rel->next = rels;
    rels = rel;
}

//-----------------------------------------------------------------------------

void rtm_dld_relocation_table::relocate (rtm_dld_section* sect) {
    rtm_dld_relocation*	rel;
    int r_code;

    r_code = 0;
    for (rel = rels; rel != NULL; rel = rel->next) {
	r_code = rel->apply(sect);
	if (r_code != 0)
	    break;
    }
    return r_code;

}

//-----------------------------------------------------------------------------

rtm_dld_section::rtm_dld_section(rtm_dld_module* v_module, char* v_name, 
				 int v_s_num, int v_f_offs, int v_f_size)
{
    // Insert self into list of module sections
    module = v_module;
    next = v_module->sect;
    v_module->sect = this;

    // Set section name if provided
    if (v_name != NULL) {
	name = new char [strlen(v_name) + 1];
	strcpy (name, v_name);
    } else {
	name = NULL;
    }
    s_name = NULL;

    // Initialize other components
    s_num = v_s_num;
    attr = 0;
    data = NULL;
    m_size = 0;
    a_size = 0;
    f_offs = v_f_offs;
    f_size = v_f_size;
}

//-----------------------------------------------------------------------------

rtm_dld_section::~rtm_dld_section () {
    if (data != NULL)
	free (data);
    if (name != NULL)
	delete [] name;
}

//-----------------------------------------------------------------------------

int rtm_dld_section::load (FILE*, int) { data = NULL; return -1; }

//-----------------------------------------------------------------------------

void* rtm_dld_symbol::operator new (size_t size, 
				    rtm_dld_module* m, 
				    short v_attr)
{
    Dnm_Stack*	stack;

    if ((v_attr & is_global) || v_attr == is_undefined)
	stack = &m->glb_syms;
    else
	stack = &m->loc_data;
    return stack->Push(size);
}

//-----------------------------------------------------------------------------

rtm_dld_symbol::rtm_dld_symbol(char* name, 
			       short v_attr, 
			       short v_type,
			       size_t v_value,
			       rtm_dld_module* m,
			       rtm_dld_section* s)
{
    rtm_dld_symbol_name_table*	snt;
    rtm_dld_symbol_name*	snm;

    attr = v_attr;
    type = v_type;
    sym_value.as_offset = v_value;
    sect = s;

    if (v_attr & is_global)
	snt = &rtm_dld_global;
    else
	snt = &m->loc_names;

    snm = snt->add (name);
    rtm_dld_symbol::name = snm;
    next = snm->syms;
    snm->syms = this;

    m->push_symbol(this);
}

//-----------------------------------------------------------------------------

bool rtm_dld_symbol::prepare_value() {
    if (! is_defined())
	return FALSE;
    if ((attr & (is_indirect | is_debug | is_sect)))
	return TRUE;
    sym_value.as_offset += (size_t)sect->data;
    return TRUE;
}

//-----------------------------------------------------------------------------

size_t rtm_dld_symbol::value () {
    if (attr & is_indirect)
	return (*sym_value.as_symbol)->value();
    else
	return sym_value.as_offset;
}

//-----------------------------------------------------------------------------

void rtm_dld_symbol::dump () {
}

//-----------------------------------------------------------------------------

int rtm_dld_relocation::apply (rtm_dld_section* sect) {
    switch (type) {
	#define DLD_DEFRELOC(NAME, CODE) case NAME: CODE
	#include "dldrl.dpp"
      default:
	Abzac ("Bad internal relocation type - %#x\n", type);
    }
    return TRUE;
}

//-----------------------------------------------------------------------------

void rtm_dld_relocation::dump () {

}
