//-----------------------------------------------------------------------------
//   Dynamic loader internal declarations for true loader
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#ifndef __DLDIN_HPP__
#define __DLDIN_HPP__

#include "dnmem.hpp"
#include "dynld.hpp"

//=============================================================================
// This file contains definitions for cases when we need to implement true
// linking loader (instead of just calling some external library routines).
// The generic linker/loader classes are defined here. Classes for specific
// object file formats are derived from them in format/machine specific
// components. To speedup the loading process and simplify code, the
// memory for most of objects is allocated in per module dynamic stacks.
//=============================================================================

class rtm_dld_symbol;
class rtm_dld_relocation;
class rtm_dld_module;
class rtm_dld_section;

//-----------------------------------------------------------------------------
// Symbol name table entry. Is a header of list of all references/definitions
//  for symbols in this name table with the same name.

class rtm_dld_symbol_name {
  private:
    rtm_dld_symbol_name*	n_next;	// Next item in hash bucket
    rtm_dld_symbol*		n_syms;	// Symbol entries
    char			n_name[1];// Start of the zero terminated name

    // Special allocator
    void* operator new (size_t size, size_t len, Dnm_Stack* stk) {
	size += len;
	size = ALIGN (size, sizeof (void*));
	return (void*)stk->Push(size);
    }

    // Constructor
    rtm_dld_symbol_name (rtm_dld_symbol_name* v_next, 
			 size_t n_len, char const* v_name)
    {
	n_next = v_next;
	n_syms = NULL;
	bcpy (n_name, v_name, n_len + 1);
    }

    friend class rtm_dld_symbol_name_table;

  public:
    // Get head of symbol defs/refs list
    rtm_dld_symbol*	syms() const { return n_syms; }

    // Add new symbol def/ref to defs/refs list.
    //  Precondition: symbol is not included into defs/refs list
    void		push(rtm_dld_symbol* symb);

    // Remove symbol def/ref from defs/refs list
    //  Precondition: symbol in defs/refs list of this
    void		drop(rtm_dld_symbol* symb);

    // Accessor to symbol name
    char const*		name() const { return n_name; }
};

//-----------------------------------------------------------------------------
// Symbol name table

class rtm_dld_symbol_name_table {
    Dnm_Stack*			stor;	// Storage for symbols
    int				h_size;	// Number of buckets in hash table
    rtm_dld_symbol_name**	h_table;// Array of bucket heads

  public:
    // Constructor
    rtm_dld_symbol_name_table(Dnm_Stack* stk, int v_h_size = 31);

    // Destructor, deletes all symbol names in this table
    ~rtm_dld_symbol_name_table();

    // Add a new name to name table, or reuse the existing one
    rtm_dld_symbol_name*	add(char const* name);

    // Search for a name
    rtm_dld_symbol_name*	search(char const* name);
};

// Symbol name table for global names
export	rtm_dld_symbol_name_table	rtm_dld_global;

//-----------------------------------------------------------------------------
// Table of sections defined within a given module

class rtm_dld_section_table {
    rtm_dld_section*	s_sect;	// Header of section table list
    rtm_dld_section*	s_abs;	// Absolute pseudo section
    int			s_num;	// Number of sections in section table
  public:
    // Add section to table.
    void	push (rtm_dld_section* s) {
	s->next = s_sect;
	s_sect = s;
    }

    // Removes section from table.
    void	pop (rtm_dld_section* s);

    // Applies relocation information to all sections
    int		relocate();

    // Finds section by name. Returns NULL if no section found
    rtm_dld_section* operator [] (char const* name);

    // Accessors
    rtm_dld_section*	abs()	const { return s_abs; }
    rtm_dld_section*	sect()	const { return s_sect; }
    int			num()	const { return s_num; }

    // Constructor
    rtm_dld_section_table() {
	s_sect = NULL;
	s_abs = NULL;
	s_num = 0;
    }
};

//-----------------------------------------------------------------------------
// Table of symbol definitions/references found in a given module

class rtm_dld_symbol_table {
    rtm_dld_symbol**	s_table;// Table of pointers to symbol definitions
    size_t		a_size;	// Number of allocated slots in symbol table
    size_t		n_used;	// Number of used slots in symbol table
  public:
    void		size_hint (size_t size);
    size_t		push(rtm_dld_symbol* symb);

    void		replace(size_t i, rtm_dld_symbol* by) {
	s_table [i] = by;
    }

    rtm_dld_symbol*	operator [] (size_t i) const { return s_table[i]; }

    int			find(rtm_dld_symbol* symb);
    int			resolve();
    
    rtm_dld_symbol_table() {
	s_table = NULL;
	a_size = 0;
	n_used = 0;
    }

    ~rtm_dld_symbol_table();
};

//-----------------------------------------------------------------------------
// module

class rtm_dld_module : public rtm_dld {
  protected:
    // All sections defined within this module
    rtm_dld_section_table	sect;

    // Local name table
    rtm_dld_symbol_name_table	loc_names;

    // Table with all defs/refs in this module
    rtm_dld_symbol_table	symb;

    Dnm_Stack		loc_data;// Memory for local data
    Dnm_Stack		glb_data;// Memory for global data
    
    size_t		f_base;	// Offset in file 
    size_t		f_size;	// Size in file

    // "Member" constructor
    rtm_dld_module(rtm_dld_module* v_master, rtm_dld_format* v_form,
		   const char* v_name, const char* v_path,
		   size_t v_f_base, size_t v_f_size);

    virtual int	run_constr()= 0;// Run constructors
    virtual int	run_destr() = 0;// Run destructors

    // Load additional information
    virtual	int	do_load(FILE* file, int what) = 0;

    // Unload all file information, run destructors, and delete object
    virtual	~rtm_dld_module();

    friend class rtm_dld_symbol;
    friend class rtm_dld_section;
  public:
    // Debugging dump for single instance
    virtual 	void 	dump();
};

//-----------------------------------------------------------------------------

class rtm_dld_relocation_table {
    rtm_dld_relocation* rels;	// Table with relocations for this section
    
  public:
    // Allocate entry in relocation table
    void push_relocation (rtm_dld_relocation* rel);

    // Apply relocation table
    int	 relocate(rtm_dld_section* sect);

    rtm_dld_relocation_table() {
	rels = NULL;
    }
};

//-----------------------------------------------------------------------------
// Section

class rtm_dld_section {
  protected:
    rtm_dld_section*	next;	// Next section
    rtm_dld_module*	module;	// Module where section is defined
    rtm_dld_symbol*	s_name;	// Symbol corresponding to section name
    char*		name;	// Name of the section
    short		s_num;	// Number of section within module

    rtm_dld_relocation_table	r_tab;

    // Values for section attributes
    enum {
	s_alloc		= 0x0001,
	s_load		= 0x0002,
	s_reloc		= 0x0004,
	s_readonly	= 0x0010,
	s_code		= 0x0020,
	s_data		= 0x0040,
	s_rom		= 0x0080,
	s_constructor	= 0x0100,
	s_cons_text	= 0x1100,
	s_cons_data	= 0x2100,
	s_cons_bss	= 0x4100,
	s_has_contents	= 0x0200,
	s_never_load	= 0x0400,
	s_shared_lib	= 0x0800,
	s_is_common	= 0x8000,
	s_debugging	= 0x0008
    };
    short		attr;	// Section attributes

    char*		data;	// Section data area
    size_t		m_size;	// Size of section in memory
    size_t		a_size;	// Size of area allocated in section
    size_t		f_offs;	// Offset within file
    size_t		f_size; // Size of section in file

    rtm_dld_section(rtm_dld_module* v_module, char* v_name, 
		    int v_s_num, int v_f_offs, int v_f_size);

    void* operator new (size_t size, rtm_dld_module* m) {
	return m->glb_data.Push(size);
    }

    void  		force_symbol();

    friend class 	rtm_dld_relocation;
    friend class	rtm_dld_module;
    friend class	rtm_dld_symbol;
    friend class	rtm_dld_section_table;

  public:
    // Apply relocations
    int			relocate() { return r_tab.relocate(this); }
};
************************************************************
************************************************************
************************************************************
************************************************************

//-----------------------------------------------------------------------------
// Symbol reference

class rtm_dld_symbol {
  protected:
    rtm_dld_symbol*		next; // Next item in chain of all omonims
    rtm_dld_symbol_name*	name; // Name block
    short			attr; // Symbol attributes
    short			type; // Additional type information

    // Values for attr field
    enum {
	is_undefined	= 0x0000,
	is_local	= 0x0001,
	is_global	= 0x0002,
	is_common	= 0x0004,
	is_debug	= 0x0008,
	is_function	= 0x0010,
	is_not_at_end	= 0x0400,
	is_constructor	= 0x0800,
	is_warning	= 0x1000,
	is_indirect	= 0x2000,
	is_weak		= 0x4000,
	is_sect		= 0x8000 // Reference to section address
    };

    friend class rtm_dld_symbol_name;
    friend class rtm_dld_symbol_table;
  public:
    // Special allocation procedure
    void* operator new (size_t size, Dnm_Stack* stk) {
	return stk->Push(size);
    }

    rtm_dld_symbol(rtm_dld_symbol* v_next, rtm_dld_module* m,
		   const char* v_name, short v_attr, short v_type);

    // Test if this is a symbol definition
    bool is_defined() {
	return attr != is_undefined && attr != (is_undefined | is_global);
    }

    // Calculate value (for some special kinds of symbols)
    virtual bool prepare_value();

    // Obtain a value
    virtual size_t value();

    // Debugging dump
    virtual void dump();
};

//-----------------------------------------------------------------------------
// Symbol definition

class rtm_dld_symbol_def : public rtm_dld_symbol {
  protected:
    rtm_dld_section*		sect; // Section 

    // Symbols value
    union {
	size_t			as_offset;
	rtm_dld_symbol**	as_symbol;
    } sym_value;

    friend class rtm_dld_module;
  public:
    // Constructor
    rtm_dld_symbol_def (rtm_dld_symbol* v_next,
			const char* v_name, short v_attr, short v_type,
			size_t v_value, rtm_dld_module* m, rtm_dld_section* s)
    : rtm_dld_symbol (v_next, m, v_name, v_attr, v_type)
    {
	sect = v_value;
	sym_value.as_offset = v_value;
    }

    // Calculate value (for some special kinds of symbols)
    virtual bool prepare_value();

    // Obtain a value
    virtual size_t value();

    // Debugging dump
    virtual void dump();
};

//-----------------------------------------------------------------------------
// Relocation 

class rtm_dld_relocation {
  protected:
    rtm_dld_relocation*	next;	// Next relocation in list.
    size_t		offs;	// Offset within section to data
    int			symb;	// Index of Symbol to which relocation refers

    rtm_dld_relocation (rtm_dld_relocation* v_next, int	v_symb, size_t v_offs)
    {
	next = v_next;
	symb = v_symb;
	offs = v_offs;
    }
  public:
    void* operator new (size_t, rtm_dld_section* sect) {
    }

    // Perform relocation
    virtual int	 apply(rtm_dld_section* sect) = 0;

    // Debugging dump
    virtual void dump();
};

//-----------------------------------------------------------------------------
// Simple relocation. Implements most typical cases of relocation

class rtm_dld_simple_relocation : public rtm_dld_relocation {
  protected:
    short		type;	// Type of relocation
    short		parm;	// Additional parameter
    size_t		add;	// Addend
  public:
    rtm_dld_simple_relocation (rtm_dld_relocation* v_next,
			       int v_symb, size_t v_offs, size_t v_add,
			       short v_type, short v_parm)
    : rtm_dld_relocation (v_next, v_symb, v_offs)
    {
	add  = v_add;
	type = v_type;
	parm = v_parm;
    }

    // Perform relocation
    virtual int	 apply(rtm_dld_section* sect);

    // Debugging dump
    virtual void dump();

    // Relocation types
    enum {
	#define DLD_DEFRELOC(NAME, CODE)  NAME,
	#include "dldrl.dpp"
	num_of_relocation_codes
    };
};

#endif
