//-----------------------------------------------------------------------------
//   Dynamic loader implementation for dlfcn.h
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <string.h>
#include <dlfcn.h>

#include "stdtp.hpp"
#include "ermsg.hpp"
#include "dynld.hpp"

#ifdef __ELF__
#include "elf.hpp"
#endif

//-----------------------------------------------------------------------------

class rtm_dld_dlfcn_format : public rtm_dld_format {
  protected:
    // Detect if the file belongs to a given format
    virtual bool detect (char* 	buf_addr, size_t  buf_size, const char* name);

    // Load file with a given name and given file stream
    virtual rtm_dld* load (FILE* file, const char* file_name, 
			   const char* module_name, int what); 

    // Get the name of format
    virtual const char*	name();
    
  public:
    rtm_dld_dlfcn_format() : rtm_dld_format(TRUE) {;}
};

//-----------------------------------------------------------------------------

class rtm_dld_dlfcn : public rtm_dld {
    void*		dl_handle; // Magic cookie returned by dlopen
  public:
    // Unload object module.
    virtual void	unload();

    // Update contence of object module. Returns FALSE if module was unchanged
    virtual bool	reload();

    virtual void	dump();	// Print this module information

    rtm_dld_dlfcn(void* handl, rtm_dld_format* v_form, 
		  const char* v_path, const char* v_name);
  protected:
    virtual int		do_load(const char* file_name, int what);
};

//-----------------------------------------------------------------------------

local rtm_dld_dlfcn_format dlfcn_fmt;

//-----------------------------------------------------------------------------

rtm_dld_dlfcn::rtm_dld_dlfcn(void* handl, rtm_dld_format* v_form, 
			     const char* v_path, const char* v_name)
  : rtm_dld(NULL, v_form, v_name, v_path)
{
    dl_handle = handl;
    m_stat |= ld_mask;
}

//-----------------------------------------------------------------------------

bool rtm_dld_dlfcn_format::detect(char* buf, size_t size, const char*)
{
    #ifdef __ELF__
    {
	elf_header* eh;

        ASSERT(size < sizeof(elf_header));
	eh = (elf_header*)buf;
	return eh->good_magic();
    }
    #else
        return TRUE;
    #endif
}

//-----------------------------------------------------------------------------

rtm_dld* rtm_dld_dlfcn_format::load (FILE*, const char* file_name, 
				     const char* module_name, int what) 
{ 
    void* handl;

    if (what & rtm_dld::ld_self) {
	handl = NULL;
    } else {
	handl = dlopen(file_name, RTLD_NOW);
	if (handl == NULL) {
	    set_error(dlerror());
	    return NULL;
	}
    }
    return new rtm_dld_dlfcn(handl, this, file_name, module_name);
}

//-----------------------------------------------------------------------------

const char* rtm_dld_dlfcn_format::name() {
    return "ELF dlfcn library";
}

//-----------------------------------------------------------------------------
// Unload object module.

void	rtm_dld_dlfcn::unload(){
    rtm_dld::unload();

    if (m_stat & ld_mask) {
	dlclose(dl_handle);
	m_stat &= ~ld_mask;
    }
}

//-----------------------------------------------------------------------------
// Update contence of object module. Returns TRUE if module was
// reloaded succesfully. Otherwise, m_error contains error message text
//-----------------------------------------------------------------------------

bool	rtm_dld_dlfcn::reload(){
    m_error = NULL;

    if (m_stat & ld_mask) {
	dlclose(dl_handle);
    }
    dl_handle = dlopen(m_path, RTLD_NOW);
    if (dl_handle == NULL) {
	m_stat &= ~ld_mask;
	return FALSE;
    }
    return TRUE;
}

//-----------------------------------------------------------------------------

int	rtm_dld_dlfcn::do_load(const char*, int){
    return 0;
}

//-----------------------------------------------------------------------------

void	rtm_dld_dlfcn::dump(){
    rtm_dld::dump();
    Trace("  dlfcn handle = %p\n", dl_handle);
}


