//-----------------------------------------------------------------------------
//   Dynamic loader common methods
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <string.h>

#include "stdtp.hpp"
#include "ermsg.hpp"
#include "dynld.hpp"
#include "mmlib.hpp"

//-----------------------------------------------------------------------------

rtm_dld*	rtm_dld::head;
rtm_dld*	rtm_dld::self;
const char*	rtm_dld::m_error;

//-----------------------------------------------------------------------------

rtm_dld::rtm_dld (rtm_dld* v_mast, rtm_dld_format* v_form, const char* v_name,
		  const char* v_path)
{
    size_t	len;

    // Initialize structural information
    if (v_mast == NULL) {
	m_next = head;
	head = this;
	m_subm = NULL;
	m_mast = NULL;
    } else {
	m_next = v_mast->m_subm;
	v_mast->m_subm = this;
	m_subm = NULL;
	m_mast = v_mast;
    }

    // Initialize dependency information
    m_deps = NULL;
    a_deps = 0;
    n_deps = 0;

    // Initialize module information
    m_form = v_form;
    len = strlen (v_name) + 1;
    m_name = new char [len];
    memcpy (m_name, v_name, len);
    len = strlen (v_path) + 1;
    m_path = new char [len];
    memcpy (m_path, v_path, len);
    m_stat = loaded;

    // Reset error message
    m_error = NULL;
}

//-----------------------------------------------------------------------------

rtm_dld::~rtm_dld() {
    rtm_dld**	rdpp;
    rtm_dld*	rdp;

    // Delete submodules
    while (m_subm != NULL)
	delete m_subm;

    // Unlink from global structure
    rdpp = (m_mast == NULL) ? &head : &m_mast->m_subm;
    while ((rdp = *rdpp) != this) {
	ASSERT (rdp != NULL);
	rdpp = &rdp->m_next;
    }
    *rdpp = m_next;

    // Free memory
    if (m_deps != NULL)
	free (m_deps);
    delete [] m_name;
    delete [] m_path;

    // Prevent self from pointing to destructed object
    if (self == this)
	self = NULL;
}

//-----------------------------------------------------------------------------

rtm_dld* rtm_dld::tree_next() const {
    rtm_dld*	p_next;

    if (m_subm != NULL)
	return m_subm;
    if (m_next != NULL)
	return m_next;
    for (p_next = m_mast; p_next != NULL; p_next = p_next->m_mast) {
	if (p_next->m_next != NULL)
	    return p_next->m_next;
    }
    return NULL;
}

//-----------------------------------------------------------------------------

void	rtm_dld::flush () {
    rtm_dld*	curs;
    rtm_dld*	p;
    rtm_dld**	m_stack;
    rtm_dld**	p_stack;
    size_t	n_mods;
    
    // Calculate number of modules
    n_mods = 0;
    for (curs = head; curs != NULL; curs = curs->tree_next())
	n_mods += 1;

    // Allocate stack
    m_stack = (rtm_dld**)xmalloc (n_mods * sizeof(rtm_dld*));
    p_stack = m_stack;

    // Push all loaded modules
    for (curs = head; curs != NULL; curs = curs->tree_next()) {
	if (curs->m_stat & loaded) {
	    *p_stack++ = curs;
	    curs->m_stat |= marked;
	}
    }

    // Recursively mark all modules required for loaded
    while (p_stack != m_stack) {
	curs = --*p_stack;
	
	// Process master
	if (curs->m_mast != NULL && (curs->m_mast->m_stat & marked) == 0) {
	    curs->m_mast->m_stat |= marked;
	    *p_stack++ = curs->m_mast;
	}

	// Process modules from which curs is dependent
	for (size_t i = 0; i < curs->n_deps; i++) {
	    p = curs->m_deps[i];
	    if ((p->m_stat & marked) == 0) {
		p->m_stat |= marked;
		*p_stack++ = p;
	    }
	}
    }
    
    // Free stack
    free (m_stack);

    // Unload all unneeded modules.
    //  since for each marked module, its master is also marked, shallow
    //  scan is sufficient
    for (curs = head; curs != NULL; curs = p) {
	p = curs->m_next;
	if ((curs->m_stat & marked) == 0)
	    delete curs;
    }

    // Unmark all modules left
    for (curs = head; curs != NULL; curs = curs->tree_next())
	curs->m_stat &= ~marked;
}

//-----------------------------------------------------------------------------

int	rtm_dld::init(const char* exe_name) {
    self = load (exe_name, "self-exe", (symbols | ld_self));
    return 0;
}

//-----------------------------------------------------------------------------

#ifndef RTM_DLD_DETECT_BUF_SIZE
#define RTM_DLD_DETECT_BUF_SIZE 1024
#endif

rtm_dld* rtm_dld::load(const char* file_name,  
		       const char* module_name, int what) 
{
    rtm_dld_format*	fmt;
    rtm_dld*		module;
    FILE*		f_stream;
    char		buf [RTM_DLD_DETECT_BUF_SIZE];
    int			b_size;

    if ((f_stream = fopen (file_name, "r")) == NULL)
	return NULL;

    module = NULL;
    b_size = fread (buf, sizeof(char), RTM_DLD_DETECT_BUF_SIZE, f_stream);

    if (b_size > 0) {
	for (fmt = rtm_dld_format::list; fmt != NULL; fmt = fmt->next) {
	    if (fmt->detect (buf, b_size, file_name)) {
		module = fmt->load (f_stream, file_name, module_name, what);
		break;
	    }
	}
    }

    fclose (f_stream);

    return module;
}

//-----------------------------------------------------------------------------

void	rtm_dld::unload() {
    m_stat &= ~loaded;
    flush();
}

//-----------------------------------------------------------------------------

bool	rtm_dld::reload() {
    m_error = NULL;
    do_load (m_path, m_stat & ld_mask);
    return TRUE;
}

//-----------------------------------------------------------------------------

int	rtm_dld::write_debug_info(const char*) {
    m_error = "not implemented";
    return -1;
}

//-----------------------------------------------------------------------------

rtm_dld* rtm_dld::find(const char* module_name) {
    rtm_dld*	curr;

    for (curr = head; curr != NULL; curr = curr->tree_next()) {
	if (curr->m_name != NULL && strcmp (module_name, curr->m_name) == 0)
	    break;
    }

    return curr;
}

//-----------------------------------------------------------------------------

rtm_dld* rtm_dld::find_by_path(const char* file_name) {
    rtm_dld*	curr;

    for (curr = head; curr != NULL; curr = curr->tree_next()) {
	if (curr->m_path != NULL && strcmp (file_name, curr->m_path) == 0)
	    break;
    }

    return curr;
}

//-----------------------------------------------------------------------------

void	rtm_dld::dump() {
    Trace ("Dump for module at %p, name = \"%s\", path = \"%s\"\n",
	   this, m_name ? m_name : "<null>", m_path ? m_path : "<null>");
    Trace ("  next = %p, subm = %s, mast = %s\n", m_next, m_subm, m_mast);

    Trace ("  stat = ");
    if (m_stat & loaded)
	Trace ("/loaded");
    if (m_stat & marked)
	Trace ("/marked");
    switch (m_stat & m_type) {
      case library :
	Trace ("/Library");
	break;
      case obj :
	Trace ("/Obj"); 
	break;
      case exe :
	Trace ("/Exe");
	break;
      default:
	Trace ("/m_type=$#x", m_stat & m_type);
	break;
    }
    if (m_stat & g_symb)
	Trace ("/global_symbols");
    if (m_stat & l_symb)
	Trace ("/local_symbols");
    if (m_stat & text)
	Trace ("/text");
    if (m_stat & debug)
	Trace ("/debug");
    if (m_stat & reloc)
	Trace ("/reloc");
    if (m_stat & constr)
	Trace ("/constr");
    if (m_stat & headers)
	Trace ("/headers");
    Trace ("\n");
}

//-----------------------------------------------------------------------------

void	rtm_dld::dump_all() {
    rtm_dld*	curr;

    for (curr = head; curr != NULL; curr = curr->tree_next()) {
	curr->dump();
    }
}

//=============================================================================

rtm_dld_format*	rtm_dld_format::list;
rtm_dld_format* rtm_dld_format::native;

//-----------------------------------------------------------------------------
// Constructor, includes into global list, maintains the native component

rtm_dld_format::rtm_dld_format(bool is_native) {
    next = list;
    list = this;
    if (is_native)
	native = this;
}

//-----------------------------------------------------------------------------
// Destructor, excludes from the global list, maintains the native component

rtm_dld_format::~rtm_dld_format() {
    rtm_dld_format**	p_fmt;
    rtm_dld_format*	fmt;

    if (native == this)
	native = NULL;
    for (p_fmt = &list; (fmt = *p_fmt) != this; p_fmt = &fmt->next) {
	ASSERT (fmt != NULL);
    } 
}

