//-----------------------------------------------------------------------------
//   Dynamic loader interface
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#ifndef __DYNLD_HPP__
#define __DYNLD_HPP__

class	rtm_dld_format;

//-----------------------------------------------------------------------------
// Class which represents dynamically loadable module. This class is the only
// interface of dynamic loader to other components of a program
//-----------------------------------------------------------------------------

class	rtm_dld {
  public:
    // Values for status field, and 'what' arguments 
    enum {
	loaded	= 0x0001,	// If this bit is clear, then module is marked
				// for unloading, but retained due to 
				// references resolved to it form other modules
	marked  = 0x0002,	// Bit used for marking
	library = 0x0010,	// This module is object library
	obj	= 0x0020,	// This module is ordinary object file
	exe	= 0x0030,	// This module is executable file
	m_type  = 0x00f0,	// Mask to access module type
	g_symb	= 0x0100,	// Information on global symbols defs/refs
	l_symb	= 0x0200,	// Information on local symbols defs/refs
	symbols = 0x0300,	// Information on all symbols
	text	= 0x0400,	// Data from sections
	debug	= 0x0800,	// Debugging information
	reloc	= 0x1000,	// Relocation information (load & apply)
	constr	= 0x2000,	// Constructors (run)
	headers = 0x4000,	// File and section headers
	ld_self	= 0x8000,	// Loading contence of our executable file
	ld_mask = 0xff00	// Bits used to determine which info should be
				// loaded from file.
    };

    // Initialize static data, load information on our executable.
    // Method 'init' should be called only once with full pathname of
    // executable being running as argument. 
    // This call must precede all other calls to dld methods.
    static int		init(const char* exe_name);

    // Load object module. Returns address of created object, or NULL, if
    // something failed
    static rtm_dld*	load(const char* file_name, const char* module_name) {
	return load (file_name, module_name,
		     (symbols | text | headers | reloc | constr));
    }

    // Unload object module.
    virtual void	unload();

    // Update contence of object module. Returns TRUE if module was
    // reloaded succesfully. Otherwise, m_error contains error message text
    virtual bool	reload();

    // Write information on debugging symbols into specified file. Writes
    // debugging information to specified file. This function could be
    // unimplemented in some environments.
    static int		write_debug_info(const char* file_name);	

    // Find module by name. Returns address of module with 
    // m_name == module_name, or NULL if no such module loaded
    static rtm_dld*	find(const char* module_name);

    // Find module by pathname. Returns address of module with
    // m_path == file_name, or NULL if no such module loaded
    static rtm_dld*	find_by_path(const char* file_name);

    // Accessors to some components
    char const* 	name() const { return m_name; }
    rtm_dld const*	subm() const { return m_subm; }
    char const*		path() const { return m_path; }
    int			stat() const { return m_stat; }
    rtm_dld_format*	form() const { return m_form; }

    // Debugging utilities
    virtual void	dump();	// Print this module information
    static  void	dump_all();// Print information on all modules

    // Error reporting. Returns text of last encountered error message.
    // NULL otherwise.
    static const char*  last_error() { return m_error; }

  protected:
    // Structural information
    rtm_dld*		m_next;	// Address of next module in global list
    rtm_dld*		m_subm;	// Header of list of submodules
    rtm_dld*		m_mast;	// Master of this module
    static rtm_dld*	head;	// Head of all currently loaded modules
    static rtm_dld*	self;	// Module corresponding to our executable

    // Next module in depth-first traverse
    rtm_dld*		tree_next() const;

    // Dependency information
    rtm_dld**		m_deps;	// Modules from which this one depends
    size_t		a_deps;	// Number of slots in dependency array
    size_t		n_deps;	// Number of modules from which this depends

    // Module information
    rtm_dld_format*	m_form; // Format of this module
    char*		m_name;	// Code name for module
    char*		m_path;	// Full pathname for file module loaded from
    int			m_stat; // Module status

    static const char*	m_error;// Text of last encountered error or NULL

    static rtm_dld*	load(const char* file_name, 
			     const char* module_name, int what);

    virtual int		do_load(const char* file_name, int what) = 0;
    static  void	flush();

    rtm_dld(rtm_dld* v_mast, rtm_dld_format* v_form, const char* v_name,
	    const char* v_path);
    virtual ~rtm_dld();

    friend class rtm_dld_format;
};

//-----------------------------------------------------------------------------
// rtm_dld_format -- interface to implementations of various object file
//		     formats supported on a current host
//-----------------------------------------------------------------------------

#include <stdio.h>

class	rtm_dld_format {
    rtm_dld_format*	next;	// Next item in the global format list

    // Object file format native to host environment
    static rtm_dld_format*	native;

    // List of all available object file formats
    static rtm_dld_format*	list;

  protected:
    // Detect if the file belongs to a given format
    virtual bool detect (char* 	buf_addr, size_t  buf_size, 
			 const char* name) = 0;

    // Load file with a given name and given file stream
    virtual rtm_dld* load (FILE* file, const char* file_name,
			   const char* module_name, int what) = 0; 

    // Get the name of format
    virtual const char*	name() = 0;

    // Setup dynamic loader error code, in human readable form
    void set_error(const char* m_error) { rtm_dld::m_error = m_error; }

    friend class rtm_dld;
  public:
    // Constructor, includes into global list, maintains the native component
    rtm_dld_format(bool is_native);

    // Destructor, excludes from the global list,maintains the native component
    virtual	~rtm_dld_format();
};

#endif
