//-----------------------------------------------------------------------------
//   Description of elf family formats
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

// This file is derived from GNU BFD elf support

//=============================================================================
// ELF Header

struct	elf_header {		
    unsigned char	e_ident[16];	// ELF "magic number"
    unsigned short	e_type;		// Identifies object file type
    unsigned short	e_machine; 	// Specifies required architecture
    unsigned int	e_version; 	// Identifies object file version
    unsigned long	e_entry; 	// Entry point virtual address
    unsigned long	e_phoff; 	// Program header table file offset
    unsigned long	e_shoff; 	// Section header table file offset
    unsigned int	e_flags; 	// Processor-specific flags
    unsigned short	e_hsize; 	// ELF header size in bytes
    unsigned short	e_phentsize; 	// Program header table entry size
    unsigned short	e_phnum; 	// Program header table entry count
    unsigned short	e_shentsize; 	// Section header table entry size
    unsigned short	e_shnum; 	// Section header table entry count
    unsigned short	e_shstrndx; 	// Section header string table index

    // Setup magic
    void set_magic() {
	e_ident[0] = 0x7f;
	e_ident[1] = 'E';
	e_ident[2] = 'L';
	e_ident[3] = 'F';
    };

    // Test for a magic
    bool good_magic () {
	return e_ident[0] == 0x7f && e_ident[1] == 'E'
	    && e_ident[2] == 'L' && e_ident[3] == 'F';
    }

    // Access to ei_class within e_ident
    unsigned char& ei_class() {
	return e_ident[4];
    }

    // Access to ei_data within e_ident
    unsigned char& ei_data() {
	return e_ident[5];
    }

    // Access to ei_version within e_ident
    unsigned char& ei_version() {
	return e_ident[6];
    }
};

// Values for ei_class
enum {
    ELFCLASSNONE	= 0,		// Invalid class
    ELFCLASS32		= 1,		// 32-bit objects
    ELFCLASS64		= 2		// 64-bit objects
};

// Values for ei_data
enum {
    ELFDATANONE		= 0,		// Invalid data encoding
    ELFDATA2LSB		= 1,		// 2's complement, little endian
    ELFDATA2MSB		= 2		// 2's complement, big endian
};

// Values for e_type, which identifies the object file type
enum	elf_e_type_val {
    ET_NONE		= 0,		// No file type
    ET_REL		= 1,		// Relocatable file
    ET_EXEC		= 2,		// Executable file
    ET_DYN		= 3,		// Shared object file
    ET_CORE		= 4,		// Core file
    ET_LOPROC		= 0xFF00,	// Processor-specific
    ET_HIPROC		= 0xFFFF	// Processor-specific
};

// Values for e_machine, which identifies the architecture
enum	elf_e_machine_val {
    EM_NONE		= 0,		// No machine
    EM_M32		= 1,		// AT&T WE 32100
    EM_SPARC		= 2,		// SUN SPARC
    EM_386		= 3,		// Intel 80386
    EM_68K		= 4,		// Motorola m68k family
    EM_88K		= 5,		// Motorola m88k family
    EM_860		= 7,		// Intel 80860
    EM_MIPS		= 8,		// MIPS R3000 (officially, big-endian)
    EM_MIPS_RS4_BE 	= 10,		// MIPS R4000 big-endian
    EM_SPARC64		= 11,		// SPARC v9 (not official) 64-bit
    EM_PARISC		= 15,		// HPPA
    EM_CYGNUS_POWERPC	= 0x9025	// Cygnus PowerPC ELF backend.
};

// Values for e_version
enum 	elf_e_version_val {
    EV_NONE		= 0,		// Invalid ELF version
    EV_CURRENT		= 1		// Current version
};

//-----------------------------------------------------------------------------
// Processor specific Values for e_flags

#if defined(HPPA)

enum {
    CPU_PA_RISC1_0	= 0x0000020b,	// PA 1.0 big endian.
    CPU_PA_RISC1_1	= 0x00000210,	// PA 1.1 big endian.
    CPU_PA_RISC1_0_LE	= 0x0000028b,	// PA 1.0 little endian.
    CPU_PA_RISC1_1_LE	= 0x00000290,	// PA 1.1 little endian.
    ELF_PARISC_TRAPNIL	= 0x00010000,	// Trap null address dereferences.
    ELF_PARISC_EXT	= 0x00020000	// .PARISC.archext section is present.
};

#elif defined(MIPS)

enum {
    // At least one .noreorder directive appears in the source.
    EF_MIPS_NOREORDER	= 0x00000001,

    // File contains position independent code.
    EF_MIPS_PIC		= 0x00000002,

    // Code in file uses the standard calling sequence for calling
    // position independent code.
    EF_MIPS_CPIC	= 0x00000004,

    EF_MIPS_ARCH	= 0xf0000000,	// Four bit MIPS architecture field.
    E_MIPS_ARCH_1	= 0x00000000,	// -mips1 code.
    E_MIPS_ARCH_2	= 0x10000000,	// -mips2 code.
    E_MIPS_ARCH_3	= 0x20000000	// -mips3 code.
};

#endif

//=============================================================================
// Program header

struct	elf_phdr {
    unsigned	int	p_type;		// Identifies program segment type
    unsigned	int	p_offset; 	// Segment file offset
    unsigned	long	p_vaddr; 	// Segment virtual address
    unsigned	long	p_paddr; 	// Segment physical address
    unsigned	long	p_filesz; 	// Segment size in file
    unsigned	long	p_memsz; 	// Segment size in memory
    unsigned	long	p_flags; 	// Segment flags
    unsigned	long	p_align; 	// Segment alignment, file & memory
};

// Values for program header, p_type field
enum	elf_p_type_val {
    PT_NULL		= 0,		// Program header table entry unused
    PT_LOAD		= 1,		// Loadable program segment
    PT_DYNAMIC		= 2,		// Dynamic linking information
    PT_INTERP		= 3,		// Program interpreter
    PT_NOTE		= 4,		// Auxiliary information
    PT_SHLIB		= 5,		// Reserved, unspecified semantics
    PT_PHDR		= 6,		// Entry for header table itself
    PT_LOPROC		= 0x70000000,	// Processor-specific
    PT_HIPROC		= 0x7FFFFFFF	// Processor-specific
};

// Processor specific program header types.
#if defined(MIPS)
enum elf_mips_p_type_val {
    // Register usage information. Identifies one .reginfo section.
    PT_MIPS_REGINFO	= 0x70000000
};
#endif

// Program segment permissions, in program header p_flags field
enum 	elf_p_flags_val {
    PF_X		= (1 << 0),	// Segment is executable
    PF_W		= (1 << 1),	// Segment is writable
    PF_R		= (1 << 2),	// Segment is readable
    PF_MASKPROC		= 0xF0000000	// Processor-specific reserved bits
};

//=============================================================================
// Section header

struct	elf_shdr {
    unsigned int	sh_name; 	// Section name, index in string tbl 
    unsigned int	sh_type; 	// Type of section
    unsigned long	sh_flags; 	// Miscellaneous section attributes
    unsigned long	sh_addr; 	// Section virtual addr at execution
    unsigned long	sh_offset; 	// Section file offset
    unsigned long	sh_size; 	// Size of section in bytes
    unsigned int	sh_link; 	// Index of another section
    unsigned int	sh_info; 	// Additional section information
    unsigned long	sh_addralign; 	// Section alignment
    unsigned long	sh_entsize; 	// Entry size if section holds table
};

// Values for section header, sh_type field
enum	elf_sh_type_val {
    SHT_NULL		= 0,		// Section header table entry unused
    SHT_PROGBITS	= 1,		// Program specific (private) data
    SHT_SYMTAB		= 2,		// Link editing symbol table
    SHT_STRTAB		= 3,		// A string table
    SHT_RELA		= 4,		// Relocation entries with addends
    SHT_HASH		= 5,		// A symbol hash table
    SHT_DYNAMIC		= 6,		// Information for dynamic linking
    SHT_NOTE		= 7,		// Information that marks file
    SHT_NOBITS		= 8,		// Section occupies no space in file
    SHT_REL		= 9,		// Relocation entries, no addends
    SHT_SHLIB		= 10,		// Reserved, unspecified semantics
    SHT_DYNSYM		= 11,		// Dynamic linking symbol table
    SHT_LOPROC		= 0x70000000,	// Processor-specific semantics, lo
    SHT_HIPROC		= 0x7FFFFFFF,	// Processor-specific semantics, hi
    SHT_LOUSER		= 0x80000000,	// Application-specific semantics
    SHT_HIUSER		= 0x8FFFFFFF	// Application-specific semantics
};

// Processor specific section types
#if defined (HPPA)
enum elf_hppa_sh_type_val {
    // Holds the global offset table, a table of pointers to external data.
    SHT_PARISC_GOT	= SHT_LOPROC+0,

    // Nonloadable section containing information in architecture
    // extensions used by the code.
    SHT_PARISC_ARCH	= SHT_LOPROC+1,

    // Section in which $global$ is defined.
    SHT_PARISC_GLOBAL	= SHT_LOPROC+2,

    // Section holding millicode routines (mul, div, rem, dyncall, etc.
    SHT_PARISC_MILLI	= SHT_LOPROC+3,

    // Section holding unwind information for use by debuggers.
    SHT_PARISC_UNWIND	= SHT_LOPROC+4,

    // Section holding the procedure linkage table.
    SHT_PARISC_PLT	= SHT_LOPROC+5,

    // Short initialized and uninitialized data.
    SHT_PARISC_SDATA	= SHT_LOPROC+6,
    SHT_PARISC_SBSS	= SHT_LOPROC+7,

    // Optional section holding argument location/relocation info.
    SHT_PARISC_SYMEXTN	= SHT_LOPROC+8,

    // Option section for linker stubs.
    SHT_PARISC_STUBS	= SHT_LOPROC+9
};
#elif defined (MIPS)
enum elf_mips_sh_type_val {
    // Section contains the set of dynamic shared objects used when
    // statically linking.
    SHT_MIPS_LIBLIST	= 0x70000000,

    // Section contains list of symbols whose definitions conflict with
    // symbols defined in shared objects.
    SHT_MIPS_CONFLICT	= 0x70000002,

    // Section contains the global pointer table.
    SHT_MIPS_GPTAB	= 0x70000003,

    // Section contains microcode information. The exact format is unspecified.
    SHT_MIPS_UCODE	= 0x70000004,

    // Section contains some sort of debugging information.  The exact
    // format is unspecified.  It's probably ECOFF symbols.
    SHT_MIPS_DEBUG	= 0x70000005,

    // Section contains register usage information.
    SHT_MIPS_REGINFO	= 0x70000006,

    // Section contains miscellaneous options (used on Irix).
    SHT_MIPS_OPTIONS	= 0x7000000d
};

#endif

// Values for section header, sh_flags field
enum elf_sh_flags_val {
    SHF_WRITE		= (1 << 0),	// Writable data during execution
    SHF_ALLOC		= (1 << 1),	// Occupies memory during execution
    SHF_EXECINSTR	= (1 << 2),	// Executable machine instructions
    SHF_MASKPROC	= 0xF0000000	// Processor-specific semantics
};

// Processor specific section flags
#if defined(HPPA)
enum elf_hppa_sh_flags_val {
    // This section is near the global data pointer and thus allows short
    // addressing modes to be used.
    SHF_PARISC_SHORT	= 0x20000000
};
#elif defined(MIPS)
/* This section must be in the global data area.  */
enum elf_mips_sh_flags_val {
    SHF_MIPS_GPREL	= 0x10000000
};
#endif

//=============================================================================
// Symbol table entry

#ifndef ELF64
struct	elf_sym {
    unsigned int	st_name; 	// Symbol name, index in string tbl
    unsigned long	st_value; 	// Value of the symbol
    unsigned long	st_size; 	// Associated symbol size
    unsigned char	st_info; 	// Type and binding attributes
    unsigned char	st_other; 	// No defined meaning, 0
    unsigned short	st_shndx; 	// Associated section index
};
#else
struct	elf_sym {
    unsigned int	st_name; 	// Symbol name, index in string tbl
    unsigned char	st_info; 	// Type and binding attributes
    unsigned char	st_other; 	// No defined meaning, 0
    unsigned short	st_shndx; 	// Associated section index
    unsigned long	st_value; 	// Value of the symbol
    unsigned long	st_size; 	// Associated symbol size
};
#endif

// These three macros disassemble and assemble a symbol table st_info field,
// which contains the symbol binding and symbol type.  The STB_ and STT_
// defines identify the binding and type.

#define ELF_ST_BIND(val)		(((unsigned int)(val)) >> 4)
#define ELF_ST_TYPE(val)		((val) & 0xF)
#define ELF_ST_INFO(bind,type)		(((bind) << 4) + ((type) & 0xF))

#define STN_UNDEF	0		/* undefined symbol index */

#define STB_LOCAL	0		/* Symbol not visible outside obj */
#define STB_GLOBAL	1		/* Symbol visible outside obj */
#define STB_WEAK	2		/* Like globals, lower precedence */
#define STB_LOPROC	13		/* Application-specific semantics */
#define STB_HIPROC	15		/* Application-specific semantics */

#define STT_NOTYPE	0		/* Symbol type is unspecified */
#define STT_OBJECT	1		/* Symbol is a data object */
#define STT_FUNC	2		/* Symbol is a code object */
#define STT_SECTION	3		/* Symbol associated with a section */
#define STT_FILE	4		/* Symbol gives a file name */
#define STT_LOPROC	13		/* Application-specific semantics */
#define STT_HIPROC	15		/* Application-specific semantics */

// Processor specific symbol types
#if defined (HPPA)
// Millicode function entry point.
#define STT_PARISC_MILLICODE	STT_LOPROC+0
#endif

// Special section indices, which may show up in st_shndx fields, among
// other places.
#define SHN_UNDEF	0		/* Undefined section reference */
#define SHN_LORESERVE	0xFF00		/* Begin range of reserved indices */
#define SHN_LOPROC	0xFF00		/* Begin range of appl-specific */
#define SHN_HIPROC	0xFF1F		/* End range of appl-specific */
#define SHN_ABS		0xFFF1		/* Associated symbol is absolute */
#define SHN_COMMON	0xFFF2		/* Associated symbol is in common */
#define SHN_HIRESERVE	0xFFFF		/* End range of reserved indices */

// Processor specific section indices
#if defined (MIPS)
/* Defined and allocated common symbol.  Value is virtual address.  If
   relocated, alignment must be preserved.  */
#define SHN_MIPS_ACOMMON	0xff00

/* Defined and allocated text symbol.  Value is virtual address.
   Occur in the dynamic symbol table of Alpha OSF/1 and Irix 5 executables.  */
#define SHN_MIPS_TEXT		0xff01

/* Defined and allocated data symbol.  Value is virtual address.
   Occur in the dynamic symbol table of Alpha OSF/1 and Irix 5 executables.  */
#define SHN_MIPS_DATA		0xff02

/* Small common symbol.  */
#define SHN_MIPS_SCOMMON	0xff03

/* Small undefined symbol.  */
#define SHN_MIPS_SUNDEFINED	0xff04
#endif

//=============================================================================
// Note segments

struct	elf_note {
    unsigned int	namesz;		// Size of entry's owner string
    unsigned int	descsz;		// Size of the note descriptor
    unsigned int	type;		// Interpretation of the descriptor
    char		name[1]; 	// Start of the name+desc data
};

// Values of note segment descriptor types for core files.
#define NT_PRSTATUS	1		/* Contains copy of prstatus struct */
#define NT_FPREGSET	2		/* Contains copy of fpregset struct */
#define NT_PRPSINFO	3		/* Contains copy of prpsinfo struct */

// Values of note segment descriptor types for object files.
// (Only for hppa right now.  Should this be moved elsewhere?)
#define NT_VERSION	1		/* Contains a version string.  */

//=============================================================================
// Relocation entries

struct	elf_rel {
    unsigned long	r_offset; 	// Loc. at which to apply the action
    unsigned long	r_info;		// index and type of relocation
};

struct	elf_rela : public elf_rel {
    unsigned long	r_addend; 	// Constant addend used to compute val
};

// relocation info handling macros
#ifndef ELF64
#define ELF_R_SYM(i)		((i) >> 8)
#define ELF_R_TYPE(i)		((i) & 0xff)
#define ELF_R_INFO(s,t)	(((s) << 8) + ((t) & 0xff))
#else
#define ELF_R_SYM(i)		((i) >> 32)
#define ELF_R_TYPE(i)		((i) & 0xffffffff)
#define ELF_R_INFO(s,t)	(((large) (s) << 32) + (large) (t))
#endif

//=============================================================================
// Dynamic section structure

struct	elf_dyn {
    unsigned long	d_tag;		// Entry tag value
    union {
	unsigned long	d_val;
	unsigned long	d_ptr;
    } d_un;
};

// Dynamic section tags
#define DT_NULL		0
#define DT_NEEDED	1
#define DT_PLTRELSZ	2
#define DT_PLTGOT	3
#define DT_HASH		4
#define DT_STRTAB	5
#define DT_SYMTAB	6
#define DT_RELA		7
#define DT_RELASZ	8
#define DT_RELAENT	9
#define DT_STRSZ	10
#define DT_SYMENT	11
#define DT_INIT		12
#define DT_FINI		13
#define DT_SONAME	14
#define DT_RPATH	15
#define DT_SYMBOLIC	16
#define DT_REL		17
#define DT_RELSZ	18
#define DT_RELENT	19
#define DT_PLTREL	20
#define DT_DEBUG	21
#define DT_TEXTREL	22
#define DT_JMPREL	23
#define DT_LOPROC	0x70000000
#define DT_HIPROC	0x7fffffff

/* Processor specific dynamic array tags.  */
#if defined (MIPS)
/* 32 bit version number for runtime linker interface.  */
#define DT_MIPS_RLD_VERSION	0x70000001

/* Time stamp.  */
#define DT_MIPS_TIME_STAMP	0x70000002

/* Checksum of external strings and common sizes.  */
#define DT_MIPS_ICHECKSUM	0x70000003

/* Index of version string in string table.  */
#define DT_MIPS_IVERSION	0x70000004

/* 32 bits of flags.  */
#define DT_MIPS_FLAGS		0x70000005

/* Base address of the segment.  */
#define DT_MIPS_BASE_ADDRESS	0x70000006

/* Address of .conflict section.  */
#define DT_MIPS_CONFLICT	0x70000008

/* Address of .liblist section.  */
#define DT_MIPS_LIBLIST		0x70000009

/* Number of local global offset table entries.  */
#define DT_MIPS_LOCAL_GOTNO	0x7000000a

/* Number of entries in the .conflict section.  */
#define DT_MIPS_CONFLICTNO	0x7000000b

/* Number of entries in the .liblist section.  */
#define DT_MIPS_LIBLISTNO	0x70000010

/* Number of entries in the .dynsym section.  */
#define DT_MIPS_SYMTABNO	0x70000011

/* Index of first external dynamic symbol not referenced locally.  */
#define DT_MIPS_UNREFEXTNO	0x70000012

/* Index of first dynamic symbol in global offset table.  */
#define DT_MIPS_GOTSYM		0x70000013

/* Number of page table entries in global offset table.  */
#define DT_MIPS_HIPAGENO	0x70000014

/* Address of run time loader map, used for debugging.  */
#define DT_MIPS_RLD_MAP		0x70000016
#endif

//=============================================================================
// Processor specific sections

#if defined (MIPS)
//-----------------------------------------------------------------------------
// A section of type SHT_MIPS_LIBLIST contains an array of the
// following structure.  The sh_link field is the section index of the
// string table.  The sh_info field is the number of entries in the section.

struct elf_mips_lib {
    // String table index for name of shared object.
    unsigned long	l_name;		// String tab i
    unsigned long	l_time_stamp;	// Time stamp.
    unsigned long	l_checksum;	// Checksum of symb names & comm sizes.
    unsigned long	l_version;	// String table index for version.
    unsigned long	l_flags;	// Flags.
};

// The l_flags may contain the following flags.
enum elf_mips_l_flags_val {
    LL_EXACT_MATCH	= 0x00000001,	// Require an exact match at runtime.
    LL_IGNORE_INT_VER	= 0x00000002	// Ignore version incomp. at runtime.
};

// A section of type SHT_MIPS_CONFLICT is an array of indices into the
// .dynsym section.  Each element has the following type.
typedef unsigned long elf_mips_conflict;

// A section of type SHT_MIPS_GPTAB contains information about how
// much GP space would be required for different -G arguments.  This
// information is only used so that the linker can provide informative
// suggestions as to the best -G value to use.  The sh_info field is
// the index of the section for which this information applies.  The
// contents of the section are an array of the following union.  The
// first element uses the gt_header field.  The remaining elements use
// the gt_entry field.

union elf_mips_gptab {
    struct {
	// -G value actually used for this object file.
	unsigned long 	gt_current_g_value;
	unsigned long 	gt_unused;	// Unused
    } gt_header;
    struct {
	// If this -G argument has been used...
	unsigned long 	gt_g_value;
	// ...this many GP section bytes would be required.
	unsigned long 	gt_bytes;
    } gt_entry;
};

// A section of type SHT_MIPS_REGINFO contains the following structure.
struct elf_mips_reginfo {
    unsigned long 	ri_gprmask;	// Mask of general purpose regs used.
    unsigned long 	ri_cprmask[4];	// Mask of co-processor regs used.
    long 		ri_gp_value;	// GP reg. value for this object file.
};
#endif
