//-----------------------------------------------------------------------------
//   Large block memory allocator 
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "stdtp.hpp"
#include "ermsg.hpp"
#include "numlb.hpp"
#include "mmlow.hpp"

//#define MMBLK_DEBUG

#ifdef MMBLK_DEBUG
    void rtm_blk_dump();
#endif

//-----------------------------------------------------------------------------

global	rtm_low_memory_info	rtm_mem;
global	rtm_hook_list		rtm_morecore_hooks;

//=============================================================================

struct	rtm_free_node {
    rtm_free_node*	next;
    rtm_free_node*	prev;
    size_t		size;
};

#define	RTM_MEM_NUM_FREE_CHAINS \
  (1 << (RTM_LOG_MEM_CLUSTER_SIZE - RTM_LOG_MEM_BLOCK_SIZE + 1))

local	rtm_free_node	free_list_headers[RTM_MEM_NUM_FREE_CHAINS + 1];
local	rtm_free_node*	free_list_curr;

//-----------------------------------------------------------------------------
// blk_pool_init -- initialize block pool data structures 
//-----------------------------------------------------------------------------

local	void	blk_pool_init() {
    rtm_free_node*	f;
    rtm_free_node*	fe;

    // Make cycles in list headers
    f = free_list_headers;
    fe = f + RTM_MEM_NUM_FREE_CHAINS + 1;
    do {
	f->next = f->prev = f;
	f += 1;
    } while (f != fe);

    // Set last searched item in variable sized blocks list to list header
    free_list_curr = f - 1;

    // Initialize low level structures
    rtm_mem_init();
}

//-----------------------------------------------------------------------------
// rtm_mem_check_init -- initialize block pool data structures if them are not
//			 already initializied
//-----------------------------------------------------------------------------

local 	inline	void	rtm_mem_check_init() {
    if (rtm_mem.page_size == 0) {
	blk_pool_init();
    }
}

//-----------------------------------------------------------------------------

#if 0
void  check_fixed_block(rtm_free_node* p) {
    rtm_free_node *q, *r;
    int i;

    Trace("checking block @%p\n", p);
    for (q = p, r = q->next; r != p; q = r, r = r->next) {
	if (r->prev != q)
	    *(int*)0 = 0;
    }
}
#define inline
#else
#define check_fixed_block(p)
#endif


//-----------------------------------------------------------------------------
// remove_fixed_block
//-----------------------------------------------------------------------------

local	inline	void	remove_fixed_block(rtm_free_node* p) {
    check_fixed_block(p);
    p->next->prev = p->prev;
    p->prev->next = p->next;
}

//-----------------------------------------------------------------------------
// remove_block
//-----------------------------------------------------------------------------

local	inline	void	remove_block(rtm_free_node* p) {
    remove_fixed_block(p);
    if (p == free_list_curr)
	free_list_curr = p->next;
}

//-----------------------------------------------------------------------------
// insert_block
//-----------------------------------------------------------------------------

local	inline	void	insert_block(rtm_free_node* p) {
    rtm_free_node* q;
    size_t 	s;

    // Point to the right list
    s = p->size >> RTM_LOG_MEM_BLOCK_SIZE;
    q = free_list_headers + ((s <= RTM_MEM_NUM_FREE_CHAINS) 
			     ? s - 1 : RTM_MEM_NUM_FREE_CHAINS);

    check_fixed_block(q);

    // Insert into the list
    p->next = q->next;
    p->prev = q;
    q->next->prev = p;
    q->next = p;
}

//-----------------------------------------------------------------------------
// rtm_blk_low_morecore
//-----------------------------------------------------------------------------

#define RTM_INITIAL_TAG_TABLE_SIZE	(8*RTM_MEM_BLOCK_SIZE)

struct rtm_free_block {
    rtm_free_block*	next;
    size_t		size;
};

#ifndef RTM_TWO_LEVEL_TAG_TABLE

local	rtm_free_block*	rtm_blk_low_morecore(size_t size) {
    char*	b;
    rtm_free_block* p;

    // Request memory from operating system
    b = rtm_mem_sys_morecore(size);

    // Recalculate the pool layout data
    if (rtm_mem.p_base == NULL)
	rtm_mem.p_base = b;
    rtm_mem.p_size = (b - rtm_mem.p_base) + size;

    // Build free block at the start of allocated region
    p = (rtm_free_block*)b;
    p->size = size;

    // Reallocate tag table if needed
    if (((rtm_mem.p_size>>RTM_LOG_MEM_BLOCK_SIZE) + 2) > rtm_mem.t_size) {
	size_t		new_t_size; // Number of elements in new tag table
	rtm_tag*	new_t_base; // Base address of new tag table

	// Calculate allocation size, it should be large enough to map
	// the data just allocated and a new tag table if required.
	// And we must allocate the multiple of RTM_MEM_BLOCK_SIZE bytes.
	new_t_size = (rtm_mem.t_size 
		      + (size >> RTM_LOG_MEM_BLOCK_SIZE) 
		      + RTM_INITIAL_TAG_TABLE_SIZE);
	new_t_size = ALIGN (new_t_size, (RTM_MEM_BLOCK_SIZE/sizeof(rtm_tag)));

	// If some strange thing will happen, and somehow, someone will
	// do sbrk(2) between the allocation of data region and tag table
	// region it is not a catastrophe, because we do not assume that
	// tag table itself is mapped by the tag table. And subsequent
	// allocation will fix this problem before we put this tag table into
	// the free block list.
	new_t_base = 
	    (rtm_tag*)rtm_mem_sys_morecore (new_t_size*sizeof(rtm_tag));

	if (rtm_mem.t_size != 0) {
	    memcpy(new_t_base, rtm_mem.t_base, rtm_mem.t_size*sizeof(rtm_tag));

	    // Append an old tag table to free block chain
	    p->next = (rtm_free_block*)rtm_mem.t_base;
	    p->next->next = NULL;
	    p->next->size = rtm_mem.t_size*sizeof(rtm_tag);
	} else {
	    *new_t_base = RTM_CONS_TAG(NULL, RTM_TAG_EDGE);
	}

	// initialize the rest of table
	{
	    rtm_tag* tp;
	    rtm_tag* tpe;

	    tp = new_t_base + rtm_mem.t_size; 
	    tpe = new_t_base + new_t_size;
	    while (tp != tpe) {
		*tp++ = RTM_CONS_TAG(NULL, RTM_TAG_USED);
	    }
	}

	// Switch to a new map
	rtm_mem.t_base = new_t_base;
	rtm_mem.t_size = new_t_size;

	// Recalculate virtual origin for tag table
	rtm_mem.v_base = new_t_base -
	    ((large)rtm_mem.p_base >> RTM_LOG_MEM_BLOCK_SIZE) + 1;

    } else {
	p->next = NULL;		// Returning single block chain
    }

    RTM_TAG(rtm_mem.p_base + rtm_mem.p_size) 
	= RTM_CONS_TAG(NULL, RTM_TAG_EDGE);
    return p;
}
#else
local	rtm_free_block*	rtm_blk_low_morecore(size_t size) {
    char*	b;
    rtm_free_block* p;

    b = rtm_mem_sys_morecore(size);
    if (rtm_mem.p_base == NULL)
	rtm_mem.p_base = b;
    rtm_mem.p_size = (b - rtm_mem.p_base) + size;

    // Build free block at the start of allocated region
    p = (rtm_free_block*)b;
    p->size = size;



    return p;
}
#error FIXME: did not implemented yet
#endif

//-----------------------------------------------------------------------------
// rtm_blk_morecore
//-----------------------------------------------------------------------------

local	char*	rtm_blk_morecore(size_t size) {
    rtm_free_block*	a;
    rtm_free_block* 	b;
    size_t		a_size;
    
    // Try to reclaim memory from upper level subsystems
    if (rtm_morecore_hooks.run_hooks() == TRUE) {
	return rtm_blk_alloc (size);
    }

    // Prevent allocation of too small portions
    a_size = size;
    if (a_size < RTM_MEM_MORECORE_QUANT)
	a_size = RTM_MEM_MORECORE_QUANT;

    // Request memory from lower levels
    if ((a = rtm_blk_low_morecore (a_size)) == NULL) {
	Abzac("No memory");
    }

    // Possibly we got several chunks of memory, insert all of them into free
    // lists using common deallocation function. 
    do {
	b = a->next;
	rtm_blk_free ((char*)a, a->size);
    } while ((a = b) != NULL);

    // And now we can safely allocate what we want
    return rtm_blk_alloc (size);
}

//-----------------------------------------------------------------------------
// rtm_blk_alloc
//-----------------------------------------------------------------------------

global	char*	rtm_blk_alloc(size_t size) {
    rtm_free_node*	f;
    rtm_free_node*	q;
    rtm_free_node*	r;

    // Ensure that we already did all the initialization
    rtm_mem_check_init(); 

    // Try to get memory from small block list
    if (size <= (RTM_MEM_BLOCK_SIZE * RTM_MEM_NUM_FREE_CHAINS)) {
	f = free_list_headers + (size >> RTM_LOG_MEM_BLOCK_SIZE) - 1;

	// Exact fit ?
	if (f->next != f) {
	    // Simply get the block
	    f = f->next;
	    remove_fixed_block(f);
	    goto found;
	}

	// Lookup for a larger block in a fixed sized blocks list
	while (++f < (free_list_headers + RTM_MEM_NUM_FREE_CHAINS)) {
	    if (f->next != f) {
		// unlink the block
		f = f->next;
		remove_fixed_block(f);

		// Split it and insert the rest of block into appropriate list
		f->size -= size;
		insert_block(f);
		RTM_TAG(f) = RTM_TAG((char*)f + f->size - 1) =
		    RTM_CONS_TAG(f, 0);

		// Point to the carved out region
		f = (rtm_free_node*)((char*)f + f->size);
		goto found;
	    }
	}
    }

    // Try to find memory in large block list
    // We are using the first fit algorithm with roving pointer 
    // and boundary tags
    q = free_list_curr;
    for (r = q; (f = r->next) != q; r = f) {
	if (f->size >= size) {
	    remove_fixed_block(f);
	    if (f->size > size) {
		f->size -= size;
		insert_block(f);
		RTM_TAG(f) = RTM_TAG((char*)f + f->size - 1) =
		    RTM_CONS_TAG(f, 0);
		f = (rtm_free_node*)((char*)f + f->size);
	    }
	    free_list_curr = r;
	    goto found;
	}
    }

    // Request memory from operating system
    return rtm_blk_morecore(size);

  found:
    // Set the boundary tags
    RTM_TAG(f) = RTM_TAG((char*)f + size - 1)
      = RTM_CONS_TAG(size, RTM_TAG_USED);

    #ifdef MMBLK_DEBUG
        Trace("leave in rtm_blk_alloc(%#x)\n", size);
        rtm_blk_dump();
    #endif

    return (char*)f;
}

//-----------------------------------------------------------------------------
// rtm_blk_realloc
//-----------------------------------------------------------------------------

global char*	rtm_blk_realloc(char* p, size_t size) {
    rtm_tag*		t;
    size_t		old_size;
    
    t = RTM_TAG_A(p);
    old_size = RTM_TAG_ADDR(*t);
    if (size < old_size) {	// Squeezing ?
	RTM_TAG(p + size - 1) = RTM_TAG_SET_ADDR(*t, size);
	rtm_blk_free(p + size, old_size - size);
        #ifdef MMBLK_DEBUG
            Trace("leave in rtm_blk_realloc %p(%#x)\n", p, size);
            rtm_blk_dump();
        #endif
	return p;      
    } else {
	if (size == old_size) { // Nothing to do
            #ifdef MMBLK_DEBUG
                Trace("leave in rtm_blk_realloc %p(%#x)\n", p, size);
                rtm_blk_dump();
            #endif
	    return p;		// Just simply return the block address
	}
    }

    // Extending --- the hardest case.
    {
	rtm_tag*		t1;

	t1 = RTM_TAG_A(p + old_size);
	if ((RTM_TAG_VAL(*t1) & (RTM_TAG_USED | RTM_TAG_EDGE)) == 0) {
	    rtm_free_node*	f;
	    
	    f = (rtm_free_node*)RTM_TAG_ADDR(*t1);
	    if (f->size + old_size >= size) {
		remove_block(f);
		RTM_TAG(p + size - 1) = RTM_TAG_SET_ADDR(*t, size);
		if (f->size + old_size > size) {
		    rtm_blk_free(p + size, f->size + old_size - size);
		}
                #ifdef MMBLK_DEBUG
                    Trace("leave in rtm_blk_realloc %p(%#x)\n", p, size);
                    rtm_blk_dump();
                #endif
		return p;
	    }
	}

	char* q;

	q = rtm_blk_alloc(size);
	memcpy(q, p, old_size);
	rtm_blk_free(p, old_size);
        #ifdef MMBLK_DEBUG
            Trace("leave in rtm_blk_realloc %p(%#x)\n", p, size);
            rtm_blk_dump();
        #endif
	return q;
    }
}

//-----------------------------------------------------------------------------
// rtm_blk_valloc
//-----------------------------------------------------------------------------

global char*	rtm_blk_valloc(size_t size, int align_log) {
    char*	p;
    char*	q;
    long 	l;
    size_t	ds;

    if (align_log <= RTM_LOG_MEM_BLOCK_SIZE) {
	return rtm_blk_alloc(size);
    }
    l = 1 << align_log;
    p = rtm_blk_alloc(size + l);

    q = (char*)(((large)p + (l-1))&~(l-1));
    ds = q - p;
    if (ds != 0) {
	rtm_blk_free(p, ds);
    }
    rtm_blk_free(q + size, size + l - ds);
    return q;
}

//-----------------------------------------------------------------------------
// rtm_blk_free -- return block to free list
//-----------------------------------------------------------------------------

global void	rtm_blk_free(char* p, size_t size) {
    rtm_tag	t;
    rtm_free_node* f;

    #ifdef MMBLK_DEBUG
        Trace("enter in rtm_blk_free(%p,%#x)\n", p, size);
        rtm_blk_dump();
    #endif
    rtm_tag_set_segm(p, (size>>RTM_LOG_MEM_BLOCK_SIZE), 
		     RTM_CONS_TAG(NULL, 0));

    // Try to coalesce with block before
    t = RTM_TAG(p-1);
    if ((RTM_TAG_VAL(t) & (RTM_TAG_USED|RTM_TAG_EDGE)) == 0) {
	f = (rtm_free_node*)RTM_TAG_ADDR(t);
	remove_block(f);
	size += f->size;
	p = (char*)f;
    }

    // Try to coalesce with block after
    t = RTM_TAG(p + size);
    if ((RTM_TAG_VAL(t) & (RTM_TAG_USED|RTM_TAG_EDGE)) == 0) {
	f = (rtm_free_node*)RTM_TAG_ADDR(t);
	remove_block(f);
	size += f->size;
    }

    // Build free block
    RTM_TAG(p) = RTM_TAG(p + size - 1) = RTM_CONS_TAG(p, 0);
    f = (rtm_free_node*)p;
    f->size = size;
    insert_block(f);

    #ifdef MMBLK_DEBUG
        Trace("leaving rtm_blk_free\n");
        rtm_blk_dump();
    #endif
}

//----------------------------------------------------------------------------

#ifdef	MMBLK_DEBUG
void	rtm_blk_dump() {
    rtm_free_node	*p, *q;
    int i;

    Trace("rtm_blk_dump:\n");
    for (i = 0; i <= RTM_MEM_NUM_FREE_CHAINS; i++) {
	Trace(". Queue #%#x\n", i);
	for (p = free_list_headers+i, q = p->next; p != q; q = q->next) {
	    Trace(". . Block at %p next = %p prev %p size = %#x\n",
		  q, q->next, q->prev, q->size);
	}
    }   
    Trace("Curr %#x\n", free_list_curr);
    Trace("end of rtm_blk_dump\n");
}

#endif

#ifdef RTM_MMLOW_DEBUG

void    rtm_mem_info_dump() {
    Trace("rtm_mem dump {:\n");
    Trace("\tpage_size = %#lx, s_limit = %#lx, u_limit = %#lx\n",
	  rtm_mem.page_size, rtm_mem.s_limit, rtm_mem.u_limit);
    Trace("\tp_base = %#lx, p_size = %#lx\n", 
	  rtm_mem.p_base, rtm_mem.p_size);
    Trace("\tt_base = %#lx, v_base = %#lx, t_size = %#lx\n",
	  rtm_mem.t_base, rtm_mem.v_base, rtm_mem.t_size);
    Trace("}\n");
}

#endif
