//-----------------------------------------------------------------------------
//   Emulation of ANSI C memory allocator 
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <string.h>
#include <stdlib.h>
#include "stdtp.hpp"
#include "numlb.hpp"
#include "ermsg.hpp"
#include "mmlow.hpp"
#include "mmlib.hpp"

local	rtm_mem_page_queue page_queues[RTM_MEM_BLOCK_SIZE + 1];
global	rtm_mem_page_queue* const rtm_mem_free_headers[RTM_MEM_BLOCK_SIZE + 1]
 =
{
#include "mmlib.dpp"
};

//-----------------------------------------------------------------------------
// rtm_new_page --- allocate new cluster from block pool, initialize it, and
//		    set as current
//-----------------------------------------------------------------------------

local	rtm_mem_slot_page* rtm_new_page(rtm_mem_page_queue* q) {
    rtm_mem_slot_page*	p;
    size_t		s_size;
    int			i;

    static bool		inited = FALSE;

    if (!inited) {
	inited = TRUE;
	int l = RTM_MEM_CLUSTER_SIZE - sizeof(rtm_mem_slot_page)
	    + sizeof(rtm_mem_free_slot);
	for (i = 1; i <= RTM_MEM_BLOCK_SIZE; i++) {
	    page_queues[i].n_slots = l/i;
	    page_queues[i].s_size = i;
	}
    }

    p = (rtm_mem_slot_page*)rtm_blk_alloc(RTM_MEM_CLUSTER_SIZE);
    rtm_tag_set_segm(p, RTM_MEM_CLUSTER_SIZE/RTM_MEM_BLOCK_SIZE,
		     RTM_CONS_TAG(p, (RTM_TAG_USED | RTM_TAG_CLUSTER)));
    q->f_page = p;
    s_size = p->s_size = q->s_size;
    i = (p->n_used = q->n_slots) - 1;
    p->s_free = NULL;

    p->p_prev = NULL;
    p->p_next = NULL;

    // Make free list
    rtm_mem_free_slot*	f_curr;
    rtm_mem_free_slot*	f_next;

    f_next = f_curr = p->b;
    q->f_slot = f_next;
    while (--i >= 0) {
	f_next = (rtm_mem_free_slot*)((char*)f_next + s_size);
	f_curr = f_curr->next = f_next;
    }
    f_curr->next = NULL;

    return p;
}

//-----------------------------------------------------------------------------

global	rtm_mem_free_slot* rtm_mem_get_slot(rtm_mem_page_queue* q) {
    rtm_mem_slot_page*	p;

    if ((p = q->f_page) == NULL) {
	p = rtm_new_page(q);
    } else {
	if (p->s_free == NULL) {
	    if (p->p_next == NULL) {
		p = rtm_new_page(q);
	    } else {
		p = q->f_page = p->p_next;
		p->p_prev = NULL;
		p->n_used = q->n_slots;
		q->f_slot = p->s_free;
		p->s_free = NULL;
	    }
	} else {
	    q->f_page = p;
	    p->n_used = q->n_slots;
	    q->f_slot = p->s_free;
	    p->s_free = NULL;
	}
    }    
    return q->f_slot;
}

//-----------------------------------------------------------------------------

inline	void*	rtm_small_alloc(size_t size) {
    rtm_mem_page_queue*	q;
    rtm_mem_free_slot*	f;

    q = rtm_mem_free_headers[size];
    if ((f = q->f_slot) == NULL) {
	f = rtm_mem_get_slot(q);
    }
    q->f_slot = f->next;
    return (void*)f;
}

//-----------------------------------------------------------------------------

extern "C" void*	malloc(size_t size) {
    if (size <= RTM_MEM_BLOCK_SIZE) {
	return rtm_small_alloc(size);
    } else {
	return rtm_blk_alloc(ALIGN(size, RTM_MEM_BLOCK_SIZE));
    }
}

//-----------------------------------------------------------------------------

extern "C" void*	calloc(size_t n_items, size_t size) {
    void* p;

    size *= n_items;
    if ((p = malloc(size)) != NULL)
	memset (p, 0, size);
    return p;
}

//-----------------------------------------------------------------------------

extern "C" void*	realloc(void* oldp, size_t size) {
    rtm_tag	t;

    t = RTM_TAG(oldp);
    if ((RTM_TAG_VAL(t) & RTM_TAG_CLUSTER) == 0) {
	return rtm_blk_realloc((char*)oldp, ALIGN(size, RTM_MEM_BLOCK_SIZE));
    } else {
	if (size <= ((rtm_mem_slot_page*)RTM_TAG_ADDR(t))->s_size){
	    return oldp;
	} else {
	    void* pn;

	    if ((pn = malloc(size)) == NULL) {
		return NULL;
	    }
	    memcpy(pn, oldp, ((rtm_mem_slot_page*)RTM_TAG_ADDR(t))->s_size);
	    free(oldp);
	    return pn;
	}
    }
}

//-----------------------------------------------------------------------------

extern "C" void	free(void* ptr) {
    rtm_tag		t;

    if (ptr == NULL)
	return;

    t = RTM_TAG(ptr);
    if ((RTM_TAG_VAL(t) & RTM_TAG_CLUSTER) == 0) {
	rtm_blk_free((char*)ptr, (size_t)RTM_TAG_ADDR(t));
    } else {
	rtm_mem_slot_page*	p;

	p = (rtm_mem_slot_page*)RTM_TAG_ADDR(t);
	((rtm_mem_free_slot*)ptr)->next = p->s_free;
	p->s_free = (rtm_mem_free_slot*)ptr;
	if (--p->n_used == 0) {
	    // p->n_used == 1 && p->prev == NULL identifies current page
	    if (p->p_prev != NULL) {
		if ((p->p_prev->p_next = p->p_next) != NULL) {
		    p->p_next->p_prev = p->p_prev;
		}
		rtm_blk_free((char*)p, RTM_MEM_CLUSTER_SIZE);
	    }
	} else {
	    if (p->p_prev == NULL) {
		rtm_mem_slot_page*	pp;

		pp = page_queues[p->s_size].f_page;
		if (pp != p) {
		    if ((p->p_next = pp->p_next) != NULL) {
			p->p_next->p_prev = p;
		    }
		    p->p_prev = pp;
		    pp->p_next = p;
		}
	    }
	}
    }	  
}

//-----------------------------------------------------------------------------

extern "C" void*	valloc(size_t size) {
    static size_t	log_p_size;

    if (log_p_size == 0) {
	size_t	p_size;

	if (rtm_mem.page_size < RTM_MEM_BLOCK_SIZE)
	    p_size = RTM_MEM_BLOCK_SIZE;
	else
	    p_size = rtm_mem.page_size;

	do {
	    log_p_size += 1;
	} while ((p_size >>= 1) != 0);
    }
    size = ALIGN(size, (1<<log_p_size));
    return rtm_blk_valloc(size, log_p_size);
}

//-----------------------------------------------------------------------------

extern "C" void*	xmalloc(size_t size) {
    void*	p;

    if ((p = malloc(size)) == NULL)
	Fatal("No memory");
    return p;
}

extern "C" void*	xcalloc(size_t N, size_t size) {
    void*	p;

    if ((p = calloc(N, size)) == NULL)
	Fatal("No memory");
    return p;
}

extern "C" void*	xrealloc(void* oldp, size_t size) {
    void*	p;

    if ((p = realloc(oldp, size)) == NULL)
	Fatal("No memory");
    return p;
}

//-----------------------------------------------------------------------------
// For some reasons GNUs do that, we blindly follow them

extern "C" void*	_malloc(size_t size) {
    return malloc(size);
}

extern "C" void*	_calloc(size_t N, size_t size) {
    return calloc(N, size);
}

extern "C" void*	_realloc(void* oldp, size_t size) {
    return realloc(oldp, size);
}

extern "C" void	_free(void* p) {
    free(p);
}

extern "C" void*	_valloc(size_t size) {
    return valloc(size);
}

//-----------------------------------------------------------------------------
