//-----------------------------------------------------------------------------
//   Low level memory allocator implementation
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>

#include "stdtp.hpp"
#include "ermsg.hpp"
#include "numlb.hpp"
#include "mmlow.hpp"

//-----------------------------------------------------------------------------
// rtm_mem_sys_morecore -- request more storage from operating system
//
// Parameter: 
//   size_t a_size -- number of bytes to allocate, should be multiple of
//		      RTM_MEM_BLOCK_SIZE. 
// Returns: 
//   address of allocated region, or NULL if allocation failed, errno
//   contains error code in the last case. Region start is aligned at 
//   RTM_MEM_BLOCK_SIZE boundary.
//-----------------------------------------------------------------------------

global	char*	rtm_mem_sys_morecore(size_t a_size){
    long	a_base;
    size_t	a_align;

    ASSERT(rtm_mem.page_size);

    // Calculate alignment 
    a_align = (size_t)sbrk(0) & (RTM_MEM_CLUSTER_SIZE - 1);
    if (a_align != 0)
	a_align = RTM_MEM_CLUSTER_SIZE - a_align;

    // Do actual allocation and alignment
    if ((a_base = (long)sbrk(a_size + a_align)) == -1)
	a_base = 0;
    else
	a_base += a_align;

    #ifdef MMLOW_TRACE
    {
	int old_errno;

	old_errno = errno;	// Save value or errno
        Trace("rtm_mem_sys_morecore[%#x,(%#x)] align = %#x\n", 
	      a_base, a_size, a_align);
	errno = old_errno;	// Restore errno
    }
    #endif
    return (char*)a_base;
}

//-----------------------------------------------------------------------------
// rtm_mem_sys_lesscore -- try return region of memory to operating system
// 
// Parameters:
//   char*	r_base -- address of memory region start
//   size_t	r_size -- size of memory region
//
// Returns:
//   TRUE on successful return.
//   FALSE in the case if region is not adjacent to the end of data
//	segment or if the brk system call fails. In the latter case, the
//	errno variable contains error code.
//-----------------------------------------------------------------------------

global	bool	rtm_mem_sys_lesscore(char* r_base, size_t r_size) {
    bool	success;

    success = FALSE;		// Be pessimistic
    if ((r_base + r_size) == (char*)sbrk(0) || brk(r_base) != 0)
	success = TRUE;

    #ifdef MMLOW_TRACE
    {
	int old_errno;

	old_errno = errno;	// Save value or errno
        Trace("rtm_mem_sys_lesscore[%#x,(%#x)] success = %#x\n", 
	      r_base, r_size, success);
	errno = old_errno;	// Restore errno
    }
    #endif

    return success;
}

//=============================================================================
// Emulate getpagesize on systems that lack it.
//=============================================================================

#ifndef HAVE_GETPAGESIZE
    #ifdef VMS
        #define getpagesize() 512
    #endif
    #ifdef _SC_PAGESIZE
        #define getpagesize() sysconf(_SC_PAGESIZE) 
    #else
        #include <sys/param.h>
        #ifdef EXEC_PAGESIZE
            #define getpagesize() EXEC_PAGESIZE
        #else
            #ifdef NBPG
                #define getpagesize() NBPG * CLSIZE
                #ifndef CLSIZE
                    #define CLSIZE 1
                #endif /* no CLSIZE */
            #else /* no NBPG */
                #ifdef NBPC
                    #define getpagesize() NBPC
                #endif /* NBPC */
            #endif /* no NBPG */
        #endif /* no EXEC_PAGESIZE */
    #endif /* no _SC_PAGESIZE */
#endif /* not HAVE_GETPAGESIZE */

//=============================================================================
// Get system limit on data section
//=============================================================================

// Some systems need this before <sys/resource.h>.
#include <sys/types.h>

#if defined(__osf__) || defined(__bsdi__) || defined(BSD4_2) || \
  defined(__linux__) || defined(__svr4__)
    extern "C" {
        #include <sys/time.h>
        #include <sys/resource.h>
    }
    #define BSD_STYLE
#else
    #ifndef USG
        #ifdef MSDOS
            #include <dpmi.h>
        #else
            #include <sys/vlimit.h>
        #endif /* not MSDOS */
    #endif /* not USG */
#endif /* BSD4_2 */

extern char etext;
#define start_of_data() &etext

#ifdef NO_LIM_DATA
    local void get_lim_data () {
        rtm_mem.s_limit = -1;
    }
#else // not NO_LIM_DATA
    #if defined(USG) || defined(__hpux__)
        extern "C" long ulimit(int cmd, long);
	extern int end;
        local void get_lim_data () {

            rtm_mem.s_limit = -1;

            // Use the ulimit call, if we seem to have it.
            #if !defined (ULIMIT_BREAK_VALUE) || defined (LINUX) || defined(__hpux__)
                rtm_mem.s_limit = ulimit(3, 0);
            #endif

	    // If that didn't work, just use the macro's value.
            #ifdef ULIMIT_BREAK_VALUE
                if (rtm_mem.s_limit == -1)
		    rtm_mem.s_limit = ULIMIT_BREAK_VALUE;
            #endif
//	    rtm_mem.s_limit -= (long) data_space_start;
	    rtm_mem.s_limit -= (long) &end;
	}
    #else // not USG
        #if !defined (BSD_STYLE)
            #ifdef MSDOS
                local void get_lim_data() {
		    _go32_dpmi_meminfo info;

		    _go32_dpmi_get_free_memory_information (&info);
		    rtm_mem.s_limit = info.available_memory;
		}
            #else // not MSDOS
                local void get_lim_data() {
		    rtm_mem.s_limit = vlimit(LIM_DATA, -1);
	        }
            #endif // not MSDOS
        #else // BSD_STYLE
            local void get_lim_data () {
	        struct rlimit XXrlimit;

	        getrlimit (RLIMIT_DATA, &XXrlimit);
                #ifdef RLIM_INFINITY
                    rtm_mem.s_limit = XXrlimit.rlim_cur & RLIM_INFINITY;
                #else
                    rtm_mem.s_limit = XXrlimit.rlim_cur;
                #endif
	    }
        #endif // BSD_STYLE
    #endif // not USG
#endif

//-----------------------------------------------------------------------------
// rtm_mem_init --- initialize low level memory management structures
//-----------------------------------------------------------------------------

global	void	rtm_mem_init() {

    // Prevent reinitialization
    if (rtm_mem.page_size != 0)
	return;

    // Try to determine page size
    rtm_mem.page_size = getpagesize();

    // Setup limit information
    get_lim_data();
    if (rtm_mem.u_limit == 0 || rtm_mem.u_limit > rtm_mem.s_limit)
	rtm_mem.u_limit = rtm_mem.s_limit;
}

//=============================================================================

#include <signal.h>
#include <sys/types.h>
#include <sys/mman.h>

#ifndef MAP_ANON
#define MAP_ANON 0
#endif

//-----------------------------------------------------------------------------

extern "C" void* rtm_mem_map (char* addr, size_t len) {
    return mmap (addr, len, PROT_EXEC|PROT_READ|PROT_WRITE,
		 MAP_ANON, -1, 0);
}

//-----------------------------------------------------------------------------

extern "C" int	rtm_mem_unmap (char* addr, size_t len) {
    return munmap (addr, len);
}

//-----------------------------------------------------------------------------

static	const int	rtm_mode_decode [4] = {
    0,					// rtm_mem_none
    PROT_READ|PROT_EXEC,		// rtm_mem_read
    PROT_WRITE,				// rtm_mem_write
    PROT_READ|PROT_WRITE|PROT_EXEC	// rtm_mem_read | rtm_mem_write
};

extern "C" int  rtm_mem_protect (char* addr, size_t len, int mode) {
    return mprotect (addr, len, rtm_mode_decode[mode]);
}

//-----------------------------------------------------------------------------
extern	"C" void rtm_mem_process_fault (char* addr, bool is_write);
extern	"C" void rtm_mem_low_prot_handler(int s_code);

#ifdef __linux__
struct rtm_linux_exception_frame {
    long	signr;
    long	gs, fs, es, ds;
    long	edi, esi, ebp, efp;
    long	ebx, edx, ecx, eax;
    long	tss_trap_no, tss_error_code;
    long	eip, cs, eflags, esp, ss;
    long	fpp_stat;
    // Two following fields are not provided by iBCS2
    long	oldmask, tss_cr2;
};

extern "C" void	rtm_mem_low_prot_handler(int s_code) {
    rtm_linux_exception_frame* efp;
    char*	addr;
    bool	is_write;
    
    efp = (rtm_linux_exception_frame*)&s_code;
    addr = (char*)efp->tss_cr2;
    is_write = (efp->tss_error_code & 0x2) == 0;
    #if 1
    fprintf (stderr, "SIGSEGV signal received, addr = %p, write = %s\n",
	     addr, is_write ? "yes" : "no");
    #endif
    rtm_mem_process_fault (addr, is_write);
    rtm_mem_setup_handler();
}
#else
extern "C" void	rtm_mem_low_prot_handler(int s_code) {
    #if 1
    fprintf (stderr, "SIGSEGV signal received, code = %d\n", s_code );
    #endif
}
#endif

extern "C" void rtm_mem_setup_handler() {
    signal (SIGSEGV, &rtm_mem_low_prot_handler);
}


