//-----------------------------------------------------------------------------
//   Optimizing generator for block sizes grouping  
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "stdtp.hpp"
#include "mmlib.hpp"
#include "mmlow.hpp"

//===== Parameters ============================================================

// Compile time parameters
#define	BUF_SIZE	1024

// Command line parameter values
char*	program_name;
char*	output_file_name;
char*	input_file_name;
int	max_n_group;
char*	table_name;
bool	sentinel;
int	alloc_quant;

//====== Global Variables and Primitive operations for them ===================

// grouping information
int	num_of_groups;
int	group_size[RTM_MEM_BLOCK_SIZE];

inline	void	push_group(int size) {
    group_size[num_of_groups++] = size;
}

double	group_prob[RTM_MEM_BLOCK_SIZE];

// Optimizer tables
double*	g_cost;

// cost(i, j) is an overhead introduced by merging j initial groups starting 
// from i-th group
inline	double& cost(int i, int j) { 
    return g_cost[i*num_of_groups + j - 1];
}

double*	g_weight;

// weight(i, j) is weight of optimal sequence
// of first i initial groups grouped into j segments
inline	double& weight(int i, int j) {
    return g_weight[(i-1)*num_of_groups + j - 1];
}

int*	g_last;

// last(i, j) is length (in initial groups) of last segment in optimal
// sequence of first i initial groups grouped into j segments 
inline	int&	last(int i, int j) {
    return g_last[(i-1)*num_of_groups + j - 1];
}

//====== Procedures ===========================================================

//-----------------------------------------------------------------------------
// build_frequencies -- read input file, and build table of group frequences
//			i.e. relative frequency (probability if normalized)
//			of allocation request for size within given group
//-----------------------------------------------------------------------------

void	build_frequencies() {
    FILE*	input;
    char	buf[BUF_SIZE];
    int		len;
    double	prob;
    int		line = 0;

    if ((input = fopen(input_file_name, "r")) == NULL) {
	fprintf(stderr, "can't open input file \"%s\"\n", input_file_name);
	exit(1);
    }
    while (fgets(buf, BUF_SIZE, input) != NULL) {
	line += 1;
	if (buf[0] == '#')
	    continue;
	if (sscanf(buf, "%d : %lg", &len, &prob) != 2) {
	    fprintf(stderr, "bad format of line %d (%s)\n", line, buf);
	    exit(1);
	}
	if (len > RTM_MEM_BLOCK_SIZE) {
	    fprintf(stderr, 
		    "warning: size %d greater than block size(%d). ignored\n",
		    len, RTM_MEM_BLOCK_SIZE);
	    continue;
	}

        int i, s;
	for (i = 0, s = 0; len > s; i++) 
	    s += group_size[i];
	group_prob[i] += prob;
    }
    fclose(input);
}

//-----------------------------------------------------------------------------
// build_sequence -- convert optimal sequence data encoded into "last" table
//		     into group_size array
//-----------------------------------------------------------------------------

void	build_sequence(int n_segms) {
    int		seq[RTM_MEM_BLOCK_SIZE];
    int* 	p_seq = seq;
    int		n, g;
    
    // Extract sequence from "last" table
    for (n = n_segms, g = num_of_groups; n > 0; n--) {
	g -= *p_seq++ = last(g, n);
    }

    // Merge groups according to extracted seguence
    {
	int	*p, *q;
	int	i,  l;

	for (p = q = group_size; p_seq != seq; p_seq--) {
	    for (i = *p_seq, l = 0; --i >= 0;) {
		l += *q++;
	    }
	    *p++ = l;
	}
	num_of_groups = n_segms;
    }
}

//-----------------------------------------------------------------------------
// build_g_wait --- build weight(,) and last(,) tables
// Thanks to R.Bellman for saving a lot of our CPU time
//-----------------------------------------------------------------------------

void	build_g_wait() {
    int i, j, k, l;
    double c, d = 0.0;

    // Allocate memory
    if ((g_weight = new double[num_of_groups * num_of_groups]) == NULL
	|| (g_last = new int[num_of_groups * num_of_groups]) == NULL) 
    {
	fprintf(stderr, "Memory allocation failure\n");
	exit(1);
    }

    // Base of induction --- diagonal
    for (i = 1; i <= num_of_groups; i++) {
	weight(i, i) = 0.0;
	last(i, i) = 1;
    }
    
    // Fill the left triangle
    for (i = 1; i < num_of_groups; i++) {
	for (j = 1; j <= (num_of_groups - i); j++) {
	    for (l = k = 1, c = weight(i - k, j - 1); (i - k) >= (j - 1); k++){
		d = weight(i - k, j - 1);
		if (d < c) {
		    c = d;
		    l = k;
		}
	    }
	    weight(i, j) = d;
	    last(i, j) = l;
	}
    }
}

//-----------------------------------------------------------------------------
// build_g_cost -- build cost table
//-----------------------------------------------------------------------------

void	build_g_cost() {
    int i, j, k, l;

    // Allocate memory
    if ((g_cost = new double[num_of_groups * num_of_groups]) == NULL) {
	fprintf(stderr, "Memory allocation failure\n");
	exit(1);
    }

    for (i = 0; i < num_of_groups; i++)
	cost(1, i) = 0.0;
    for (i = 2; i <= num_of_groups; i++) {
	for (j = 0; j <= (num_of_groups - i); j++) {
	    for (k = 1, l = 0; k < i; k++)
		l += group_size[k + j] - group_size[k + j - 1];
	    cost(i, j) = group_prob[j]*l + cost(i - i, j + 1);
	}
    }
}

//-----------------------------------------------------------------------------
// init_groups -- build initial groups
//-----------------------------------------------------------------------------

void	init_groups(int	quant, int usable) {
    int i, d;

    for (i = quant, d = 0; i <= RTM_MEM_BLOCK_SIZE; i += quant) {
	if ((usable/i) != d) {
	    d = usable/i;
	    push_group(i);
	}
    }

    // ensure that last group was closed 
    if (group_size[num_of_groups - 1] != RTM_MEM_BLOCK_SIZE)
	push_group(RTM_MEM_BLOCK_SIZE);
}

//-----------------------------------------------------------------------------
// write_groups -- generate rounding table by group sizes
//-----------------------------------------------------------------------------

void	write_groups() {
    int		i;
    int		g;
    FILE*	output;

    if ((output = fopen(output_file_name, "w")) == NULL) {
	fprintf(stderr, "Can't open output file \"%s\"\n", output_file_name);
	exit(1);
    }
    fprintf(output, "// Do not edit, this file is generated automagically\n");
    for (i = 0, g = 0; i <= RTM_MEM_BLOCK_SIZE; i++) {
	if (i > group_size[g])
	    g += 1;
	fprintf(output, "%s+%d%s", table_name, group_size[g],
		(i&3) ? ", " : ",\n");
    }    
    fclose(output);
}

//-----------------------------------------------------------------------------
// usage --- print usage information, and exit program
//-----------------------------------------------------------------------------

void 	usage() {
    fprintf(stderr, 
	    "This is Garret C-Talk memory queues size distribution "
	    "optimizer\n"
	    "Usage: %s -o ofile -t table [-n maxn] [-s] [histogram]\n"
	    "  -o ofile  -- name of output file\n"
	    "  -t table  -- name of table in which to index\n"
	    "  -n maxn   -- limit on number of groups\n"
	    "  -s	 -- use sentinel\n"
	    "  -q number -- minimal allocation quant\n"
	    "  histogram -- name of file with size frequences\n",
	    program_name);
    exit(0);
}

//-----------------------------------------------------------------------------
// process_options -- parse and validate command line options
//-----------------------------------------------------------------------------

void	process_options(int argc, char** argv) {

    // Store program name
    program_name = *argv++; 
    argc -= 1;

    // Parse options
    for (; --argc >= 0; argv++) {
	if (**argv == '-') {
	    switch((*argv)[1]) {
	      case 'o':
		if (argc == 0 || output_file_name != NULL)
		    usage();
		argv++; argc--;
		output_file_name = *argv;
		break;
	      case 't':
		if (argc == 0 || table_name != NULL)
		    usage();
		argv++; argc--;
		table_name = *argv;
		break;
	      case 'n':
		if (argc == 0 || max_n_group != 0)
		    usage();
		argv++; argc--;
		max_n_group = atoi(*argv);
		break;
	      case 's':
		sentinel = TRUE;
		break;
	      case 'q':
		if (argc == 0 || alloc_quant != 0)
		    usage();
		argv++; argc--;
		alloc_quant = atoi(*argv);
		break;
	      default :
		usage();
	    }
	} else {
	    if (input_file_name != NULL) {
		usage();
	    } else {
		input_file_name = *argv;
	    }
	}
    }

    // Validate options
    if (output_file_name == NULL || table_name == NULL)
	usage();
}

//-----------------------------------------------------------------------------
// main -- top level rouutine
//-----------------------------------------------------------------------------

int main(int argc, char** argv) {
    process_options(argc, argv);
    init_groups(alloc_quant == 0 ? sizeof(void*) : alloc_quant,
		RTM_MEM_CLUSTER_SIZE - sizeof(rtm_mem_slot_page) +
		  sizeof(rtm_mem_free_slot) - 
		(sentinel ? sizeof(void*) : 0) );
    if (input_file_name != NULL) {
	build_frequencies();
	build_g_cost();
	build_g_wait();
	build_sequence(max_n_group);
    }
    write_groups();
}

