//-----------------------------------------------------------------------------
//  Pointer swizzler implementation
//  Copyright (C)  1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <string.h>
#include <stdlib.h>
#include "stdtp.hpp"
#include "memmv.hpp"
#include "swizl.hpp"
#include "numlb.hpp"
#include "ermsg.hpp"
#include "mmlow.hpp"
#include "mmlib.hpp"

//-----------------------------------------------------------------------------

global	rtm_swz_pool*	rtm_swz_pool::list;

//-----------------------------------------------------------------------------
// rtm_mem_process_fault -- is called from machine and system dependent level
//			    when memory protection exception is raised
//			    (such as SIGSEGV)

extern	"C" void rtm_mem_process_fault (char* addr, bool is_write) {
    rtm_swz_pool*	pool;
    rtm_swz_slot*	slot;

    for (pool = rtm_swz_pool::list; pool != NULL; pool = pool->next) {
	if ((size_t)(addr - pool->base) < pool->size) {
	    slot = pool->get_slot(addr);
	    if (slot->usr & rtm_swz_slot::allocated) {
		if ((slot->usr & rtm_swz_slot::mapped) == 0) {
		    rtm_mem_protect ((char*)slot->v_addr, 
				     slot->size << pool->log_page_size,
				     rtm_mem_read|rtm_mem_write);
		    pool->reader (pool, slot);
		    pool->swizzler (pool, slot);
		    if (is_write) {
			slot->usr |= (rtm_swz_slot::mapped
				      | rtm_swz_slot::modified);
		    } else {
			slot->usr |= rtm_swz_slot::mapped;
			rtm_mem_protect ((char*)slot->v_addr, 
					 slot->size << pool->log_page_size,
					 rtm_mem_read);
		    }
		} else {
		    // assert
		    slot->usr |= rtm_swz_slot::modified;
		    rtm_mem_protect ((char*)slot->v_addr, 
				     slot->size << pool->log_page_size,
				     rtm_mem_read|rtm_mem_write);
		}
		return;
	    } else {
		break;
	    }
	}
    }
    // FIXME: call here default handler ...
    abort();
}

//-----------------------------------------------------------------------------

void	rtm_swz_pool::sweep() {
    rtm_swz_slot**	pps;
    rtm_swz_slot*	ps;
    rtm_swz_slot*	slot;
    int			i;

    for (i = 0; i < h_table_size; i += 1) {
	for (pps = h_table + i; (ps = *pps) != NULL;) {
	    if (ps->usr & rtm_swz_slot::marked) {
		ps->usr &= ~rtm_swz_slot::marked;
		pps = &ps->next;
	    } else {
		*pps = ps->next;
		if (ps->usr & rtm_swz_slot::mapped) {
		    if (ps->usr & rtm_swz_slot::modified) {
			unswizzler (this, ps);
			writer (this, ps);
		    } else {
			flusher (this, ps);
		    }
		    rtm_mem_protect ((char*)ps->v_addr, 
				     ps->size << log_page_size,
				     rtm_mem_read|rtm_mem_write);
		}
		slot = ps + ps->size;

		do {
		    slot -= 1;
		    slot->next = free;
		    free = slot;
		    slot->usr = 0;
		} while (slot != ps);
	    }
	}
    }
}

//-----------------------------------------------------------------------------

rtm_swz_slot*	rtm_swz_pool::find (rtm_swz_key& key) {
    unsigned 		idx;
    rtm_swz_slot*	slot;

    idx = key.hash()%h_table_size;
    for (slot = h_table[idx]; slot != NULL; slot = slot->next) {
	if (slot->key == key)
	    break;
    }
    return slot;
}

//-----------------------------------------------------------------------------

rtm_swz_slot*	rtm_swz_pool::alloc(rtm_swz_key& key) {
    unsigned 		idx;
    rtm_swz_slot*	slot;

    // Search if slot with key given was already allocated
    idx = key.hash()%h_table_size;
    for (slot = h_table[idx]; slot != NULL; slot = slot->next) {
	if (slot->key == key)
	    return slot;
    }

    // Allocate new slot
    slot = free;
    if (slot == NULL) {
	// is it enough ???
	return NULL;
    }
    free = slot->next;

    // Build new slot
    slot->next = h_table[idx];
    h_table[idx] = slot;
    slot->key = key;
    slot->usr = rtm_swz_slot::allocated;
    slot->size = 1;

    return slot;
}

//-----------------------------------------------------------------------------

rtm_swz_pool::rtm_swz_pool (size_t		v_size,
			    size_t		v_log_page_size,
			    void* 		v_user_key,
			    rtm_swz_reader	v_reader,
			    rtm_swz_writer	v_writer,
			    rtm_swz_flusher	v_flusher,
			    rtm_swz_swizzler	v_swizzler,
			    rtm_swz_unswizzler	v_unswizzler)
{
    int			n_slots;
    rtm_swz_slot*	p_slot;
    size_t		page_size;
    char*		p_area;

    // Initialize data fields
    size = v_size;
    log_page_size = v_log_page_size;
    user_key = v_user_key;
    reader = v_reader;
    writer = v_writer;
    flusher = v_flusher;
    swizzler = v_swizzler;
    unswizzler = v_unswizzler;

    // Map and protect data area
    base = (char*)rtm_mem_map (NULL, size);
    rtm_mem_protect (base, size, rtm_mem_none);

    // Create and initialize hash table
    h_table = (rtm_swz_slot**)malloc (h_table_size * sizeof(rtm_swz_slot*));
    pzero (h_table, h_table_size);

    // Create and initialize slot table
    n_slots = size >> log_page_size;
    slot = new rtm_swz_slot [n_slots];
    free = slot;
    page_size = 1 << log_page_size;
    for (p_area = base, p_slot = slot; --n_slots >= 0; p_slot++, p_area++) {
	p_slot->next = p_slot + 1;
	p_slot->v_addr = p_area;
	p_slot->usr = 0;
    }
    (p_slot - 1)->next = NULL;

    // Insert self into global list
    // CS::in(list)
    next = list;
    list = this;
    // CS::out(list)
}

//-----------------------------------------------------------------------------

rtm_swz_pool::~rtm_swz_pool () {
    rtm_swz_pool**	ppp;
    rtm_swz_pool*	pp;

    // CS::in(list)
    // Find self in global list
    for (ppp = &list; (pp = *ppp) != this; ppp = &pp->next) {
	if (pp == NULL)
	    abort();
    }
    
    // Exclude self from global list
    *ppp = next;
    // CS::out(list)

    delete [] slot;
    delete [] h_table;
    rtm_mem_unmap (base, size);
}

