//
//-< FILES.HPP >-----------------------------------------------------*--------*
// Standard class          Version 1.0           (c) 1992  GARRET    *     ?  *
//   library                                                         *   /\|  *
//                                                                   *  /  \  *
// File system interface                                             * / [] \ *
//                                                                   * GARRET *
//-------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#ifndef __FILES_HPP__
    #define __FILES_HPP__

    class   File_Name;
    class   File_Info;
    class   File_System_Info;

    class   Gaw_Editor;
    class   Gaw_Box;
    class   Gaw_Window;

    // Formats to print file name. Formats specific to each file system could
    // be added in file system specific code.

    enum    File_Form   {
        Ffm_Short,              // Shortest way to print file name
        Ffm_Medium,             // Medium
        Ffm_Long,               // Long ( usually 80 syms ) file info
        Ffm_Full,               // Full file information
        Ftm_Last                // Unused, but must be present
    };

    //        I N T E R F A C E   W I T H   F I L E   S Y S T E M

    // Implementation of this functions are duty of the system dependent part
    // of library

    class  Fsi  {
      public :
        // General

        static  bool        Init( void );           // Start the system
        static  char const* Error( void );          // Get the text for the
                                                    // system error
        // System information

        static  char    const*  User_Name;
        static  char    const*  Node_Name;
        static  char    const*  Sys_Name;
        static  char    const*  Ver_Name;
        static  char    const*  Rel_Name;
        static  char    const*  Mach_Name;

        // Files

        static  int         Create( char const* Path );
        static  int         Delete( char const* Path );
        static  int         Alias ( char const* From, char const* To );
        static  int         Copy  ( char const* From, char const* To );
        static  int         Move  ( char const* From, char const* To );

        // Directory

        static  File_Name*  Root_Dir;               // Root directory
        static  File_Name*  Curr_Dir;               // Current directory
        static  File_Name*  Home_Dir;               // User's home directory
        static  bool        CD( char const* Name ); // Change current directory
        static  bool        CD( File_Name* Name );  // Another form of CD
        static  int         Mkdir( char const* Path );
        static  int         Rmdir( char const* Path );
        static  int         Mvdir( char const* From, char const* To );
        static  int         Cpdir( char const* From, char const* To );
        static  int         Prune( char const* Path );
    };

    //  F I L E    N A M E

    // Bit definition for file name format mask

    enum    Fnm_Spec        {
       Fnm_Dev     =       0x01,
       Fnm_Dir     =       0x02,
       Fnm_Nam     =       0x04,
       Fnm_Typ     =       0x08,
       Fnm_Oth     =       0x10,
       Fnm_All     =       0x1F
    };

    enum   Fns_Status   {
        Fns_Null        = 0x00,
        Fns_Static      = 0x01
    };

    class   File_Name   {
      protected :
        File_Name( char const* Name, bool Static );
      public :
        File_Name*              Nxt;
        File_Name*              Nxd;
        File_Name*              Dir;
        File_Info*              Inf;

        char const*             Nam;
        short                   Usd;
        tiny                    Flg;
        tiny                    Typ;
        Gaw_Editor*             Edi;

        virtual                 ~File_Name( void );
        bool                    Forget( void );
        void                    Fix( void );
        void                    Unfix( void );

        virtual char const*     Name( enum Fnm_Spec Form = Fnm_All ) = 0;

        // File manipulations :

        static  File_Name*      New( char const* Name );
        static  File_Name*      New( char const* Name, char const*  Def_Dir );
        static  File_Name*      New( char const* Name, File_Name* Def_Dir );

        virtual int             Print( char* Buf, int Len, char const* Format ) = 0;
        #if DEBUG >= DUMP
            virtual void        Dump( int Depth = 0 );
            static  void        Dump_Tree( File_Name* Node = Fsi::Root_Dir );
        #endif

        Gaw_Editor*     Edit( Gaw_Box& Box, Gaw_Window* w, char* Options=NULL );
        int             Read_Dir( void );
    };

    //   F I L E   I N F O

    class   File_Info : public L2_Elem {
      public :
        #define FDS_EDITING     0x0001          // File is being edited now
        #define FDS_OBSOLETE    0x0002          // File was changed since editor
                                                // started
        #define FDS_LOCK        0x0004          // File info is locked
        #define FDS_TAG         0x0008          // File is tagged
        #define FDS_INC         0x0010          // (For directory) Info on sub-
                                                // files is incomplete
        #define FDS_DIR         0x0020          // Directory
        #define FDS_DEV         0x0040          // Device
        #define FDS_GHOST       0x0080          // ...

        short                   Stat;           // Status of file entry
        tiny                    Type;           // Generic type of file system
        tiny                    Vers;           // Subtype of file system
        File_Name*              Names;          // Names of this file
        File_Name*              Sub_Files;      // Used for directory files only

        File_Info( void );
        virtual ~File_Info( void );

        int                     N_Files( void );
        virtual int             Read_Dir    ( void ) = 0;
        virtual bool            Changed     ( void ) = 0;
        virtual bool            Update      ( bool Force = FALSE ) = 0;
        virtual bool            Forget      ( void );

        #if DEBUG >= DUMP
            virtual     void    Dump( int Depth = 1 );
        #endif

        static  File_Info*      Get( char const* Path );
    };

    inline  int   File_Name::Read_Dir( void ) { return Inf->Read_Dir(); }

    // F I L E   S Y S T E M   I N F O R M A T I O N

    // Some predefined types of file systems

    enum    Fsys_Type   {
        Fls_Unix,               // Classical unix file system
        Fls_11,                 // Vax/Vms Files-11 file system
        Fls_Dos                 // Ms_Dos file system
    };

    // File_System_Info -- information on mounted file system. Abstract class

    class   File_System_Info    : public L2_Elem {
      protected :
        File_System_Info( void );
      public :
        char*           Sys_Name;       // File system name
        char*           Dev_Name;       // Name of device where system mounted
        large           Blk_Total;      // Total number of blocks
        large           Blk_Used;       // Number of allocated blocks
        large           Blk_Free;       // Number of free blocks
        large           Max_Files;      // Maximal number of files
        large           Usd_Files;      // Number of existing files
        short           Blk_Size;       // Block size
        tiny            Stat;
        tiny            Type;
        int             Used;

        // Define some bits of the Stat
        #define         FSI_OBSOLETE    0x01    // Information is obsolete
        #define         FSI_REMOTE      0x02    // System is mounted via network
        #define         FSI_NATIVE      0x04    // File system is native for
                                                // operating system

        virtual                 ~File_System_Info( void );
        virtual int             Update( void ) = 0;
        virtual char const*     Sys_Type_Name( void ) = 0;

        // Debugging tools
        #if DEBUG >= DUMP
            virtual     void    Dump( void );
            static      void    Dump_All( void );
        #endif

        static  int     Load_List( void );
        static  void*   Get( char  const* Name );
        static  void*   Get( int   Idn  );

        static  L2_Elem List;
    };

#endif
