#include    <stdlib.h>
#include    "stdtp.hpp"
#include    "ermsg.hpp"
#include    "dnmem.hpp"
#include    "semph.hpp"
#include    "ctxsw.hpp"
#include    "timer.hpp"

#define TRACING   0

//-----------------------------------------------------------------------------

local   void    prc_def_hook(char const* msg) { Abzac(msg); }

//---------------

Task       *Task::tsk_head,      *Task::tsk_tail;
Task       *Task::run_head,      *Task::run_tail;
Task       *Task::current;

void       (*Task::err_hook)(char const*) = prc_def_hook;

//-----------------------------------------------------------------------------
// Critical section stuff

int        Task_Critical = 0;
int        Critical_Request = 0;

inline  void    cs_enter() { Task_Critical++; }

inline  void    cs_leave() {
  Task_Critical--;
  if (Critical_Request) {
      int i = Critical_Request;

      Critical_Request = 0;
      tmr_proc_event(i);
  }
}

//-----------------------------------------------------------------------------

void    Timer_Handler(void* arg){
  Task*  t = (Task*)arg;

  t->state = Task::Ready;
  t->run_queue_ins();
}

//-----------------------------------------------------------------------------

Task::Task(Task* p){

    if ((parent = p) != NULL) {
        brother = p->subproc;
        p->subproc = this;
    } else
        brother = NULL;
    subproc = NULL;

    #if TRACING
        Trace("Task::Task this = %#x\n", this);
    #endif
    tsk_next = NULL;
    if ((tsk_prev = tsk_head) == NULL)
        tsk_tail = this;
    else
        tsk_head->tsk_next = this;
    tsk_head = this;
    pri = 0;
    ctx_area = NULL;
    ctx_size = 0;
    trm = sem = NULL;
    mutexs = NULL; 
    save_hook = load_hook = exit_hook = NULL;
}

//-----------------------------------------------------------------------------

Task::~Task(){
    #if TRACING
        Trace("Task::~Task (%#x)\n", this);
    #endif

    if (state == Zomby)
        err_hook("proc:zomby ~Task");
    state = Zomby;

    // Call user hook, if present
    if (exit_hook != NULL)
        exit_hook(this);

    // Unlink from global task list
    if (tsk_next == NULL) tsk_head = tsk_prev;
    else tsk_next->tsk_prev = tsk_prev;
    if (tsk_prev == NULL) tsk_tail = tsk_next;
    else tsk_prev->tsk_next = tsk_next;

    if (trm != NULL)
        trm->open();

    // Free context buffer, if neccesary
    if (ctx_area != NULL)
        Dnm_Fre_Block(ctx_area);

    // Release all mutexs hold by deleted task
    while( mutexs != NULL ) mutexs->release(); 
}

//-----------------------------------------------------------------------------

void    Task::run_queue_ins(){
    cs_enter();
    if (run_head == NULL) {
        #if TRACING
            Trace("Task::run_queue_ins 0 this = %#x\n", this);
        #endif
        run_head = run_tail = this;
        run_next = run_prev = NULL;
    } else if (pri > run_head->pri) {
        #if TRACING
            Trace("Task::run_queue_ins 1 this = %#x\n", this);
        #endif
        run_head->run_next = this;
        run_prev = run_head;
        run_next = NULL;
        run_head = this;
    } else if (pri <= run_tail->pri) {
        #if TRACING
            Trace("Task::run_queue_ins 2 this = %#x\n", this);
        #endif
        run_tail->run_prev = this;
        run_next = run_tail;
        run_tail = this;
        run_prev = NULL;
    } else {
        Task* t;
        #if TRACING
            Trace("Task::run_queue_ins 3 this = %#x\n", this);
        #endif
        for (t = run_head; t->pri >= pri; t = t->run_prev);
        run_prev = t;
        run_next = t->run_next;
        t->run_next = run_next->run_prev = this;
    }
    cs_leave();
}

void    Task::run_queue_rem(){
    cs_enter();
    #if TRACING
        Trace("Task::run_queue_rem this = %#x run_head = %#x run_tail = %#x\n",
           this, run_head, run_tail);
    #endif
    if (run_head == this)   run_head = run_prev;
    else run_next->run_prev = run_prev;
    if (run_tail == this)   run_tail = run_next;
    else run_prev->run_next = run_next;
    cs_leave();
}

Task*   Task::fork(){
    Task* t = new Task(current);

    #if TRACING
        Trace("Task::fork\n");
    #endif
    t->pri = current->pri;
    t->state = Ready;
    t->run_queue_ins();
    return t->save_ctx() ? t : (Task*)0;
}

int Task::run
  (int (*func)(int argc, char** argv), int argc, char** argv, bool wait, int pr)
{
  Semaphore* s              = wait ? new Semaphore : (Semaphore*)NULL;
  int        exit_status    = 0;
  Task*      t;

  if((t = Task::fork()) == NULL){
      // child process

      exit_status = func(argc, argv);
      // if (wait) {
      // s->open();
      // current->prune();
      // } else
      Task::exit();
  } else {
      // parent process

      t->trm = s;
      if (pr != current->pri) {
          t->run_queue_rem();
          t->pri = pr;
          t->run_queue_ins();
      }
      if (wait) s->wait();
      else Task::reschedule();
  }
  return exit_status;
}

int Task::run
  (int (*func)(void*), void* args, bool wait, int pr)
{
  Semaphore* s              = wait ? new Semaphore : (Semaphore*)NULL;
  int        exit_status    = 0;
  Task*      t;

  if((t = Task::fork()) == NULL){
      // child process

      exit_status = func(args);
      // if (wait) {
      //  current->prune();
      // } else
      Task::exit();
  } else {
      // parent process

      t->trm = s;
      if (pr != current->pri) {
          t->run_queue_rem();
          t->pri = pr;
          t->run_queue_ins();
      }
      if (wait) s->wait();
      else Task::reschedule();
  }
  return exit_status;
}

int     Task::save_ctx(){
    int i;

    #if TRACING
        Trace("save_ctx -> %#x %#x\n", this, current);
    #endif
    if (save_hook != NULL)
        save_hook(this);
    if ((i = Tsk_Stack_Size()) > ctx_size)
        ctx_area = Dnm_Realloc_Block(ctx_area, ctx_size = i);
    if (((i = Tsk_Save_Stack(ctx_area)) == 0) && (load_hook != NULL))
        load_hook(this);
    #if TRACING
        Trace("save_ctx -> %d %#x %#x\n", i, this, current);
    #endif
    return i;
}

//-----------------------------------------------------------------------------

void    Task::load_ctx()
  { Tsk_Load_Stack(ctx_area, 0); }

//-----------------------------------------------------------------------------

void    Task::reschedule(){
    Task*       t;

    #if TRACING
        Trace("Task::reschedule\n");
    #endif
    if ((t = current) != NULL) {
        if (t->state == Current) {
            t->state = Ready;
            t->run_queue_ins();
        }
        if (t->save_ctx() == 0){
            #if TRACING
                Trace("resch return %#x\n", t);
            #endif
            return;
        }
    }
    if ((t = run_head) == NULL)
        abort();
    t->run_queue_rem();
    t->state = Current;
    current = t;
    #if TRACING
        Trace("loading %#x\n", t);
    #endif
    t->load_ctx();
    //. not reached
    abort();
}

//-----------------------------------------------------------------------------

void    Task::init(int i) {
    Task*       main_proc;

    ASSERT(Tsk_Stack_Base == NULL);
    Tsk_Stack_Base = &i;
    main_proc = new Task(NULL);
    main_proc->pri = 300;       // Some random value (yet ?)
    main_proc->state = Current;
    current = main_proc;
}

//-----------------------------------------------------------------------------

void    Task::sleep(int delay) {
  current->timer_entry = tmr_setup_event(delay, Timer_Handler, (void*)current);
  current->state = Sleep;
  reschedule();
}

//-----------------------------------------------------------------------------

void    Task::kill() {
    if (parent != NULL) {
        Task **pp, *p;

        for (pp = &parent->subproc; (p = *pp) != NULL; pp = &p->brother) {
            if (p == this) {
                *pp = p->brother;
                goto found;
            }
        }
        abort();
      found:;
        if (subproc != NULL) {
            p = subproc;
            while (TRUE) {
                p->parent = parent;
                if (p->brother == NULL) {
                    p->brother = parent->subproc;
                    parent->subproc = subproc;
                    break;
                } else
                    p = p->brother;
            }
        }
    } else {
        Task *p, *q;
        for (p = subproc; p != NULL; p = q) {
            q = p->brother;
            p->parent = p->brother = NULL;
        }
    }
    switch (this->state) {
      case Current:
        current = NULL;
        delete this;
        reschedule();
      case Ready:
        run_queue_rem();
        break;
      case Suspended:
        break;
      case Sem_Wait:
        if (sem->head == this) sem->head = sem_prev;
        else sem_next->sem_prev = sem_prev;
        if (sem->tail == this) sem->tail = sem_next;
        else sem_prev->sem_next = sem_next;
        sem->count -= 1;
        break;
      case Sleep:
        tmr_remove_event(timer_entry);
        break;
      case Msg_Wait:
      case Zomby:
      default:
        err_hook("proc: zomby kill");
    }
    delete this;
}

global  void    Task::prune() {
    Task  *p, *q;
    bool  suicide = FALSE;

    p = this;
    do {
        if (p->subproc != NULL)
            p = p->subproc;
        else {
            q = p;
            if (p == this)
                p = NULL;
            else
                p = p->brother != NULL ? p->brother : p->parent;
            if (q != current)
                q->kill();
            else
                suicide = TRUE;
        }
    } while (p != NULL);
    if (suicide)
        Task::exit();
}

global  void    Task::resume() {
    if (state != Suspended)
        err_hook("proc:resume of not suspended");
    state = Ready;
    run_queue_ins();
    if (pri > current->pri)
        reschedule();
}

global  void    Task::suspend() {
    switch(state) {
      case Ready :
        run_queue_rem();
      case Suspended :
        state = Suspended;
        break;
      case Current :
        state = Suspended;
        reschedule();
        break;
      default :
        err_hook("proc:suspend of blocked process");
    }
}

global  void    Task::chpri(int new_pri, bool resch) {
    switch(state) {
      case Current :
        pri = new_pri;
        if (resch && run_head != NULL && run_head->pri >= new_pri)
            reschedule();
        break;
      case Ready :
        run_queue_rem();
        pri = new_pri;
        run_queue_ins();
        if (resch && new_pri >= current->pri)
            reschedule();
        break;
      default :
        pri = new_pri;
    }
}


